import { Nullable, IMatrix, IQuaternion } from "../interface/index";
import { Scene } from "../render/scene";
import { RuntimeAnimation } from "./runtimeAnimation";
import { Animatable } from "./animatable";
import { Animation } from "./animation";
import { Node } from "../render/node";
import { AnimationGroup } from "./animationGroup";
import { IAnimationKey } from "./animationKey";
import { AnimationEvent } from "./animationEvent";

export class AnimConstants {
    /**
     * Use matrix interpolation instead of using direct key value when animating matrices
     */
    public static AllowMatricesInterpolation = false;

    /**
     * When matrix interpolation is enabled, this boolean forces the system to use Matrix.DecomposeLerp instead of Matrix.Lerp. Interpolation is more precise but slower
     */
    public static AllowMatrixDecomposeForInterpolation = true;

    // Statics
    /**
     * Float animation type
     */
    public static readonly ANIMATIONTYPE_FLOAT = 0;
    /**
     * Vector3 animation type
     */
    public static readonly ANIMATIONTYPE_VECTOR3 = 1;
    /**
     * Quaternion animation type
     */
    public static readonly ANIMATIONTYPE_QUATERNION = 2;
    /**
     * Matrix animation type
     */
    public static readonly ANIMATIONTYPE_MATRIX = 3;
    /**
     * Color3 animation type
     */
    public static readonly ANIMATIONTYPE_COLOR3 = 4;
    /**
     * Color3 animation type
     */
    public static readonly ANIMATIONTYPE_COLOR4 = 7;
    /**
     * Vector2 animation type
     */
    public static readonly ANIMATIONTYPE_VECTOR2 = 5;
    /**
     * Size animation type
     */
    public static readonly ANIMATIONTYPE_SIZE = 6;
    // PI_BEGIN 20200904 - 带精度计算的动画数据处理
    /**
     * Size animation type
     */
    public static readonly ANIMATIONTYPE_FLOAT_PRECISION = 8;
    // PI_END
    /**
     * Relative Loop Mode
     */
    public static readonly ANIMATIONLOOPMODE_RELATIVE = 0;
    /**
     * Cycle Loop Mode
     */
    public static readonly ANIMATIONLOOPMODE_CYCLE = 1;
    /**
     * Constant Loop Mode
     */
    public static readonly ANIMATIONLOOPMODE_CONSTANT = 2;

    /** @hidden */
    public static _UniversalLerp(left: any, right: any, amount: number): any {
        let constructor = left.constructor;
        if (constructor.Lerp) { // Lerp supported
            return constructor.Lerp(left, right, amount);
        } else if (constructor.Slerp) { // Slerp supported
            return constructor.Slerp(left, right, amount);
        } else if (left.toFixed) { // Number
            return left * (1.0 - amount) + amount * right;
        } else { // Blending not supported
            return right;
        }
    }
}

/**
 * @hidden
 */
export class _IAnimationState {
    key: number;
    repeatCount: number;
    workValue?: any;
    loopMode?: number;
    offsetValue?: any;
    highLimitValue?: any;
}

export interface IAnimation {
    runtimeAnimations: IRuntimeAnimation[];
    _runtimeAnimations: IRuntimeAnimation[];
    hasRunningRuntimeAnimations: boolean;
    /**Name of the animation */
    name: string,
    /**Property to animate */
    targetProperty: string,
    targetPropertyPath: string[];
    /**The frames per second of the animation */
    framePerSecond: number,
    /**The data type of the animation */
    dataType: number,
    /**The loop mode of the animation */
    loopMode?: number,
    /**Specifies if blending should be enabled */
    enableBlending?: boolean,
    blendingSpeed: number;
    addEvent(event: AnimationEvent): void,
    removeEvents(frame: number): void,
    getEvents(): AnimationEvent[],
    createRange(name: string, from: number, to: number): void,
    deleteRange(name: string, deleteFrames?: boolean): void,
    getKeys(): Array<IAnimationKey>,
    _interpolate(currentFrame: number, state: _IAnimationState): any;
    _getKeyValue(key: string): any;
}

export interface IRuntimeAnimation {
    isStopped(): boolean;
}

export interface IAnimatable {
    isAdditive: boolean;
    syncRoot: Nullable<Animatable>;
    onAnimationEnd?: Nullable<() => void>;
}

export interface IAnimationScene extends Scene {
    /** @hidden */
    _registerTargetForLateAnimationBinding(runtimeAnimation: RuntimeAnimation, originalValue: any): void;

    /** @hidden */
    _processLateAnimationBindingsForMatrices(holder: {
        totalWeight: number,
        totalAdditiveWeight: number,
        animations: RuntimeAnimation[],
        additiveAnimations: RuntimeAnimation[],
        originalValue: IMatrix
    }): any;

    /** @hidden */
    _processLateAnimationBindingsForQuaternions(holder: {
        totalWeight: number,
        totalAdditiveWeight: number,
        animations: RuntimeAnimation[],
        additiveAnimations: RuntimeAnimation[],
        originalValue: IQuaternion
    }, refQuaternion: IQuaternion): IQuaternion;

    /** @hidden */
    _processLateAnimationBindings(): void;

    /**
     * Will start the animation sequence of a given target
     * @param target defines the target
     * @param from defines from which frame should animation start
     * @param to defines until which frame should animation run.
     * @param weight defines the weight to apply to the animation (1.0 by default)
     * @param loop defines if the animation loops
     * @param speedRatio defines the speed in which to run the animation (1.0 by default)
     * @param onAnimationEnd defines the function to be executed when the animation ends
     * @param animatable defines an animatable object. If not provided a new one will be created from the given params
     * @param targetMask defines if the target should be animated if animations are present (this is called recursively on descendant animatables regardless of return value)
     * @param onAnimationLoop defines the callback to call when an animation loops
     * @param isAdditive defines whether the animation should be evaluated additively (false by default)
     * @returns the animatable object created for this animation
     */
    beginWeightedAnimation(target: any, from: number, to: number, weight: number, loop?: boolean, speedRatio?: number,
        onAnimationEnd?: () => void, animatable?: Animatable, targetMask?: (target: any) => boolean, onAnimationLoop?: () => void, isAdditive?: boolean): Animatable;

    /**
     * Will start the animation sequence of a given target
     * @param target defines the target
     * @param from defines from which frame should animation start
     * @param to defines until which frame should animation run.
     * @param loop defines if the animation loops
     * @param speedRatio defines the speed in which to run the animation (1.0 by default)
     * @param onAnimationEnd defines the function to be executed when the animation ends
     * @param animatable defines an animatable object. If not provided a new one will be created from the given params
     * @param stopCurrent defines if the current animations must be stopped first (true by default)
     * @param targetMask defines if the target should be animate if animations are present (this is called recursively on descendant animatables regardless of return value)
     * @param onAnimationLoop defines the callback to call when an animation loops
     * @param isAdditive defines whether the animation should be evaluated additively (false by default)
     * @returns the animatable object created for this animation
     */
    beginAnimation(target: any, from: number, to: number, loop?: boolean, speedRatio?: number,
        onAnimationEnd?: () => void, animatable?: Animatable, stopCurrent?: boolean,
        targetMask?: (target: any) => boolean, onAnimationLoop?: () => void, isAdditive?: boolean): Animatable;

    /**
     * Will start the animation sequence of a given target and its hierarchy
     * @param target defines the target
     * @param directDescendantsOnly if true only direct descendants will be used, if false direct and also indirect (children of children, an so on in a recursive manner) descendants will be used.
     * @param from defines from which frame should animation start
     * @param to defines until which frame should animation run.
     * @param loop defines if the animation loops
     * @param speedRatio defines the speed in which to run the animation (1.0 by default)
     * @param onAnimationEnd defines the function to be executed when the animation ends
     * @param animatable defines an animatable object. If not provided a new one will be created from the given params
     * @param stopCurrent defines if the current animations must be stopped first (true by default)
     * @param targetMask defines if the target should be animated if animations are present (this is called recursively on descendant animatables regardless of return value)
     * @param onAnimationLoop defines the callback to call when an animation loops
     * @param isAdditive defines whether the animation should be evaluated additively (false by default)
     * @returns the list of created animatables
     */
    beginHierarchyAnimation(target: any, directDescendantsOnly: boolean, from: number, to: number, loop?: boolean, speedRatio?: number,
        onAnimationEnd?: () => void, animatable?: Animatable, stopCurrent?: boolean,
        targetMask?: (target: any) => boolean, onAnimationLoop?: () => void, isAdditive?: boolean): Animatable[];

    /**
     * Begin a new animation on a given node
     * @param target defines the target where the animation will take place
     * @param animations defines the list of animations to start
     * @param from defines the initial value
     * @param to defines the final value
     * @param loop defines if you want animation to loop (off by default)
     * @param speedRatio defines the speed ratio to apply to all animations
     * @param onAnimationEnd defines the callback to call when an animation ends (will be called once per node)
     * @param onAnimationLoop defines the callback to call when an animation loops
     * @param isAdditive defines whether the animation should be evaluated additively (false by default)
     * @returns the list of created animatables
     */
    beginDirectAnimation(target: any, animations: Animation[], from: number, to: number, loop?: boolean, speedRatio?: number, onAnimationEnd?: () => void, onAnimationLoop?: () => void, isAdditive?: boolean): Animatable;

    /**
     * Begin a new animation on a given node and its hierarchy
     * @param target defines the root node where the animation will take place
     * @param directDescendantsOnly if true only direct descendants will be used, if false direct and also indirect (children of children, an so on in a recursive manner) descendants will be used.
     * @param animations defines the list of animations to start
     * @param from defines the initial value
     * @param to defines the final value
     * @param loop defines if you want animation to loop (off by default)
     * @param speedRatio defines the speed ratio to apply to all animations
     * @param onAnimationEnd defines the callback to call when an animation ends (will be called once per node)
     * @param onAnimationLoop defines the callback to call when an animation loops
     * @param isAdditive defines whether the animation should be evaluated additively (false by default)
     * @returns the list of animatables created for all nodes
     */
    beginDirectHierarchyAnimation(target: Node, directDescendantsOnly: boolean, animations: Animation[], from: number, to: number, loop?: boolean, speedRatio?: number, onAnimationEnd?: () => void, onAnimationLoop?: () => void, isAdditive?: boolean): Animatable[];

    /**
     * Gets the animatable associated with a specific target
     * @param target defines the target of the animatable
     * @returns the required animatable if found
     */
    getAnimatableByTarget(target: any): Nullable<Animatable>;

    /**
     * Gets all animatables associated with a given target
     * @param target defines the target to look animatables for
     * @returns an array of Animatables
     */
    getAllAnimatablesByTarget(target: any): Array<Animatable>;

    /**
    * Stops and removes all animations that have been applied to the scene
    */
    stopAllAnimations(): void;

    /**
     * Gets the current delta time used by animation engine
     */
    deltaTime: number;
    animationGroupsTS: AnimationGroup[];
    addAnimationGroup(group: AnimationGroup): void;
}