import { BABYLON } from "../render3d/babylon";
import { CylinderBuilder } from "./cylinder";

export class Arrow extends BABYLON.TransformNode {
    lineMesh: BABYLON.AbstractMesh;
}
export class ArrowInstance extends Arrow {
    declare lineMesh: BABYLON.InstancedMesh;
    color(r: number, g: number, b: number) {
        this.lineMesh.instanceAttribute("InsColor4", r, g, b, 1.);
    }
}

export class ArrowBuilder {
    /** @hidden */
    public static _CreateArrow(scene: BABYLON.Scene, material: BABYLON.DefaultMaterial, thickness: number = 1, isCollider = false): Arrow {
        var cone = CylinderBuilder.CreateCylinder({
            diameterTop: 0,
            height: 0.25,
            diameterBottom: 0.2,
            tessellation: 6,
        });
        var line = CylinderBuilder.CreateCylinder({
            diameterTop: 0.05,
            height: 0.75,
            diameterBottom: 0.05,
            tessellation: 6,
        });

        let coneVertexCount = cone.positions.length / 3;
        let lineVertexCount = line.positions.length / 3;
        let coneIndicesCount = cone.indices.length;
        let lineIndicesCount = line.indices.length;

        let positions = new Float32Array(cone.positions.length + line.positions.length);
        let normals = new Float32Array(cone.normals.length + line.normals.length);
        let uvs = new Float32Array(cone.uvs.length + line.uvs.length);
        let indices = new Uint16Array(cone.indices.length + line.indices.length);

        let coneOffsetX = (1.0 - 0.25 / 2);
        let lineOffsetX = 0.75 / 2;
        for (let i = 0; i < coneVertexCount; i++) {
            let j3 = i * 3;
            let j2 = i * 2;
            let i3 = i * 3;
            let i2 = i * 2;
    
            positions[j3 + 0] = cone.positions[i3 + 0];
            positions[j3 + 1] = cone.positions[i3 + 2];
            positions[j3 + 2] = cone.positions[i3 + 1] + coneOffsetX;
            
            normals[j3 + 0] = cone.normals[i3 + 0];
            normals[j3 + 1] = cone.normals[i3 + 2];
            normals[j3 + 2] = cone.normals[i3 + 1];

            uvs[j2 + 0] = cone.uvs[i2 + 0];
            uvs[j2 + 1] = cone.uvs[i2 + 1];
        }
        for (let i = 0; i < lineVertexCount; i++) {
            let j3 = (i + coneVertexCount) * 3;
            let j2 = (i + coneVertexCount) * 2;
            let i3 = i * 3;
            let i2 = i * 2;

            positions[j3 + 0] = line.positions[i3 + 0];
            positions[j3 + 1] = line.positions[i3 + 2];
            positions[j3 + 2] = line.positions[i3 + 1] + lineOffsetX;
            
            normals[j3 + 0] = line.normals[i3 + 0];
            normals[j3 + 1] = line.normals[i3 + 2];
            normals[j3 + 2] = line.normals[i3 + 1];
            
            uvs[j2 + 0] = line.uvs[i2 + 0];
            uvs[j2 + 1] = line.uvs[i2 + 1];

        }

        for (let i = 0; i < coneIndicesCount; i++) {
            indices[i] = cone.indices[i];
        }
        for (let i = 0; i < lineIndicesCount; i++) {
            indices[i + coneIndicesCount] = line.indices[i] + coneVertexCount;
        }

        let vertexData = new BABYLON.VertexData();
        vertexData.indices      = indices;
        vertexData.positions    = positions;
        vertexData.normals      = normals;
        vertexData.uvs          = uvs;
        vertexData.apply();

        // let cylinderMesh = BABYLON.BoxBuilder.CreateBox("AA", scene, [], false, true);
        let arrowMesh = new BABYLON.Mesh(
            "arrow", scene,
            [{ key: "InsColor4", code: "A_COLOR4 = InsColor4;", vtype: BABYLON.Constants.INSTANCE_ATTR_VEC4 }],
            true, true
        );
        let geocylinder = new BABYLON.Geometry("arrow", scene);
        vertexData.applyToGeometry(geocylinder);
        geocylinder.applyToMesh(arrowMesh);
        arrowMesh.alwaysSelectAsActiveMesh = true;

        var arrow = new Arrow("arrow", scene);
        arrow.lineMesh = arrowMesh;

        // Position arrow pointing in its drag axis
        arrowMesh.rotation.y = Math.PI;
        arrowMesh.parent = arrow;
        arrowMesh.setMaterial(material);
        // cylinderMesh.rotation.x = Math.PI / 2;
        // cylinderMesh.position.z += 0.3;

        if (isCollider) {
            arrowMesh.visibility = 0;
        }
        return arrow;
    }
}