import { BABYLON } from "../render3d/babylon";
import { Arrow, ArrowBuilder, ArrowInstance } from "./arrow";
import { AxisDragGizmo } from "./axisDragGizmo";

/**
 * The Axes viewer will show 3 axes in a specific point in space
 */
export class AxesViewer {
    private _base: Arrow;
    private _xAxis: Arrow;
    private _yAxis: Arrow;
    private _zAxis: Arrow;
    private _scaleLinesFactor = 1;
    private _instanced = false;

    /**
     * Gets the hosting scene
     */
    public scene: BABYLON.Nullable<BABYLON.Scene> = null;

    /**
     * Gets or sets a number used to scale line length
     */
    public scaleLines = 1;

    /**
     * Creates a new AxesViewer
     * @param scene defines the hosting scene
     * @param scaleLines defines a number used to scale line length (1 by default)
     * @param renderingGroupId defines a number used to set the renderingGroupId of the meshes (2 by default)
     * @param xAxis defines the node hierarchy used to render the x-axis
     * @param yAxis defines the node hierarchy used to render the y-axis
     * @param zAxis defines the node hierarchy used to render the z-axis
     */
    constructor(scene: BABYLON.Scene, scaleLines = 1, renderingGroupId: BABYLON.Nullable<number> = 2, arrow?: Arrow) {
        this.scaleLines = scaleLines;
        this.scene = scene;

        if (!arrow) {
            arrow = ArrowBuilder._CreateArrow(scene, new BABYLON.DefaultMaterial("Arrow", scene));
            // arrow.cylinderMesh.setEnabled(false);
            // arrow.lineMesh.setEnabled(false);
            this._base = arrow;
        } else {
            this._xAxis = AxisDragGizmo._CreateArrowInstance(this.scene!, arrow);
            this._xAxis.scaling.setAll(this.scaleLines * this._scaleLinesFactor);
            (<ArrowInstance>this._xAxis).color(1, 0, 0);
            this._xAxis.lineMesh.alphaIndex = arrow.lineMesh.alphaIndex;
            this._yAxis = AxisDragGizmo._CreateArrowInstance(this.scene!, arrow);
            this._yAxis.scaling.setAll(this.scaleLines * this._scaleLinesFactor);
            (<ArrowInstance>this._yAxis).color(0, 1, 0);
            this._yAxis.lineMesh.alphaIndex = arrow.lineMesh.alphaIndex;
            this._zAxis = AxisDragGizmo._CreateArrowInstance(this.scene!, arrow);
            this._zAxis.scaling.setAll(this.scaleLines * this._scaleLinesFactor);
            (<ArrowInstance>this._zAxis).color(0, 0, 1);
            this._zAxis.lineMesh.alphaIndex = arrow.lineMesh.alphaIndex;
        }

        if (renderingGroupId != null) {
            AxesViewer._SetRenderingGroupId(arrow, renderingGroupId);
        }

        this.update(new BABYLON.Vector3(), BABYLON.Vector3.Right(), BABYLON.Vector3.Up(), BABYLON.Vector3.Forward());
    }

    /**
     * Force the viewer to update
     * @param position defines the position of the viewer
     * @param xaxis defines the x axis of the viewer
     * @param yaxis defines the y axis of the viewer
     * @param zaxis defines the z axis of the viewer
     */
    public update(position: BABYLON.Vector3, xaxis: BABYLON.Vector3, yaxis: BABYLON.Vector3, zaxis: BABYLON.Vector3): void {
        if (this._xAxis) {
        this._xAxis.position.copyFrom(position);
        this._xAxis.setDirection(xaxis);
        this._xAxis.scaling.setAll(this.scaleLines * this._scaleLinesFactor);
        }

        if (this._yAxis) {
        this._yAxis.position.copyFrom(position);
        this._yAxis.setDirection(yaxis);
        this._yAxis.scaling.setAll(this.scaleLines * this._scaleLinesFactor);
        }

        if (this._zAxis) {
        this._zAxis.position.copyFrom(position);
        this._zAxis.setDirection(zaxis);
        this._zAxis.scaling.setAll(this.scaleLines * this._scaleLinesFactor);
        }

        if (this._base) {
        // this._base.position.copyFrom(position);
        // this._base.setDirection(zaxis);
        // this._base.scaling.setAll(this.scaleLines * this._scaleLinesFactor);
        }
    }

    /**
     * Creates an instance of this axes viewer.
     * @returns a new axes viewer with instanced meshes
     */
    public createInstance(): AxesViewer {
        const axesViewer = new AxesViewer(this.scene!, this.scaleLines, null, this._base);
        axesViewer._instanced = true;
        return axesViewer;
    }

    /** Releases resources */
    public dispose() {
        if (this._xAxis) {
            this._xAxis.dispose();
        }

        if (this._yAxis) {
            this._yAxis.dispose();
        }

        if (this._zAxis) {
            this._zAxis.dispose();
        }
        if (this._base) {
            this._base.dispose();
        }

        this.scene = null;
    }

    private static _SetRenderingGroupId(node: BABYLON.TransformNode, id: number) {
        node.getChildMeshes().forEach((mesh) => {
            (<BABYLON.Mesh>mesh).alphaIndex = 999999 + id;
        });
    }
}
