import { ICustomGeometry, Constants } from "../interface/index";
import { BABYLON } from "../render3d/babylon";

export class CustomGeometry implements ICustomGeometry {
    private kinds: (typeof Constants.VERTEX_ATTR_POSITION | typeof Constants.VERTEX_ATTR_COLOR4 | typeof Constants.VERTEX_ATTR_UV)[] = [];
    private kindSize: number[] = [];
    private kindOffset: number[] = [];
    private perVertxNumber = 1;
    private mPositions: Float32Array;
    private mIndices: Uint16Array;
    private bufferid: string;
    private indicesBufferid: string;
    private vertexDirty = true;
    private indicesDirty = true;
    private custmGeometry: BABYLON.Geometry;
    private static kind(kind: (typeof Constants.VERTEX_ATTR_POSITION | typeof Constants.VERTEX_ATTR_COLOR4 | typeof Constants.VERTEX_ATTR_UV)) {
        if (kind == Constants.VERTEX_ATTR_POSITION)  { return { kind: BABYLON.Constants.VERTEX_ATTR_POSITION, format: BABYLON.Constants.VERTEX_FORMAT_F32X3 } }
        if (kind == Constants.VERTEX_ATTR_UV)        { return { kind: BABYLON.Constants.VERTEX_ATTR_UV, format: BABYLON.Constants.VERTEX_FORMAT_F32X2 } }
        if (kind == Constants.VERTEX_ATTR_COLOR4)    { return { kind: BABYLON.Constants.VERTEX_ATTR_COLOR4, format: BABYLON.Constants.VERTEX_FORMAT_F32X4 } }
    }
    constructor(private mesh: BABYLON.Mesh, private unique: string, kinds: (typeof Constants.VERTEX_ATTR_POSITION | typeof Constants.VERTEX_ATTR_COLOR4 | typeof Constants.VERTEX_ATTR_UV)[], maxVertex: number, maxtriangle: number, private useIndices: boolean = true) {
        this.kinds.length = 0;
        let offset = 0;
        let attrs = [];
        kinds.forEach((kind, idx) => {
            this.kinds.push(kind);
            let size = 0;
            if (kind == Constants.VERTEX_ATTR_POSITION) { size = 3; }
            if (kind == Constants.VERTEX_ATTR_UV)       { size = 2; }
            if (kind == Constants.VERTEX_ATTR_COLOR4)   { size = 4; }
            attrs.push(CustomGeometry.kind(kind));

            this.kindOffset[idx] = offset;
            this.kindSize[idx] = size;
            offset += size;
        });
        this.perVertxNumber = offset;
        
        this.mPositions = new Float32Array(maxVertex * this.perVertxNumber);
        this.bufferid       = `${unique}`;
        this.indicesBufferid= `${unique}#Indices`;

        this.custmGeometry  = new BABYLON.Geometry(this.bufferid, mesh._scene);
        
        this.custmGeometry.setVerticesBuffer(
            <BABYLON.IVertexBuffer> {
                name: this.bufferid,
                range: undefined,
                attrs
            }
        );
        
        if (useIndices) {
            this.mIndices       = new Uint16Array(maxtriangle * 3);
            this.custmGeometry.setIndices({
                name: this.indicesBufferid,
                range: null,
                as_u16: true,
            });
        }

        this.custmGeometry.applyToMesh(this.mesh);
    }
    update(pointCount: number): void {
        if (this.vertexDirty) {
            BABYLON.Geometry.create_vertex_buffer(this.bufferid, this.mPositions);
        }
        if (this.indicesDirty && this.useIndices) {
            BABYLON.Geometry.create_indices_buffer(this.indicesBufferid, this.mIndices);
        }
        if (this.useIndices) {
            this.mesh.indexStart = 0;
            this.mesh.indexCount = pointCount;
        } else {
            this.mesh.vertexStart = 0;
            this.mesh.vertexCount = pointCount;
        }
        
        this.vertexDirty = false;
        this.indicesDirty = false;
    }
    vertices(kind: (typeof Constants.VERTEX_ATTR_POSITION | typeof Constants.VERTEX_ATTR_COLOR4 | typeof Constants.VERTEX_ATTR_UV)): [Float32Array, number, number] {
        let idx = this.kinds.indexOf(kind);
        if (idx >= 0) {
            this.vertexDirty = true;
            return [this.mPositions, this.perVertxNumber, this.kindOffset[idx]]
        }
        return undefined;
    }
    indices() {
        this.indicesDirty = true;
        return this.mIndices;
    }
    dispose() {
        this.mesh.dispose();
    }
}
