import { Vector4 } from "../maths/math.vector";
import { BABYLON } from "../render3d/babylon";

export class CylinderBuilder {
    public static CreateCylinder = function(options: { height?: number, diameterTop?: number, diameterBottom?: number, diameter?: number, tessellation?: number, subdivisions?: number, arc?: number, faceColors?: BABYLON.Color4[], faceUV?: BABYLON.Vector4[], hasRings?: boolean, enclose?: boolean, cap?: number, sideOrientation?: number, frontUVs?: Vector4, backUVs?: Vector4 }): BABYLON.VertexData {
        var height: number = options.height || 2;
        var diameterTop: number = (options.diameterTop === 0) ? 0 : options.diameterTop || options.diameter || 1;
        var diameterBottom: number = (options.diameterBottom === 0) ? 0 : options.diameterBottom || options.diameter || 1;
        diameterTop = diameterTop || 0.00001; // Prevent broken normals
        diameterBottom = diameterBottom || 0.00001; // Prevent broken normals
        var tessellation: number = options.tessellation || 24;
        var subdivisions: number = options.subdivisions || 1;
        var hasRings: boolean = options.hasRings ? true : false;
        var enclose: boolean = options.enclose ? true : false;
        var cap = (options.cap === 0) ? 0 : options.cap || BABYLON.Mesh.CAP_ALL;
        var arc: number = options.arc && (options.arc <= 0 || options.arc > 1) ? 1.0 : options.arc || 1.0;
        var sideOrientation: number = (options.sideOrientation === 0) ? 0 : options.sideOrientation || BABYLON.VertexData.DEFAULTSIDE;
        var faceUV: BABYLON.Vector4[] = options.faceUV || new Array<BABYLON.Vector4>(3);
        var faceColors = options.faceColors;
        // default face colors and UV if undefined
        var quadNb: number = (arc !== 1 && enclose) ? 2 : 0;
        var ringNb: number = (hasRings) ? subdivisions : 1;
        var surfaceNb: number = 2 + (1 + quadNb) * ringNb;
        var f: number;
    
        for (f = 0; f < surfaceNb; f++) {
            if (faceColors && faceColors[f] === undefined) {
                faceColors[f] = new BABYLON.Color4(1, 1, 1, 1);
            }
        }
        for (f = 0; f < surfaceNb; f++) {
            if (faceUV && faceUV[f] === undefined) {
                faceUV[f] = new BABYLON.Vector4(0, 0, 1, 1);
            }
        }
    
        var indices = new Array<number>();
        var positions = new Array<number>();
        var normals = new Array<number>();
        var uvs = new Array<number>();
        var colors = new Array<number>();
    
        var angle_step = Math.PI * 2 * arc / tessellation;
        var angle: number;
        var h: number;
        var radius: number;
        var tan = (diameterBottom - diameterTop) / 2 / height;
        var ringVertex: BABYLON.Vector3 = BABYLON.Vector3.Zero();
        var ringNormal: BABYLON.Vector3 = BABYLON.Vector3.Zero();
        var ringFirstVertex: BABYLON.Vector3 = BABYLON.Vector3.Zero();
        var ringFirstNormal: BABYLON.Vector3 = BABYLON.Vector3.Zero();
        var quadNormal: BABYLON.Vector3 = BABYLON.Vector3.Zero();
        var Y: BABYLON.Vector3 = BABYLON.Axis.Y;
    
        // positions, normals, uvs
        var i: number;
        var j: number;
        var r: number;
        var ringIdx: number = 1;
        var s: number = 1;      // surface index
        var cs: number = 0;
        var v: number = 0;
    
        for (i = 0; i <= subdivisions; i++) {
            h = i / subdivisions;
            radius = (h * (diameterTop - diameterBottom) + diameterBottom) / 2;
            ringIdx = (hasRings && i !== 0 && i !== subdivisions) ? 2 : 1;
            for (r = 0; r < ringIdx; r++) {
                if (hasRings) {
                    s += r;
                }
                if (enclose) {
                    s += 2 * r;
                }
                for (j = 0; j <= tessellation; j++) {
                    angle = j * angle_step;
    
                    // position
                    ringVertex.x = Math.cos(-angle) * radius;
                    ringVertex.y = -height / 2 + h * height;
                    ringVertex.z = Math.sin(-angle) * radius;
    
                    // normal
                    if (diameterTop === 0 && i === subdivisions) {
                        // if no top cap, reuse former normals
                        ringNormal.x = normals[normals.length - (tessellation + 1) * 3];
                        ringNormal.y = normals[normals.length - (tessellation + 1) * 3 + 1];
                        ringNormal.z = normals[normals.length - (tessellation + 1) * 3 + 2];
                    }
                    else {
                        ringNormal.x = ringVertex.x;
                        ringNormal.z = ringVertex.z;
                        ringNormal.y = Math.sqrt(ringNormal.x * ringNormal.x + ringNormal.z * ringNormal.z) * tan;
                        ringNormal.normalize();
                    }
    
                    // keep first ring vertex values for enclose
                    if (j === 0) {
                        ringFirstVertex.copyFrom(ringVertex);
                        ringFirstNormal.copyFrom(ringNormal);
                    }
    
                    positions.push(ringVertex.x, ringVertex.y, ringVertex.z);
                    normals.push(ringNormal.x, ringNormal.y, ringNormal.z);
                    if (hasRings) {
                        v = (cs !== s) ? faceUV[s].y : faceUV[s].w;
                    } else {
                        v = faceUV[s].y + (faceUV[s].w - faceUV[s].y) * h;
                    }
                    uvs.push(faceUV[s].x + (faceUV[s].z - faceUV[s].x) * j / tessellation, v);
                    if (faceColors) {
                        colors.push(faceColors[s].r, faceColors[s].g, faceColors[s].b, faceColors[s].a);
                    }
                }
    
                // if enclose, add four vertices and their dedicated normals
                if (arc !== 1 && enclose) {
                    positions.push(ringVertex.x, ringVertex.y, ringVertex.z);
                    positions.push(0, ringVertex.y, 0);
                    positions.push(0, ringVertex.y, 0);
                    positions.push(ringFirstVertex.x, ringFirstVertex.y, ringFirstVertex.z);
                    BABYLON.Vector3.CrossToRef(Y, ringNormal, quadNormal);
                    quadNormal.normalize();
                    normals.push(quadNormal.x, quadNormal.y, quadNormal.z, quadNormal.x, quadNormal.y, quadNormal.z);
                    BABYLON.Vector3.CrossToRef(ringFirstNormal, Y, quadNormal);
                    quadNormal.normalize();
                    normals.push(quadNormal.x, quadNormal.y, quadNormal.z, quadNormal.x, quadNormal.y, quadNormal.z);
                    if (hasRings) {
                        v = (cs !== s) ? faceUV[s + 1].y : faceUV[s + 1].w;
                    } else {
                        v = faceUV[s + 1].y + (faceUV[s + 1].w - faceUV[s + 1].y) * h;
                    }
                    uvs.push(faceUV[s + 1].x, v);
                    uvs.push(faceUV[s + 1].z, v);
                    if (hasRings) {
                        v = (cs !== s) ? faceUV[s + 2].y : faceUV[s + 2].w;
                    } else {
                        v = faceUV[s + 2].y + (faceUV[s + 2].w - faceUV[s + 2].y) * h;
                    }
                    uvs.push(faceUV[s + 2].x, v);
                    uvs.push(faceUV[s + 2].z, v);
                    if (faceColors) {
                        colors.push(faceColors[s + 1].r, faceColors[s + 1].g, faceColors[s + 1].b, faceColors[s + 1].a);
                        colors.push(faceColors[s + 1].r, faceColors[s + 1].g, faceColors[s + 1].b, faceColors[s + 1].a);
                        colors.push(faceColors[s + 2].r, faceColors[s + 2].g, faceColors[s + 2].b, faceColors[s + 2].a);
                        colors.push(faceColors[s + 2].r, faceColors[s + 2].g, faceColors[s + 2].b, faceColors[s + 2].a);
                    }
                }
                if (cs !== s) {
                    cs = s;
                }
    
            }
    
        }
    
        // indices
        var e: number = (arc !== 1 && enclose) ? tessellation + 4 : tessellation;     // correction of number of iteration if enclose
        var s: number;
        i = 0;
        for (s = 0; s < subdivisions; s++) {
            let i0: number = 0;
            let i1: number = 0;
            let i2: number = 0;
            let i3: number = 0;
            for (j = 0; j < tessellation; j++) {
                i0 = i * (e + 1) + j;
                i1 = (i + 1) * (e + 1) + j;
                i2 = i * (e + 1) + (j + 1);
                i3 = (i + 1) * (e + 1) + (j + 1);
                indices.push(i0, i1, i2);
                indices.push(i3, i2, i1);
            }
            if (arc !== 1 && enclose) {      // if enclose, add two quads
                indices.push(i0 + 2, i1 + 2, i2 + 2);
                indices.push(i3 + 2, i2 + 2, i1 + 2);
                indices.push(i0 + 4, i1 + 4, i2 + 4);
                indices.push(i3 + 4, i2 + 4, i1 + 4);
            }
            i = (hasRings) ? (i + 2) : (i + 1);
        }
    
        // Caps
        var createCylinderCap = (isTop: boolean) => {
            var radius = isTop ? diameterTop / 2 : diameterBottom / 2;
            if (radius === 0) {
                return;
            }
    
            // Cap positions, normals & uvs
            var angle;
            var circleVector;
            var i: number;
            var u: BABYLON.Vector4 = (isTop) ? faceUV[surfaceNb - 1] : faceUV[0];
            var c: BABYLON.Nullable<BABYLON.Color4> = null;
            if (faceColors) {
                c = (isTop) ? faceColors[surfaceNb - 1] : faceColors[0];
            }
            // cap center
            var vbase = positions.length / 3;
            var offset = isTop ? height / 2 : -height / 2;
            var center = new BABYLON.Vector3(0, offset, 0);
            positions.push(center.x, center.y, center.z);
            normals.push(0, isTop ? 1 : -1, 0);
            uvs.push(u.x + (u.z - u.x) * 0.5, u.y + (u.w - u.y) * 0.5);
            if (c) {
                colors.push(c.r, c.g, c.b, c.a);
            }
    
            var textureScale = new BABYLON.Vector2(0.5, 0.5);
            for (i = 0; i <= tessellation; i++) {
                angle = Math.PI * 2 * i * arc / tessellation;
                var cos = Math.cos(-angle);
                var sin = Math.sin(-angle);
                circleVector = new BABYLON.Vector3(cos * radius, offset, sin * radius);
                var textureCoordinate = new BABYLON.Vector2(cos * textureScale.x + 0.5, sin * textureScale.y + 0.5);
                positions.push(circleVector.x, circleVector.y, circleVector.z);
                normals.push(0, isTop ? 1 : -1, 0);
                uvs.push(u.x + (u.z - u.x) * textureCoordinate.x, u.y + (u.w - u.y) * textureCoordinate.y);
                if (c) {
                    colors.push(c.r, c.g, c.b, c.a);
                }
            }
            // Cap indices
            for (i = 0; i < tessellation; i++) {
                if (!isTop) {
                    indices.push(vbase);
                    indices.push(vbase + (i + 1));
                    indices.push(vbase + (i + 2));
                }
                else {
                    indices.push(vbase);
                    indices.push(vbase + (i + 2));
                    indices.push(vbase + (i + 1));
                }
            }
        };
    
        // add caps to geometry based on cap parameter
        if ((cap === BABYLON.Mesh.CAP_START)
            || (cap === BABYLON.Mesh.CAP_ALL)) {
            createCylinderCap(false);
        }
        if ((cap === BABYLON.Mesh.CAP_END)
            || (cap === BABYLON.Mesh.CAP_ALL)) {
            createCylinderCap(true);
        }
    
        // Sides
        BABYLON.VertexData._ComputeSides(sideOrientation, positions, indices, normals, uvs, options.frontUVs, options.backUVs);

        let vertexData = new BABYLON.VertexData();
        vertexData.indices      = indices;
        vertexData.positions    = positions;
        vertexData.normals      = normals;
        vertexData.uvs          = uvs;
        if (faceColors) {
            vertexData.colors   = colors;
        }
    
        return vertexData;
    };
}