import { BABYLON } from "../render3d/babylon";
import { DiscBuidler } from "./disc";
import { SphereBuilder } from "./sphere";

/**
 * Class containing static functions to help procedurally build meshes
 */
export class HemisphereBuilder {
    /**
     * Creates a hemisphere mesh
     * @param name defines the name of the mesh
     * @param options defines the options used to create the mesh
     * @param scene defines the hosting scene
     * @returns the hemisphere mesh
     */
    public static CreateHemisphere(name: string, options: { segments?: number, diameter?: number, sideOrientation?: number }, scene: BABYLON.Scene): BABYLON.Mesh {
        if (!options.diameter) {
            options.diameter = 1;
        }
        if (!options.segments) {
            options.segments = 16;
        }

        var halfSphere = SphereBuilder._CreateSphere({slice: 0.5, diameter: options.diameter, segments: options.segments});
        var disc = DiscBuidler._CreateDisc({
            radius: options.diameter / 2,
            tessellation: (options.segments * 3) + (4 - options.segments),
            sideOrientation: options.sideOrientation,
        });


        let halfSphereVertexCount = halfSphere.positions.length / 3;
        let discVertexCount = disc.positions.length / 3;
        let halfSphereIndicesCount = halfSphere.indices.length;
        let discIndicesCount = disc.indices.length;

        let positions = new Float32Array(halfSphere.positions.length + disc.positions.length);
        let normals = new Float32Array(halfSphere.normals.length + disc.normals.length);
        let uvs = new Float32Array(halfSphere.uvs.length + disc.uvs.length);
        let indices = new Uint16Array(halfSphere.indices.length + disc.indices.length);

        let coneOffsetX = 0;
        let lineOffsetX = 0;
        for (let i = 0; i < halfSphereVertexCount; i++) {
            let j3 = i * 3;
            let j2 = i * 2;
            let i3 = i * 3;
            let i2 = i * 2;
    
            positions[j3 + 0] = halfSphere.positions[i3 + 0] + coneOffsetX;
            positions[j3 + 1] = halfSphere.positions[i3 + 1];
            positions[j3 + 2] = halfSphere.positions[i3 + 2];
            
            normals[j3 + 0] = halfSphere.normals[i3 + 0];
            normals[j3 + 1] = halfSphere.normals[i3 + 1];
            normals[j3 + 2] = halfSphere.normals[i3 + 2];

            uvs[j2 + 0] = halfSphere.uvs[i2 + 0];
            uvs[j2 + 1] = halfSphere.uvs[i2 + 1];
        }
        for (let i = 0; i < discVertexCount; i++) {
            let j3 = (i + halfSphereVertexCount) * 3;
            let j2 = (i + halfSphereVertexCount) * 2;
            let i3 = i * 3;
            let i2 = i * 2;

            positions[j3 + 0] = disc.positions[i3 + 0] + lineOffsetX;
            positions[j3 + 1] = disc.positions[i3 + 2];
            positions[j3 + 2] = disc.positions[i3 + 1];
            
            normals[j3 + 0] = disc.normals[i3 + 0];
            normals[j3 + 1] = disc.normals[i3 + 2];
            normals[j3 + 2] = disc.normals[i3 + 1];
            
            uvs[j2 + 0] = disc.uvs[i2 + 0];
            uvs[j2 + 1] = disc.uvs[i2 + 1];

        }

        for (let i = 0; i < halfSphereIndicesCount; i++) {
            indices[i] = halfSphere.indices[i];
        }
        for (let i = 0; i < discIndicesCount; i++) {
            indices[i + halfSphereIndicesCount] = disc.indices[i] + halfSphereVertexCount;
        }

        let vertexData = new BABYLON.VertexData();
        vertexData.indices      = indices;
        vertexData.positions    = positions;
        vertexData.normals      = normals;
        vertexData.uvs          = uvs;
        vertexData.apply();

        let result = new BABYLON.Mesh(
            name, scene,
            // [{ key: "InsColor4", code: "A_COLOR4 = InsColor4;", vtype: BABYLON.Constants.INSTANCE_ATTR_VEC4 }],
            // true, true
            [],
            false, false
        );
        let geocylinder = new BABYLON.Geometry("Vector", scene);
        vertexData.applyToGeometry(geocylinder);
        geocylinder.applyToMesh(result);

        return result;
    }
}