import { BABYLON } from "../render3d/babylon";

export class LinesBuilder {
    public static _CreateLineSystem = function(options: { lines: BABYLON.Vector3[][], colors?: BABYLON.Nullable<BABYLON.Color4[][]> }): BABYLON.VertexData {
        var indices = [];
        var positions = [];
        var lines = options.lines;
        var colors = options.colors;
        var vertexColors = [];
        var idx = 0;
    
        for (var l = 0; l < lines.length; l++) {
            var points = lines[l];
            for (var index = 0; index < points.length; index++) {
                positions.push(points[index].x, points[index].y, points[index].z);
                if (colors) {
                    var color = colors[l];
                    vertexColors.push(color[index].r, color[index].g, color[index].b, color[index].a);
                }
                if (index > 0) {
                    indices.push(idx - 1);
                    indices.push(idx);
                }
                idx++;
            }
        }
        var vertexData = new BABYLON.VertexData();
        vertexData.indices = indices;
        vertexData.positions = positions;
        if (colors) {
            vertexData.colors = vertexColors;
        }
        return vertexData;
    }
    public static CreateDashedLines = function(options: { points: BABYLON.Vector3[], dashSize?: number, gapSize?: number, dashNb?: number }): BABYLON.VertexData {
        var dashSize = options.dashSize || 3;
        var gapSize = options.gapSize || 1;
        var dashNb = options.dashNb || 200;
        var points = options.points;
    
        var positions = new Array<number>();
        var indices = new Array<number>();
    
        var curvect = BABYLON.Vector3.Zero();
        var lg = 0;
        var nb = 0;
        var shft = 0;
        var dashshft = 0;
        var curshft = 0;
        var idx = 0;
        var i = 0;
        for (i = 0; i < points.length - 1; i++) {
            points[i + 1].subtractToRef(points[i], curvect);
            lg += curvect.length();
        }
        shft = lg / dashNb;
        dashshft = dashSize * shft / (dashSize + gapSize);
        for (i = 0; i < points.length - 1; i++) {
            points[i + 1].subtractToRef(points[i], curvect);
            nb = Math.floor(curvect.length() / shft);
            curvect.normalize();
            for (var j = 0; j < nb; j++) {
                curshft = shft * j;
                positions.push(points[i].x + curshft * curvect.x, points[i].y + curshft * curvect.y, points[i].z + curshft * curvect.z);
                positions.push(points[i].x + (curshft + dashshft) * curvect.x, points[i].y + (curshft + dashshft) * curvect.y, points[i].z + (curshft + dashshft) * curvect.z);
                indices.push(idx, idx + 1);
                idx += 2;
            }
        }
    
        // Result
        var vertexData = new BABYLON.VertexData();
        vertexData.positions = positions;
        vertexData.indices = indices;
    
        return vertexData;
    }
    // /**
    //  * Creates a line system mesh. A line system is a pool of many lines gathered in a single mesh
    //  * * A line system mesh is considered as a parametric shape since it has no predefined original shape. Its shape is determined by the passed array of lines as an input parameter
    //  * * Like every other parametric shape, it is dynamically updatable by passing an existing instance of LineSystem to this static function
    //  * * The parameter `lines` is an array of lines, each line being an array of successive Vector3
    //  * * The optional parameter `instance` is an instance of an existing LineSystem object to be updated with the passed `lines` parameter
    //  * * The optional parameter `colors` is an array of line colors, each line colors being an array of successive Color4, one per line point
    //  * * The optional parameter `useVertexAlpha` is to be set to `false` (default `true`) when you don't need the alpha blending (faster)
    //  * * Updating a simple Line mesh, you just need to update every line in the `lines` array : https://doc.babylonjs.com/how_to/how_to_dynamically_morph_a_mesh#lines-and-dashedlines
    //  * * When updating an instance, remember that only line point positions can change, not the number of points, neither the number of lines
    //  * * The mesh can be set to updatable with the boolean parameter `updatable` (default false) if its internal geometry is supposed to change once created
    //  * @see https://doc.babylonjs.com/how_to/parametric_shapes#line-system
    //  * @param name defines the name of the new line system
    //  * @param options defines the options used to create the line system
    //  * @param scene defines the hosting scene
    //  * @returns a new line system mesh
    //  */
    // public static CreateLineSystem(name: string, options: { lines: BABYLON.Vector3[][], updatable?: boolean, instance?: BABYLON.Nullable<LinesMesh>, colors?: BABYLON.Nullable<BABYLON.Color4[][]>, useVertexAlpha?: boolean }, scene: BABYLON.Nullable<BABYLON.Scene>): LinesMesh {
    //     var instance = options.instance;
    //     var lines = options.lines;
    //     var colors = options.colors;

    //     if (instance) { // lines update
    //         var positions = instance.getVerticesData(VertexBuffer.PositionKind)!;
    //         var vertexColor;
    //         var lineColors;
    //         if (colors) {
    //             vertexColor = instance.getVerticesData(VertexBuffer.ColorKind)!;
    //         }
    //         var i = 0;
    //         var c = 0;
    //         for (var l = 0; l < lines.length; l++) {
    //             var points = lines[l];
    //             for (var p = 0; p < points.length; p++) {
    //                 positions[i] = points[p].x;
    //                 positions[i + 1] = points[p].y;
    //                 positions[i + 2] = points[p].z;
    //                 if (colors && vertexColor) {
    //                     lineColors = colors[l];
    //                     vertexColor[c] = lineColors[p].r;
    //                     vertexColor[c + 1] = lineColors[p].g;
    //                     vertexColor[c + 2] = lineColors[p].b;
    //                     vertexColor[c + 3] = lineColors[p].a;
    //                     c += 4;
    //                 }
    //                 i += 3;
    //             }
    //         }
    //         instance.updateVerticesData(VertexBuffer.PositionKind, positions, false, false);
    //         if (colors && vertexColor) {
    //             instance.updateVerticesData(VertexBuffer.ColorKind, vertexColor, false, false);
    //         }
    //         return instance;
    //     }

    //     // line system creation
    //     var useVertexColor = (colors) ? true : false;
    //     var lineSystem = new LinesMesh(name, scene, null, undefined, undefined, useVertexColor, options.useVertexAlpha);
    //     var vertexData = VertexData.CreateLineSystem(options);
    //     vertexData.applyToMesh(lineSystem, options.updatable);
    //     return lineSystem;
    // }
}