import { BABYLON } from "../render3d/babylon";

export class SphereBuilder {
    public static _CreateSphere = function(options: { segments?: number, diameter?: number, diameterX?: number, diameterY?: number, diameterZ?: number, arc?: number, slice?: number, sideOrientation?: number, frontUVs?: BABYLON.Vector4, backUVs?: BABYLON.Vector4, dedupTopBottomIndices?: boolean }): BABYLON.VertexData {
        var segments: number = options.segments || 32;
        var diameterX: number = options.diameterX || options.diameter || 1;
        var diameterY: number = options.diameterY || options.diameter || 1;
        var diameterZ: number = options.diameterZ || options.diameter || 1;
        var arc: number = options.arc && (options.arc <= 0 || options.arc > 1) ? 1.0 : options.arc || 1.0;
        var slice: number = options.slice && (options.slice <= 0) ? 1.0 : options.slice || 1.0;
        var sideOrientation = (options.sideOrientation === 0) ? 0 : options.sideOrientation || BABYLON.VertexData.DEFAULTSIDE;
        var dedupTopBottomIndices = !!options.dedupTopBottomIndices;
    
        var radius = new BABYLON.Vector3(diameterX / 2, diameterY / 2, diameterZ / 2);
    
        var totalZRotationSteps = 2 + segments;
        var totalYRotationSteps = 2 * totalZRotationSteps;
    
        var indices = [];
        var positions = [];
        var normals = [];
        var uvs = [];
    
        for (var zRotationStep = 0; zRotationStep <= totalZRotationSteps; zRotationStep++) {
            var normalizedZ = zRotationStep / totalZRotationSteps;
            var angleZ = normalizedZ * Math.PI * slice;
    
            for (var yRotationStep = 0; yRotationStep <= totalYRotationSteps; yRotationStep++) {
                var normalizedY = yRotationStep / totalYRotationSteps;
    
                var angleY = normalizedY * Math.PI * 2 * arc;
    
                var rotationZ = BABYLON.Matrix.RotationZ(-angleZ);
                var rotationY = BABYLON.Matrix.RotationY(angleY);
                var afterRotZ = BABYLON.Vector3.TransformCoordinates(BABYLON.Vector3.Up(), rotationZ);
                var complete = BABYLON.Vector3.TransformCoordinates(afterRotZ, rotationY);
    
                var vertex = complete.multiply(radius);
                var normal = complete.divide(radius).normalize();
    
                positions.push(vertex.x, vertex.y, vertex.z);
                normals.push(normal.x, normal.y, normal.z);
                uvs.push(normalizedY, normalizedZ);
            }
    
            if (zRotationStep > 0) {
                var verticesCount = positions.length / 3;
                for (var firstIndex = verticesCount - 2 * (totalYRotationSteps + 1); (firstIndex + totalYRotationSteps + 2) < verticesCount; firstIndex++) {
                    if (dedupTopBottomIndices) {
                        if (zRotationStep > 1) {
                            indices.push((firstIndex));
                            indices.push((firstIndex + 1));
                            indices.push(firstIndex + totalYRotationSteps + 1);
                        }
                        if (zRotationStep < totalZRotationSteps || slice < 1.0) {
                            indices.push((firstIndex + totalYRotationSteps + 1));
                            indices.push((firstIndex + 1));
                            indices.push((firstIndex + totalYRotationSteps + 2));
                        }
                    }
                    else {
                        indices.push(firstIndex);
                        indices.push(firstIndex + 1);
                        indices.push(firstIndex + totalYRotationSteps + 1);
    
                        indices.push(firstIndex + totalYRotationSteps + 1);
                        indices.push(firstIndex + 1);
                        indices.push(firstIndex + totalYRotationSteps + 2);
                    }
                }
            }
        }
    
        // Sides
        BABYLON.VertexData._ComputeSides(sideOrientation, positions, indices, normals, uvs, options.frontUVs, options.backUVs);
    
        // Result
        var vertexData = new BABYLON.VertexData();
    
        vertexData.indices = indices;
        vertexData.positions = positions;
        vertexData.normals = normals;
        vertexData.uvs = uvs;
    
        return vertexData;
    }
    
    /**
     * Creates a sphere mesh
     * * The parameter `diameter` sets the diameter size (float) of the sphere (default 1)
     * * You can set some different sphere dimensions, for instance to build an ellipsoid, by using the parameters `diameterX`, `diameterY` and `diameterZ` (all by default have the same value of `diameter`)
     * * The parameter `segments` sets the sphere number of horizontal stripes (positive integer, default 32)
     * * You can create an unclosed sphere with the parameter `arc` (positive float, default 1), valued between 0 and 1, what is the ratio of the circumference (latitude) : 2 x PI x ratio
     * * You can create an unclosed sphere on its height with the parameter `slice` (positive float, default1), valued between 0 and 1, what is the height ratio (longitude)
     * * You can also set the mesh side orientation with the values : BABYLON.Mesh.FRONTSIDE (default), BABYLON.Mesh.BACKSIDE or BABYLON.Mesh.DOUBLESIDE
     * * If you create a double-sided mesh, you can choose what parts of the texture image to crop and stick respectively on the front and the back sides with the parameters `frontUVs` and `backUVs` (Vector4). Detail here : https://doc.babylonjs.com/babylon101/discover_basic_elements#side-orientation
     * * The mesh can be set to updatable with the boolean parameter `updatable` (default false) if its internal geometry is supposed to change once created
     * @param name defines the name of the mesh
     * @param options defines the options used to create the mesh
     * @param scene defines the hosting scene
     * @returns the sphere mesh
     * @see https://doc.babylonjs.com/how_to/set_shapes#sphere
     */
    public static CreateSphere(name: string, options: { segments?: number, diameter?: number, diameterX?: number, diameterY?: number, diameterZ?: number, arc?: number, slice?: number, sideOrientation?: number, frontUVs?: BABYLON.Vector4, backUVs?: BABYLON.Vector4, updatable?: boolean }, scene: BABYLON.Nullable<BABYLON.Scene> = null): BABYLON.Mesh {
        var sphere = new BABYLON.Mesh(name, scene);

        // options.sideOrientation = BABYLON.Mesh._GetDefaultSideOrientation(options.sideOrientation);
        // sphere._originalBuilderSideOrientation = options.sideOrientation;

        var vertexData = SphereBuilder._CreateSphere(options);

        let geocylinder = new BABYLON.Geometry(name, scene);
        vertexData.applyToGeometry(geocylinder);
        geocylinder.applyToMesh(sphere);

        return sphere;
    }
}