import { BABYLON } from "./render3d/babylon";
import { BABYLONAction, CustomScene, ITempFrameBuffer, RenderTargetFactory, Copy, ExtensionTools, getContentWH, RenderTarget, SceneRenderFlags, EFilter, EWrap, GlContext } from "./index";

export interface IFullScreenScene {
    name: string;
    zIndex: number;
    clearColor: [number, number, number, number] | null;
    cleardepth: number | null;
}

export interface ICustomSizeScene {
    name: string;
    clearColor: [number, number, number, number] | null;
    cleardepth: number | null;
    framebuffer: ITempFrameBuffer;
}
/**
 * 全屏渲染的场景的管理器
 * * 项目可能有多层场景
 * * 多层场景间需要层级控制
 * * 场景可能在UI之下,场景可能在UI之上
 * * UI之下下的场景可能会在渲染前清空屏幕, UI之上的场景一般不会清空屏幕(即会保留UI内容)
 * * 层级设计
 *   * 使用zIndex配置层级
 *   * 按从小到大顺序进行从下到上渲染
 *   * zIndex=0 预留给UI渲染
 *   * zIndex<0 为UI之下场景渲染
 *   * zIndex>0 为UI之上场景渲染
 * * 场景清屏配置
 *   * clearColor 进行赋值即标识需要清屏(UI之下的一般都需要设置,默认为[0,0,0,0],值为 0~1)
 *   * clearDepth 进行赋值即标识需要清屏深度数据(一般都需要设置,默认设置为1)
 * * 更新场景配置
 *   * 调用 FullScreenSceneManager.FullScreenScene 创建/更新配置 场景
 * @example
 * ```ts
 *  const [curScene2D, effScene, uiEffectScene] = FullScreenSceneManager.Initial([
 *      { zIndex: -3, name: SceneControl.mainScene2DName, clearColor: [0, 0, 0, 0], cleardepth: 1, },
 *      { zIndex: -2, name: SceneControl.effectSceneName, clearColor: [0, 0, 0, 0], cleardepth: 1, },
 *      { zIndex:  2, name: SceneControl.uiEffectSceneName, clearColor: null, cleardepth: 1, },
 *  ]);
 *  
 *  const mainScene = FullScreenSceneManager.FullScreenScene({ zIndex: -4, name: SceneControl.mainSceneName, clearColor: null, cleardepth: 1, }, BackGroundScene.scene);
 * ```
 */
export class FullScreenSceneManager {
    private static FullScreenSceneMap: Map<string, CustomScene> = new Map();
    private static FullScreenSceneQueue: Map<string, [number, boolean]> = new Map();
    private static Copy: Copy;
    private static CopyUnder0: Copy;
    private static useDepth: boolean;
    public static Initial(scenes?: IFullScreenScene[], useDepth: boolean = false): CustomScene[] | undefined {
        ExtensionTools.setPermanent(FullScreenSceneManager.Render);
        FullScreenSceneManager.Copy = new Copy();
        FullScreenSceneManager.CopyUnder0 = new Copy(undefined, true, BABYLON.Constants.ALPHA_INVERT_COMBINE);
        FullScreenSceneManager.useDepth = useDepth;
        if (scenes) {
            let result: CustomScene[] = [];
            scenes.forEach(function(param) {
                let scene = FullScreenSceneManager.FullScreenScene(param);
                result.push(scene);
            });
            return result;
        }
    }
    public static FullScreenScene(param: IFullScreenScene, scene?: CustomScene): CustomScene {
        if (scene == undefined) {
            scene = FullScreenSceneManager.FullScreenSceneMap.get(param.name);
            if (scene == undefined) {
                scene = new CustomScene(param.name);
                FullScreenSceneManager.FullScreenSceneMap.set(param.name, scene);
            }
        } else {
            let old = FullScreenSceneManager.FullScreenSceneMap.get(param.name);
            if (old !== undefined && old != scene) {
                console.error(`Scene ${param.name} Create Again!!!`);
                return;
            }
            FullScreenSceneManager.FullScreenSceneMap.set(param.name, scene);
        }
        if (param.clearColor != null) {
            scene.renderpipeline.opaqueRenderer.autoClearColor(true);
            scene.renderpipeline.opaqueRenderer.clearColor(...param.clearColor);
        } else {
            scene.renderpipeline.opaqueRenderer.autoClearColor(false);
        }
        if (param.cleardepth != null) {
            scene.renderpipeline.opaqueRenderer.autoClearDepth(true);
            scene.renderpipeline.opaqueRenderer.clearDepth(param.cleardepth);
        } else {
            scene.renderpipeline.opaqueRenderer.autoClearDepth(false);
        }
        FullScreenSceneManager.FullScreenSceneQueue.set(param.name, [param.zIndex, param.clearColor != null]);

        return scene;
    }
    public static Get(name: string): CustomScene | undefined {
        return FullScreenSceneManager.FullScreenSceneMap.get(name);
    }
    public static Remove(name: string) {
        let scene = FullScreenSceneManager.FullScreenSceneMap.get(name);
        if (scene) {
            scene.dispose();
        }
        FullScreenSceneManager.FullScreenSceneMap.delete(name);
        FullScreenSceneManager.FullScreenSceneQueue.delete(name);
    }
    public static RenderStart(resultTarget: RenderTarget) {
        // let _context = RenderTargetFactory.getContext();
        // let gl = _context.gl;
        // gl.bindFramebuffer(gl.FRAMEBUFFER, resultTarget.frameBuffer);
        // gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT | gl.STENCIL_BUFFER_BIT);
    }
    public static RenderForZIndex0(resultTarget: RenderTarget) {

    }
    public static Render() {
        let vdocument = window["vdocument"];
        let info = getContentWH();
        let resultTarget = RenderTargetFactory.createRenderTarget(info.renderWidth, info.renderHeight, FullScreenSceneManager.useDepth);

        let customrenderer = null;
        {
            CustomRenderTargetScene.List.forEach((item) => {
                item.renderTarget.frameBuffer._renderer = customrenderer;
                item.render(undefined);
                customrenderer = item.renderTarget.frameBuffer._renderer;
            });
        }

        resultTarget.frameBuffer._renderer = customrenderer;
        FullScreenSceneManager.RenderStart(resultTarget);

        let queue: [number, boolean, CustomScene][] = [];
        FullScreenSceneManager.FullScreenSceneQueue.forEach((item, key) => {
            queue.push([item[0], item[1], FullScreenSceneManager.FullScreenSceneMap.get(key)]);
        });
        queue.sort((a, b) => a[0] - b[0]);

        let startIdx = Number.MIN_SAFE_INTEGER;

        let currIndex: number = Number.MIN_SAFE_INTEGER;
        let len = queue.length;
        for (let idx = len - 1; idx >= 0; idx--) {
            let val = queue[idx];
            if (val[0] < 0 && val[1] && val[2].renderFlag == SceneRenderFlags.active && val[2].impl.rootNodes.length > 1) {
                startIdx = idx;
                break;
            }
        }
        for (let idx = 0; idx < len; idx++) {
            if (idx < startIdx) continue;

            let [zIndex, _autoclear, scene] = queue[idx];
            if (scene && scene.impl.rootNodes.length > 1) {
                if (currIndex < 0 && 0 < zIndex) {
                    FullScreenSceneManager.RenderForZIndex0(resultTarget);
                    let screen = RenderTargetFactory.createScreenRenderTarget(info.contextWidth, info.contextHeight, true);
                    let rootframe: ITempFrameBuffer = {
                        rect: [info.renderX, info.renderY, info.renderWidth, info.renderHeight],
                        ctx: screen.ctx,
                        frameBuffer: screen.frameBuffer
                    };
                    FullScreenSceneManager.CopyUnder0.mat.tilloffset(
                        info.renderWidth / info.contextWidth, info.renderHeight / info.contextHeight,
                        info.renderX / info.contextWidth, info.renderY / info.contextHeight, 
                    );
                    FullScreenSceneManager.CopyUnder0.apply({ etype: "Copy", intensity: 1, alpha: 1 }, [resultTarget], rootframe, RenderTargetFactory.getContext());
                    screen.frameBuffer._renderer = resultTarget.frameBuffer._renderer;
                    RenderTargetFactory.unuseRenderTarget(resultTarget);
                    resultTarget = screen;
                }
                currIndex = zIndex;

                let rootframe: ITempFrameBuffer = {
                    rect: [0, 0, info.renderWidth, info.renderHeight],
                    ctx: resultTarget.ctx,
                    frameBuffer: resultTarget?.frameBuffer
                };
                BABYLONAction.updateEngineDimensions(info.renderWidth, info.renderHeight);
                // scene.impl.clientRect.left      = 0;
                // scene.impl.clientRect.top       = 0;
                // scene.impl.clientRect.width     = vdocument.width;
                // scene.impl.clientRect.height    = vdocument.height;
                scene.resize(info.renderWidth, info.renderHeight);
                scene.render(undefined, rootframe);
            }
        }

        if (currIndex < 0) {
            FullScreenSceneManager.RenderForZIndex0(resultTarget);
            let screen = RenderTargetFactory.createScreenRenderTarget(info.contextWidth, info.contextHeight, true);
            let rootframe: ITempFrameBuffer = {
                rect: [info.renderX, info.renderY, info.renderWidth, info.renderHeight],
                ctx: screen.ctx,
                frameBuffer: screen.frameBuffer
            };

            FullScreenSceneManager.CopyUnder0.mat.tilloffset(
                info.renderWidth / info.contextWidth, info.renderHeight / info.contextHeight,
                info.renderX / info.contextWidth, info.renderY / info.contextHeight, 
            );
            FullScreenSceneManager.CopyUnder0.apply({ etype: "Copy", intensity: 1, alpha: 1 }, [resultTarget], rootframe, RenderTargetFactory.getContext());
            screen.frameBuffer._renderer = resultTarget.frameBuffer._renderer;
            RenderTargetFactory.unuseRenderTarget(resultTarget);
            resultTarget = screen;
        }

        // FullScreenSceneManager.RenderToScreen(resultTarget);
        // RenderTargetFactory.unuseRenderTarget(resultTarget);
    }
    public static RenderToScreen(resultTarget: RenderTarget) {
        // RenderTargetFactory.unuseRenderTarget(resultTarget);
        // let _context = RenderTargetFactory.getContext();
        // let info =  getContentWH();
        // let target: ITempFrameBuffer = {
        //     rect: [info.renderX, info.renderY, info.renderWidth, info.renderHeight],
        //     ctx: undefined,
        //     frameBuffer: undefined
        // };

        // _context.gl.enable(_context.gl.BLEND);
        // _context.gl.blendFunc(_context.gl.ONE, _context.gl.ONE_MINUS_SRC_ALPHA);
        // _context.gl.viewport(...target.rect);
        // _context.gl.scissor(...target.rect);
        // _context.gl.disable(_context.gl.DEPTH_TEST);
        // _context.gl.disable(_context.gl.CULL_FACE);
        // FullScreenSceneManager.Copy.apply({ etype: "Copy", intensity: 1, alpha: 1 }, [resultTarget], <any>{ x: info.renderX, y: info.renderY, width: target.rect[2], height: target.rect[3], framebuffer: target.frameBuffer }, RenderTargetFactory.getContext());
    }
}

export class CustomRenderTargetScene extends CustomScene {
    public static List: CustomRenderTargetScene[] = [];
    public renderTarget: RenderTarget;
    constructor(
        name: string,
        width: number, height: number,
        shadowsEnabled: boolean = false,
        blockMaterialDirtyMechanism: boolean = true,
        needDepth: boolean = false,
        magFilter: EFilter = EFilter.LINEAR, minFilter: EFilter = EFilter.LINEAR, wrapS: EWrap = EWrap.CLAMP_TO_EDGE, wrapT: EWrap = EWrap.CLAMP_TO_EDGE, mipmap: boolean = true
    ) {
        super(name, width, height, shadowsEnabled, blockMaterialDirtyMechanism);

        this.renderTarget = RenderTargetFactory.createRenderTarget(width, height, needDepth, magFilter, minFilter, wrapS, wrapT, mipmap);

        CustomRenderTargetScene.List.push(this);
    }
    getTexture(): BABYLON.Texture {
        let texture = new BABYLON.Texture(null, this.impl, true, false, BABYLON.Texture.LINEAR_LINEAR);
        texture._texture = this.renderTarget;
        return texture;
    }
    render(ctx: GlContext, target?: ITempFrameBuffer, postprocess?: BABYLON.IPostprocessParam[], blend?: BABYLON.Constants.ALPHA_MODE, forceResetContext?: boolean): boolean {
        super.render(undefined, {
            rect: [0, 0, this.renderTarget.width, this.renderTarget.height],
            frameBuffer: this.renderTarget.frameBuffer,
            ctx: undefined,
        });
        return true;
    }
    dispose(): void {
        let idx = CustomRenderTargetScene.List.indexOf(this);
        if (idx >= 0) {
            CustomRenderTargetScene.List.splice(idx, 1);
        }
        super.dispose();
    }
}