import { ResTab } from "../interface/index";
import { Scene } from "../render/scene";
import { UriLoader } from "./uri";
import * as GLTF2 from "./glTF2Interface";
import { IGLTFNew } from "./interface";
import * as BABYLON from "../render/index";
import { AccessorLoader } from "./accessor";
import { ETempNodeAnimationDataKey, ITempNodeAnimationAbout, LoaderRecord } from "./base";
import { LoadExtraMaterial } from "./Extras/material";
import { IParticleSystemConfig } from "./partilce_system";
import { MeshParticleSystem } from "../mesh_particle/mesh_particle_system";
import { LoadExtraAnimation } from "./Extras/animation";
import { ExtensionTools } from "../interface/index";

/**
 * [自定义名称, 要修改的gltf.images 列表序号, 目标图片相对模型文件的路径, 图片变纹理时是否需要反转Y轴]
 */
export type IImageSolt = [string, number, string, boolean?];
/**
 * [自定义名称, 要修改的gltf.materials 列表序号, [对应extension.PI_material 的属性名,属性值] ]
 */
export type IMaterialSolt = [string, number, [string, any][]];

export interface IMaterialReplace {
    /**
     * 替换的材质的 index
     * 如果没设置则认为为所有材质做该替换
     */
    index?: number;
    /**
     * 替换后材质
     */
    extensions: {
        PI_material?: any;
        name: string;
    };
}

export class Loader {
    _scene: Scene;
    resTab: ResTab;

    gltf: IGLTFNew;
    rootUrl: string;
    rootDir: string;
    displayName: string;
    public imageSolts: string[] | IImageSolt[];
    public materialSolts: string[] | IMaterialSolt[];
    public materialReplace: IMaterialReplace[];

    includeNodes: number[];
    includeNodeNames: string[];

    private loaders: LoaderRecord;

    private animationFPS: number;

    public static MeshParticle(loaders: LoaderRecord, cfg: IParticleSystemConfig, geo: BABYLON.Geometry, mesh: BABYLON.Mesh, index: number): BABYLON.AbstractMesh {
        let mp = new MeshParticleSystem(cfg.name, loaders.scene);
        
        // formatMeshParticle(cfg, mp);
        if (loaders.gltfres) {
            mp.formatByGLTF(loaders.gltfres, index);
        }

        // mp.setSourceMesh(mesh);

        // geo.setInstanceBuffer(BABYLON.EInstanceKind.TSR);
        // geo.setInstanceBuffer(BABYLON.EInstanceKind.Color);
        // geo.setInstanceBuffer(BABYLON.EInstanceKind.Tilloff);

        return mp;
    }

    public static GLTFFormat(gltfUrl: string, gltf: IGLTFNew) : IGLTFNew {
        return gltf;
    }

    constructor(scene: BABYLON.Scene, rootPath: string, rootUrl: string, displayName: string, includeNodes?: number[]) {
        this._scene = scene;
        this.resTab = ExtensionTools.CreateResTab();
        this.animationFPS = scene.animationFPS;
        this.includeNodes = includeNodes;
        this.displayName = displayName;
        this.rootDir = rootPath;
        this.rootUrl = rootPath + rootUrl;
        this.loaders = new LoaderRecord();
        this.loaders.scene = this._scene;
        this.loaders.rootUrl = rootPath + rootUrl;
        this.loaders.rootDir = rootPath;
        this.loaders.resTab = this.resTab;
        this.loaders.animationFPS = this.animationFPS;
    }

    dispose() {
        this.loaders.dispose();
        this.resTab.clear();
        this.gltf = undefined;
        this._scene = undefined;
    }

    public loadAsync(js: boolean = false) {
        return UriLoader._loadGLTF(this.rootUrl, this.resTab).then((data) => {
            if (this._scene.isDisposed) {
                this.dispose();
                return;
            }
            // const str = PISYS.UTF8.utf8Decode(data);
            // let json = JSON.parse(str);
            this.gltf = Loader.GLTFFormat(this.rootUrl, data);

            if (this.imageSolts) {
                _gltfImageSolt(this.gltf, this.imageSolts);
            }
            if (this.materialSolts) {
                _gltfMaterialSolts(this.gltf, this.materialSolts);
            }
            if (this.materialReplace) {
                _gltfMaterialReplace(this.gltf, this.materialReplace);
            }

            this.loaders.gltf = this.gltf;

            if (js) {
                return this._analy();
            } else {
                return UriLoader.CreateGLTF(this._scene, this.displayName, this.rootUrl, "").then((res) => {
                    // console.error("New GLTF");
                    this.loaders.gltfres = res.gltf;
                    this.loaders.rootNode = res.root;
                    if (this._scene.isDisposed) {
                        this.dispose();
                        throw new Error("GLTFLoad: Scene Disposed!");
                    } else {
                        return this._analy().then((result) => {
                            result.gltf = this.loaders.gltfres;
                            this.loaders.gltfres = undefined;
                            this.loaders.rootNode = undefined;
                            this.loaders.skeletons = undefined;
                            this.dispose();
                            return result;
                        });
                    }
                })
                .catch((e) => {
                    this.dispose();
                    throw new Error(e);
                });
            }
        });
    }

    private _analy() {
        if (this.gltf.asset.extras?.VersionNumber != "20230626") {
            return Promise.reject("GLTF Version Error: Need 20230626 But Give " + this.gltf.asset.extras?.VersionNumber);
        }

        let accessorVerticeNeedLoad: number[] = [];
        let accessorIndicesNeedLoad: number[] = [];
        let materialIndexsNeedLoad: number[] = [];

        let nodes: BABYLON.Node[] = [];
        let matrixs: BABYLON.Matrix[] = [];
        let meshes: BABYLON.AbstractMesh[] = [];
        let meshParticles: MeshParticleSystem[] = [];
        let animationGroups: BABYLON.AnimationGroupRS[] = [];
        let materials: BABYLON.Material[] = [];
        let skeletons: BABYLON.Skeleton[] = this.loaders.skeletons;
        
        //#region Nodes
        let _root_ = this.loaders.rootNode;
        if (!_root_) {
            _root_ = new BABYLON.TransformNode(this.displayName, this._scene);
        }
        _root_.setEnabled(false);

        let skins = this.gltf.skins || [];
        let isbones = [];
        skins.forEach((skin) => {
            skin.joints.forEach((val) => {
                isbones[val] = true;
            });
        });
        let skinRef: Map<number, number> = new Map();

        this.gltf.nodes?.forEach((nodeInfo, nodeIndex) => {
            let flag = true;
            if (this.includeNodes) {
                if (!this.includeNodes.includes(nodeIndex)) {
                    flag = false;
                }
            }
            if (flag && this.includeNodeNames) {
                if (!this.includeNodeNames.includes(nodeInfo.name)) {
                    flag = false;
                }
            }
            if (flag == false) {
                return;
            }

            let animationAbout: ITempNodeAnimationAbout = {
                [ETempNodeAnimationDataKey.Node]: null,
                Mesh: [],
                Material: [],
                MainTex: [],
                EmissiveTex: [],
                OpacityTex: [],
                MaskTex: [],
            };
            this.loaders.nodeAnimationAbout.set(nodeIndex, animationAbout);

            let tempMeshes: BABYLON.Mesh[] = [];
            if (nodeInfo.mesh != undefined) {

                let meshParticleCfg = nodeInfo.extras?.meshParticle;
                let meshInfo = this.gltf.meshes[nodeInfo.mesh];
                if (meshInfo) {
                    meshInfo.primitives.forEach((item, index) => {

                        let meta = undefined;
                        let mesh: BABYLON.Mesh; //  = new BABYLON.Mesh(nodeInfo.name, this._scene);
                        if (meshParticleCfg) {
                            let mp = new MeshParticleSystem(nodeInfo.name, this._scene);
                            mp.formatByGLTF(this.loaders.gltfres, nodeIndex);
                            // Loader.MeshParticle(this.loaders, meshParticleCfg, geometry, mesh, nodeIndex);
                            
                            // geometry.setInstanceBuffer(BABYLON.EInstanceKind.TSR);
                            // geometry.setInstanceBuffer(BABYLON.EInstanceKind.Color);
                            // geometry.setInstanceBuffer(BABYLON.EInstanceKind.Tilloff);

                            if (meshParticleCfg.renderMode != 4) {
                                meta = BABYLON.QuadBuilder.meta();
                            }

                            meshParticles.push(<any>mp);
                            mesh = mp;
                        } else {
                            mesh = new BABYLON.Mesh(nodeInfo.name, this._scene);
                        }

                        //#region Geometry
                        let geometry = new BABYLON.Geometry(nodeInfo.name, this._scene);
                        if (meta == undefined) {
                            for (let attr in item.attributes) {
                                let indexAccessor = item.attributes[attr];
                                if (!accessorVerticeNeedLoad.includes(indexAccessor)) {
                                    accessorVerticeNeedLoad.push(indexAccessor);
                                }
                                let accessor = this.gltf.accessors[indexAccessor];
                                geometry.setVerticesBuffer({
                                    name: `${this.rootUrl}#${indexAccessor}`,
                                    range: [0, accessor.count * BABYLON.VertexBuffer.GetTypeByteLength(accessor.componentType) * AccessorLoader._GetNumComponents(accessor.type)],
                                    attrs: [
                                        { kind: BABYLON.Geometry.vertex_kind(attr), format: BABYLON.Geometry.vertex_format(accessor) }
                                    ]
                                });
                            }
                            if (item.targets) {
                                let morphcount = item.targets.length;
                                for (let i = 0; i < morphcount; i++) {
                                    let target = item.targets[i];
                                    for (let attr in target) {
                                        let indexAccessor = target[attr];
                                        if (!accessorVerticeNeedLoad.includes(indexAccessor)) {
                                            accessorVerticeNeedLoad.push(indexAccessor);
                                        }
                                        let accessor = this.gltf.accessors[indexAccessor];
                                        geometry.setVerticesBuffer({
                                            name: `${this.rootUrl}#${indexAccessor}`,
                                            range: [0, accessor.count * BABYLON.VertexBuffer.GetTypeByteLength(accessor.componentType) * AccessorLoader._GetNumComponents(accessor.type)],
                                            attrs: [
                                                { kind: BABYLON.Geometry.morph_vertex_kind(attr, i), format: BABYLON.Geometry.vertex_format(accessor) }
                                            ]
                                        });
                                    }
                                }
                            }
                            if (item.indices != undefined) {
                                let indexAccessor = item.indices;
                                if (!accessorIndicesNeedLoad.includes(indexAccessor)) {
                                    accessorIndicesNeedLoad.push(indexAccessor);
                                }
                                let accessor = this.gltf.accessors[indexAccessor];
                                geometry.setIndices({
                                    name: `${this.rootUrl}#${indexAccessor}`,
                                    range: [0, accessor.count * BABYLON.VertexBuffer.GetTypeByteLength(accessor.componentType) * AccessorLoader._GetNumComponents(accessor.type)],
                                    as_u16: true
                                });
                            }
                        }
                        //#endregion

                        if (animationAbout[ETempNodeAnimationDataKey.Node] == undefined) {
                            animationAbout[ETempNodeAnimationDataKey.Node] = mesh;
                        }

                        animationAbout.Mesh.push(mesh);
                        tempMeshes.push(mesh);
                        meshes.push(mesh);
                        //#region Geometry
                        {
                            geometry.applyToMesh(mesh, meta);
                        }
                        //#endregion
                        //#region Material
                        {
                            if (item.material != undefined) {
                                let indexMat = item.material;
                                let data = this.loaders.materials.get(indexMat);
                                if (!data) {
                                    LoadExtraMaterial.loadAsync(this.loaders, indexMat, nodeIndex, (mat, state) => { 
                                        materials[indexMat] = mat;
                                        this.loaders.materials.set(indexMat, { material: mat, state: state });
                                    });
                                    data = this.loaders.materials.get(indexMat);
                                }
                                
                                animationAbout[ETempNodeAnimationDataKey.Material].push(data.material);
                                let pass = this._scene.queryPassTag(data.state.renderQueue)
                                mesh.setMaterial(data.material);
                                mesh.depthTest = data.state.depthTest ? BABYLON.Constants.LEQUAL : BABYLON.Constants.ALWAYS;
                                mesh.depthWrite = data.state.depthWrite;
                                mesh.alphaMode = data.state.alphaMode;
                                mesh.cullMode = data.state.cullMode;
                                mesh.alphaIndex = data.state.renderQueue;
                            }
                        }
                        //#endregion

                        if (item.extras?.bakeframes) {
                            let animegroup = LoadExtraAnimation.IndiceRenderRange(this.loaders, nodeIndex, index, item.extras?.bakeframes, mesh, _root_);
                            animationGroups.push(animegroup);
                        }
                        if (item.extras?.bakeframesNew) {
                            let animegroup = LoadExtraAnimation.IndiceRenderRangeNew(this.loaders, nodeIndex, index, item.extras?.bakeframesNew, mesh, _root_);
                            animationGroups.push(animegroup);
                        }
                    });
                    if (meshInfo.weights) {
                        let weightLen = meshInfo.weights?.length;
                        for (let i = 0; i < weightLen; i++) {
                            animationAbout.Mesh.forEach((mesh) => {
                                if (mesh.morphTargetInfluenceWeight) {
                                    mesh.morphTargetInfluenceWeight(i, meshInfo.weights[i]);
                                }
                            });
                        }
                    }
                    if (nodeInfo.skin != undefined) {
                        skinRef.set(nodeIndex, nodeInfo.skin);
                    }

                    let node = tempMeshes[0];
                    node.parent = _root_;
                    nodes[nodeIndex] = node;

                    let len = tempMeshes.length;
                    for (let i = 1; i < len; i++) {
                        tempMeshes[i].parent = node;
                    }
                } else {
                    throw new Error(`Mesh Index Error !`);
                }
                
            } else {
                let node = new BABYLON.TransformNode(nodeInfo.name, this._scene);
                node.parent = _root_;
                nodes[nodeIndex] = node;
            }

            if (animationAbout[ETempNodeAnimationDataKey.Node] == undefined) {
                animationAbout[ETempNodeAnimationDataKey.Node] = nodes[nodeIndex];
            }
            
            if (animationAbout[ETempNodeAnimationDataKey.Node]) {
                let position = BABYLON.Vector3.Zero();
                let scaling = BABYLON.Vector3.One();
                let rotation = BABYLON.Quaternion.Identity();
                let matrix = BABYLON.Matrix.Identity();
                let node: BABYLON.ITransformNode = <any>animationAbout[ETempNodeAnimationDataKey.Node];
                if (nodeInfo.translation) {
                    BABYLON.Vector3.FromArrayToRef(nodeInfo.translation, 0, position);
                    node.position = position.clone();
                }
                if (nodeInfo.scale) {
                    BABYLON.Vector3.FromArrayToRef(nodeInfo.scale, 0, scaling);
                    node.scaling = scaling.clone();
                }
                if (nodeInfo.rotation) {
                    BABYLON.Quaternion.FromArrayToRef(nodeInfo.rotation, 0, rotation);
                    node.rotationQuaternion = rotation.clone();
                }
                if (nodeInfo.matrix) {
                    BABYLON.Matrix.FromArrayToRef(nodeInfo.matrix, 0, matrix);
                    let position = new BABYLON.Vector3();
                    let scaling = new BABYLON.Vector3();
                    let quaternion = new BABYLON.Quaternion();
                    if (matrix.decompose(scaling, quaternion, position)) {
                        node.position = position;
                        node.scaling = scaling;
                        node.rotation = quaternion.toEulerAngles();
                    }
                } else {
                    BABYLON.Matrix.ComposeToRef(scaling, rotation, position, matrix);
                }
                matrixs[nodeIndex] = matrix;
            }

            if (nodeInfo.camera != undefined) {

            }
        });

        
        this.gltf.nodes?.forEach((nodeInfo, index) => {
            if (nodeInfo.children) {
                nodeInfo.children.forEach((indexChild) => {
                    if (nodes[indexChild] && nodes[index]) {
                        nodes[indexChild].parent = nodes[index];
                    }
                });
            }
        });

        skinRef.forEach((skinidx, nodeindex) => {
            let skin = skins[skinidx];
            let meshNode = nodes[nodeindex];

            let bones: BABYLON.Bone[] = [];
            let totalbones: BABYLON.Bone[] = [];
            skin.joints.forEach((nodeidx, idx) => {
                let node = nodes[nodeidx];
                let boneParent: BABYLON.Bone;
                let bone = <BABYLON.Bone>totalbones[nodeidx];
                if (bone == undefined) {
                    totalbones[nodeidx] = new BABYLON.Bone(node.name, node._scene);
                    totalbones[nodeidx]._linkedTransformNode = node;
                    bone = totalbones[nodeidx];
                }
                bone.setBindPose(matrixs[nodeidx]);
                bones.push(bone);
                while (node.parent != _root_ && node.parent != meshNode.parent && node.parent != meshNode) {
                    node = <BABYLON.Node>node.parent;
                    let id = nodes.indexOf(node);
                    boneParent = totalbones[id];
                    if (id >= 0 && boneParent == undefined) {
                        totalbones[id] = new BABYLON.Bone(node.name, node._scene);
                        totalbones[id]._linkedTransformNode = node;
                        totalbones[id].setBindPose(matrixs[id]);
                        boneParent = totalbones[id];
                    }
                    bone.parent = boneParent;
                    bone = boneParent;
                }
            });
            let rootBone;
            totalbones.forEach((bone) => {
                if (rootBone == undefined && bone) {
                    rootBone = bone;
                }
            });
            let skeleton = new BABYLON.Skeleton(skin.name || `${skinidx}`, rootBone, bones);
            skeleton._totalBones = totalbones;
            skeletons[skinidx] = skeleton;
        });

        // this.gltf.nodes?.forEach((nodeInfo, index) => {
        //     if (nodeInfo.children) {
        //         nodeInfo.children.forEach((indexChild) => {
        //             if (totalbones[indexChild] && totalbones[index]) {
        //                 totalbones[indexChild].parent = totalbones[index];
        //             }
        //         });
        //     }
        // });

        skinRef.forEach((idxskin, nodeindex) => {
            let temp = this.loaders.nodeAnimationAbout.get(nodeindex);
            let skin = skeletons[idxskin];
            if (temp && skin) {
                temp[ETempNodeAnimationDataKey.Mesh]?.forEach((mesh) => {
                    skin.applyToMesh(mesh);
                });
            }
        });
        //#endregion
        
        skins.forEach((skin, idx) => {
            if (skin.inverseBindMatrices != undefined) {
                AccessorLoader.loadAsyncFloat32(this.loaders, skin.inverseBindMatrices, this.gltf.accessors[skin.inverseBindMatrices]);
            }
        });

        if (!this.loaders.gltfres) {
            this.gltf.accessors.forEach((accessor, idx) => {
                switch (accessor.componentType) {
                    case (GLTF2.AccessorComponentType.UNSIGNED_SHORT):  {
                        AccessorLoader.loadAsyncUint16(this.loaders, idx, accessor);
                        break;
                    }
                    case (GLTF2.AccessorComponentType.UNSIGNED_INT): {
                        AccessorLoader.loadAsyncUint32(this.loaders, idx, accessor);
                        break;
                    }
                    case (GLTF2.AccessorComponentType.FLOAT): {
                        AccessorLoader.loadAsyncFloat32(this.loaders, idx, accessor);
                        break;
                    }
                    default: {
                        throw new Error(``);
                    }
                }
            });
        }

        if (!this.loaders.gltfres) {
            return this.loaders.accessors().then(() => {
                let matrix = BABYLON.Matrix.Identity();
                skins.forEach((skin, skinidx) => {
                    console.log(">>>>>>>>>>>>>>");
                    let data: Float32Array;
                    if (skin.inverseBindMatrices != undefined) {
                        data = this.loaders.f32_accessors.get(skin.inverseBindMatrices);
                    }
                    skeletons[skinidx]._totalBones.forEach((bone, idx) => {
                        if (bone) {
                            let jointIdx = skin.joints.indexOf(idx);
                            if (jointIdx >= 0) {
                                if (data) {
                                    BABYLON.Matrix.FromArrayToRef(data, jointIdx * 16, matrix);
                                    matrix.invertToRef(matrix);
                                    // matrix.transposeToRef(matrix);
                                } else {
                                    BABYLON.Matrix.IdentityToRef(matrix);
                                }
                            } else {
                                BABYLON.Matrix.IdentityToRef(matrix);
                            }
                            if (bone.parent) {
                                matrix.multiplyToRef((<BABYLON.Bone>bone.parent)._invertedAbsoluteTransform, matrix);
                            }
                            bone.setBindPose(matrix);
                        }
                    });
                });

                //#region Accessor BufferView Buffer
                accessorVerticeNeedLoad.forEach((indexAccessor) => {
                    let accessor = this.gltf.accessors[indexAccessor];
                    if (accessor) {
                        let key = `${this.rootUrl}#${indexAccessor}`;
                        switch (accessor.componentType) {
                            case (GLTF2.AccessorComponentType.UNSIGNED_SHORT):  {
                                break;
                            }
                            case (GLTF2.AccessorComponentType.UNSIGNED_INT): {
                                break;
                            }
                            case (GLTF2.AccessorComponentType.FLOAT): {
                                let data = this.loaders.f32_accessors.get(indexAccessor);
                                if (data) {
                                    BABYLON.Geometry.create_vertex_buffer(key, data);
                                }
                                break;
                            }
                            default: {
                                throw new Error(``);
                            }
                        }
                    }
                });
                accessorIndicesNeedLoad.forEach((indexAccessor) => {
                    let accessor = this.gltf.accessors[indexAccessor];
                    if (accessor) {
                        let key = `${this.rootUrl}#${indexAccessor}`;
                        switch (accessor.componentType) {
                            case (GLTF2.AccessorComponentType.UNSIGNED_SHORT):  {
                                let data = this.loaders.u16_accessors.get(indexAccessor);
                                if (data) {
                                    BABYLON.Geometry.create_indices_buffer(key, data);
                                }
                                break;
                            }
                            case (GLTF2.AccessorComponentType.UNSIGNED_INT): {
                                break;
                            }
                            case (GLTF2.AccessorComponentType.FLOAT): {
                                break;
                            }
                            default: {
                                throw new Error(``);
                            }
                        }
                    }
                });
                //#region 
    
                this._analyAnimations(_root_, animationGroups);
    
                return this.loaders.promise().then(() => {
                    return <IGLTFLoadResult>{ _root_, nodes, meshes, animationGroups, particles: meshParticles, skeletons };
                });
            });
        } else {
            this._analyAnimations(_root_, animationGroups);

            return Promise.all(
                [
                    this.loaders.promise(),
                    this.loaders.textureLoaders()
                ]
            ).then(() => {
                let matrix = BABYLON.Matrix.Identity();
                skins.forEach((skin, skinidx) => {
                    console.log(">>>>>>>>>>>>>>");
                    let data: Float32Array;
                    if (skin.inverseBindMatrices != undefined) {
                        data = this.loaders.f32_accessors.get(skin.inverseBindMatrices);
                    }
                    skeletons[skinidx]._totalBones.forEach((bone, idx) => {
                        if (bone) {
                            let jointIdx = skin.joints.indexOf(idx);
                            if (jointIdx >= 0) {
                                if (data) {
                                    BABYLON.Matrix.FromArrayToRef(data, jointIdx * 16, matrix);
                                    matrix.invertToRef(matrix);
                                    // matrix.transposeToRef(matrix);
                                } else {
                                    BABYLON.Matrix.IdentityToRef(matrix);
                                }
                            } else {
                                BABYLON.Matrix.IdentityToRef(matrix);
                            }
                            if (bone.parent) {
                                matrix.multiplyToRef((<BABYLON.Bone>bone.parent)._invertedAbsoluteTransform, matrix);
                            }
                            bone.setBindPose(matrix);
                        }
                    });
                });
                return <IGLTFLoadResult>{ _root_, nodes, meshes, animationGroups, particles: meshParticles, materials, skeletons };
            });
        }
    }

    private _analyAnimations(root: BABYLON.TransformNode, animationGroups: BABYLON.AnimationGroupRS[]) {
        if (this.gltf.animations) {
            this.gltf.animations.forEach((item, idxGroup) => {
                let animationGroup = new BABYLON.AnimationGroupRS(item.name, this._scene, this.animationFPS);

                let samplers = item.samplers;
                item.channels.forEach((channel, idx) => {
                    let sampler = samplers[channel.sampler];
                    LoadExtraAnimation.load(this.loaders, idxGroup, idx, channel, sampler, animationGroup);
                });
                // 

                animationGroups.push(animationGroup);
            });
        }
    }

    protected _analySkin() {
        
    }
}

export interface IGLTFLoadResult {
    _root_: BABYLON.TransformNode;
    nodes: BABYLON.TransformNode[];
    meshes: BABYLON.AbstractMesh[];
    particles: MeshParticleSystem[];
    animationGroups: BABYLON.AnimationGroupRS[];
    materials: BABYLON.Material[];
    skeletons: BABYLON.Skeleton[];
    gltf?: BABYLON.IGLTFRes;
}

// PI_BEGIN - 20200309 扩展可变纹理数据
function _gltfImageSolt(data: IGLTFNew, imageSolts: string[] | IImageSolt[]) {
    let solts = data.extras?.imageSolts;
    const images = data.images;
    const textures = data.textures;
    if (data && images) {
        const imagesLen = images.length;
        if (typeof imageSolts[0] === 'string') {
            if (solts) {
                for (let i = 0, len = imageSolts.length; i < len; i++) {
                    const solt = solts.find((s) => {
                                        return s[0] === imageSolts[i];
                                    });
                    if (solt && solt[1] <= imagesLen && solt[1] >= 0) {
                        images[solt[1]].uri = solt[2];
                        const tex = textures.find((tex) => {
                            return tex.source === solt[1];
                        });
                        tex && ((<any>tex).invertY = !!solt[3]);
                    }
                }
            }
        } else {
            solts = <IImageSolt[]>imageSolts;
            for (let i = 0, len = solts.length; i < len; i++) {
                const solt = solts[i];
                if (solt && solt[1] <= imagesLen && solt[1] >= 0) {
                    images[solt[1]].uri = solt[2];
                    const tex = textures.find((tex) => {
                        return tex.source === solt[1];
                    });
                    tex && ((<any>tex).invertY = !!solt[3]);
                }
            }
        }
    }
}
// PI_END

// PI_BEGIN - 20200310 - 拓展gltf materialSolts
function _gltfMaterialSolts(data: IGLTFNew, materialSolts: string[] | IMaterialSolt[]) {
    let solts = data.extras?.materialSolts;
    const materials = data.materials;
    if (data && materials) {
        const materialLen = materials.length;

        if (typeof materialSolts[0] === 'string') {
            if (solts) {
                let material;
                for (let i = 0, len = materialSolts.length; i < len; i++) {
                    material = undefined;
                    const solt = solts.find((s) => {
                                        return s[0] === materialSolts[i];
                                    });

                    if (solt && solt[1] <= materialLen && solt[1] >= 0) {
                        material = (<any>materials[solt[1]]).extensions.PI_material;
                        if (material) {
                            const attrList = solt[2];
                            if (attrList) {
                                this._gltfMaterialSolt(attrList, material);
                            }
                        }
                    }
                }
            }
        } else {
            solts = <IMaterialSolt[]>materialSolts;
            let material = null;
            for (let i = 0, len = solts.length; i < len; i++) {
                const solt = solts[i];
                material = null;
                if (solt && solt[1] <= materialLen && solt[1] >= 0 && materials[solt[1]]) {
                    material = materials[solt[1]].extensions.PI_material;
                }
                if (material) {
                    const attrList = solt[2];
                    if (attrList) {
                        this._gltfMaterialSolt(attrList, material);
                    }
                }
            }
        }
    }
}

function _gltfMaterialReplace(data: IGLTFNew, materialReplaces: IMaterialReplace[]) {
    let replacedIndexs: number[] = [];

    let gltf = data;

    let count = gltf.materials.length;
    let replaceCount = materialReplaces.length;

    let noIndexReplace = null;

    for (let i = 0; i < replaceCount; i++) {
        let replace = materialReplaces[i];
        let replaceIndex = replace.index;
        if (replaceIndex === undefined) {
            noIndexReplace = replace;
        } else {
            gltf.materials[replaceIndex].extras[replace.extensions.name] = replace.extensions.PI_material;
            replacedIndexs.push(replaceIndex);
        }
    }

    if (noIndexReplace != null) {
        for (let i = 0; i < count; i++) {
            if (replacedIndexs.indexOf(i) < 0) {
                gltf.materials[i].extras[noIndexReplace.extensions.name] = noIndexReplace.extensions.PI_material;
            }
        }
    }
}