import { SceneManagerData } from "./scene";
import * as BABYLON from "./render3d/context";
import { SceneStruct } from "./scene_struct";
import { Constants, IBakedKeyframes, IPrefabPool, resortPostprocessParams, IPostprocessParam, KeyPostprocess, ITempFrameBuffer } from "./interface/index";
import { PrefabModelObjPool } from "./model_prefab";
import { MaterialLoaderOpacityClipMaterial, OpacityClipMaterial } from "./material/opacity_clip";
import { IPostProcess } from "./postprocess/base";
import { IModelOpt, SceneRenderFlags } from "./interface/index";
import { Copy } from "./postprocess/copy";
import { ExtensionTools, TLoadCall } from "./interface/index";

import { Logger } from 'pi_sys/util/logger';
import { PISYS } from "pi_sys/index";
import { TaskMgr } from 'pi_ucore/util/task_mgr';
import { register, ResTab } from 'pi_res/res_mgr';
import { getLogger } from "pi_tools/logger";
import { FramCall } from 'pi_wasm/dom_init';
import { getAtom } from "pi_wasm/atom";
import * as P3D from "pi_wasm/export_interface";

import { ModelObj } from "./model_obj";
import { EFilter, RenderTargetFactory } from "./postprocess/render_target";
import { TimeoutPool } from "./util/timeout_pool";
import { IAtom } from "./context";
import * as API from "./context";
import { IColliderItem } from "interface/extends";
import { PiCollider } from "scene_pick";
import { StandardMaterialData } from "./material/standardmaterial";
import { MathDataPool } from "./render3d/context";

export { CustomGeometry } from "./customgeometry/custom_geometry";
export { Sprite } from "./sprite/sprite";
export { SpriteManager } from "./sprite/sprite_manager";

export * from "./model_prefab";
export * from "./model_obj";
export * from "./scene_tool";
export * from "./scene_struct";
export * from "./interface/index";
export * from "./postprocess/copy";
export * from "./postprocess/render_target";
export * from "./postprocess/webgl";

export { AnimationRS as Animation } from "./render/animations/animation";
export { MainOpacityMaterial,  } from "./render3d/context";
export { OpacityClipMaterial } from "./material/opacity_clip";
export { AnimationGroupRS as AnimationGroup } from "./render/animations/animationGroup";
export { FrameCurve, FrameCurveAlpha, FrameCurveEuler, FrameCurvePosition, FrameCurveRotationQuaternion, FrameCurveScaling } from "./render/animations/curves/index";

// export function init() {
//     BABYLON.Engine.Initial();
//     BABYLON.Engine.registBeforeRun(BABYLON.GLTF2.UriLoader.CheckGLTFLoad);
//     BABYLON.Engine.registBeforeRun(BABYLON.BaseTexture.Query);
//     SceneManagerData.engine = BABYLON.Engine.getEngine();
// }

export function getContentWH() {
    let adaptiveResult = ExtensionTools.AdaptiveResult();
    let contextWidth = adaptiveResult.renderSize.width;
    let contextHeight = adaptiveResult.renderSize.height;

    let renderX = Math.round(adaptiveResult.styleLeft / adaptiveResult.styleSize.width * adaptiveResult.renderSize.width);
    let renderY = Math.floor(adaptiveResult.styleBottom / adaptiveResult.styleSize.height * adaptiveResult.renderSize.height);
    let renderWidth = Math.max(Math.floor((1.0 - (adaptiveResult.styleLeft + adaptiveResult.styleRight) / adaptiveResult.styleSize.width) * adaptiveResult.renderSize.width), 0);
    let renderHeight = Math.max(Math.floor((1.0 - (adaptiveResult.styleTop + adaptiveResult.styleBottom) / adaptiveResult.styleSize.height) * adaptiveResult.renderSize.height), 0);

    return {
        contextWidth,
        contextHeight,
        renderX,
        renderY,
        renderWidth,
        renderHeight,
        layoutSize: adaptiveResult.layoutSize,
        styleTop: adaptiveResult.styleTop,
        styleBottom: adaptiveResult.styleBottom,
        styleLeft: adaptiveResult.styleLeft,
        styleSize: adaptiveResult.styleSize,
        styleRight: adaptiveResult.styleRight
    }
}

declare var window: any;

export enum TCullMode {
    Off,
    Front,
    Back,
}
export enum TFrontFace {
    Ccw,
    Cw,
}
export enum TAlphaMode {
    ALPHA_COMBINE,
}

export interface GlContext {
    canvas: {
        setBrush(value: number): void;
        width: number;
        height: number;
    }
}

export interface IRefScene {
    camera: BABYLON.Camera;
    scene: SceneStruct;
    prefabPool: IPrefabPool;
    render(uicanvas: any): boolean;
    resize(): void;
    renderWidth: number;
    renderHeight: number;
}

export const ALPHA_DISABLE =  0;
export const ALPHA_ADD =  1;
export const ALPHA_COMBINE =  2;
export const ALPHA_SUBTRACT =  3;
export const ALPHA_MULTIPLY =  4;
export const ALPHA_MAXIMIZED =  5;
export const ALPHA_ONEONE =  6;
export const ALPHA_PREMULTIPLIED =  7;
export const ALPHA_PREMULTIPLIED_PORTERDUFF =  8;
export const ALPHA_INTERPOLATE =  9;
export const ALPHA_SCREENMODE = 10;
export const PI_ALPHA_PREMULTIPLIED = 11;
export const ALPHA_ONEONE_ONEONE = 11;
export const ALPHA_ALPHATOCOLOR = 12;
export const ALPHA_REVERSEONEMINUS = 13;
export const ALPHA_SRC_DSTONEMINUSSRCALPHA = 14;
export const ALPHA_ONEONE_ONEZERO = 15;
export const ALPHA_EXCLUSION = 16;
export const PI_ALPHA_SRCCOLOR_ONE = 20;
export const PI_ALPHA_ONE_ONEMINUSSRCALPHA = 21;
export const PI_ALPHA_SRCCOLOR_ONEMINUSSRCALPHA = 22;
export const PI_ALPHA_SRCCOLOR_ONEMINUSSRCCOLOR = 23;

export type ALPHA_MODE = typeof ALPHA_DISABLE
                        | typeof ALPHA_ADD
                        | typeof ALPHA_COMBINE
                        | typeof ALPHA_SUBTRACT
                        | typeof ALPHA_MULTIPLY
                        | typeof ALPHA_MAXIMIZED
                        | typeof ALPHA_ONEONE
                        | typeof ALPHA_PREMULTIPLIED
                        | typeof ALPHA_PREMULTIPLIED_PORTERDUFF
                        | typeof ALPHA_INTERPOLATE
                        | typeof ALPHA_SCREENMODE
                        | typeof PI_ALPHA_PREMULTIPLIED
                        | typeof ALPHA_ONEONE_ONEONE
                        | typeof ALPHA_ALPHATOCOLOR
                        | typeof ALPHA_REVERSEONEMINUS
                        | typeof ALPHA_SRC_DSTONEMINUSSRCALPHA
                        | typeof ALPHA_ONEONE_ONEZERO
                        | typeof ALPHA_EXCLUSION
                        | typeof PI_ALPHA_SRCCOLOR_ONE
                        | typeof PI_ALPHA_ONE_ONEMINUSSRCALPHA
                        | typeof PI_ALPHA_SRCCOLOR_ONEMINUSSRCALPHA
                        | typeof PI_ALPHA_SRCCOLOR_ONEMINUSSRCCOLOR
;

export const  REPEAT = Constants.REPEAT;
export const  CLAMP_TO_EDGE = Constants.CLAMP_TO_EDGE;
export const  MIRRORED_REPEAT = Constants.MIRRORED_REPEAT;
export type AddressMode = typeof REPEAT | typeof CLAMP_TO_EDGE | typeof MIRRORED_REPEAT;

export const MATERIAL_OPAQUE	= 0;
export const MATERIAL_ALPHATEST = 1;	
export const MATERIAL_ALPHABLEND = 2;
export const MATERIAL_ALPHATESTANDBLEND	= 3;
export type TransparentcyMode = typeof MATERIAL_OPAQUE | typeof MATERIAL_ALPHATEST | typeof MATERIAL_ALPHABLEND | typeof MATERIAL_ALPHATESTANDBLEND;

export interface IInitialOption {
    pbr: boolean;
    KHR_lights_punctual: boolean;
    PI_environment: boolean;
    PI_particle: boolean;
    onoffCollectInfo: boolean;
    DefaultWhitePath: string;
    width: number;
    height: number;
    textureTilloffBundle?: boolean;
    canvas: HTMLCanvasElement,
}

interface IUINode {
    offsetLeft: number;
    offsetTop: number;
    offsetWidth: number;
    offsetHeight: number;
}
interface IUIBoundingClientRect {
    left: number;
    top: number;
    width: number;
    height: number;
}
export interface IActionD3 {
    initial(opt: IInitialOption): void;
    engine(): BABYLON.Engine;
    piSizeForWidth(camera: BABYLON.FreeCamera, val: boolean): void;
    piOrthSize(camera: BABYLON.FreeCamera, val: number): void;
    cullMode(mesh: BABYLON.Mesh, val: TCullMode): void;
    frontFace(mesh: BABYLON.Mesh, val: TFrontFace): void;
    depthWrite(mesh: BABYLON.Mesh, val: boolean): void;
    depthTest(mesh: BABYLON.Mesh, val: boolean): void;
    alphaMode(mesh: BABYLON.Mesh, val: ALPHA_MODE): void;
    alpha(mesh: { material: BABYLON.Material, visibility: number }, val: number): void;
    autoClearColor(camera: BABYLON.Camera, val: boolean): void;
    autoClearDepth(camera: BABYLON.Camera, val: boolean): void;
    clearColor(camera: BABYLON.Camera, val: BABYLON.Color4): void;
    clearDepth(camera: BABYLON.Camera, val: number): void;
    activeCamera(camera: BABYLON.Camera, val: boolean): void;
    aspect(camera: BABYLON.Camera, val: number): void;
    render(scene: SceneStruct, val: GlContext): void;
    isDisposed(scene: BABYLON.Scene): void;
    alphaCutOff(mesh: { material: BABYLON.Material}, val: number): void;
    mainColor(material: BABYLON.Material, r: number, g: number, b: number): void;
    useMaterial(mesh: BABYLON.Mesh, material: BABYLON.Material, transparency: TransparentcyMode, alphaMode: ALPHA_MODE): void;
    wrapU(texture: BABYLON.Texture, val: AddressMode): void;
    wrapV(texture: BABYLON.Texture, val: AddressMode): void;
    disableDepthWrite(mesh: BABYLON.Mesh, val: boolean): void;
    useAlphaFromDiffuseTexture(mesh: BABYLON.Mesh, val: boolean): void;
    hasAlpha(texture: BABYLON.Texture, val: boolean): void;
    transparencyMode(mesh: BABYLON.Mesh, val: TransparentcyMode): void;
    updateEngineDimensions(width: number, height: number): void;
    backFaceCulling(mesh: BABYLON.Mesh, val: boolean): void;
    /**
     * 
     * @param mesh 
     * @param texturePropertyName 目标纹理属性名: "diffuseTexture" 、 "opacityTexture" ...
     * @param val 
     */
    uScale(mesh: BABYLON.Mesh, texturePropertyName: string, val: number);
    /**
     * 
     * @param mesh 
     * @param texturePropertyName 目标纹理属性名: "diffuseTexture" 、 "opacityTexture" ...
     * @param val 
     */
    vScale(mesh: BABYLON.Mesh, texturePropertyName: string, val: number);
    /**
     * 
     * @param mesh 
     * @param texturePropertyName 目标纹理属性名: "diffuseTexture" 、 "opacityTexture" ...
     * @param val 
     */
    uOffset(mesh: BABYLON.Mesh, texturePropertyName: string, val: number);
    /**
     * 
     * @param mesh 
     * @param texturePropertyName 目标纹理属性名: "diffuseTexture" 、 "opacityTexture" ...
     * @param val 
     */
    vOffset(mesh: BABYLON.Mesh, texturePropertyName: string, val: number);
    /**
     * UI 节点映射到场景中的位置和缩放
     * @param content 与相机渲染范围对应的容器节点, 与UI完整对齐可使用root节点
     * @param target 要计算的目标节点, 与UI完整对齐可使用root节点
     * @param camera 目标相机
     * @param resultPosition 映射的全局位置
     * @param resultSize 映射的全局缩放
     */
    uiNodeTransformToSceneRef(content: IUINode, target: IUINode, camera: BABYLON.Camera, resultPosition: BABYLON.Vector3, resultSize: BABYLON.Vector3);
    /**
     * 场景坐标映射到目标节点下相对坐标
     * @param content 与相机渲染范围对应的容器节点, 与UI完整对齐可使用root节点
     * @param target 定位的目标父节点, 与UI完整对齐可使用root节点
     * @param camera 目标相机
     * @param x 
     * @param y 
     * @param z 
     * @param result 映射的局部UI坐标
     */
    scenePositionToUILeftTopRef(content: IUINode, target: IUINode, camera: BABYLON.Camera, x: number, y: number, z: number, result: [number, number]);
    /**
     * UI 节点映射到场景中的位置和缩放
     * @param content 与相机渲染范围对应的容器节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param target 要计算的目标节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param camera 目标相机
     * @param resultPosition 映射的全局位置
     * @param resultSize 映射的全局缩放
     */
    uiBoundingClientRectTransformToSceneRef(content: IUIBoundingClientRect, target: IUIBoundingClientRect, camera: BABYLON.Camera, resultPosition: BABYLON.Vector3, resultSize: BABYLON.Vector3);
    /**
     * 场景坐标映射到目标节点下相对坐标
     * @param content 与相机渲染范围对应的容器节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param target 定位的目标父节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param camera 目标相机
     * @param x 
     * @param y 
     * @param z 
     * @param result 映射的局部UI坐标
     */
    scenePositionToUIBoundingClientRectLeftTopRef(content: IUIBoundingClientRect, target: IUIBoundingClientRect, camera: BABYLON.Camera, x: number, y: number, z: number, result: [number, number]);
}

export interface ICustomScene {
    readonly prefabPool: PrefabModelObjPool;
    readonly camera: BABYLON.Camera;
    readonly renderWidth: number;
    readonly renderHeight: number;
    renderQuality: number;
    resize(width: number, height: number): void;
    render(ctx: GlContext, target?: ITempFrameBuffer, postprocess?: IPostprocessParam[]): boolean;
}

export class _BABYLONAction implements IActionD3 {
    
    piOrthSize(camera: BABYLON.FreeCamera, val: number): void {
        camera.piOrthSize = val;
    }
    piSizeForWidth(camera: BABYLON.FreeCamera, val: boolean) {
        camera.piSizeForWidth = val;
    }

    initial(opt: IInitialOption) {
        ExtensionTools.AdaptiveResult = function() {
            return PISYS.Env.get(`adaptiveResult`);
        }
        ExtensionTools.AddTask = function(func: Function, args: any[], priority: number, type: 0 | 1 | 2) {
            TaskMgr.set(func, args, priority, type);
        }
        ExtensionTools.RES_FILE = "RES_FILE";
        ExtensionTools.RES_IMAGE = "RES_IMAGE";
        ExtensionTools.CreateResTab = function () {
            return new ResTab();
        }
        ExtensionTools.Domains = function() {
            return PISYS.Env.get(`domains`);
        }
        ExtensionTools.FileNameWithoutDirectory = function(file: string) {
            let i = file.lastIndexOf("/");
            return (i < 0) ? file : file.slice(i + 1);
        }
        ExtensionTools.GetLogger = function(module: string, tag: string) {
            return getLogger(module, tag);
        }
        ExtensionTools.GetPureResPath = function(url: string) {
            return PISYS.Depend.getPureResPath(url);
        }
        ExtensionTools.QueryRuntimeFullName = function(url: string) {
            return PISYS.Depend.queryRuntimeFullName(url);
        }
        ExtensionTools.Logger = function(module: string, tag: string) {
            return new Logger(module, tag);
        }
        ExtensionTools.QuerySupportCompressedTextureFormat = function() {
            return PISYS.Env.get(`compressed_tex_format`);
        }
        ExtensionTools.ResLoadRegist = function<T, A, D>(type: string, load: TLoadCall<T, A, D>, destroy: (link: any) => void, sizeOf?: (link: any) => number, replace?: boolean, timeout?: number, min_capacity?: number, max_capacity?: number) {
            return register(type, load, destroy, sizeOf, replace, timeout, min_capacity, max_capacity);
        }
        ExtensionTools.UTF8Decode = function(data: ArrayBuffer) {
            return PISYS.UTF8.utf8Decode(data);
        }
        ExtensionTools.setPermanentBefore = function(func: () => void) {
            FramCall.addFramCallBeforeInterval(func);
        }
        ExtensionTools.setPermanent = function(func: () => void) {
            FramCall.addFramCallInterval(func);
        }
        ExtensionTools.clearPermanent = function(func: () => void) {
            FramCall.removeFramCallInterval(func);
        }
        ExtensionTools.getAtom = function(v: number | string | IAtom) {
            return getAtom(v);
        }
        ExtensionTools.LoadFile = function(resTab: BABYLON.ResTab, url: string): Promise<ArrayBuffer | ArrayBufferView> {
            return resTab.load(ExtensionTools.RES_FILE, url, []).then((res) => {
                let data = res.link;
                resTab.delete(res);
                return data;
            });
        }
        ExtensionTools.setTimeout = TimeoutPool.setTimeout;
        ExtensionTools.clearTimeout = TimeoutPool.clearTimeout;
        TimeoutPool.Init();
        API.p3d = P3D;
        ExtensionTools.Ready();

        BABYLON.Engine.Initial(opt.canvas);
        BABYLON.Engine.registBeforeRun(BABYLON.GLTF2.UriLoader.CheckGLTFLoad);
        BABYLON.Engine.registBeforeRun(BABYLON.BaseTexture.Query);
        BABYLON.Engine.GetRenderContent = getContentWH;
        SceneManagerData.engine = BABYLON.Engine.getEngine();

        BABYLON.Engine.ErrorState((v, v1) => {}, true);

        if (!opt.textureTilloffBundle) {
            BABYLON.NodeMaterialBlockManager.Init();
            let mgr = BABYLON.NodeMaterialBlockManager.Single;
            mgr.regist(new BABYLON.UniformMainRenderTarget());
            mgr.regist(new BABYLON.UniformMainTexture());
            mgr.regist(new BABYLON.UniformMaskTexture());
            mgr.regist(new BABYLON.UniformEmissiveTexture());
            mgr.regist(new BABYLON.UniformOpacityTexture());
            mgr.regist(new BABYLON.UniformMainTextureSimple());
        }

        BABYLON.MainOpacityMaterial.regist();
        BABYLON.GLTF2.MaterialLoaderMainOpacity.init();

        // BABYLON.MainOpacityEmissiveFresnelMaterial.regist();
        BABYLON.GLTF2.MaterialLoaderMainOpacityEmissiveFresnel.init();

        // BABYLON.MainOpacityOpacityFresnelMaterial.regist();
        BABYLON.GLTF2.MaterialLoaderMainOpacityOpacityFresnel.init();

        // BABYLON.TwoOpacityMixMaterial.regist();
        BABYLON.GLTF2.MaterialLoaderTwoOpacityMix.init();

        // BABYLON.DistortionUVMaterial.regist();
        BABYLON.GLTF2.MaterialLoaderDistortionUV.init();

        // BABYLON.MainOpacityFresnelMaterial.regist();
        BABYLON.PluginMeshParticleSystem.setup();

        // OpacityClipMaterial.regist();
        MaterialLoaderOpacityClipMaterial.init();
    }
    engine(): BABYLON.Engine {
        return BABYLON.Engine.getEngine();
    }
    disableDepthWrite(mesh: BABYLON.Mesh, val: boolean): void {
        mesh.depthWrite = !val;
    }
    useAlphaFromDiffuseTexture(mesh: BABYLON.Mesh, val: boolean): void {
        // throw new Error("Method not implemented.");
    }
    hasAlpha(texture: BABYLON.Texture, val: boolean): void {
        // throw new Error("Method not implemented.");
    }
    transparencyMode(mesh: BABYLON.Mesh, val: TransparentcyMode): void {
        mesh.transparencyMode = val;
    }
    cullMode(mesh: BABYLON.Mesh, val: TCullMode) {
        switch (val) {
            case(TCullMode.Off): {
                mesh.cullMode = BABYLON.Constants.NONE;
                break;
            }
            case(TCullMode.Front): {
                mesh.cullMode = BABYLON.Constants.FRONT;
                break;
            }
            case(TCullMode.Back): {
                mesh.cullMode = BABYLON.Constants.BACK;
                break;
            }
        }
    }
    frontFace(mesh: BABYLON.Mesh, val: TFrontFace) {
        switch (val) {
            case(TFrontFace.Ccw): {
                mesh.frontFace = BABYLON.Constants.CCW;
                break;
            }
            case(TFrontFace.Cw): {
                mesh.frontFace = BABYLON.Constants.CW;
                break;
            }
        }
    }
    depthWrite(mesh: BABYLON.Mesh, val: boolean) {
        mesh.depthWrite = val;
    }
    depthTest(mesh: BABYLON.Mesh, val: boolean) {
        mesh.depthTest = val ? BABYLON.Constants.LEQUAL : BABYLON.Constants.ALWAYS;
    }
    alphaMode(mesh: BABYLON.Mesh, val: number) {
        mesh.alphaMode = val;
    }
    alpha(mesh: { material: BABYLON.Material, visibility: number }, val: number) {
        if(mesh.material instanceof BABYLON.Material) {
            mesh.material.alpha = val;
        }else {
            // 非自定义材质的透明度变化
            mesh.visibility = val;
        }
    }
    autoClearColor(camera: BABYLON.Camera, val: boolean) {
        // camera.autoClearColor = val;
    }
    autoClearDepth(camera: BABYLON.Camera, val: boolean) {
        // camera.autoClearDepth = val;
    }
    clearColor(camera: BABYLON.Camera, val: BABYLON.Color4) {
        // camera.clearColor = val;
    }
    clearDepth(camera: BABYLON.Camera, val: number) {
        // camera.clearDepth = val;
    }
    activeCamera(camera: BABYLON.Camera, val: boolean) {
        camera.active = val;
    }
    aspect(camera: BABYLON.Camera, val: number) {
        camera.aspect = val;
    }
    render(scene: SceneStruct, val: GlContext) {
        // return scene?.render(val);
    }
    isDisposed(scene: BABYLON.Scene) {
        return scene.isDisposed;
    }
    alphaCutOff(mesh: { material: BABYLON.Material}, val: number) {
        mesh.material.alphaCutOff = val;
    }
    mainColor(material: BABYLON.Material, r: number, g: number, b: number) {
        (<BABYLON.StandardMaterial>material).mainColor(r, g, b);
    }
    useMaterial(mesh: BABYLON.Mesh, material: BABYLON.Material, transparency: TransparentcyMode, alphaMode: ALPHA_MODE) {
        let queue = 0;
        if (transparency == MATERIAL_OPAQUE) {
            queue = 2000;
        } else if (transparency == MATERIAL_ALPHABLEND || transparency == MATERIAL_ALPHATESTANDBLEND ) {
            queue = 3000;
        } else {
            queue = 2450;
        }
        mesh.setMaterial(material);
        mesh.alphaMode = alphaMode;
    }
    wrapU(texture: BABYLON.Texture, val: AddressMode) {
        texture.wrapU = val;
    }
    wrapV(texture: BABYLON.Texture, val: AddressMode) {
        texture.wrapV = val;
    }
    updateEngineDimensions(width: number, height: number) {
    }
    
    backFaceCulling(mesh: BABYLON.Mesh, val: boolean): void {
        
    }
    /**
     * 
     * @param zIndex 值越大 层级越高
     */
    createFrameBuffer(zIndex: number): ITempFrameBuffer {
        return undefined;
    }
    
    uScale(mesh: { material: BABYLON.Material, materialData?: StandardMaterialData }, texturePath: string, val: number) {
        if ((mesh.materialData || mesh.material)[texturePath]) {
            (mesh.materialData || mesh.material)[texturePath].uScale = val;
        }
    }
    vScale(mesh: { material: BABYLON.Material, materialData?: StandardMaterialData }, texturePath: string, val: number) {
        if ((mesh.materialData || mesh.material)[texturePath]) {
            (mesh.materialData || mesh.material)[texturePath].vScale = val;
        }
    }
    uOffset(mesh: { material: BABYLON.Material, materialData?: StandardMaterialData }, texturePath: string, val: number) {
        if ((mesh.materialData || mesh.material)[texturePath]) {
            (mesh.materialData || mesh.material)[texturePath].uOffset = val;
        }
    }
    vOffset(mesh: { material: BABYLON.Material, materialData?: StandardMaterialData }, texturePath: string, val: number) {
        if ((mesh.materialData || mesh.material)[texturePath]) {
            (mesh.materialData || mesh.material)[texturePath].vOffset = val;
        }
    }
    uiNodeTransformToSceneRef(content: IUINode, target: IUINode, camera: BABYLON.Camera, resultPosition: BABYLON.Vector3, resultSize: BABYLON.Vector3) {
        let matrix = MathDataPool.Matrix();
        let source = MathDataPool.Vector3();

        // 渲染范围的矩形信息 - 0~1 - 左上角为原点,Y轴指向下方
        let dx = (target.offsetLeft   - content.offsetLeft  ) / content.offsetWidth ;
        let dy = (target.offsetTop    - content.offsetTop   ) / content.offsetHeight;
        let dw = target.offsetWidth  / content.offsetWidth ;
        let dh = target.offsetHeight / content.offsetHeight;
        // 渲染范围的矩形信息 - -1~1 - 渲染中心为原点,Y轴指向上方
        let rw = dw;
        let rh = dw;
        let rx =      (dx + dw * 0.5)  * 2 - 1;
        let ry = (1 - (dy + dh * 0.5)) * 2 - 1;

        camera.getTransformationMatrix().invertToRef(matrix);
        
        source.copyFromFloats(rx,       ry,         0);
        BABYLON.Vector3._UnprojectFromInvertedMatrixToRef(source, matrix, resultPosition);
        source.copyFromFloats(rx + rw,  ry + rh,    0);
        BABYLON.Vector3._UnprojectFromInvertedMatrixToRef(source, matrix, resultSize);
        resultSize.subtractInPlace(resultPosition);

        MathDataPool.Recycle(matrix);
        MathDataPool.Recycle(source);
    }
    scenePositionToUILeftTopRef(content: IUINode, target: IUINode, camera: BABYLON.Camera, x: number, y: number, z: number, result: [number, number]) {
        let matrix = MathDataPool.Matrix();
        let source = MathDataPool.Vector3();

        matrix.copyFrom(camera.getTransformationMatrix());
        
        // 世界坐标转视口坐标 - -1~1
        BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(x, y, z, matrix, source);
        let ux = (source.x + 1) * 0.5;
        let uy = 1.0 - (source.y + 1) * 0.5;
        let left = ux * content.offsetWidth  - target.offsetLeft;
        let top  = uy * content.offsetHeight - target.offsetTop;

        result[0] = left;
        result[1] = top;

        MathDataPool.Recycle(matrix);
        MathDataPool.Recycle(source);
    }
    /**
     * UI 节点映射到场景中的位置和缩放
     * @param content 与相机渲染范围对应的容器节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param target 要计算的目标节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param camera 目标相机
     * @param resultPosition 映射的全局位置
     * @param resultSize 映射的全局缩放
     */
    uiBoundingClientRectTransformToSceneRef(content: IUIBoundingClientRect, target: IUIBoundingClientRect, camera: BABYLON.Camera, resultPosition: BABYLON.Vector3, resultSize: BABYLON.Vector3) {
        BABYLONAction.uiNodeTransformToSceneRef(
            { offsetLeft: content.left, offsetTop: content.top, offsetWidth: content.width, offsetHeight: content.height },
            { offsetLeft: target.left, offsetTop: target.top, offsetWidth: target.width, offsetHeight: target.height },
            camera,
            resultPosition,
            resultSize
        );
    }
    /**
     * 场景坐标映射到目标节点下相对坐标
     * @param content 与相机渲染范围对应的容器节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param target 定位的目标父节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param camera 目标相机
     * @param x 
     * @param y 
     * @param z 
     * @param result 映射的局部UI坐标
     */
    scenePositionToUIBoundingClientRectLeftTopRef(content: IUIBoundingClientRect, target: IUIBoundingClientRect, camera: BABYLON.Camera, x: number, y: number, z: number, result: [number, number]) {
        BABYLONAction.scenePositionToUILeftTopRef(
            { offsetLeft: content.left, offsetTop: content.top, offsetWidth: content.width, offsetHeight: content.height },
            { offsetLeft: target.left, offsetTop: target.top, offsetWidth: target.width, offsetHeight: target.height },
            camera,
            x, y, z,
            result
        );
    }
}

export const BABYLONAction = new _BABYLONAction;

export class SpineBakeModel extends ModelObj {
    protected _changeAnim() {
        super._changeAnim();
        if (!this.isLoaded || this.isDisposed) { return }
        if (this.animationMap && this.animOpt) {
            this.meshes.forEach((mesh) => {
                let enable = (<any>mesh) == this.rootImpl || mesh.name == this.animOpt.animName;
                mesh.setEnabled(enable);
            });
        }
    }
    public static Creation(name: string, scene: CustomScene, opt: IModelOpt = {}) {
        return new SpineBakeModel(name, scene, opt);
    }
}

export class CustomScene extends SceneStruct implements ICustomScene {
    public readonly prefabPool: PrefabModelObjPool;
    public readonly renderpipeline: BABYLON.ForwardRenderPipeline;
    private _renderQuality: number = 1;
    public readonly postprocesses: Map<KeyPostprocess, IPostProcess> = new Map();
    private _rendertarget: BABYLON.CustomRenderTarget;
    private width: number;
    private height: number;
    private _copy: Copy;
    get renderWidth (): number {
        return this.width;
    }
    get renderHeight(): number {
        return this.height;
    }
    set renderQuality(val: 1.2 | 1.5 | 2.0 | 1.0 | 0.8 | 0.5) {
        // this._renderQuality = val;
        // this.resize();
    }
    resize(width: number, height: number): void {
        // width   = Math.floor(width  * this._renderQuality / 2) * 2;
        // height  = Math.floor(height * this._renderQuality / 2) * 2;

        // if (this._rendertarget && (this._rendertarget.width != width || this._rendertarget.height != height)) {
        //     this._rendertarget.dispose();
        //     this._rendertarget = undefined;
        // }
        // this.width  = width;
        // this.height = height;
        // if (this._rendertarget == undefined) {
        //     this._rendertarget = new BABYLON.CustomRenderTarget(width, height, BABYLON.Constants.RGBA8, BABYLON.Constants.DEPTH32F, BABYLON.Constants.LINEAR, BABYLON.Constants.CLAMP_TO_EDGE, BABYLON.Constants.ANISOTROPY_ONE);
        //     this.renderpipeline.opaqueRenderer.setRenderTarget(this._rendertarget);
        //     this.renderpipeline.transparentRenderer.setRenderTarget(this._rendertarget);
        //     BABYLONAction.aspect(<BABYLON.FreeCamera>this.camera, width / height);
        // }
    }
    render(ctx: GlContext, target?: ITempFrameBuffer, postprocess: IPostprocessParam[] = [], blend: Constants.ALPHA_MODE = Constants.ALPHA_PREMULTIPLIED, forceResetContext: boolean = true): boolean {
        postprocess = resortPostprocessParams(postprocess, this.postprocesses);
        this.impl._active = this.renderFlag === SceneRenderFlags.active;
        if (this.renderFlag === SceneRenderFlags.active) {
            
            this.impl.ctx = undefined;
            this.beforeRenderCallList.forEach((call) => {
                try {
                    call(this);
                } catch (error) {
                    
                }
            });

            if (postprocess.length > 0) {
                
            } else {
                if (target.ctx) {
                    // this.impl.ctx = target.ctx;
                    // let info: BABYLON.TAutoRenderTarget = { width: target.rect[2], height: target.rect[3], colorFormat: BABYLON.Constants.RGBA8, depthStencilFormat: BABYLON.Constants.DEPTH32F, autoTarget: true };

                    // // let width = target.rect[2];
                    // // let height = target.rect[3];
                    // // let renderTarget = RenderTargetFactory.createRenderTarget(width, height, true);
                    // // let renderer = this._copy;
                    // // if (renderer == undefined) {
                    // //     renderer = new Copy();
                    // //     renderer.init(this.impl, this.renderpipeline.PASS_TRANSPARENT, true);
                    // //     this._copy = renderer;
                    // // }
                    // this.renderpipeline.opaqueRenderer.setRenderTarget(info);
                    // this.renderpipeline.transparentRenderer.setRenderTarget(info);
                    // // this.renderpipeline.transparentRenderer.next(renderer.renderer);
                    // // renderer.renderer.autoClearColor(true);
                    // // renderer.renderer.autoClearDepth(true);
                    // // renderer.renderer.clearDepth(1.0);
                    
                    // this.impl.render();

                    // // let info: BABYLON.TAutoRenderTarget = { width: target.rect[2], height: target.rect[3], colorFormat: BABYLON.Constants.RGBA8, depthStencilFormat: BABYLON.Constants.DEPTH32F, autoTarget: true };
                    // // // let info: BABYLON.ICustomRenderTarget = { key: undefined, dispose(){} };
                    // // renderer.apply({ etype: "Copy", alpha: 1, intensity: 1 }, [this._rendertarget], info);

                    // target.ctx.canvas.setBrush(this.renderpipeline.transparentRenderer.entity);

                    // // RenderTargetFactory.unuseRenderTarget(renderTarget);

                    let width = target.rect[2];
                    let height = target.rect[3];
                    let rendertarget: BABYLON.TAutoRenderTarget = { width: target.rect[2], height: target.rect[3], colorFormat: BABYLON.Constants.RGBA8, depthStencilFormat: BABYLON.Constants.DEPTH32F, autoTarget: true }; // RenderTargetFactory.createRenderTarget(width, height, true);
                    // this._rendertarget = new BABYLON.CustomRenderTarget(width, height, BABYLON.Constants.RGBA8, BABYLON.Constants.DEPTH32F, BABYLON.Constants.LINEAR, BABYLON.Constants.CLAMP_TO_EDGE, BABYLON.Constants.ANISOTROPY_ONE);
                    this.renderpipeline.opaqueRenderer.setRenderTarget(rendertarget);
                    this.renderpipeline.transparentRenderer.setRenderTarget(rendertarget);
                    BABYLONAction.aspect(<BABYLON.FreeCamera>this.camera, width / height);

                    // console.error(`Render setBrush ${this.renderpipeline.transparentRenderer.uniqueId}`);
                    target.ctx.canvas.setBrush(this.renderpipeline.transparentRenderer.entity);

                    // this.renderpipeline.opaqueRenderer.setRenderTarget(this._rendertarget);
                    // this.renderpipeline.transparentRenderer.setRenderTarget(this._rendertarget);
                    this.impl.render();

                    // {
                    //     let renderer = this._copy;
                    //     if (renderer == undefined) {
                    //         renderer = new Copy();
                    //         renderer.init(this.impl, this.renderpipeline.PASS_TRANSPARENT, true);
                    //         this._copy = renderer;
                    //     }
                    //     this.renderpipeline.transparentRenderer.next(renderer.renderer);
                    //     renderer.renderer.autoClearColor(true);
                    //     renderer.renderer.autoClearDepth(true);
                    //     renderer.renderer.clearDepth(1.0);

                    //     if (target) {
                    //         if (target.ctx) {
                    //             let info: BABYLON.TAutoRenderTarget = { width: target.rect[2], height: target.rect[3], colorFormat: BABYLON.Constants.RGBA8, depthStencilFormat: BABYLON.Constants.DEPTH32F, autoTarget: true };
                    //             // let info: BABYLON.ICustomRenderTarget = { key: undefined, dispose(){} };
                    //             renderer.apply({ etype: "Copy", alpha: 1, intensity: 1 }, [this._rendertarget], info);
                    //             target.ctx.canvas.setBrush(renderer.renderer.entity);
                    //         } else {
                    //             let info = getContentWH();
                    //             renderer.renderer.viewport(info.renderX / info.contextWidth, info.renderY / info.contextHeight, info.renderWidth / info.contextWidth, info.renderHeight / info.contextHeight);
                    //             renderer.apply({ etype: "Copy", alpha: 1, intensity: 1 }, [this._rendertarget], null);
                    //         }
                    //     }
                    // }
                    // RenderTargetFactory.unuseRenderTarget(this._rendertarget);
                    this._rendertarget = undefined;
                } else {
                    this.renderpipeline.opaqueRenderer.setRenderTarget(target);
                    this.renderpipeline.transparentRenderer.setRenderTarget(target);

                    let framebuffer: BABYLON.PIWebGLFramebuffer = <BABYLON.PIWebGLFramebuffer>target.frameBuffer;

                    if (framebuffer?._renderer) {
                        (<BABYLON.Renderer>framebuffer._renderer).next(<BABYLON.Renderer>this.renderpipeline.first);
                    }
                    let width = target.rect[2];
                    let height = target.rect[3];
                    BABYLONAction.aspect(<BABYLON.FreeCamera>this.camera, width / height);
                    this.impl.render();
                    framebuffer._renderer = <BABYLON.Renderer>this.renderpipeline.final;
                }
            }

            this.afterRenderCallList.forEach((call) => {
                try {
                    call(this);
                } catch (error) {
                    
                }
            });

            // {
            //     let renderer = this._copy;
            //     if (renderer == undefined) {
            //         renderer = new Copy();
            //         renderer.init(this.impl, this.renderpipeline.PASS_TRANSPARENT, true);
            //         this._copy = renderer;
            //     }
            //     this.renderpipeline.transparentRenderer.next(renderer.renderer);
            //     renderer.renderer.autoClearColor(true);
            //     renderer.renderer.autoClearDepth(true);
            //     renderer.renderer.clearDepth(1.0);

            //     if (target) {
            //         if (target.ctx) {
            //             let info: BABYLON.TAutoRenderTarget = { width: target.rect[2], height: target.rect[3], colorFormat: BABYLON.Constants.RGBA8, depthStencilFormat: BABYLON.Constants.DEPTH32F, autoTarget: true };
            //             // let info: BABYLON.ICustomRenderTarget = { key: undefined, dispose(){} };
            //             renderer.apply({ etype: "Copy", alpha: 1, intensity: 1 }, [this._rendertarget], info);
            //             target.ctx.canvas.setBrush(renderer.renderer.entity);
            //         } else {
            //             let info = getContentWH();
            //             renderer.renderer.viewport(info.renderX / info.contextWidth, info.renderY / info.contextHeight, info.renderWidth / info.contextWidth, info.renderHeight / info.contextHeight);
            //             renderer.apply({ etype: "Copy", alpha: 1, intensity: 1 }, [this._rendertarget], null);
            //         }
            //     }
            // }

            return true;
        } else {
            return false;
        }
    }
    constructor(
        name: string, 
        width?: number, height?: number,
        shadowsEnabled: boolean = false,
        blockMaterialDirtyMechanism: boolean = true,
    ) {
        let stype = name;
        super(stype, BABYLONAction.engine())
        if (!this.impl.metadata) {
            this.impl.metadata = {};
        }
        // 设置动画的旋转方向
        this.impl.metadata.gltfAsLeftHandedSystem = true;
        // this.impl.autoClear = false;
        // this.impl.autoClearDepthAndStencil = false;

        this.camera = new BABYLON.FreeCamera(`${stype}_camera`, new BABYLON.Vector3(0, 0, -50), this.impl);
        // 设置相机模式——正交相机
        // camera.mode = BABYLON.FreeCamera.ORTHOGRAPHIC_CAMERA;
        (<BABYLON.FreeCamera>this.camera).mode = BABYLON.Constants.ORTHOGRAPHIC_CAMERA;
        BABYLONAction.piOrthSize(<BABYLON.FreeCamera>this.camera, 20);
        this.addCamera(this.camera);
        this.setCurrCamera(this.camera.name);

        this.prefabPool = new PrefabModelObjPool(this);

        this.renderpipeline = new BABYLON.ForwardRenderPipeline(this.impl, this.camera, undefined, false, true, false, false, true);
        this.renderpipeline.transparentRenderer.autoClearColor(false);
        this.renderpipeline.transparentRenderer.autoClearDepth(false);
        this.renderpipeline.transparentRenderer.autoClearStencil(false);
        this.renderpipeline.transparentRenderer.clearColor(0, 0, 0, 0);
        this.renderpipeline.transparentRenderer.clearDepth(1);
        this.renderpipeline.opaqueRenderer.autoClearColor(true);
        this.renderpipeline.opaqueRenderer.autoClearDepth(true);
        this.renderpipeline.opaqueRenderer.clearColor(0, 0, 0, 0);
        this.renderpipeline.opaqueRenderer.clearDepth(1);

        this.impl.renderPipeline = this.renderpipeline;

        if (width && height) {
            this.resize(width, height);
        }
    }
    dispose() {
        this.renderpipeline.dispose();
        this.prefabPool.clearPrefabs();
        super.dispose();
    }
    
    public pick(e: PointerEvent, items?: IColliderItem[]): BABYLON.Node | undefined {
        let info = getContentWH();
        let pointerX = (e.clientX - info.styleLeft) / (info.styleSize.width - info.styleLeft - info.styleRight);
        let pointerY = (e.clientY - info.styleTop)  / (info.styleSize.height - info.styleBottom - info.styleTop);
        pointerX =  pointerX * 2 - 1;
        pointerY = (1 - pointerY) * 2 - 1;
        let pickresult = this.impl.pickWithColliderInfo(pointerX, pointerY, this.camera);
        return <BABYLON.Node>pickresult.pickedMesh;
    }
    public onPointerDown = (e: PointerEvent) => {
        let info = getContentWH();
        let pointerX = (e.clientX - info.styleLeft) / (info.styleSize.width - info.styleLeft - info.styleRight);
        let pointerY = (e.clientY - info.styleTop)  / (info.styleSize.height - info.styleBottom - info.styleTop);
        pointerX =  pointerX * 2 - 1;
        pointerY = (1 - pointerY) * 2 - 1;
        let pickresult = this.impl.pickWithColliderInfo(pointerX, pointerY, this.camera);
        pickresult.pickedMesh?.onPointerObservable.notifyObservers(new BABYLON.PointerInfo(BABYLON.PointerEventTypes.POINTERDOWN, <any>e, pickresult));
    }
    public onPointerUp = (e: PointerEvent) => {
        let info = getContentWH();
        let pointerX = (e.clientX - info.styleLeft) / (info.styleSize.width - info.styleLeft - info.styleRight);
        let pointerY = (e.clientY - info.styleTop)  / (info.styleSize.height - info.styleBottom - info.styleTop);
        pointerX =  pointerX * 2 - 1;
        pointerY = (1 - pointerY) * 2 - 1;
        let pickresult = this.impl.pickWithColliderInfo(pointerX, pointerY, this.camera);
        pickresult.pickedMesh?.onPointerObservable.notifyObservers(new BABYLON.PointerInfo(BABYLON.PointerEventTypes.POINTERUP, <any>e, pickresult));
    }
    public clientPositionToScene(clientX: number, clientY: number, target?: BABYLON.Vector3): BABYLON.Vector3 {
        let info = getContentWH();
        let pointerX = (clientX - info.styleLeft) / (info.styleSize.width - info.styleLeft - info.styleRight);
        let pointerY = (clientY - info.styleTop)  / (info.styleSize.height - info.styleBottom - info.styleTop);
        pointerX =  pointerX * 2 - 1;
        pointerY = (1 - pointerY) * 2 - 1;

        let temp = MathDataPool.Vector3();
        let matrix = MathDataPool.Matrix();
        let source = MathDataPool.Vector3();
        if (target == undefined) {
            temp.copyFromFloats(0, 0, 0);
        } else {
            temp.copyFrom(target);
        }

        let transform = this.camera.getTransformationMatrix();
        transform.invertToRef(matrix);
        BABYLON.Vector3.TransformCoordinatesToRef(temp, transform, temp); 
        
        source.copyFromFloats(pointerX,      pointerY,         temp.z);
        BABYLON.Vector3._UnprojectFromInvertedMatrixToRef(source, matrix, temp);

        MathDataPool.Recycle(matrix);
        MathDataPool.Recycle(source);

        return temp;
    }
}