import { BABYLON } from "render3d/babylon";
import { IOfflineAnimationCfg } from "./animation";
import { INodeCollider } from "./collider";
import { IVector3 } from "./math";
import { IMeshMaterial, IMeshRenderQueue } from "./mesh";
import { ITransformNode } from "./transform";
import { Nullable } from "./types";
import { INodeD3 } from "../render/index";

/**
 * 封装项目层使用的 场景命中信息
 * * 底层处理过命中检测 - 只通过包围盒检测
 */
// tslint:disable-next-line:interface-name
export interface IRayCastInfo extends BABYLON.PickingInfo {
    hit: boolean;
    /**
     * 命中的目标mesh
     */
    target: INodeD3;
    rayID: number;
}

/**
 * 项目框架下 - 场景中数据结构
 */
export interface Scene3DEventInfo {
    rayInfo: IRayCastInfo;
    e?: any;
    s: any;
}

export enum CAMERATYPES {
    ArcRotateCamera = 'ArcRotateCamera',
    UniversalCamera = 'UniversalCamera',
    TargetCamera = 'TargetCamera'
}

export enum SceneRenderFlags {
    active = 'active',
    pause = 'pause',
    dispose = 'dispose'
}

export enum LIGHTTYPES {
    HemisphericLight = 'HemisphericLight'
}

/**
 * [自定义名称, 要修改的gltf.images 列表序号, 目标图片相对模型文件的路径, 图片变纹理时是否需要反转Y轴]
 */
export type IImageSolt = [string, number, string, boolean?];
/**
 * [自定义名称, 要修改的gltf.materials 列表序号, [对应extension.PI_material 的属性名,属性值] ]
 */
export type IMaterialSolt = [string, number, [string, any][]];

export interface IMaterialReplace {
    /**
     * 替换的材质的 index
     * 如果没设置则认为为所有材质做该替换
     */
    index?: number;
    /**
     * 替换后材质
     */
    extensions: {
        PI_material?: any;
        name: string;
    };
}

/**
 * 模型加载配置
 */
// tslint:disable-next-line:interface-name
export interface IModelOpt {
    /**
     * rayID
     */
    rayID?: number;
    /**
     * 要加载的模型名称
     * * 美术资源(GLTF) 中定义的模型名称
     */
    modelName?: string | string[];
    /**
     * 要加载的模型在所属资源管理路径下的子路径
     * * 资源文件子路径
     */
    path?: string;
    /**
     * 要加载的模型的资源文件名称
     * * 资源文件名称
     */
    fileName?: string;
    absolutePath?: string;
    absoluteFileName?: string;
    isEffect?: boolean;
    /**
     * 自动播放动画 - 默认为第一个
     */
    animDefault?: boolean;
    /**
     * 自动播放粒子
     */
    particleAutoStart?: boolean;
    /**
     * 使用深度预通过
     */
    needDepthPrePass?: boolean;
    /**
     * 控制当前动画完全结束才会切换到目标动画
     */
    workWithAnimationGroupChange?: boolean;
    /**
     * 模型纹理槽数据
     */
    imageSolts?: string[] | IImageSolt[];
    /**
     * 模型纹理槽数据
     */
    materialSolts?: string[] | IMaterialSolt[];
    /**
     * 加载成功的回调
     */
    insertedCall?(model: any): any;
    /**
     * 是否具有骨骼动画优化
     */
    skeletonAnimOptimize?: boolean;
    /**
     * 动画切换前是否重置各属性为无动画状态
     */
    initialWhileAnimChange?: boolean;
    sklaiPath?: string;
    sklaiName?: string;
    /**
     * 材质替换
     */
    materialReplace?: IMaterialReplace[];
    appendAnimation?: IOfflineAnimationCfg;
}

/**
 * 3D 节点变换配置
 */
// tslint:disable-next-line:interface-name
export interface ITransformCfg {
    /**
     * 3D 节点定位
     */
    position?: number[];
    /**
     * 3D 节点旋转
     */
    rotate?: number[];
    /**
     * 3D 节点缩放
     */
    scale?: number[];
}

// tslint:disable-next-line:interface-name
export interface IModelChildOpt {
    name: string;
    /**
     * 模型加载参数
     */
    modelOpt?: IModelOpt;
    /**
     * 模型
     */
    model?: IModelObj;
    /**
     * MESH
     */
    mesh?: ITransformNode;
    /**
     * 变换设置
     */
    transform?: ITransformCfg;
    /**
     * 附加成功标识
     */
    isFinished?: boolean;
    /**
     * 加載好的動畫狀態
     */
    isLoop?: boolean;

    successCall?(OPT: IModelChildOpt);
}

// tslint:disable-next-line:interface-name
export interface IModelAttachOpt {
    name: string;
    /**
     * 模型加载参数
     */
    modelOpt?: IModelOpt;
    /**
     * 模型
     */
    model?: IModelObj;
    /**
     * MESH
     */
    mesh?: ITransformNode;
    /**
     * 附加到目标 skeleton
     */
    skeletonName: string;
    /**
     * 附加到目标 骨头
     */
    boneName: string;
    /**
     * 变换设置
     */
    transform?: ITransformCfg;
}


/**
 * 预制内容的动画指令
 */
export interface IPrefabAnime {
    animName: string;
    isLoop: boolean;
    delay?: number;
    speed?: number;
    endCall?: (data?: any) => void;
    startCall?: (data: any) => void;
    loopCall?: (data: any) => void;
    frameCall?: (framedata: any) => void;
}

export interface IModel {
    /**
     * 创建时的拓展数据 - 自定义作用
     */
    extendData: any;
    rootImpl: ITransformNode;
    loadPromise?: Promise<any>;
    readonly isDisposed: boolean;
    /**
     * * 同步操作
     * @param x 
     * @param y 
     * @param z 
     */
    setPosition(x: number | number[], y?: number, z?: number): void;
    /**
     * * 同步操作
     * @param x 
     * @param y 
     * @param z 
     */
    setScale(x: number | number[], y?: number, z?: number): void;
    /**
     * * 同步操作
     * @param x 
     * @param y 
     * @param z 
     */
    setRotate(x: number | number[], y?: number, z?: number): void;
    setEnabled(val: boolean): void;
    isEnabled(): boolean;
    /**
     * * 同步操作
     * @param r 
     * @param g 
     * @param b 
     */
    setColor(r: number, g: number, b: number): void;
    setAlpha(val: number): void;
    dispose(disposeMaterialAndTextures?: boolean): void;
    /**
     * 添加动画事件监听
     * @deprecated use animationGroup.onAnimationGroupFrameEventObserva
     * @param key 动画事件关键字
     * @param func 监听 - 返回值为 true 表示该监听仅生效一次
     */
    addAnimationEventListen(key: string, func: () => boolean): void;
    pickAboutItems(): INodeCollider[];
    /**
     * 启动指定某个动画 - （模型）一次只播一个动画
     * @param animOpt
     */
    setAnim(animOpt: IPrefabAnime): void;
    setAnimeSpeed(param: number): void;
    get position(): [number, number, number];
    get scale(): [number, number, number];
    get rotate(): [number, number, number];
    rayID: any;
    setParent(parent: IModel, idx?: number): void;
    getChild(index?: number): ITransformNode;

    addEventListener(etype: "pointerup" | "pointerdown", func: Function);
    removeEventListener(etype: "pointerup" | "pointerdown", func: Function);
    collider(min: BABYLON.Vector3, max: BABYLON.Vector3);
    colliderSphereTreshold(val: number);

    /**
     * 手动控制动画进度
     * @param animName 动画名
     * @param progress 动画进度 0~1
     */
    setAnimeFrame(animName: string, progress: number): IModel;
}

export interface IModelObj extends IModel {
    name: string;
    path: string;
    fileName: string;
    animations: { name: string }[];
    rootImpl: ITransformNode;
    renderingGroupId: number;
    alphaIndex: number;
    loadPromise?: Promise<any>;
    readonly isDisposed: boolean;
    meshMap: Map<string, any>;
    animationMap: Map<string, any>;
    scene: any;
    /**
     * * 同步操作
     * @param x 
     * @param y 
     * @param z 
     */
    setPosition(x: number | number[], y?: number, z?: number): void;
    /**
     * * 同步操作
     * @param x 
     * @param y 
     * @param z 
     */
    setScale(x: number | number[], y?: number, z?: number): void;
    /**
     * * 同步操作
     * @param x 
     * @param y 
     * @param z 
     */
    setRotate(x: number | number[], y?: number, z?: number): void;
    setEnabled(val: boolean): void;
    isEnabled(): boolean;
    /**
     * * 同步操作
     * @param r 
     * @param g 
     * @param b 
     */
    setColor(r: number, g: number, b: number): void;
    setAlpha(val: number): void;
    dispose(disposeMaterialAndTextures?: boolean): void;
    loadAsync(opt: IModelOpt): void;
    /**
     * 添加动画事件监听
     * @deprecated use animationGroup.onAnimationGroupFrameEventObserva
     * @param key 动画事件关键字
     * @param func 监听 - 返回值为 true 表示该监听仅生效一次
     */
    addAnimationEventListen(key: string, func: () => boolean): void;
    renderQueueItems(): IMeshRenderQueue[];
    renderMaterialItems(): IMeshMaterial[];
    pickAboutItems(): INodeCollider[];
    /**
     * 启动指定某个动画 - （模型）一次只播一个动画
     * @param animOpt
     */
    setAnim(animOpt: IPrefabAnime): void;
    setAnimeSpeed(param: number): void;
    setLookAt(x: number | number[], y?: number, z?: number): void;
    stopAnim(): void;
    pauseAnim(): void;
    /**
     * @deprecated
     */
    startAllAnim(): void;
    /**
     * 停止模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 start 配对
     * @deprecated
     */
    stopAllAnim(): void;
    /**
     * 停止模型所有粒子
     * @deprecated
     */
    stopAllParticle(): void;
    /**
     * 暂停模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 play 配对
     * @deprecated
     */
    pauseAllAnim(): void;
    /**
     * 继续播放模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 pause 配对
     * @param loop 是否循环
     * @deprecated
     */
    playAllAnim(loop?: boolean): void;
    /**
     * 重新播放模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 pause 配对,从第0帧开始
     * @deprecated
     */
    restartAllAnim(): void;
    /**
     * 重新播放模型所有粒子
     * @description 与 pause 配对,从第0帧开始
     * @deprecated
     */
    restartAllParticle(): void;
    /**
     * 重新播放模型所有粒子
     * @description 与 pause 配对,从第0帧开始
     * @deprecated
     */
    startAllParticle(): void;
    get position(): [number, number, number];
    get scale(): [number, number, number];
    get rotate(): [number, number, number];
    animOpt: IPrefabAnime;
    rayID: any;
    setParent(parent: IModelObj): void;
}