import { BABYLON } from "../render3d/babylon";
import { Constants, CustomGeometry, CustomScene, MainOpacityMaterial } from "../index";
import { ILine, ILineRenderer } from "./interface";
import { ForwardRenderPipeline } from "../render/renderer";

export class Line implements ILine {
    protected static TempNormal: BABYLON.Vector3 = BABYLON.Vector3.Zero();
    protected static TempDirect: BABYLON.Vector3 = BABYLON.Vector3.Zero();
    protected static TempExtend: BABYLON.Vector3 = BABYLON.Vector3.Zero();
    startx: number;
    starty: number;
    startz: number;
    endx: number;
    endy: number;
    endz: number;
    ex: number;
    ey: number;
    ez: number;
    r: number;
    g: number;
    b: number;
    a: number;
    image: string;
    repeatU: number;
    repeatV: number;

    constructor(public readonly unique: number, protected renderer: LineMeshRenderer) {

    }
    extend(normalx: number, normaly: number, normalz: number, width: number, repeatULen: number) {
        Line.TempNormal.copyFromFloats(normalx, normaly, normalz).normalize();
        Line.TempDirect.copyFromFloats(this.endx - this.startx, this.endy - this.starty, this.endz - this.startz);
        let distance = Line.TempDirect.length();
        Line.TempDirect.normalize()
        BABYLON.Vector3.CrossToRef(Line.TempDirect, Line.TempNormal, Line.TempExtend);

        this.ex = Line.TempExtend.x * width;
        this.ey = Line.TempExtend.y * width;
        this.ez = Line.TempExtend.z * width;

        this.repeatU = distance / repeatULen;
    }
    dispose() {
        this.renderer.del(this);
    }
}


class LineMeshRenderer {
    protected counter = Number.MIN_SAFE_INTEGER;
    protected lines: Map<number, Line> = new Map();
    protected mesh: BABYLON.Mesh;
    protected material: BABYLON.MainOpacityMaterial;


    protected _dirty: boolean = false;
    protected customGeometry: CustomGeometry;

    constructor(protected scene: CustomScene, protected image: string, protected maxCount = 128, alphaIndex: number = 3000) {
        let material = new BABYLON.MainOpacityMaterial(`Line#${image}`, this.scene.impl, undefined, (<ForwardRenderPipeline>this.scene.impl.renderPipeline).transparentRenderer.pass);
        let texture = new BABYLON.Texture(null, this.scene.impl, true, false, BABYLON.Texture.LINEAR_LINEAR);
        texture.uScale = 1.01;
        texture.updateURL(image);
        texture.hasAlpha = true;
        material.diffuseTexture = texture;
        // material.backFaceCulling = false;
        // material.emissiveColor.copyFromFloats(1, 1, 1);
        // material.disableLighting = true;
        // material.forceDepthWrite = false;
        // material.alphaMode = 2;
        // material.useAlphaFromDiffuseTexture = true;

        this.material = material;


        this.mesh = new BABYLON.Mesh("meshSource", this.scene.impl);
        this.customGeometry = new CustomGeometry(this.mesh, `${image}#`, [Constants.VERTEX_ATTR_POSITION, Constants.VERTEX_ATTR_COLOR4, Constants.VERTEX_ATTR_UV], this.maxCount * 4, this.maxCount * 2);
        // this.mesh.material = material;
        // this.mesh.useVertexColors = true;
        // this.mesh.hasVertexAlpha = true;

        this.mesh.alwaysSelectAsActiveMesh = true;
        this.mesh.alphaIndex = alphaIndex;
        this.mesh.depthTest = Constants.NEVER;

        // BABYLONAction.useMaterial(this.mesh, material, MATERIAL_ALPHABLEND, 2);
        this.mesh.alphaMode = 2;
        this.mesh.setMaterial(material);

        this.mesh.depthWrite = false;
        this.mesh.cullMode = Constants.NONE;

        let indices = this.customGeometry.indices();
        for (let i = 0; i < maxCount; i++) {
            let idx = i * 2 * 3;
            let ind = i * 4;
            indices[idx + 0] = ind + 0;
            indices[idx + 1] = ind + 1;
            indices[idx + 2] = ind + 2;
            indices[idx + 3] = ind + 3;
            indices[idx + 4] = ind + 2;
            indices[idx + 5] = ind + 1;
        }
    }
    set alphaIndex(val: number) {
        this.mesh.alphaIndex = val;
    }
    set alphaMode(val: number) {
        this.mesh.alphaMode = val;
    }
    set uvSpeedU(val: number) {
        if (this.material?.diffuseTexture) {
            (<BABYLON.Texture>this.material.diffuseTexture).offsetSpeedU = val;
        }
    }
    set uvSpeedV(val: number) {
        if (this.material?.diffuseTexture) {
            (<BABYLON.Texture>this.material.diffuseTexture).offsetSpeedV = val;
        }
    }
    create(image: string, startx: number, starty: number, startz: number, endx: number, endy: number, endz: number, normalx: number, normaly: number, normalz: number, r: number, g: number, b: number, a: number, width: number, repeatULen: number, repeatV: number) {
        let unique = this.counter++;
        let line = new Line(unique, this);
        line.startx = startx;
        line.starty = starty;
        line.startz = startz;
        line.endx = endx;
        line.endy = endy;
        line.endz = endz;
        line.r = r;
        line.g = g;
        line.b = b;
        line.a = a;
        line.image = image;
        line.repeatV = repeatV;
        line.extend(normalx, normaly, normalz, width, repeatULen);

        this.lines.set(unique, line);
        this._dirty = true;
        return line;
    }
    del(line: Line) {
        let item = this.lines.get(line.unique);
        if (item == line) {
            this._dirty = true;
            this.lines.delete(line.unique);
        } else {
            console.error(`Line Dispose Error !`);
        }
    }
    update() {
        if (this._dirty) {
            let idx = 0;
            let [uvs, uvstrip, uvoffset] = this.customGeometry.vertices(Constants.VERTEX_ATTR_UV);
            let [positions, posstrip, posoffset] = this.customGeometry.vertices(Constants.VERTEX_ATTR_POSITION);
            let [colors, colorstrip, coloroffset] = this.customGeometry.vertices(Constants.VERTEX_ATTR_COLOR4);
            this.lines.forEach((line) => {
                {
                    let off = idx * 4 * uvstrip;
                    uvs[off + 0 + 0 * uvstrip + uvoffset] = 0 * line.repeatU;
                    uvs[off + 1 + 0 * uvstrip + uvoffset] = 0 * line.repeatV;
                    uvs[off + 0 + 1 * uvstrip + uvoffset] = 1 * line.repeatU;
                    uvs[off + 1 + 1 * uvstrip + uvoffset] = 0 * line.repeatV;
                    uvs[off + 0 + 2 * uvstrip + uvoffset] = 0 * line.repeatU;
                    uvs[off + 1 + 2 * uvstrip + uvoffset] = 1 * line.repeatV;
                    uvs[off + 0 + 3 * uvstrip + uvoffset] = 1 * line.repeatU;
                    uvs[off + 1 + 3 * uvstrip + uvoffset] = 1 * line.repeatV
                }

                {
                    let off = idx * 4 * posstrip;
                    positions[off + 0 + 0 * posstrip + posoffset] = line.startx - line.ex;
                    positions[off + 1 + 0 * posstrip + posoffset] = line.starty - line.ey;
                    positions[off + 2 + 0 * posstrip + posoffset] = line.startz - line.ez;

                    positions[off + 0 + 1 * posstrip + posoffset] = line.endx - line.ex;
                    positions[off + 1 + 1 * posstrip + posoffset] = line.endy - line.ey;
                    positions[off + 2 + 1 * posstrip + posoffset] = line.endz - line.ez;


                    positions[off + 0 + 2 * posstrip + posoffset] = line.startx + line.ex;
                    positions[off + 1 + 2 * posstrip + posoffset] = line.starty + line.ey;
                    positions[off + 2 + 2 * posstrip + posoffset] = line.startz + line.ez;

                    positions[off + 0 + 3 * posstrip + posoffset] = line.endx + line.ex;
                    positions[off + 1 + 3 * posstrip + posoffset] = line.endy + line.ey;
                    positions[off + 2 + 3 * posstrip + posoffset] = line.endz + line.ez;
                }

                {
                    let off = idx * 4 * colorstrip;
                    colors[off + 0 + 0 * colorstrip + coloroffset] = line.r;
                    colors[off + 1 + 0 * colorstrip + coloroffset] = line.g;
                    colors[off + 2 + 0 * colorstrip + coloroffset] = line.b;
                    colors[off + 3 + 0 * colorstrip + coloroffset] = line.a;
                    colors[off + 0 + 1 * colorstrip + coloroffset] = line.r;
                    colors[off + 1 + 1 * colorstrip + coloroffset] = line.g;
                    colors[off + 2 + 1 * colorstrip + coloroffset] = line.b;
                    colors[off + 3 + 1 * colorstrip + coloroffset] = line.a;
                    colors[off + 0 + 2 * colorstrip + coloroffset] = line.r;
                    colors[off + 1 + 2 * colorstrip + coloroffset] = line.g;
                    colors[off + 2 + 2 * colorstrip + coloroffset] = line.b;
                    colors[off + 3 + 2 * colorstrip + coloroffset] = line.a;
                    colors[off + 0 + 3 * colorstrip + coloroffset] = line.r;
                    colors[off + 1 + 3 * colorstrip + coloroffset] = line.g;
                    colors[off + 2 + 3 * colorstrip + coloroffset] = line.b;
                    colors[off + 3 + 3 * colorstrip + coloroffset] = line.a;
                }
                idx++;
            });

            this._dirty = false;
            this.customGeometry.update(idx * 6);
            this.mesh.setEnabled(idx > 0);
        }
    }
    dispose() {
        this.lines.clear();
        this.mesh.dispose(false, true);
        this.material.dispose();
    }
}

export class LineRenderer implements ILineRenderer {
    protected meshes: Map<string, LineMeshRenderer> = new Map();
    constructor(protected scene: CustomScene, protected maxCount: number = 128) {

        this.scene.registerBeforeRenderCall(this.update);
    }
    protected _alphaIndex: number;
    set alphaIndex(val: number) {
        this.meshes.forEach((mesh) => {
            mesh.alphaIndex = val;
        });
        this._alphaIndex = val;
    }
    protected _alphaMode: number;
    set alphaMode(val: number) {
        this.meshes.forEach((mesh) => {
            mesh.alphaMode = val;
        });
        this._alphaMode = val;
    }
    protected _uvSpeedU: number;
    /**
     * 每多少ms变化一轮
     */
    set uvSpeedU(val: number) {
        this.meshes.forEach((mesh) => {
            mesh.uvSpeedU = val;
        });
        this._uvSpeedU = val;
    }
    protected _uvSpeedV: number;
    /**
     * 每多少ms变化一轮
     */
    set uvSpeedV(val: number) {
        this.meshes.forEach((mesh) => {
            mesh.uvSpeedV = val;
        });
        this._uvSpeedV = val;
    }
    protected update = () => {
        this.meshes.forEach((mesh) => {
            mesh.update();
        });
    }
    create(image: string, startx: number, starty: number, startz: number, endx: number, endy: number, endz: number, normalx: number, normaly: number, normalz: number, r: number, g: number, b: number, a: number, width: number, repeatULen: number, repeatV: number): ILine {
        let mesh = this.meshes.get(image);
        if (mesh == undefined) {
            mesh = new LineMeshRenderer(this.scene, image, this.maxCount);
            mesh.alphaIndex = this._alphaIndex;
            mesh.alphaMode = this._alphaMode;
            mesh.uvSpeedU = this._uvSpeedU;
            mesh.uvSpeedV = this._uvSpeedV;
            this.meshes.set(image, mesh);
        }

        return mesh.create(image, startx, starty, startz, endx, endy, endz, normalx, normaly, normalz, r, g, b, a, width, repeatULen, repeatV);
    }
    dispose() {
        this.meshes.forEach((item) => {
            item.dispose();
        });
        this.scene.unregisterBeforeRenderCall(this.update);
        this.meshes.clear();
    }
}

export class LineRenderer2D extends LineRenderer {
    constructor(protected scene: CustomScene, protected viewDirection: BABYLON.Vector3, protected maxCount: number = 128) {
        super(scene, maxCount);
        this.viewDirection.normalize();
    }

    /**
     * 
     * @param image image 因为可能有UV流动需求, 因此此处不使用精灵图，且图片尺寸需要2的幂次
     * @param startx 
     * @param starty 
     * @param startz 
     * @param endx 
     * @param endy 
     * @param endz 
     * @param r 
     * @param g 
     * @param b 
     * @param a 
     * @param width 线宽度
     * @param repeatULen 每多少距离纹理重复一次
     * @param repeatV 纹理重复多少次
     * @returns 
     */
    create(image: string, startx: number, starty: number, startz: number, endx: number, endy: number, endz: number, r: number, g: number, b: number, a: number, width: number, repeatULen: number, repeatV: number): ILine {
        let mesh = this.meshes.get(image);
        if (mesh == undefined) {
            mesh = new LineMeshRenderer(this.scene, image, this.maxCount);
            mesh.alphaIndex = this._alphaIndex;
            this.meshes.set(image, mesh);
        }

        return mesh.create(image, startx, starty, startz, endx, endy, endz, this.viewDirection.x, this.viewDirection.y, this.viewDirection.z, r, g, b, a, width, repeatULen, repeatV);
    }
}