import { BABYLON } from "../render3d/babylon";
import { ExtensionTools, InstanceAttribute } from "../interface/index";
import { p3d } from "../render/types";


export class TextureHSVMaterial extends BABYLON.MainTextureMaterial {
    public static ShaderName = "HSVTEX";
    
    constructor(name: string, scene: BABYLON.Scene, shaderName = TextureHSVMaterial.ShaderName, _pass?: BABYLON.Constants.PassTag) {
        TextureHSVMaterial.regist();
        super(name, scene, shaderName, _pass);
    }
    
    public static regist() {
        if (p3d.p3d_check_shader(BABYLON.Engine.native(), BABYLON.Engine.param(), TextureHSVMaterial.ShaderName)) {
            return;
        }
        let uniforms: p3d.MaterialUniformDefines = p3d.MaterialUniformDefines.create();
        let includes: p3d.NodematerialIncludes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
                // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(BABYLON.UniformMainTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTextureUVOffsetSpeed);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacity);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTexture);
        // p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(BABYLON.UniformOpacityTexture.BLOCK_NAME));
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTextureUVOffsetSpeed);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTexture);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTextureUVOffsetSpeed);

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec2", "v_uv"],
            ["vec3", "vHSV"],
            ["float", "vAlpha"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });
        let binds_defines_base = BABYLON.Constants.BIND_MODEL | BABYLON.Constants.BIND_VIEWER | BABYLON.Constants.BIND_EFFECT_VALUE | BABYLON.Constants.BIND_EFFECT_TEXTURE;

        p3d.p3d_regist_material(
            BABYLON.Engine.native(), BABYLON.Engine.param(),
            TextureHSVMaterial.ShaderName,
            uniforms,
            `
            `,
            `
vec3 rgb2hsv(vec3 c)
{
    vec4 K = vec4(0.0, -1.0 / 3.0, 2.0 / 3.0, -1.0);
    vec4 p = mix(vec4(c.bg, K.wz), vec4(c.gb, K.xy), step(c.b, c.g));
    vec4 q = mix(vec4(p.xyw, c.r), vec4(c.r, p.yzx), step(p.x, c.r));

    float d = q.x - min(q.w, q.y);
    float e = 1.0e-10;
    return vec3(abs(q.z + (q.w - q.y) / (6.0 * d + e)), d / (q.x + e), q.x);
}

vec3 hsv2rgb(vec3 c)
{
    vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
    vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
    return c.z * mix(K.xxx, clamp(p - K.xxx, vec3(0.0), vec3(1.0)), c.y);
}

vec3 applyHSV(vec3 c, vec3 pHSV) {
    vec3 hsv = rgb2hsv(c.rgb);
    hsv.r += pHSV.r;
    c.rgb = hsv2rgb(hsv);

    float gray = max(c.r, max(c.g, c.b)) + min(c.r, min(c.g, c.b));
    c.rgb = mix(c.rgb, vec3(0.5 * gray), -pHSV.g);

    return mix(c.rgb * (1.0 + pHSV.b), mix(c.rgb, vec3(1.0), pHSV.b), step(0., pHSV.b));
}
layout(location = 0) out vec4 gl_FragColor;
            `,
            `
    mat4 finalWorld = PI_ObjectToWorld;
    vec4 position =  vec4(A_POSITION, 1.);
    vec4 worldPos =  finalWorld * position;
    gl_Position = PI_MATRIX_VP * worldPos;
            
    v_uv    = A_UV;
    vHSV = A_COLOR4.rgb;
    vAlpha = A_COLOR4.a;
            `,
            `
    vec4 textureColor           = mainTexture(v_uv, applyUVOffsetSpeed(matParam.uMainUVOS), matParam);
    textureColor.rgb           *= mainStrength(matParam);

    textureColor.rgb = applyHSV(textureColor.rgb, vHSV);
    textureColor.a *= vAlpha;
    gl_FragColor = textureColor;   
            `,
            includes,
            "",
            varyings,
            binds_defines_base
        );

        varyings.free && varyings.free();
        uniforms.free && uniforms.free();
        includes.free && includes.free();
    }
}
