import { BABYLON } from "../render3d/babylon";
import { ExtensionTools, InstanceAttribute } from "../interface/index";
import { p3d } from "../render/types";


export class TextureHueSaturateMaterial extends BABYLON.MainTextureMaterial {
    public static ShaderName = "HueSaturate";
    
    constructor(name: string, scene: BABYLON.Scene, shaderName = TextureHueSaturateMaterial.ShaderName, _pass?: BABYLON.Constants.PassTag) {
        TextureHueSaturateMaterial.regist();
        super(name, scene, shaderName, _pass);
    }
    protected _alpha: number = 1;
    public get alpha() {
        return this._alpha;
    }
    public set alpha(value: number) {
        this._alpha = value;
        this.baseColor(this._emissiveColor.r, this._emissiveColor.g, this._emissiveColor.b, this._alpha);
    }
    
    public _emissiveColor: BABYLON.Color3 = new BABYLON.Color3(1, 1, 1);
    public get emissiveColor() {
        return this._emissiveColor;
    }
    public set emissiveColor(value: BABYLON.Color3) {
        this._emissiveColor = value;
        this.baseColor(this._emissiveColor.r, this._emissiveColor.g, this._emissiveColor.b, this._alpha);
    }
    protected _diffuseTexture: BABYLON.Texture;
    public set diffuseTexture(tex: BABYLON.Texture) {
        if (this._diffuseTexture != tex) {
            this._diffuseTexture?.dispose();
        }
        this.mainTexture = tex;
        this._diffuseTexture = tex;
    }
    public get diffuseTexture() {
        return this._diffuseTexture;
    }
    public static regist() {
        if (p3d.p3d_check_shader(BABYLON.Engine.native(), BABYLON.Engine.param(), TextureHueSaturateMaterial.ShaderName)) {
            return;
        }
        let uniforms = p3d.MaterialUniformDefines.create();
        let includes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
                // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(BABYLON.UniformMainTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTextureUVOffsetSpeed);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacity);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTexture);
        // p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(BABYLON.UniformOpacityTexture.BLOCK_NAME));
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTextureUVOffsetSpeed);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTexture);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTextureUVOffsetSpeed);

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec2", "v_uv"],
            ["vec3", "vHSV"],
            ["float", "vAlpha"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });
        let binds_defines_base = BABYLON.Constants.BIND_MODEL | BABYLON.Constants.BIND_VIEWER | BABYLON.Constants.BIND_EFFECT_VALUE | BABYLON.Constants.BIND_EFFECT_TEXTURE;

        p3d.p3d_regist_material(
            BABYLON.Engine.native(), BABYLON.Engine.param(),
            TextureHueSaturateMaterial.ShaderName,
            uniforms,
            `
            `,
            `

const float A = 1.0/6.0;
const float B = 1.0/2.0;
const float C = 2.0/3.0;
float hue2rgb(float p, float q, float t) {
    float a = step(0., t);
    float e = step(1., t);
    t = mix(t + 1.0, mix(t, t - 1.0, e), a);

    float b = step(A, t);
    float c = step(B, t);
    float d = step(C, t);

    return mix(
        p + (q - p) * 6.0 * t,
        mix( q, mix( p + (q - p) * (2.0/3.0 - t) * 6.0, p, d ), c ),
        b
    );
}
vec3 hslToRgb (float h, float s, float l) {
    float q = mix( l * (1.0 + s), l + s - l * s, step(0.5, l));
    float p = 2.0 * l - q;

    return mix(
        vec3(l, l, l),
        vec3(hue2rgb(p, q, h + 1.0/3.0), hue2rgb(p, q, h), hue2rgb(p, q, h - 1.0/3.0)),
        step(0.000001, abs(s))
    );
}
layout(location = 0) out vec4 gl_FragColor;
            `,
            `
    mat4 finalWorld = PI_ObjectToWorld;
    vec4 position =  vec4(A_POSITION, 1.);
    vec4 worldPos =  finalWorld * position;
    gl_Position = PI_MATRIX_VP * worldPos;
            
    v_uv    = A_UV;
    vHSV    = A_COLOR4.rgb;
    vAlpha  = A_COLOR4.a;
            `,
            `
    vec4 src           = mainTexture(v_uv, applyUVOffsetSpeed(matParam.uMainUVOS), matParam);
    src.rgb            *= mainStrength(matParam);

    float gray = max(src.r, max(src.g, src.b)) + min(src.r, min(src.g, src.b));
    vec3 rgb = hslToRgb(vHSV.x, vHSV.y, gray * 0.5);
    rgb = mix(rgb.rgb * (1.0 + vHSV.b), mix(rgb.rgb, vec3(1.0), vHSV.b), step(0., vHSV.b));
    gl_FragColor = vec4(rgb, src.w * vAlpha);
            `,
            includes,
            "",
            varyings,
            binds_defines_base
        );

        varyings.free && varyings.free();
        uniforms.free && uniforms.free();
        includes.free && includes.free();
    }
}
