import { ExtensionTools } from "../interface/index";
import { IOpacityClipMaterial } from "../gltf2/interface";
import { BABYLON } from "../render3d/babylon";
import { p3d } from "../render/types";
import { StandardEffectMaterial } from "../render/materials/standard_effect_material";

export class OpacityClipMaterial extends BABYLON.MainOpacityMaterial {

    public static ShaderName = "OpacityClip";
    
    constructor(name: string, scene: BABYLON.Scene, shaderName = OpacityClipMaterial.ShaderName) {
        super(name, scene, shaderName);
    }
    
    public static regist() {
        if (p3d.p3d_check_shader(BABYLON.Engine.native(), BABYLON.Engine.param(), OpacityClipMaterial.ShaderName)) {
            return;
        }
        let uniforms = p3d.MaterialUniformDefines.create();
        let includes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
                // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(BABYLON.UniformMainTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTextureUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacity);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(BABYLON.UniformOpacityTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTextureUVOffsetSpeed);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTexture);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTextureUVOffsetSpeed);

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec3", "v_pos"],
            ["vec3", "v_normal"],
            ["vec4", "v_color"],
            ["vec2", "v_uv"],
            ["vec2", "v_uv2"],
            ["vec2", "v_uv3"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });
        let binds_defines_base = BABYLON.Constants.BIND_MODEL | BABYLON.Constants.BIND_VIEWER | BABYLON.Constants.BIND_EFFECT_VALUE | BABYLON.Constants.BIND_EFFECT_TEXTURE;

        p3d.p3d_regist_material(
            BABYLON.Engine.native(), BABYLON.Engine.param(),
            OpacityClipMaterial.ShaderName,
            uniforms,
            `
            `,
            `
            layout(location = 0) out vec4 gl_FragColor;
            `,
            `
            mat4 finalWorld = PI_ObjectToWorld;
        
            vec4 position =  vec4(A_POSITION, 1.);
            vec4 worldPos =  finalWorld * position;
            // vec4 worldPos =  position;
        
            gl_Position = PI_MATRIX_VP * worldPos;
            gl_Position.z = gl_Position.z * 0.5 + 0.5;
            // gl_Position = position;
        
            v_pos = worldPos.xyz;
        
            mat3 normalWorld = mat3(finalWorld);
            v_normal = A_NORMAL;
            
            v_uv    = A_UV;
            v_uv2   = A_UV;
            v_uv3   = A_UV;
            v_color = A_COLOR4;
            `,
            `
            vec4 baseColor = v_color;
            float alpha = opacity(matParam);
            alpha   *= v_color.a;
        
            float glossiness 			= 0.;
            vec3 diffuseColor 		    = vec3(0., 0., 0.);
            vec3 baseAmbientColor		= vec3(1., 1., 1.);
            vec4 refractionColor		= vec4(0., 0., 0., 1.);
            vec4 reflectionColor 		= vec4(0., 0., 0., 1.);
        
            // ----------------------------------------------------------
        
            vec4 mainTextureColor   = mainTexture(v_uv, applyUVOffsetSpeed(matParam.uMainUVOS), matParam);
            baseColor.rgb           *= mainTextureColor.rgb * mainStrength(matParam);
            alpha                   *= mainTextureColor.a;
        
            diffuseColor            = mainColor(matParam);
        
            vec4 opacityData        = opacityTexture(v_uv2, applyUVOffsetSpeed(matParam.uOpacityUVOS), matParam);
            if (cutoff(opacityChannel(opacityData, matParam), matParam)) {
                discard;
            }
        
            // vec3 emissiveColor = emissiveColor(matParam);
            // emissiveColor *= emissiveTexture(v_uv3, applyUVOffsetSpeed(matParam.uEmissiveUVOS), matParam).rgb * emissiveStrength(matParam);
        
            vec3 finalSpecular 		= vec3(0., 0., 0.);
            vec3 finalDiffuse       = diffuseColor * baseColor.rgb;
            
            vec4 finalColor 		= vec4(
                finalDiffuse * baseAmbientColor
                + 
                finalSpecular
                +
                refractionColor.rgb
                + 
                reflectionColor.rgb
                ,
                alpha
            );
        
            finalColor				= max(finalColor, 0.0);
        
            // Premulty
            // finalColor.rgb          *= finalColor.a;
        
            gl_FragColor = finalColor;        
            `,
            includes,
            "",
            varyings,
            binds_defines_base
        );

        varyings.free && varyings.free();
        uniforms.free && uniforms.free();
        includes.free && includes.free();
    }
}

export class MaterialLoaderOpacityClipMaterial {
    public static init() {
        BABYLON.GLTF2.LoadExtraMaterial.regist(
            OpacityClipMaterial.ShaderName,
            MaterialLoaderOpacityClipMaterial.load
        );
    }
    public static load(loaders: BABYLON.GLTF2.LoaderRecord, matdata: IOpacityClipMaterial, indexNode: number, callback: (mat: BABYLON.Material, state: BABYLON.GLTF2.IRenderState) => void): BABYLON.Material {

            // StandardEffectMaterial.regist();
            // let mat = new StandardEffectMaterial("Mat", loaders.scene);
            // mat.clipByOpacity = true;

            OpacityClipMaterial.regist();
            let mat = new OpacityClipMaterial("Mat", loaders.scene);

            let info = matdata;

            if (info.diffuseTexture) {
                BABYLON.GLTF2.LoadExtraMaterial.diffuseTexture(mat, loaders, indexNode, info);
            }

            // if (info.emissionTexture) {
            //     LoadExtraMaterial.emissiveTexture(mat, loaders, indexNode, info);
            // }

            if (info.opacityTexture) {
                BABYLON.GLTF2.LoadExtraMaterial.opacityTexture(mat, loaders, indexNode, info);
            }

            let numbers = info.diffuseColor;
            // if (numbers) {
            //     mat.diffuseColor = new Color3(numbers[0], numbers[1], numbers[2]);
            // }
            numbers = info.emissionColor;
            if (numbers) {
                mat.diffuseColor = new BABYLON.Color3(numbers[0], numbers[1], numbers[2]);
            }

            let state = BABYLON.GLTF2.LoadExtraMaterial.state(mat, info);

            state.alphaMode = state.alphaMode || 7;

            callback(mat, state);

            return mat;
    }
}