import { Scene } from "../render3d/scene";
import { BABYLON } from "../render3d/babylon";
import { MathDataPool } from "../render3d/extends/math";

export interface TextureTillOffDataScene extends Scene {
    __TextureTillOffDatas: Map<number, TextureTillOffData>;
    __TextureTillOffPool: TextureTillOffData[];
}

export class TextureTillOffData {
    private static Identity = BABYLON.Matrix.FromValues(1, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1);
    private static isIdentityAs3x2 = function() { return false; }
    private static Counter: number = Number.MIN_SAFE_INTEGER;

    private _uScale: number = 1;
    private _vScale: number = 1;
    private _uOffset: number = 0;
    private _vOffset: number = 0;
    get uScale() { return this._uScale };
    get vScale() { return this._vScale };
    get uOffset() { return this._uOffset };
    get vOffset() { return this._vOffset };
    set uScale(val: number) { this._uScale = val; this._dirty = true; };
    set vScale(val: number) { this._vScale = val; this._dirty = true; };
    set uOffset(val: number) { this._uOffset = val; this._dirty = true; };
    set vOffset(val: number) { this._vOffset = val; this._dirty = true; };
    animations: Array<BABYLON.Animation> = [];
    _cachedTextureMatrix: BABYLON.Matrix;

    private _dirty: boolean = true;
    public get dirty() {
        return this._dirty;
    }
    public readonly uid: number;

    public static Create(scene: TextureTillOffDataScene) {
        if (scene.__TextureTillOffDatas == undefined) {
            scene.__TextureTillOffDatas = new Map();
            scene.__TextureTillOffPool = [];
            scene.onDisposeObservable.addOnce(function(scene: TextureTillOffDataScene) {
                scene.__TextureTillOffDatas.forEach(function(item) {
                    item.dispose();
                });
                scene.__TextureTillOffPool = [];
                scene.__TextureTillOffDatas.clear();
            });
            scene.onAfterRenderObservable.add(function(scene: TextureTillOffDataScene) {
                scene.__TextureTillOffDatas.forEach(function(item) {
                    item._dirty = false;
                });
            });
        }

        let result = scene.__TextureTillOffPool.pop();
        if (result == undefined) {
            result = new TextureTillOffData(scene);
        }
        scene.__TextureTillOffDatas.set(result.uid, result);

        return result;
    }

    constructor(private _scene: TextureTillOffDataScene) {
        this.uid = TextureTillOffData.Counter++;
    }

    // PI_BEGIN
    /**
     * 纹理offset自动变化周期 - 每 offsetSpeedU 秒循环一次
     */
    public offsetSpeedU: number = 0;
    /**
     * 纹理offset自动变化周期 - 每 offsetSpeedV 秒循环一次
     */
    public offsetSpeedV: number = 0;
    // PI_END

    // PI_BEGIN
    public computeTextureOffsetWithSpeedUV() {
        if (this._scene?.animationsEnabled) {
            if (Math.abs(this.offsetSpeedU) > 0.0001) {
                this.uOffset = Math.round(this._scene._animationTime) / this.offsetSpeedU;
            }
            if (Math.abs(this.offsetSpeedV) > 0.0001) {
                this.vOffset = Math.round(this._scene._animationTime) / this.offsetSpeedV;
            }
        }
    }
    // PI_END

    getTextureMatrix(): BABYLON.Matrix {
        this.computeTextureOffsetWithSpeedUV();

        if (this.uScale == 1 && this.vScale == 1 && this.uOffset == 0 && this.vOffset == 0) {
            TextureTillOffData.Identity.isIdentityAs3x2 = TextureTillOffData.isIdentityAs3x2;
            return TextureTillOffData.Identity;
        } else {
            if (!this._cachedTextureMatrix) {
                this._cachedTextureMatrix = MathDataPool.Matrix();
                BABYLON.Matrix.FromValuesToRef(1, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1, this._cachedTextureMatrix);
                this._cachedTextureMatrix.isIdentityAs3x2 = TextureTillOffData.isIdentityAs3x2;
            }
            const m = (<any>this._cachedTextureMatrix)._m;
    
            m[0 * 4 + 0] = this.uScale;
            m[1 * 4 + 1] = this.vScale;
            m[2 * 4 + 2] = 0;
            
            m[3 * 4 + 0] = this.uOffset;
            m[3 * 4 + 1] = this.vOffset;
            m[3 * 4 + 2] = 0;
            this._cachedTextureMatrix.setRowFromFloats(2, this.uOffset, this.vOffset, 0, 1);
            this._cachedTextureMatrix.markAsUpdated();

            return this._cachedTextureMatrix;
        }

    }

    dispose() {
        if (this._scene?.__TextureTillOffPool) {
            if (this._scene.__TextureTillOffDatas.delete(this.uid)) {
                this._scene.__TextureTillOffPool.push(this);
                this.animations.length = 0;
                this.uScale = 1;
                this.vScale = 1;
                this.uOffset = 0;
                this.vOffset = 0;
                this.offsetSpeedU = 0;
                this.offsetSpeedV = 0;
                this._dirty = true;
                MathDataPool.Recycle(this._cachedTextureMatrix);
                this._cachedTextureMatrix = undefined;
            }
        }
    }

    copyFrom(tex: TextureTillOffData) {
        this.uScale = tex.uScale;
        this.vScale = tex.vScale;
        this.uOffset = tex.uOffset;
        this.vOffset = tex.vOffset;
        this.offsetSpeedU = tex.offsetSpeedU;
        this.offsetSpeedV = tex.offsetSpeedV;
    }
}

interface StandardMaterialDataScene extends Scene {
    __StandardMaterialData: Map<number, StandardMaterialData>;
}
export class StandardMaterialData {
    private static Counter: number = Number.MIN_SAFE_INTEGER;
    public static Create(_scene: Scene) {
        let scene = <StandardMaterialDataScene>_scene;
        if (scene.__StandardMaterialData == undefined) {
            scene.__StandardMaterialData = new Map();
            scene.onDisposeObservable.addOnce(() => {
                scene.__StandardMaterialData.forEach((item) => {
                    item.dispose();
                });
                scene.__StandardMaterialData.clear();
            });
            scene.onAfterRenderObservable.add(function(scene: StandardMaterialDataScene) {
                scene.__StandardMaterialData.forEach((item) => {
                    item._dirty = false;
                });
            });
        }

        let result = new StandardMaterialData(scene);

        scene.__StandardMaterialData.set(result.uid, result);

        return result;

    }

    emissiveColor: BABYLON.Color3 = BABYLON.Color3.White();
    diffuseTexture: TextureTillOffData;
    emissiveTexture: TextureTillOffData;
    opacityTexture: TextureTillOffData;
    maskTexture: TextureTillOffData;
    animations: Array<BABYLON.Animation> = [];

    public readonly uid: number;

    private _dirty: boolean = true;
    public get dirty() {
        return this._dirty || this.diffuseTexture.dirty || this.emissiveTexture.dirty || this.opacityTexture.dirty || this.maskTexture.dirty;
    }

    constructor(protected scene: BABYLON.Scene) {
        this.uid = StandardMaterialData.Counter++;

        this.diffuseTexture     = TextureTillOffData.Create(<TextureTillOffDataScene>scene);
        this.emissiveTexture    = TextureTillOffData.Create(<TextureTillOffDataScene>scene);
        this.opacityTexture     = TextureTillOffData.Create(<TextureTillOffDataScene>scene);
        this.maskTexture        = TextureTillOffData.Create(<TextureTillOffDataScene>scene);
    }
    fromMaterial(material: StandardMaterialData) {
        if (material.emissiveColor) {
            this.emissiveColor.copyFrom(material.emissiveColor);
        }
        if (material.diffuseTexture) {
            this.diffuseTexture.copyFrom(material.diffuseTexture);
        }
        if (material.opacityTexture) {
            this.opacityTexture.copyFrom(material.opacityTexture);
        }
        if (material.emissiveTexture) {
            this.emissiveTexture.copyFrom(material.emissiveTexture);
        }
        if (material.maskTexture) {
            this.maskTexture.copyFrom(material.maskTexture);
        }
        this.alpha = material.alpha;
        this.alphaCutOff = material.alphaCutOff;
        this.maskCutoff = material.maskCutoff;
    }
    // alpha: number = 1;
    private _alpha: number = 1;
    get alpha() {
        return this._alpha;
    }
    set alpha(val: number) {
        this._alpha = val;
        this._dirty = true;
    }

    // alphaCutOff: number = 0;
    private _alphaCutOff: number = 0;
    get alphaCutOff() {
        return this._alphaCutOff;
    }
    set alphaCutOff(val: number) {
        this._alphaCutOff = val;
        this._dirty = true;
    }

    maskCutoff: number = 0;

    dispose = () => {
        this.diffuseTexture     ?.dispose();
        this.emissiveTexture    ?.dispose();
        this.opacityTexture     ?.dispose();
        this.maskTexture        ?.dispose();
        this.diffuseTexture     = undefined;
        this.emissiveTexture    = undefined;
        this.opacityTexture     = undefined;
        this.maskTexture        = undefined;
    }
}

export interface StandardMaterialMesh extends BABYLON.Mesh {
    materialData: StandardMaterialData;
}

export function initStandardMaterialOpt() {
}