export * from "./math.scalar";
export * from "./math.functions";
export * from "./math.polar";
export * from "./math";
export * from "./sphericalPolynomial";

import { DeepImmutable } from "../interface/types";
import { Color3, Color4 } from "./math.color";
import { Vector3, Vector4, Quaternion, Matrix } from "./math.vector";

export class MathDataPool {
    private static _Color3: Color3[] = [];
    private static _Color4: Color4[] = [];
    private static _Vector3: Vector3[] = [];
    private static _Vector4: Vector4[] = [];
    private static _Quaternion: Quaternion[] = [];
    private static _Matrix: Matrix[] = [];
    public static MAX_CACHE: number = 10000;
    public static Init() {
        Vector3.Zero = function() {
            return MathDataPool.Vector3(0, 0, 0);
        }
        Vector3.One = function() {
            return MathDataPool.Vector3(1, 1, 1);
        }
        Vector3.FromArray = function(array: DeepImmutable<ArrayLike<number>>, offset: number = 0) {
            return MathDataPool.Vector3(array[offset + 0], array[offset + 1], array[offset + 2]);
        }
        Color3.White = function() {
            return MathDataPool.Color3(1, 1, 1);
        }
        Color3.Black = function() {
            return MathDataPool.Color3(0, 0, 0);
        }
        Color3.FromArray = function(array: DeepImmutable<ArrayLike<number>>, offset: number = 0) {
            return MathDataPool.Color3(array[offset + 0], array[offset + 1], array[offset + 2]);
        }
        Color4.FromArray = function(array: DeepImmutable<ArrayLike<number>>, offset: number = 0) {
            return MathDataPool.Color4(array[offset + 0], array[offset + 1], array[offset + 2], array[offset + 3]);
        }
        Vector4.Zero = function() {
            return MathDataPool.Vector4(0, 0, 0, 0);
        }
        Vector4.One = function() {
            return MathDataPool.Vector4(1, 1, 1, 1);
        }
        Quaternion.Zero = function() {
            return MathDataPool.Quaternion(0, 0, 0, 0);
        }
        Quaternion.Identity = function() {
            return MathDataPool.Quaternion(0, 0, 0, 1);
        }
        Matrix.Identity = function() {
            return MathDataPool.Matrix();
        }
        Matrix.Zero = function() {
            let result = MathDataPool.Matrix();
            Matrix.FromValuesToRef(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, result);
            return result;
        }
        Matrix.FromArray = function(array: DeepImmutable<ArrayLike<number>>, offset: number = 0) {
            let result = MathDataPool.Matrix();
            Matrix.FromArrayToRef(array, offset, result);
            return result;
        }
        Matrix.FromValues = function(initialM11: number, initialM12: number, initialM13: number, initialM14: number, initialM21: number, initialM22: number, initialM23: number, initialM24: number, initialM31: number, initialM32: number, initialM33: number, initialM34: number, initialM41: number, initialM42: number, initialM43: number, initialM44: number) {
            let result = MathDataPool.Matrix();
            Matrix.FromValuesToRef(initialM11, initialM12, initialM13, initialM14, initialM21, initialM22, initialM23, initialM24, initialM31, initialM32, initialM33, initialM34, initialM41, initialM42, initialM43, initialM44, result);
            return result;
        }
    }
    public static Color3(r: number = 0, g: number = 0, b: number = 0) {
        let result = MathDataPool._Color3.pop();
        if (result == undefined) {
            result = new Color3();
        }

        result.copyFromFloats(r, g, b);

        return result;
    }
    public static Color4(r: number = 0, g: number = 0, b: number = 0, a: number = 0) {
        let result = MathDataPool._Color4.pop();
        if (result == undefined) {
            result = new Color4();
        }

        result.copyFromFloats(r, g, b, a);

        return result;
    }
    public static Vector3(x: number = 0, y: number = 0, z: number = 0) {
        let result = MathDataPool._Vector3.pop();
        if (result == undefined) {
            result = new Vector3();
        }

        result.copyFromFloats(x, y, z);

        return result;
    }
    public static Quaternion(x: number = 0, y: number = 0, z: number = 0, w: number = 1) {
        let result = MathDataPool._Quaternion.pop();
        if (result == undefined) {
            result = new Quaternion();
        }

        result.copyFromFloats(x, y, z, w);

        return result;
    }
    public static Matrix() {
        let result = MathDataPool._Matrix.pop();
        if (result == undefined) {
            result = Matrix.Identity();
        } else {
            Matrix.IdentityToRef(result);
        }

        return result;
    }
    public static Vector4(x: number = 0, y: number = 0, z: number = 0, w: number = 0) {
        let result = MathDataPool._Vector4.pop();
        if (result == undefined) {
            result = new Vector4(x, y, z, w);
        }

        result.copyFromFloats(x, y, z, w);

        return result;
    }
    public static Recycle(item: Color3 | Color4 | Vector3 | Vector4 | Matrix | Quaternion) {
        if (!item) return;
        if (item instanceof Color3) {
            if (MathDataPool._Color3.length > MathDataPool.MAX_CACHE) return;
            MathDataPool._Color3.push(item);
        } else if (item instanceof Color4) {
            if (MathDataPool._Color4.length > MathDataPool.MAX_CACHE) return;
            MathDataPool._Color4.push(item);
        } else if (item instanceof Vector3) {
            if (MathDataPool._Vector3.length > MathDataPool.MAX_CACHE) return;
            MathDataPool._Vector3.push(item);
        } else if (item instanceof Vector4) {
            if (MathDataPool._Vector4.length > MathDataPool.MAX_CACHE) return;
            MathDataPool._Vector4.push(item);
        } else if (item instanceof Quaternion) {
            if (MathDataPool._Quaternion.length > MathDataPool.MAX_CACHE) return;
            MathDataPool._Quaternion.push(item);
        } else if (item instanceof Matrix) {
            if (MathDataPool._Matrix.length > MathDataPool.MAX_CACHE) return;
            MathDataPool._Matrix.push(item);
        }
    }
}