
import * as BABYLON from "../../render/index";
// import * as BABYLON from "../../pi_babylon/render3d/babylon";
import { TShapeArcModeBurstSpread, TShapeArcModeLoop, TShapeArcModePingPong, TShapeArcModeRandom } from "../iparticle_system_config";

/**
 * Shape emitter represents a volume emitting particles.
 * This is the responsibility of the implementation to define the volume shape like cone/sphere/box.
 */
export interface IShapeEmitterType {
    rotateByQuaternionAroundPoint: (position: BABYLON.Vector3, rotateQuat: BABYLON.Quaternion, center: BABYLON.Vector3, diffOffset: BABYLON.Vector3) => void;
    /**
     * Called by the particle System when the direction is computed for the created particle.
     */
    startDirectionFunction(worldMatrix: BABYLON.Matrix, directionToUpdate: BABYLON.Vector3, position: BABYLON.Vector3, localPosition: BABYLON.Vector3, isLocal: boolean): void;

    /**
     * Called by the particle System when the position is computed for the created particle.
     * @param worldMatrix 
     * @param positionToUpdate 
     * @param localPosition 
     * @param emissionLoop 当前发射事件所属的发射循环数
     * @param emissionProgress 当前发射事件的时间进度
     * @param emissionIndex 目标在此次发射中的序号
     * @param esmissionTotal 此次发射的总量
     */
    startPositionFunction(worldMatrix: BABYLON.Matrix, positionToUpdate: BABYLON.Vector3, emissionLoop: number, emissionProgress: number, emissionIndex: number, emissionTotal: number, isLocal: boolean): void;

    /**
     * Returns a string representing the class name
     * @returns a string containing the class name
     */
    getClassName(): string;
    dispose(): void;

    position: BABYLON.Vector3;
    rotation: BABYLON.Vector3;
    scaling: BABYLON.Vector3;
    localMatrix: BABYLON.Matrix;
    alignDirection: boolean;
    randomizeDirection: number;
    spherizeDirection: number;
    randomizePosition: number;
}

export enum EShapeEmitterArcMode {
    /**
     * 弧形周围随机生成粒子
     */
    Random = TShapeArcModeRandom,
    /**
     * 形状的弧形周围依序生成粒子，并在每个周期结束时循环回到起点
     */
    Loop = TShapeArcModeLoop,
    /**
     * 每个连续循环的发生方向与上一个循环相反
     */
    PingPong = TShapeArcModePingPong,
    /**
     * 在形状周围均匀分布粒子生成位置
     */
    BurstsSpread = TShapeArcModeBurstSpread
}
/**
 * 形状发射器创建模式
 */
export enum EShapeEmitterDirectionMode {
    Unity = 0,
}

export enum EBoxShapeMode {
    Volume = 0,
    Shell = 1,
    Edge = 2
}

export function computeRadians(emissionLoop: number, emissionProgress: number, emissionIndex: number, emissionTotal: number, arcTotalValue: number, arcValue: number, arcSpread: number, arcSpeed: number, arcMode: EShapeEmitterArcMode) {
    const arcSpreadLimit = 0.001;
    let s = 0;
    let spread = arcSpread;
    emissionProgress = emissionProgress * arcSpeed;

    switch (arcMode) {
        case (EShapeEmitterArcMode.Loop): {

            let radians = (emissionLoop + emissionProgress) * arcTotalValue;
            let loopCount = Math.floor(radians / arcValue);
            s = radians % arcValue;

            if (spread > arcSpreadLimit) {
                spread = spread * arcValue;

                s = Math.round(s / spread) * spread;
            }
            break;
        }
        case (EShapeEmitterArcMode.PingPong): {
            let radians = (emissionLoop + emissionProgress) * arcTotalValue;
            let loopCount = Math.floor(radians / arcValue);
            s = radians % arcValue;

            if (spread > arcSpreadLimit) {
                spread = spread * arcValue;

                s = Math.round(s / spread) * spread;
            }

            if (loopCount % 2 == 1) {
                s = arcValue - s;
            }

            break;
        }
        case (EShapeEmitterArcMode.BurstsSpread): {

            emissionProgress = emissionIndex / emissionTotal;
            if (spread > arcSpreadLimit) {
                spread = spread * arcValue;

                emissionProgress = Math.round(emissionProgress / spread) * spread;
            }

            s = arcValue * emissionProgress;
            break;
        }
        default: {
            s = 0 + Math.random() * (0 - arcValue);
            break;
        }
    }

    return s;
}