import * as BABYLON from "../render/index";
import { BoxShapeEmitter } from "./emitter/box_shape_emitter";
import { CircleShapeEmitter } from "./emitter/circle_shape_emitter";
import { ConeShapeEmitter } from "./emitter/cone_shape_emitter";
import { EdgeShapeEmitter } from "./emitter/edge_shape_emitter";
import { PointShapeEmitter } from "./emitter/point_shape_emitter";
import { RectangleShapeEmitter } from "./emitter/rectangle_shape_emitter";
import { SphereShapeEmitter } from "./emitter/sphere_shape_emitter";
import { HemisphereShapeEmitter } from "./emitter/hemisphere_shape_emitter";
import { FloatInterpolation, parseFloatInterpolation } from "./interpolation";
import { DefaultValue, IParticleSystemConfig, IShape, TEmissionIndexBursts, TEmissionIndexRate, TParamGravity, TParamLimitVelocityOverLifetime, TParamStartLifetime
    , TParamStartSpeed, TParamTextureSheet, TShapeTypeBox, TShapeTypeCircle, TShapeTypeCone, TShapeTypeEdge, TShapeTypeHemisphere, TShapeTypeRectangle, TShapeTypeSphere
} from "./iparticle_system_config";
import { EBoxShapeMode, IShapeEmitterType } from "./emitter/ishape_emitter_type";
import { MeshParticleSystem } from "./mesh_particle_system";
import { EMeshParticleSpaceMode } from "./particle_system_tool";
import { SizeOverLifetime } from "./modifier/size_over_lifetime";
import { StartSize } from "./modifier/start_size";
import { StartRotation } from "./modifier/start_rotation";
import { StartColor } from "./modifier/start_color";
import { VelocityOverLifetime } from "./modifier/velocity_over_lifetime";
import { ForceOverLifetime } from "./modifier/force_over_lifetime";
import { Color4Interpolate } from "./modifier/base";
import { RotationOverLifetime } from "./modifier/rotation_over_lifetime";
import { GLTF2 } from "../gltf2/lib";

// import { ModelObj } from "../model_obj";

export class PluginMeshParticleSystem {
    public static setup = initial;
}

export function initial() {
    GLTF2.Loader.MeshParticle = (loaders: GLTF2.LoaderRecord, cfg: GLTF2.IParticleSystemConfig, geo: BABYLON.Geometry, mesh: BABYLON.Mesh, index: number) => {
        let mp = new MeshParticleSystem(cfg.name, loaders.scene);
        
        formatMeshParticle(cfg, mp);

        mp.setSourceMesh(mesh);

        // geo.setInstanceBuffer(BABYLON.EInstanceKind.TSR);
        // geo.setInstanceBuffer(BABYLON.EInstanceKind.Color);
        // geo.setInstanceBuffer(BABYLON.EInstanceKind.Tilloff);

        return mp;
    };

    // (<any>ModelObj.prototype).animEndCB = function(this: ModelObj) {
    //     if (this.animOpt) {
    //         const func = this.animOpt.endCall;
    //         this.animOpt = undefined;
    //         if (func) {
    //             let maxWaitTime = 0;
    //             this.particleSysList.forEach((ps: any) => {
    //                 if ((<MeshParticleSystem>ps).psTool && !(<MeshParticleSystem>ps).psTool.looping && (<MeshParticleSystem>ps).psTool.maxLifetimeOverage) {
    //                     maxWaitTime = Math.max((<MeshParticleSystem>ps).psTool.maxLifetimeOverage, maxWaitTime);
    //                 }
    //             });

    //             if (maxWaitTime > 5) {
    //                 setTimer(func, [], maxWaitTime, 1);
    //             } else {
    //                 func();
    //             }
    //         }
    //     }
    // };
}
/**
 * 格式化粒子系统
 * @param config josn描述
 * @param mp 目标粒子系统
 */
export function formatMeshParticle(config: IParticleSystemConfig, mp: MeshParticleSystem) {
    const ps = mp.psTool;

    ps.looping = config.looping == 1;
    ps.duration = config.duration * 1000;
    ps.startDelay = config.startDelay * 1000;
    ps.maxParticles = config.maxParticles;
    ps.prewarm = config.prewarm == 1;
    ps.rateOverTime = config.emission[TEmissionIndexRate];
    ps.simulationSpace = config.simulationSpaceIsWorld ? EMeshParticleSpaceMode.World : EMeshParticleSpaceMode.Local;
    ps.scalingSpace = config.scalingMode;
    ps.renderAlignment = config.renderAlignment;
    ps.renderMode = config.renderMode;
    ps.stretchedLengthScale = config.stretchedLengthScale || 1;
    ps.stretchedVelocityScale = config.stretchedVelocityScale || 0;

    // bursts
    if (config.emission[TEmissionIndexBursts]) {
        const len = config.emission[TEmissionIndexBursts].length;
        for (let i = 0; i < len; i++) {
            ps.bursts[i] = config.emission[TEmissionIndexBursts][i];
        }
    }

    // emitter Shape
    ps.emitterShape = formatShape(config.shape);

    // startLifetime
    parseFloatInterpolation(<FloatInterpolation>ps.startLifetimeInterpolation, config.lifetime, TParamStartLifetime, 1000);

    // startSpeed
    parseFloatInterpolation(<FloatInterpolation>ps.startSpeedInterpolation, config.startSpeed, TParamStartSpeed, 1);

    // startSize
    StartSize.format(config.startSize, ps.startSizeInterpolation);

    // startRotation
    StartRotation.format(config.startRotation, ps.startRotationInterpolation);

    // startColor
    StartColor.format(config.startColor, ps.startColorInterpolation);

    // gravity
    parseFloatInterpolation(<FloatInterpolation>ps.gravityInterpolation.interpolation, config.gravity, TParamGravity, 1.0);

    // velocityOverLifetime
    if (config.velocityOverLifetime) {
        VelocityOverLifetime.format(config.velocityOverLifetime, ps.velocityOverLifetimeInterpolation);
        ps.velocityOverLifetimeInterpolation.isLocalSpace = (config.velocityOverLifetimeIsLocal !== 0);
        ps.enableVelocityOverLifeTime = true;
    }

    if (config.orbtialVelocity) {
        VelocityOverLifetime.format(config.orbtialVelocity, ps.localPositionModifier.orbitalRotateSpeed);
    }
    if (config.orbitalOffset) {
        VelocityOverLifetime.format(config.orbitalOffset, ps.localPositionModifier.orbitalOffset);
    }
    if (config.orbitalRadial) {
        parseFloatInterpolation(<FloatInterpolation>ps.localPositionModifier.radial, config.orbitalRadial, TParamStartSpeed, 1.0);
    }
    if (config.speedModifier) {
        parseFloatInterpolation(<FloatInterpolation>ps.localPositionModifier.speedModifier, config.speedModifier, TParamStartSpeed, 1.0);
    }

    // limitVelocityOverLifetime
    if (config.limitVelocityOverLifetime) {
        parseFloatInterpolation(<FloatInterpolation>ps.limitVelocityOverLifetimeInterpolation.interpolation, config.limitVelocityOverLifetime, TParamLimitVelocityOverLifetime, 1.0);
        ps.limitVelocityOverLifetimeInterpolation.dampen = config.limitVelocityOverLifetimeDampen || DefaultValue.limitVelocityOverLifetimeDampen;
        ps.enableLimitVelocityOverLifeTime = true;
    }

    // forceOverLifetime
    if (config.forceOverLifetime) {
        ForceOverLifetime.format(config.forceOverLifetime, ps.forceOverLifetimeInterpolation);
        ps.forceOverLifetimeInterpolation.isLocalSpace = (config.forceSpaceIsLocal !== 0);
        ps.enableForceOverLifeTime = true;
    }

    // colorOverLifetime
    if (config.colorOverLifetime) {
        Color4Interpolate.format(config.colorOverLifetime, ps.colorOverLifetimeInterpolation);
        ps.enableColorOverLifeTime = true;
    }
    if (config.colorBySpeed) {
        Color4Interpolate.format(config.colorBySpeed[0], ps.colorBySpeedInterpolation);
        ps.colorBySpeedInterpolation.rangeX = config.colorBySpeed[1];
        ps.colorBySpeedInterpolation.rangeY = config.colorBySpeed[2];
        ps.enableColorBySpeed = true;
    }

    // sizeOverLifetime
    if (config.sizeOverLifetime) {
        SizeOverLifetime.format(config.sizeOverLifetime, ps.sizeOverLifetimeInterpolation);
        ps.enableSizeOverLifeTime = true;
    }
    if (config.sizeBySpeed) {
        SizeOverLifetime.format(config.sizeBySpeed[0], ps.sizeBySpeedInterpolation);
        ps.sizeBySpeedInterpolation.rangeX = config.sizeBySpeed[1];
        ps.sizeBySpeedInterpolation.rangeY = config.sizeBySpeed[2];
        ps.enableSizeBySpeed = true;
    }

    // rotationOverLifetime
    if (config.rotationOverLifetime) {
        RotationOverLifetime.format(config.rotationOverLifetime, ps.rotationOverLifetimeInterpolation);
        ps.enableRotationOverLifeTime = true;
    }
    if (config.rotationBySpeed) {
        RotationOverLifetime.format(config.rotationBySpeed[0], ps.rotationBySpeedInterpolation);
        ps.rotationBySpeedInterpolation.rangeX = config.rotationBySpeed[1];
        ps.rotationBySpeedInterpolation.rangeY = config.rotationBySpeed[2];
        ps.enableRotationBySpeed = true;
    }

    if (config.custom1) {
        parseFloatInterpolation(ps.customDataForMainUV.uScale, config.custom1[0], TParamStartLifetime, 1.0);
        parseFloatInterpolation(ps.customDataForMainUV.vScale, config.custom1[1], TParamStartLifetime, 1.0);
        parseFloatInterpolation(ps.customDataForMainUV.uOffset, config.custom1[2], TParamStartLifetime, 1.0);
        parseFloatInterpolation(ps.customDataForMainUV.vOffset, config.custom1[3], TParamStartLifetime, 1.0);
        ps.enableCustomDataForMainUV = true;
    }

    /**
     * 导出时 有 textureSheet 必然对应 导出材质名称 为 SHADER_PI_SHADER_PS
     */
    if (config.textureSheet) {
        const tSheet = ps.textureSheetInterpolation;
        parseFloatInterpolation(tSheet.frameOverTime, config.textureSheet.frameOverTime, TParamTextureSheet, 1.0);
        tSheet.animMode = config.textureSheet.animMode;
        tSheet.customRow = config.textureSheet.customRow;
        tSheet.cycles = config.textureSheet.cycles;
        tSheet.rowMode = config.textureSheet.rowMode;
        parseFloatInterpolation(tSheet.startFrame, config.textureSheet.startFrame, TParamTextureSheet, 1.0);
        tSheet.tilesX = config.textureSheet.tilesX;
        tSheet.tilesY = config.textureSheet.tilesY;
        tSheet.timeMode = config.textureSheet.timeMode;
        ps.enableTextureSheet = true;
    }

    if (config.trail) {
        // ps.trail = new TrailModifier(mp.getScene());
        // ps.trail.mode = config.trail.mode;
        // ps.trail.ratio = config.trail.ratio;
        // parseFloatInterpolation(ps.trail.lifetime, config.trail.lifetime, null, 1.0);
        // ps.trail.ribbonCount = config.trail.ribbonCount;
        // ps.trail.attachRibbonsToTransfoem = config.trail.attachRTT == 1;
        // ps.trail.minimunVertexDistance = config.trail.minDist;
        // ps.trail.worldSpace = config.trail.worldSpace == 1;
        // ps.trail.dieWithParticle = config.trail.dieWith == 1;
        // ps.trail.textureMode = config.trail.texMode;
        // ps.trail.sizeAffectsWidth = config.trail.sizeAWidth == 1;
        // ps.trail.sizeAffectsLifetime = config.trail.sizeALifetime == 1;
        // ps.trail.inheritParticleColor = config.trail.inheritColor == 1;
        // parseColor4Gradient(ps.trail.colorOverLifetime.gradient, config.trail.colorOverLife, null);
        // parseFloatInterpolation(ps.trail.widthOverTrail, config.trail.widthOverTrail, null, 1.0);
        // parseColor4Gradient(ps.trail.colorOverTrail.gradient, config.trail.colorOverTrail, null);
        // ps.enableTrail = true;
    }

    if (config.renderPivot) {
        ps.renderPivot.copyFromFloats(...config.renderPivot);
    }
}
/**
 *
 * @param shape 形状发射器json描述
 * @returns
 */
function formatShape(shape: IShape | undefined): IShapeEmitterType {
    if (!shape) {
        return new PointShapeEmitter();
    }
    else {
        let shapeEmitter: IShapeEmitterType;
        switch (shape.type) {
            case TShapeTypeCircle: {
                const temp = new CircleShapeEmitter(shape.radius, shape.radiusThickness);
                temp.arcMode = shape.arc.mode;
                temp.arcValue = shape.arc.value * Math.PI / 180;
                temp.arcSpread = shape.arc.spread;
                temp.arcSpeed = shape.arc.speed;
                shapeEmitter = <IShapeEmitterType>temp;
                break;
            }
            case TShapeTypeCone: {
                const temp = new ConeShapeEmitter(shape.radius, shape.angle * Math.PI / 180);
                temp.radiusRange = shape.radiusThickness;
                temp.height = shape.height;
                temp.heightRange = shape.emitAsVolume ? 1.0 : 0.0;
                temp.arcMode = shape.arc.mode;
                temp.arcValue = shape.arc.value * Math.PI / 180;
                temp.arcSpread = shape.arc.spread;
                temp.arcSpeed = shape.arc.speed;
                shapeEmitter = <IShapeEmitterType>temp;
                break;
            }
            case TShapeTypeHemisphere: {
                const temp = new HemisphereShapeEmitter(shape.radius, shape.radiusThickness);
                temp.arcMode = shape.arc.mode;
                temp.arcValue = shape.arc.value * Math.PI / 180;
                temp.arcSpread = shape.arc.spread;
                temp.arcSpeed = shape.arc.speed;
                shapeEmitter = <IShapeEmitterType>temp;
                break;
            }
            case TShapeTypeEdge: {
                const temp = new EdgeShapeEmitter();
                temp.size = shape.radius;
                temp.arcMode = shape.arc.mode;
                temp.arcValue = shape.arc.value * Math.PI / 180;
                temp.arcSpread = shape.arc.spread;
                temp.arcSpeed = shape.arc.speed;
                shapeEmitter = <IShapeEmitterType>temp;
                break;
            }
            case TShapeTypeSphere: {
                const temp = new SphereShapeEmitter(shape.radius, shape.radiusThickness);
                temp.arcMode = shape.arc.mode;
                temp.arcValue = shape.arc.value * Math.PI / 180;
                temp.arcSpread = shape.arc.spread;
                temp.arcSpeed = shape.arc.speed;
                shapeEmitter = <IShapeEmitterType>temp;
                break;
            }
            case TShapeTypeBox: {
                const temp = new BoxShapeEmitter();
                temp.emitMode = shape.boxEmitMode || EBoxShapeMode.Volume;
                shapeEmitter = <IShapeEmitterType>temp;
                break;
            }
            case TShapeTypeRectangle: {
                const temp = new RectangleShapeEmitter();
                shapeEmitter = <IShapeEmitterType>temp;
                break;
            }
        }

        if (shape.position && shape.rotation && shape.scale) {
            shapeEmitter.position.copyFromFloats(...shape.position);
            shapeEmitter.rotation.copyFromFloats(...shape.rotation);
            shapeEmitter.scaling.copyFromFloats(...shape.scale);
            // Unity Shape 旋转是绕轴 顺时针为正
            shapeEmitter.rotation.scaleInPlace(3.14159265359 / 180);
            BABYLON.Quaternion.FromEulerVectorToRef(shapeEmitter.rotation, BABYLON.TmpVectors.Quaternion[0]);
            BABYLON.Matrix.ComposeToRef(shapeEmitter.scaling, BABYLON.TmpVectors.Quaternion[0], shapeEmitter.position, shapeEmitter.localMatrix);
        }

        shapeEmitter.alignDirection = !!shape.alignDir;

        //
        if (shape.randomize) {
            shapeEmitter.randomizeDirection = shape.randomize[0];
            shapeEmitter.spherizeDirection = shape.randomize[1];
            shapeEmitter.randomizePosition = shape.randomize[2];
        }

        return shapeEmitter;
    }
}