import * as BABYLON from "../render/index";
// import * as BABYLON from "../pi_babylon/render3d/babylon";
import { Color4Gradient, FloatInterpolation, IInterpolation, UpdateRandom } from "./interpolation";
import { EInterpolationCurveMode } from "./iparticle_system_config";
import { IShapeEmitterType } from "./emitter/ishape_emitter_type";
import { IParticleModifier, TempVector3A, TempVector3B, TempVector3C, TempVector3E } from "./modifier/base";
import { ColorOverLifetime } from "./modifier/color_over_lifetime";
import { ForceOverLifetime } from "./modifier/force_over_lifetime";
import { Gravity } from "./modifier/gravity";
import { LimitVelocityOverLifetime } from "./modifier/limit_velocity_over_lifetime";
import { LocalPosition } from "./modifier/local_position";
import { RotationOverLifetime } from "./modifier/rotation_over_lifetime";
import { SizeOverLifetime } from "./modifier/size_over_lifetime";
import { StartColor } from "./modifier/start_color";
import { StartRotation } from "./modifier/start_rotation";
import { StartSize } from "./modifier/start_size";
import { VelocityOverLifetime } from "./modifier/velocity_over_lifetime";
import { Particle } from "./particle";
import { Pool } from "./pool";
import { TextureSheet } from "./modifier/texture_sheet";
import { HorizontalBillboard, RenderAlignmentFacing, RenderAlignmentLocal, RenderAlignmentVelocity, RenderAlignmentView, RenderAlignmentWorld, StretchedBillboard, TRenderWorldMatrixCompute, VerticalBillboard } from "./modifier/render_world_matrix";
import { Noise } from "./modifier/noise";
import { TrailModifier } from "./modifier/trail";
import { ColorBySpeed } from "./modifier/color_by_speed";
import { SizeBySpeed } from "./modifier/size_by_speed";
import { RotationBySpeed } from "./modifier/rotation_by_speed";
import { directionToQuaternion } from "./math";
import { CustomDataForUV } from "./modifier/custom_data_for_uv";

export enum EMeshParticleSpaceMode {
    Local = 0,
    /**
     * 发射在世界空间时, 父级尽量不要有旋转动画, 因为 动画 与 粒子动画的衔接有误差，无法完美适配
     */
    World = 1,
}

export enum EMeshParticleScaleMode {
    Hierarchy = 0,
    Local = 1,
    Shape = 2,
}

export enum ERenderAlignment {
    /**
     * 粒子面向相机平面。
     */
    View    = 0,
    /**
     * 粒子与世界轴对齐
     */
    World   = 1,
    /**
     * Local 无需额外适配
     */
    Local   = 2,
    /**
     * 粒子面向相机的位置点
     */
    Facing = 3,
    /**
     * 粒子朝向它们的速度方向。
     */
    Velocity = 4
}

export enum ERenderMode {
    Billboard    = 0,
    StretchedBillboard   = 1,
    HorizontalBillboard   = 2,
    VerticalBillboard   = 3,
    Mesh = 4,
    None = 5,
}

/**
 * 对接 Unity ParticleSystem Mesh Mode
 */
export class ParticleSystemTool {

    public static sqrt3 = Math.sqrt(3);

    private static OneVector3 = BABYLON.Vector3.One();
    /**
     * 计算的间隔时间
     */
    public computeDeltaTime = 15;
    /**
     * 最大粒子数
     */
     public maxParticles: number = 100;
    // private _maxParticles: number = 100;
    // public get maxParticles() {
    //     return this._maxParticles;
    // }
    // public set maxParticles(value: number) {
    //     this._maxParticles = value + 1;
    // }
    /**
     * 是否循环
     */
    public looping: boolean = true;
    /**
     * 是否预热
     */
    public prewarm: boolean = false;
    /**
     * 粒子系统发射持续时间
     */
    public duration: number = 5.0;
    private _loopCount: number = 0;
    /**
     * 延时时间 - 毫秒
     */
    public startDelay: number = 0;
    public emitterShape: IShapeEmitterType | undefined;
    /**
     * 一次发射循环的时间 - 毫秒
     */
    public emissionTime: number = 1000;
    /**
     * 当前已循环次数
     */
    public emissionLoop: number = 0;
    /**
     * 当前发射循环的进度
     */
    public emissionProgress: number = 0;
    /**
     * 每秒发射的粒子数目
     */
    public rateOverTime: number = 10;
    /**
     * 指定时间点 开始 持续 指定时间内 发射指定数目
     * [开始时间, 发射数目, 循环次数, 间隔]
     * @tip 需要按 开始时间从小到大排序
     */
    public bursts: [number, number, number, number][] = [];
    private _burstsLoopCount: number[] = [];

    /**
     * 发射空间
     * * Local - 本地: 发射开始时的方向受父级影响, 发射后的位置也受父级影响
     * * World - 世界空间: 发射开始时的方向受父级影响, 发射后的位置不再受父级影响
     */
    public simulationSpace: EMeshParticleSpaceMode = EMeshParticleSpaceMode.Local;
    public scalingSpace: EMeshParticleScaleMode = EMeshParticleScaleMode.Hierarchy;
    private _renderAlignment: ERenderAlignment = ERenderAlignment.Local;
    public set renderAlignment(value: ERenderAlignment) {
        this._renderAlignment = value;
        this._updateParticleWorldMatrixCompute();
    }
    public get renderAlignment() {
        return this._renderAlignment;
    }

    private _renderMode: ERenderMode = ERenderMode.Billboard;
    public set renderMode(value: ERenderMode) {
        this._renderMode = value;
        this._updateParticleWorldMatrixCompute();
    }
    public get renderMode() {
        return this._renderMode;
    }

    public renderPivot: BABYLON.Vector3 = BABYLON.Vector3.Zero();

    public startLifetimeInterpolation: FloatInterpolation = new FloatInterpolation();
    public startSpeedInterpolation: FloatInterpolation = new FloatInterpolation();
    public startSizeInterpolation: StartSize = new StartSize(new FloatInterpolation(), new FloatInterpolation(), new FloatInterpolation());
    public startRotationInterpolation: StartRotation = new StartRotation(new FloatInterpolation(), new FloatInterpolation(), new FloatInterpolation());
    public startColorInterpolation: StartColor = new StartColor(new Color4Gradient());

    public gravityInterpolation: Gravity = new Gravity(new FloatInterpolation());
    /**
     * 位移速度
     * @tip 默认模式 - ConstantsUnite - XYZ 共用一个随机数
     * @tip 为 Constants | ConstantsUnite 模式应用插值时, 仅创建时使用, 意义为: 创建时速度基础上附加速度
     * @tip 为 Curves 模式时, 运行过程中使用, 意义为: 创建时速度基础上附加速度 - 每次循环累加 此次循环 与 上次循环 速度值的差值
     */
    public velocityOverLifetimeInterpolation: VelocityOverLifetime = new VelocityOverLifetime(new FloatInterpolation(), new FloatInterpolation(), new FloatInterpolation());
    public enableVelocityOverLifeTime = false;

    public limitVelocityOverLifetimeInterpolation: LimitVelocityOverLifetime = new LimitVelocityOverLifetime(new FloatInterpolation());
    public enableLimitVelocityOverLifeTime = false;

    public forceOverLifetimeInterpolation: ForceOverLifetime = new ForceOverLifetime(new FloatInterpolation(), new FloatInterpolation(), new FloatInterpolation());
    public enableForceOverLifeTime = false;

    public colorOverLifetimeInterpolation: ColorOverLifetime = new ColorOverLifetime(new Color4Gradient());
    public enableColorOverLifeTime = false;
    public colorBySpeedInterpolation: ColorBySpeed = new ColorBySpeed(new Color4Gradient());
    public enableColorBySpeed = false;
    /**
     * 缩放
     * @tip 默认模式 - ConstantsUnite - XYZ 共用一个随机数
     * @tip 为 Constants | ConstantsUnite 模式应用插值时, 仅创建时使用, 意义为: 仅对创建时大小的影响倍数
     * @tip 为 Curves 模式时, 运行过程中使用, 意义为: 创建时大小的影响倍数
     */
    public sizeOverLifetimeInterpolation: SizeOverLifetime = new SizeOverLifetime(new FloatInterpolation(), new FloatInterpolation(), new FloatInterpolation());
    public enableSizeOverLifeTime = false;
    public sizeBySpeedInterpolation: SizeBySpeed = new SizeBySpeed(new FloatInterpolation(), new FloatInterpolation(), new FloatInterpolation());
    public enableSizeBySpeed = false;

    public localPositionModifier: LocalPosition = new LocalPosition();
    /**
     * 旋转
     * @tip 默认模式 - ConstantsUnite - XYZ 共用一个随机数
     * @tip Unity 中单一 Angle 对应 只处理 Y 轴
     * @tip 与 Unity 编辑时不同, 此处单位为 弧度
     * @tip 为 Constants | ConstantsUnite 模式时, 仅创建时使用, 意义为: 每秒旋转速度
     * @tip 为 Curves 模式时, 运行过程中使用, 意义为: 每秒旋转速度
     */
    public rotationOverLifetimeInterpolation: RotationOverLifetime = new RotationOverLifetime(new FloatInterpolation(), new FloatInterpolation(), new FloatInterpolation());
    public enableRotationOverLifeTime = false;
    public rotationBySpeedInterpolation: RotationBySpeed = new RotationBySpeed(new FloatInterpolation(), new FloatInterpolation(), new FloatInterpolation());
    public enableRotationBySpeed = false;
    /**
     * Texture Sheet Animation
     */
    public textureSheetInterpolation: TextureSheet = new TextureSheet(new FloatInterpolation(), new FloatInterpolation());
    public enableTextureSheet: boolean = false;

    public customDataForMainUV: CustomDataForUV = new CustomDataForUV();
    public enableCustomDataForMainUV: boolean = false;

    public noiseModifier: Noise = new Noise();
    public enableNoise: boolean = false;

    public stretchedVelocityScale = 0;
    public stretchedLengthScale = 1;

    // 局部空间数据 - 每粒子
    private overLifetimeModifierList: IParticleModifier[] = [];

    private particleList: Particle[] = [];
    private activeParticleList: Particle[] = [];

    private _particleSystemAge: number = 0;

    private _lastTime: number = 0;
    private _deltaTime: number = 0;
    private _lastCreateTime: number = 0;

    private _isPlaying: boolean = false;
    private _isComputeable: boolean = false;

    private scene: BABYLON.Scene;
    public trail: TrailModifier;
    private _enableTrail: boolean = false;
    public set enableTrail(value: boolean) {
        this._enableTrail = value;
        if (this.trail) {
            this.trail.enable = value;
            // Trail
            this._trailLifetimeScale = 0;
            if (value && !this.trail.dieWithParticle) {
                this._trailLifetimeScale = 1;
            }
        }
    }
    public get enableTrail() {
        return this._enableTrail;
    }

    private _trailLifetimeScale: number = 0;

    public particleWorldMatrixCompute: TRenderWorldMatrixCompute = RenderAlignmentLocal.compute;

    public getCameraRotationMatrixInvert: () => BABYLON.Matrix = BABYLON.Matrix.Identity;
    public getCameraPosition: () => BABYLON.Vector3 = BABYLON.Vector3.Zero;
    public getParentWorldMatrix: () => BABYLON.Matrix = BABYLON.Matrix.Identity;
    public getWorldMatrix: () => BABYLON.Matrix = BABYLON.Matrix.Identity;
    public getLocalMatrix: () => BABYLON.Matrix = BABYLON.Matrix.Identity;
    public getCameraMatrix: () => BABYLON.Matrix = BABYLON.Matrix.Identity;

    private _isDisposed: boolean = false;

    constructor(scene: BABYLON.Scene) {
        this.scene = scene;
    }

    public dispose() {
        if (this._isDisposed) {
            return;
        }
        this.stop();

        this.trail && this.trail.dispose();
        this.trail = undefined;

        this._mpMatrixList = undefined;
        this._mpColorData = undefined;

        this.startLifetimeInterpolation             && this.startLifetimeInterpolation            .dispose();
        this.startSpeedInterpolation                && this.startSpeedInterpolation               .dispose();

        this.bursts                 .length = 0,    // this.bursts                 = undefined;
        this._isDisposed = true;

        this.activeParticleList.length = 0;

        let len = this.particleList.length;
        for (let i = len - 1; i >= 0; i--) {
            Pool.recycleParticle(this.particleList[i]);
        }
        this.particleList.length = 0;

        return;
    }

    private _mpMatrixList: Float32Array | undefined;
    private _mpColorData: Float32Array | undefined;
    private _mpUVData: Float32Array | undefined;

    public get mpMatrixList() {
        return this._mpMatrixList;
    }
    public get mpColorData() {
        return this._mpColorData;
    }
    public get mpUVData() {
        return this._mpUVData;
    }

    private _maxId: number = 0;
    public build() {
        this._mpMatrixList = new Float32Array(16 * this.maxParticles);
        this._mpColorData = new Float32Array(4 * this.maxParticles);
        this._mpUVData = new Float32Array(4 * this.maxParticles);
        // this._mpUVSheetData = new Float32Array(4 * this.maxParticles);

        for (let i = this._maxId; i < this.maxParticles; i++) {
            this.particleList.push(Pool.createParticle());
        }
        this._maxId = this.maxParticles;
    }

    public start() {
        if (!this._isPlaying) {
            if (this.startDelay > 0 && !this.prewarm) {
                setTimeout(this._startCall, this.startDelay);
            }
            else {
                this._startCall();
            }
        }

    }

    private _startCall = () => {
        if (this._isDisposed) return;

        this._isPlaying = true;
        this._isComputeable = true;
        this._particleSystemAge = 0;
        this._currTime = 0;
        this._lastTime = 0;
        this._lastCreateTime = 0;

        let count = this.particleList.length;
        for (let i = 0; i < count; i++) {
            this.particleList[i].reset();
        }

        if (this.prewarm) {
            let deltaTime = this.computeDeltaTime;
            let counter = this.duration / deltaTime;

            this._lastTime = 0 - this.duration;
            this._lastCreateTime = this._lastTime;

            this._currTime = 0 - this.duration;

            for (let i = 1; i <= counter; i++) {
                this.varCompute(deltaTime);
            }
        }

        if (this.trail) {
            this.trail.enable = this._enableTrail;
        }
    }

    public stop() {
        this._isComputeable = false;

        if (!this._isDisposed) {
            this._stop();
        }

        this._isPlaying = false;
    }

    private _stop() {
        if (this._isPlaying) {
            this._burstsLoopCount.length = 0;
        }

        this._isPlaying = false;
        if (this.trail) {
            this.trail.enable = false;
        }
    }

    public mpUpdate(): number {
        return this._update();
    }

    private _updateParticleWorldMatrixCompute() {
        switch (this.renderMode) {
            case (ERenderMode.StretchedBillboard): {
                this.particleWorldMatrixCompute = StretchedBillboard.compute;
                break;

            }
            case (ERenderMode.HorizontalBillboard): {
                this.particleWorldMatrixCompute = HorizontalBillboard.compute;
                break;
            }
            case (ERenderMode.VerticalBillboard): {
                this.particleWorldMatrixCompute = VerticalBillboard.compute;
                break;
            }
            default: {
                switch (this.renderAlignment) {
                    case (ERenderAlignment.View): {
                        this.particleWorldMatrixCompute = RenderAlignmentView.compute;
                        break;
                    }
                    case (ERenderAlignment.World): {
                        this.particleWorldMatrixCompute = RenderAlignmentWorld.compute;
                        break;
                    }
                    case (ERenderAlignment.Facing): {
                        this.particleWorldMatrixCompute = RenderAlignmentFacing.compute;
                        break;
                    }
                    case (ERenderAlignment.Velocity): {
                        this.particleWorldMatrixCompute = RenderAlignmentVelocity.compute;
                        break;
                    }
                    case (ERenderAlignment.Local): {
                        this.particleWorldMatrixCompute = RenderAlignmentLocal.compute;
                        break;
                    }
                }
                break;
            }
        }
    }

    public static TempVector3Scaling = BABYLON.Vector3.Zero();
    public static TempVector3Translation = BABYLON.Vector3.Zero();
    public static TempVector3Rotation = BABYLON.Vector3.Zero();
    public static TempQuaternion_0 = BABYLON.Quaternion.Zero();
    public static TempMatrix_1 = BABYLON.Matrix.Zero();

    private _start_p_s = BABYLON.Vector3.Zero();
    private _start_p_r = BABYLON.Quaternion.Identity();
    private _start_p_p = BABYLON.Vector3.Zero();

    private _start_l_s = BABYLON.Vector3.Zero();
    private _start_l_r = BABYLON.Quaternion.Identity();
    private _start_l_p = BABYLON.Vector3.Zero();

    private _maxLifetimeOverage = 0;
    public get maxLifetimeOverage() {
        return this._maxLifetimeOverage;
    }

    private _formatStartInfo(parentWM: BABYLON.Matrix, localWM: BABYLON.Matrix, resultStartMatrix: BABYLON.Matrix) {
        // 分解出 TRS
        if (parentWM) {
            parentWM.decompose(this._start_p_s, this._start_p_r, this._start_p_p);
            // console.log("GloblPos ", this._start_p_p);
        }
        else {
            this._start_p_p.x = 0, this._start_p_p.y = 0, this._start_p_p.z = 0;
            this._start_p_r.x = 0, this._start_p_r.y = 0, this._start_p_r.z = 0, this._start_p_r.w = 1;
            this._start_p_s.x = 1, this._start_p_s.y = 1, this._start_p_s.z = 1;
        }
        localWM.decompose(this._start_l_s, this._start_l_r, this._start_l_p);

        // if (this.renderAlignment == ERenderAlignment.View) {
        //     this._start_p_r.copyFromFloats(0, 0, 0, 1);
        //     this._start_l_r.copyFromFloats(0, 0, 0, 1);
        // }
        // 忽略粒子节点上的本地旋转 - 仅发射后的位移受此影响
        if (this.renderAlignment == ERenderAlignment.World) {
            this._start_l_r.copyFromFloats(0, 0, 0, 1);
        }
        else {
            //
        }
        if (this.scalingSpace == EMeshParticleScaleMode.Hierarchy) {
            //
        }
        else if (this.scalingSpace == EMeshParticleScaleMode.Local) {
            this._start_p_s.copyFromFloats(1, 1, 1);
        }
        else {
            this._start_p_s.copyFromFloats(1, 1, 1);
            this._start_l_s.copyFromFloats(1, 1, 1);
        }
        BABYLON.Matrix.ComposeToRef(this._start_p_s, this._start_p_r, this._start_p_p, resultStartMatrix);
        BABYLON.Matrix.ComposeToRef(this._start_l_s, this._start_l_r, this._start_l_p, ParticleSystemTool.TempMatrix_1);
        ParticleSystemTool.TempMatrix_1.multiplyToRef(resultStartMatrix, resultStartMatrix);
    }

    private _update(): number {

        if (!this._mpColorData || !this._mpMatrixList) {
            return 0;
        }

        // const cameraWM = this.getCameraMatrix();
        const cameraRotationMatrix = this.getCameraRotationMatrixInvert();
        const cameraGlobalPos = this.getCameraPosition();
        let _start_parent_m = this.getParentWorldMatrix();
        let _start_local_m = this.getLocalMatrix();
        let startWorldMatrix = this.getWorldMatrix();
        let emitWorldMatrix = BABYLON.Matrix.Identity();
        this._formatStartInfo(_start_parent_m, _start_local_m, emitWorldMatrix);

        let realStartMatrix = BABYLON.Matrix.Identity();
        let _start_w_m = BABYLON.Matrix.Identity();
        let tempStartMatrix = BABYLON.Matrix.Identity();

        let activeCount = 0;
        // if (this.renderMode != ERenderMode.None) {
            const len = this.activeParticleList.length;
            for (let i = 0; i < len; i++) {
                let particle = this.activeParticleList[i];
                if (particle.age > particle.lifetime + this._trailLifetimeScale * particle.trialLifetime) {
                    continue;
                }
                const ii = activeCount * 4;

                const _color            = particle.color;
                const _uv               = particle.uv;
                /**
                 * 仅 Pi/Simple 导出为 带纹理动画的粒子
                 */
                this._mpColorData[ii + 0] = _color.r;
                this._mpColorData[ii + 1] = _color.g;
                this._mpColorData[ii + 2] = _color.b;
                this._mpColorData[ii + 3] = _color.a;
                
                this._mpUVData[ii + 0] = _uv[0];
                this._mpUVData[ii + 1] = _uv[1];
                this._mpUVData[ii + 2] = _uv[2];
                this._mpUVData[ii + 3] = _uv[3];

                if (this.simulationSpace == EMeshParticleSpaceMode.Local) {
                    realStartMatrix.copyFrom(emitWorldMatrix);
                    _start_w_m.copyFrom(startWorldMatrix);
                }
                else {
                    realStartMatrix.copyFrom(particle.emitWorldMatrix);
                    _start_w_m.copyFrom(particle.startWorldMatrix);
                }

                ParticleSystemTool.TempVector3Rotation.copyFrom(particle.rotation);
                if (this.renderMode === ERenderMode.Billboard) {
                    ParticleSystemTool.TempVector3Rotation.z *= -1;
                }
                if (this.emitterShape.alignDirection) {
                    ParticleSystemTool.TempVector3Rotation.addInPlace(particle.emitRotation);
                }

                let pivot = TempVector3E;
                particle.scaling.multiplyToRef(this.renderPivot, pivot);
                pivot.addInPlace(particle.position);
                this.particleWorldMatrixCompute(
                    _start_w_m,
                    realStartMatrix,
                    tempStartMatrix,
                    cameraRotationMatrix,
                    cameraGlobalPos,
                    particle.readldirection,
                    particle.directionLength,
                    pivot,
                    particle.scaling,
                    ParticleSystemTool.TempVector3Rotation,
                    this.stretchedLengthScale,
                    this.stretchedVelocityScale
                );
                if (this.trail && this.enableTrail) {
                    let viewDirection = TempVector3C;
                    viewDirection.copyFromFloats(0, 0, 1);
                    BABYLON.Vector3.TransformCoordinatesToRef(viewDirection, cameraRotationMatrix, viewDirection);
                    tempStartMatrix.decompose(undefined, undefined, ParticleSystemTool.TempVector3Translation);
                    this.trail.recordTrailPath(viewDirection, tempStartMatrix, ParticleSystemTool.TempVector3Translation, particle.position, particle);
                }
                if (this.renderMode != ERenderMode.None) {
                    tempStartMatrix.copyToArray(this._mpMatrixList, activeCount * 16);
                    activeCount++;
                }
            }
        // }

        if (this.trail && this.enableTrail) {
            this.trail.build();
        }

        return activeCount;
    }

    private _currTime: number = 16;
    // 变量计算 - 异步计算
    public varCompute = (deltaTime: number) => {
        const now = this._currTime + deltaTime;
        this._currTime = now;
        const delta = this._currTime - this._lastTime;
        if (delta < this.computeDeltaTime) {
            return;
        }

        this.activeParticleList.length = 0;

        if (this._isPlaying && this._isComputeable) {

            this._lastTime = now;
            this._particleSystemAge += delta;
            this._deltaTime = delta / 1000;

            const scaleUpdateSpeed = Math.min(this._deltaTime, 50 / 1000);

            const createDelta = now - this._lastCreateTime;

            // this.emitRate / this.duration;
            const tempLoop = Math.floor(this._particleSystemAge / this.duration);

            let canNew = true;
            if (!this.looping) {
                if (this._particleSystemAge > this.duration) {
                    // 等待已发射粒子的消失
                    if (this._maxLifetimeOverage <= 0) {
                        this.stop();
                        return;
                    }
                    canNew = false;
                }
            }
            else {
                this._maxLifetimeOverage = 0;
            }

            const localTimeDiff = this._particleSystemAge % this.duration;
            let burstCreateCount = 0;

            this.emissionLoop = Math.floor(this._particleSystemAge / this.emissionTime);
            this.emissionProgress = this._particleSystemAge % this.emissionTime / this.emissionTime;

            const burstsCount = this.bursts.length;
            // 新的一轮循环
            if (this._loopCount < tempLoop) {
                // 剩余 Bursts 全部激活
                for (let i = 0 ; i < burstsCount; i++) {
                    const [burstTime, burstCount, burstLoop, interval] = this.bursts[i];

                    if (this.duration > burstTime) {
                        if (!this._burstsLoopCount[i]) {
                            this._burstsLoopCount[i] = 0;
                        }

                        const tempBurstLoop = Math.floor((this.duration - burstTime) / interval);

                        // 预估爆发次数 大于 已爆发次数
                        if (tempBurstLoop > this._burstsLoopCount[i]) {
                            // 可爆发次数 大于 已爆发次数
                            while (burstLoop > this._burstsLoopCount[i]) {
                                this._burstsLoopCount[i] += 1;
                                burstCreateCount += burstCount;
                            }
                        }
                    }

                    this._burstsLoopCount[i] = 0;
                }

                this._loopCount = tempLoop;
            }

            for (let i = 0 ; i < burstsCount; i++) {
                const [burstTime, burstCount, burstLoop, interval] = this.bursts[i];
                if (localTimeDiff > burstTime) {
                    if (!this._burstsLoopCount[i]) {
                        this._burstsLoopCount[i] = 0;
                    }

                    const tempBurstLoop = Math.floor((localTimeDiff - burstTime) / interval);

                    // 预估爆发次数 大于 已爆发次数
                    if (tempBurstLoop > this._burstsLoopCount[i]) {
                        // 可爆发次数 大于 已爆发次数
                        if (burstLoop > this._burstsLoopCount[i]) {
                            this._burstsLoopCount[i] += 1;
                            burstCreateCount += burstCount;
                        }
                    }
                }
            }

            let newCount = 0;
            if (this.rateOverTime > 0) {
                let perRateNeedTime = 1000 / this.rateOverTime;
                newCount = Math.floor((createDelta + perRateNeedTime * 0.5) * this.rateOverTime / 1000);
                if (newCount > 0) {
                    this._lastCreateTime += newCount * perRateNeedTime;
                }

            } else {
                this._lastCreateTime = now;
            }
            newCount += burstCreateCount;

            if (!canNew) {
                newCount = 0;
            }

            let newParticleList: Particle[] = [];
            let deltaAge = 0;
            for (let i = 0; i < this.maxParticles; i++) {
                let particle = this.particleList[i];

                if (this.trail && this._enableTrail) {
                    deltaAge = particle.trialLifetime * this._trailLifetimeScale;
                }
                if (particle.age - particle.lifetime >= deltaAge) {
                    if (newCount > 0) {
                        newParticleList.push(particle);
                        this.activeParticleList.push(particle);
                        newCount--;
                    }
                } else {
                    this.activeParticleList.push(particle);
                }

            }
            this.activeParticleList.sort(this.activeParticlesSort);

            const emitWorldMatrix = BABYLON.Matrix.Identity();
            const _startWorldRM = BABYLON.Matrix.Identity();
            let parentWorldMatrix = this.getParentWorldMatrix();
            let worldMatrix = this.getWorldMatrix();
            let localMatrix = this.getLocalMatrix();

            worldMatrix.getRotationMatrixToRef(_startWorldRM);
            _startWorldRM.invertToRef(_startWorldRM);

            this._formatStartInfo(parentWorldMatrix, localMatrix, emitWorldMatrix);

            this.applyModifier();
            this.varInit(newParticleList, worldMatrix, _startWorldRM, emitWorldMatrix);
            this.varPre(this.activeParticleList, delta, scaleUpdateSpeed, worldMatrix, _startWorldRM, emitWorldMatrix);

            // if (this.trail && this.enableTrail) {
            //     this.trail.modify(this.getCameraMatrix().getTranslation(), this.activeParticleList);
            // }
        }
    }

    private applyModifier() {
        this.overLifetimeModifierList.length = 0;
        if (this.enableColorOverLifeTime) {
            this.overLifetimeModifierList.push(this.colorOverLifetimeInterpolation);
        }
        if (this.enableForceOverLifeTime) {
            this.overLifetimeModifierList.push(this.forceOverLifetimeInterpolation);
        }
        if (this.enableRotationOverLifeTime) {
            this.overLifetimeModifierList.push(this.rotationOverLifetimeInterpolation);
        }
        if (this.enableSizeOverLifeTime) {
            this.overLifetimeModifierList.push(this.sizeOverLifetimeInterpolation);
        }
        if (this.enableVelocityOverLifeTime) {
            this.overLifetimeModifierList.push(this.velocityOverLifetimeInterpolation);
        }

        if (this.enableNoise) {
            this.overLifetimeModifierList.push(this.noiseModifier);
        }

        this.overLifetimeModifierList.push(this.localPositionModifier);

        if (this.enableColorBySpeed) {
            this.overLifetimeModifierList.push(this.colorBySpeedInterpolation);
        }
        if (this.enableRotationBySpeed) {
            this.overLifetimeModifierList.push(this.rotationBySpeedInterpolation);
        }
        if (this.enableSizeBySpeed) {
            this.overLifetimeModifierList.push(this.sizeBySpeedInterpolation);
        }

        if (this.enableLimitVelocityOverLifeTime) {
            this.overLifetimeModifierList.push(this.limitVelocityOverLifetimeInterpolation);
        }
        if (this.enableTextureSheet) {
            this.overLifetimeModifierList.push(this.textureSheetInterpolation);
        }
        if (this.enableCustomDataForMainUV) {
            this.overLifetimeModifierList.push(this.customDataForMainUV);
        }
    }

    /**
     * 初始化
     */
    public varInit(newParticleList: Particle[], worldMatrix: BABYLON.Matrix, startWorldMatrixInvert: BABYLON.Matrix, emitWorldMatrix: BABYLON.Matrix) {
        let newCount = newParticleList.length;

        const progress = this._particleSystemAge % this.duration / this.duration;

        for (let i = 0; i < newCount; i++) {
            let particle = newParticleList[i];
            particle.reset();

            particle.globalRecordList.length = 0;
            particle.localRecordList.length = 0;
            particle.startWorldMatrix.copyFrom(worldMatrix);
            particle.startWorldMatrixInvert.copyFrom(startWorldMatrixInvert);
            particle.emitWorldMatrix.copyFrom(emitWorldMatrix);

            this.initNew(particle, progress, this.emissionLoop, this.emissionProgress, i, newCount);
        }
    }

    public initNew(particle: Particle, progress: number, emissionLoop: number, emissionProgress: number, emissionIndex: number, emissionTotal: number) {
        UpdateRandom();

        let startSpeed = this.startSpeedInterpolation.interpolate(progress, Math.random());
        let startLocalPosition = TempVector3A;
        (<IShapeEmitterType >this.emitterShape).startPositionFunction(particle.startWorldMatrix, startLocalPosition, emissionLoop, emissionProgress, emissionIndex, emissionTotal, true);

        let startDirection = TempVector3B;
        (<IShapeEmitterType >this.emitterShape).startDirectionFunction(particle.startWorldMatrix, startDirection, undefined, startLocalPosition, true);

        // 发射器局部矩阵影响
        BABYLON.Vector3.TransformCoordinatesToRef(startLocalPosition, this.emitterShape.localMatrix, startLocalPosition);
        // 发射器局部矩阵影响
        BABYLON.Vector3.TransformNormalToRef(startDirection, this.emitterShape.localMatrix, startDirection);

        startDirection.scaleInPlace(startSpeed);

        particle.age = 0;
        particle.lifetime = this.startLifetimeInterpolation.interpolate(progress, Math.random());

        particle.position.copyFrom(startLocalPosition);
        particle.direction.copyFrom(startDirection);

        // emitRotation
        const xAxis = startDirection.x;
        const yAxis = startDirection.y;
        const zAxis = startDirection.z;
        let yaw = -Math.atan2(zAxis, xAxis) + Math.PI / 2;
        let len = Math.sqrt(xAxis * xAxis + zAxis * zAxis);
        let pitch = -Math.atan2(yAxis, len);
        particle.emitRotation.copyFromFloats(pitch, yaw, 0);

        this.startColorInterpolation.modify(particle, progress, 0);
        this.startRotationInterpolation.modify(particle, progress, 0);
        this.startSizeInterpolation.modify(particle, progress, 0);
        if (this.enableTextureSheet) {
            this.textureSheetInterpolation.runAsStart = true;
            this.textureSheetInterpolation.modify(particle, progress, 0);
        }

        particle.trialLifetime = 0;
        particle.trialWidth = 1;
        if (this.trail && this.enableTrail) {
            if (Math.random() <= this.trail.ratio) {
                particle.trialLifetime = this.trail.lifetime.interpolate(progress, Math.random()) * particle.lifetime;
                if (this.trail.sizeAffectsLifetime) {
                    particle.trialLifetime *= particle.startScaling.length() * TrailModifier._Scale;
                }
            }
            if (this.trail.widthOverTrail.mode == EInterpolationCurveMode.Constant || this.trail.widthOverTrail.mode == EInterpolationCurveMode.TwoConstants) {
                particle.trialWidth = this.trail.widthOverTrail.interpolate(progress, Math.random());
            }
        }
        startDirection.normalize();
        startDirection.scaleInPlace(0.01);
        particle.localRecordList.push([-0.002, startLocalPosition.x - startDirection.x, startLocalPosition.y - startDirection.y, startLocalPosition.z - startDirection.z]);
        particle.localRecordList.push([-0.001, startLocalPosition.x, startLocalPosition.y, startLocalPosition.z]);
        BABYLON.Vector3.TransformCoordinatesToRef(startLocalPosition, particle.startWorldMatrix, startLocalPosition);
        BABYLON.Vector3.TransformCoordinatesToRef(startDirection, particle.startWorldMatrix, startDirection);
        particle.globalRecordList.push([-0.002, startLocalPosition.x - startDirection.x, startLocalPosition.y - startDirection.y, startLocalPosition.z - startDirection.z]);
        particle.globalRecordList.push([-0.001, startLocalPosition.x, startLocalPosition.y, startLocalPosition.z]);
    }

    /**
     * 变量预处理
     * @param index
     */
    public varPre(activeParticles: Particle[], deltaTimeMS: number, scaleUpdateSpeed: number, worldMatrix: BABYLON.Matrix, startWorldMatrixInvert: BABYLON.Matrix, emitWorldMatrix: BABYLON.Matrix) {
        let count = activeParticles.length;
        let globalAmount = this._particleSystemAge % this.duration / this.duration;
        globalAmount = Math.max(0, Math.min(1, globalAmount));

        this.localPositionModifier.rotateByQuaternionAroundPoint = this.emitterShape.rotateByQuaternionAroundPoint;

        this._maxLifetimeOverage = 0.0;
        for (let i = 0; i < count; i++) {
            let particle = activeParticles[i];
            particle.age += deltaTimeMS;
            this._maxLifetimeOverage = Math.max(this._maxLifetimeOverage, particle.lifetime - particle.age + particle.trialLifetime * this._trailLifetimeScale);

            if (particle.age <= particle.lifetime) {
                if (this.simulationSpace == EMeshParticleSpaceMode.Local) {
                    particle.startWorldMatrix.copyFrom(worldMatrix);
                    particle.startWorldMatrixInvert.copyFrom(startWorldMatrixInvert);
                    particle.emitWorldMatrix.copyFrom(emitWorldMatrix);
                }

                let particleAmount = particle.age / particle.lifetime;
                particleAmount = Math.min(1.0, particleAmount);

                this.gravityInterpolation.modify(particle, globalAmount, scaleUpdateSpeed);

                this.textureSheetInterpolation.runAsStart = false;

                // 重置scaling
                particle.color.copyFrom(particle.startColor);
                particle.scaling.copyFrom(particle.startScaling);

                let modifierCount = this.overLifetimeModifierList.length;
                for (let j = 0; j < modifierCount; j++) {
                    let modifier = this.overLifetimeModifierList[j];
                    modifier.modify(particle, particleAmount, scaleUpdateSpeed);
                }
            }
        }
    }

    /**
     * 渲染结束后回收
     */
    public recycle = () => {
    }

    /**
     * 粒子排序
     */
    private activeParticlesSort(a: Particle, b: Particle) {
        return -(a.age - b.age);
    }

    private _updateIndex: number = 0;
}