import { IImageSolt, IMaterialReplace, IMaterialSolt } from "../interface/index";
import { BABYLON } from "../render3d/babylon";

export function applyGLTFLoaderDataModifier() {
    (<any>BABYLON.GLTF2.GLTFLoader.prototype)._loadData = function(this: BABYLON.GLTF2.GLTFLoader, data: BABYLON.IGLTFLoaderData): void {

        let imageSolts: string[] | IImageSolt[] = (<any>this)._metadata.imageSolts;
        if (imageSolts) {
            _gltfImageSolt(data, imageSolts);
        }

        let materialSolts: string[] | IMaterialSolt[] = (<any>this)._metadata.materialSolts;
        if (materialSolts) {
            _gltfMaterialSolts(data, materialSolts);
        }

        let materialReplace: IMaterialReplace[] = (<any>this)._metadata.materialReplace;
        if (materialReplace) {
            _gltfMaterialReplace(data, materialReplace);
        }

        (<any>this)._gltf = data.json as BABYLON.GLTF2.Loader.IGLTF;
        (<any>this)._setupData();

        if (data.bin) {
            const buffers = (<any>this)._gltf.buffers;
            if (buffers && buffers[0] && !buffers[0].uri) {
                const binaryBuffer = buffers[0];
                if (binaryBuffer.byteLength < data.bin.byteLength - 3 || binaryBuffer.byteLength > data.bin.byteLength) {
                    BABYLON.Logger.Warn(`Binary buffer length (${binaryBuffer.byteLength}) from JSON does not match chunk length (${data.bin.byteLength})`);
                }

                (<any>this)._bin = data.bin;
            }
            else {
                BABYLON.Logger.Warn("Unexpected BIN chunk");
            }
        }
    }
}

// PI_BEGIN - 20200309 扩展可变纹理数据
function _gltfImageSolt(data: BABYLON.IGLTFLoaderData, imageSolts: string[] | IImageSolt[]) {
    let solts = (<BABYLON.GLTF2.Loader.IGLTF>data.json).imageSolts;
    const images = (<BABYLON.GLTF2.Loader.IGLTF>data.json).images;
    const textures = (<BABYLON.GLTF2.Loader.IGLTF>data.json).textures;
    if (data.json && images) {
        const imagesLen = images.length;
        if (typeof imageSolts[0] === 'string') {
            if (solts) {
                for (let i = 0, len = imageSolts.length; i < len; i++) {
                    const solt = solts.find((s) => {
                                        return s[0] === imageSolts[i];
                                    });
                    if (solt && solt[1] <= imagesLen && solt[1] >= 0) {
                        images[solt[1]].uri = solt[2];
                        const tex = (<BABYLON.GLTF2.Loader.ITexture[]>textures).find((tex) => {
                            return tex.source === solt[1];
                        });
                        tex && (tex.invertY = !!solt[3]);
                    }
                }
            }
        } else {
            solts = <IImageSolt[]>imageSolts;
            for (let i = 0, len = solts.length; i < len; i++) {
                const solt = solts[i];
                if (solt && solt[1] <= imagesLen && solt[1] >= 0) {
                    images[solt[1]].uri = solt[2];
                    const tex = (<BABYLON.GLTF2.Loader.ITexture[]>textures).find((tex) => {
                        return tex.source === solt[1];
                    });
                    tex && (tex.invertY = !!solt[3]);
                }
            }
        }
    }
}
// PI_END

// PI_BEGIN - 20200310 - 拓展gltf materialSolts
function _gltfMaterialSolts(data: BABYLON.IGLTFLoaderData, materialSolts: string[] | IMaterialSolt[]) {
    let solts = (<BABYLON.GLTF2.Loader.IGLTF>data.json).materialSolts;
    const materials = (<BABYLON.GLTF2.Loader.IGLTF>data.json).materials;
    if (data.json && materials) {
        const materialLen = materials.length;

        if (typeof materialSolts[0] === 'string') {
            if (solts) {
                let material;
                for (let i = 0, len = materialSolts.length; i < len; i++) {
                    material = undefined;
                    const solt = solts.find((s) => {
                                        return s[0] === materialSolts[i];
                                    });

                    if (solt && solt[1] <= materialLen && solt[1] >= 0) {
                        material = (<any>materials[solt[1]]).extensions.PI_material;
                        if (material) {
                            const attrList = solt[2];
                            if (attrList) {
                                this._gltfMaterialSolt(attrList, material);
                            }
                        }
                    }
                }
            }
        } else {
            solts = <IMaterialSolt[]>materialSolts;
            let material = null;
            for (let i = 0, len = solts.length; i < len; i++) {
                const solt = solts[i];
                material = null;
                if (solt && solt[1] <= materialLen && solt[1] >= 0 && materials[solt[1]]) {
                    material = (<BABYLON.GLTF2.Loader.IMaterial>materials[solt[1]]).extensions.PI_material;
                }
                if (material) {
                    const attrList = solt[2];
                    if (attrList) {
                        this._gltfMaterialSolt(attrList, material);
                    }
                }
            }
        }
    }
}

function _gltfMaterialReplace(data: BABYLON.IGLTFLoaderData, materialReplaces: IMaterialReplace[]) {
    let replacedIndexs: number[] = [];

    let gltf: BABYLON.GLTF2.IGLTF = <any>data.json;

    let count = gltf.materials.length;
    let replaceCount = materialReplaces.length;

    let noIndexReplace = null;

    for (let i = 0; i < replaceCount; i++) {
        let replace = materialReplaces[i];
        let replaceIndex = replace.index;
        if (replaceIndex === undefined) {
            noIndexReplace = replace;
        } else {
            gltf.materials[replaceIndex].extensions = <any>replace.extensions;
            replacedIndexs.push(replaceIndex);
        }
    }

    if (noIndexReplace != null) {
        for (let i = 0; i < count; i++) {
            if (replacedIndexs.indexOf(i) < 0) {
                gltf.materials[i].extensions = <any>noIndexReplace.extensions;
            }
        }
    }
}