import { ITempFrameBuffer } from "../interface/base";
import { BABYLON } from "../render3d/babylon";
import { IPostprocessParam } from "./interface/index";
import { RenderTarget } from "./render_target";
import { ManagedWebGLRenderingContext } from "./webgl";
import { SceneStruct } from "../scene_struct";

export interface IPostProcess {
    readonly renderer: BABYLON.Renderer;
    apply(param: IPostprocessParam, source: (BABYLON.Texture | BABYLON.CustomRenderTarget)[], target: BABYLON.ICustomRenderTarget | BABYLON.TAutoRenderTarget | null, _ctx?: any) : void;
    dispose(): void;
}

export enum BlendMode {
    None,
    Normal,
    Additive,
    Multiply,
    Screen
}

export enum RenderMode {
    Local = 1,
    Canvas,
    Temp,
    TempScreen,
    TargetFBO
}

export interface IRenderData {
    /**
     * 源 纹理
     */
    sourceTex: WebGLTexture;
    /**
     * 源 FBO - 当 renderMode = RenderMode.Locel  时必须需要 sourceFBO
     */
    sourceFBO?: WebGLFramebuffer;
    /**
     * 源 FBO - 当 renderMode = RenderMode.TargetFBO 时必须需要 targetFBO
     */
    targetFBO?: WebGLFramebuffer;
    /**
     * 源纹理宽度
     */
    texWidth: number;
    /**
     * 源纹理高度
     */
    texHeight: number;
    /**
     * 渲染模式
     */
    renderMode?: RenderMode;
    /**
     * 渲染目标缩放值 - 相比源目标
     */
    scale?: number;
    sourceTex2?: WebGLTexture;
}

export interface IRenderDataNew {
    /**
     * 源 纹理
     */
    sourceRT: RenderTarget;
    /**
     * 源 纹理
     */
    targetRT?: RenderTarget;
    canvas?: HTMLCanvasElement;
}

export interface Restorable {
    restore(): void;
}

export class PostprocessBase {
    public renderer: BABYLON.Renderer;
    public camera: BABYLON.Camera;
    public quad: BABYLON.Mesh;
    protected scene: SceneStruct;
    constructor(name: string, pass_tag: BABYLON.Constants.PassTag = BABYLON.Constants.PASS_TAG_06, transparent: boolean = true) {
        this.scene = new SceneStruct(name, BABYLON.Engine.getEngine());
        this.camera = new BABYLON.FreeCamera(name, new BABYLON.Vector3(0, 0, -10), this.scene.impl);
        this.camera.layerMask = 0;
        this.quad = BABYLON.QuadBuilder.CreateQuad(name, { size: 1 }, this.scene.impl);
        this.quad.layerMask = 0;
        this.camera.forceInclude(this.quad, true);
        this.renderer = new BABYLON.Renderer(this.camera, `{name}${this.camera.uniqueId}`, pass_tag, transparent);
        this.renderer.autoClearColor(false);
        this.renderer.autoClearDepth(true);
        this.renderer.clearDepth(1.0);
        this.camera.active = true;
    }
    protected _apply(source: (BABYLON.Texture | BABYLON.CustomRenderTarget)[], target: ITempFrameBuffer | BABYLON.TAutoRenderTarget | null, _ctx?: ManagedWebGLRenderingContext) {
        if ((<BABYLON.ICustomRenderTarget>source[0]).frameBuffer) {
            let renderer = <BABYLON.Renderer>(<BABYLON.PIWebGLFramebuffer>(<BABYLON.ICustomRenderTarget>source[0]).frameBuffer)._renderer;
            if (renderer) {
                this.renderer.pre(renderer);
                // renderer.next(this.renderer);
                (<BABYLON.PIWebGLFramebuffer>(<BABYLON.ICustomRenderTarget>source[0]).frameBuffer)._renderer = renderer;
            }
        }

        let customTarget = <BABYLON.ICustomRenderTarget>target;
    
        if ((<BABYLON.PIWebGLFramebuffer>customTarget.frameBuffer)) {
            let renderer = <BABYLON.Renderer>(<BABYLON.PIWebGLFramebuffer>customTarget.frameBuffer)._renderer;
            if (renderer) {
                this.renderer.pre(renderer);
                // renderer.next(this.renderer);
                (<BABYLON.PIWebGLFramebuffer>customTarget.frameBuffer)._renderer = renderer;
            }

            this.renderer.setRenderTarget(target);
        } else if (customTarget.ctx) {
            let rendertarget: BABYLON.TAutoRenderTarget = { width: customTarget.rect[2], height: customTarget.rect[3], colorFormat: BABYLON.Constants.RGBA8, depthStencilFormat: BABYLON.Constants.DEPTH32F, autoTarget: true };
            this.renderer.setRenderTarget(rendertarget);
            customTarget.ctx.canvas.setBrush(this.renderer.entity);
        }
        this.scene.impl.render();
    }
    dispose() {
        this.renderer.dispose();
        this.quad.dispose();
        this.camera.dispose();
        this.scene.dispose();
    }
}