import { p3d } from "../render/types";
import { BABYLON } from "../render3d/babylon";
import { IPostProcess, PostprocessBase } from "./base";
import { KeyCopy, ICopy } from "./interface/copy";
import { ExtensionTools } from "../base";
import { ManagedWebGLRenderingContext } from "./webgl";
import { ITempFrameBuffer } from "../interface/base";
import { Engine, UniformMainRenderTarget } from "../render/index";

export class CopyMaterial extends BABYLON.EffectMaterial {
    public static ShaderName = "CopyMaterial";
    public set mainTexture(val: BABYLON.Texture | BABYLON.CustomRenderTarget) {
        UniformMainRenderTarget.texture(this.entity, val);
    }
    public baseColor(r: number, g: number, b: number, a: number) {
        UniformMainRenderTarget.color(this.entity, r, g, b, a);
    }

    public static RECT = "uRECT";
    public tilloffset(uScale: number, vScale: number, uOffset: number, vOffset: number) {
        p3d.p3d_material_uniform_vec4(Engine.commands(), this.entity, ExtensionTools.getAtom(CopyMaterial.RECT), uScale, vScale, uOffset, vOffset);
    }
    public static regist() {
        if (BABYLON.Engine.CheckShader(CopyMaterial.ShaderName)) { return };
        let uniforms: p3d.MaterialUniformDefines = p3d.MaterialUniformDefines.create();
        let includes: p3d.NodematerialIncludes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacity);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformMainRenderTarget.BLOCK_NAME));
        p3d.p3d_shader_uniform_vec4(uniforms, CopyMaterial.RECT, 1, 1, 0, 0, false);

        // let varying = BABYLON.Constants.VARYING_POSITION_WS | BABYLON.Constants.VARYING_NORMAL_WS | BABYLON.Constants.VARYING_UV | BABYLON.Constants.ARYING_COLOR;
        let binds_defines_base = BABYLON.Constants.BIND_EFFECT_VALUE | BABYLON.Constants.BIND_EFFECT_TEXTURE;

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec4", "v_color"],
            ["vec2", "v_uv"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });

        BABYLON.Engine.RegistShader(
            CopyMaterial.ShaderName,
            uniforms,
            `
            `,
            `
            layout(location = 0) out vec4 gl_FragColor;
            `,
            `
            vec2 xy = A_POSITION.xy + vec2(0.5, 0.5);
            xy = xy * matParam.uRECT.xy + matParam.uRECT.zw;
            gl_Position = vec4((xy - vec2(0.5, 0.5)) * 2.0, 0., 1.0);

            v_uv    = A_UV;
            v_color = A_COLOR4;
            `,
            `
            vec4 baseColor = v_color;
            float alpha = opacity(matParam);
            alpha   *= v_color.a;
        
            vec4 mainTextureColor   = mainTexture(v_uv, vec2(0., 0.), matParam);
            baseColor.rgb           *= mainTextureColor.rgb * mainStrength(matParam);
            alpha                   *= mainTextureColor.a;
        
            if (cutoff(alpha, matParam)) {
                discard;
            }

            vec4 finalColor 		= vec4(
                mainColor(matParam) * baseColor.rgb,
                alpha
            );
        
            finalColor				= max(finalColor, 0.0);
        
            // Premulty
            finalColor.rgb          *= finalColor.a;
        
            gl_FragColor = finalColor;        
            `,
            includes,
            ``,
            varyings,
            binds_defines_base
        );
    }
    constructor(name: string, scene: BABYLON.Scene, _pass: BABYLON.Constants.PassTag) {
        super(name, scene, CopyMaterial.ShaderName, _pass);
    }
}

export class Copy extends PostprocessBase implements IPostProcess {
    public mat: CopyMaterial;
    constructor(pass_tag: BABYLON.Constants.PassTag = BABYLON.Constants.PASS_TAG_06, transparent: boolean = true, blendmode: number = 0) {
        CopyMaterial.regist();
        super(KeyCopy, pass_tag, transparent);
        let mat = new CopyMaterial(KeyCopy, this.scene.impl, pass_tag);
        this.mat = mat;
        this.quad.setMaterial(this.mat);
        if (transparent) {
            this.quad.applyAlphaMode(blendmode, pass_tag);
        }
    }
    apply(param: ICopy, source: (BABYLON.Texture | BABYLON.CustomRenderTarget)[], target: ITempFrameBuffer | BABYLON.TAutoRenderTarget | null, _ctx?: ManagedWebGLRenderingContext) {
        this.mat.mainTexture = source[0];
        this.mat.alpha = param.alpha;
        BABYLON.UniformMainTexture.level(this.mat.entity, param.alpha);
        // this.renderer.reset();
        super._apply(source, target, _ctx);
    }
    dispose() {
        this.mat.dispose();
        super.dispose();
    }
}