
import { ExtensionTools } from "../base";
import { BABYLON } from "../render3d/babylon";
import { IPostProcess, PostprocessBase } from "./base";
import { KeyRadiusWave, IRadiusWave } from "./interface/radius_wave";
import { p3d } from "../render/types";
import { ManagedWebGLRenderingContext } from "./webgl";
import { ITempFrameBuffer } from "../interface/base";

export class RadiusWaveMaterial extends BABYLON.MainTextureMaterial {
    public static ShaderName = KeyRadiusWave;
    public static WaveInfo = 'waveInfo';
    public static ScreenInfo = 'screenInfo';

    public set mainTexture(val: BABYLON.Texture | BABYLON.CustomRenderTarget) {
        BABYLON.UniformMainTexture.texture(this.entity, val);
    }
    public baseColor(r: number, g: number, b: number, a: number) {
        BABYLON.UniformMainTexture.color(this.entity, r, g, b, a);
    }
    public static regist() {
        if (BABYLON.Engine.CheckShader(RadiusWaveMaterial.ShaderName)) { return };
        let uniforms: p3d.MaterialUniformDefines = p3d.MaterialUniformDefines.create();
        let includes: p3d.NodematerialIncludes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);

        p3d.p3d_shader_uniform_vec2(uniforms, RadiusWaveMaterial.ScreenInfo, 1, 1, false);
        p3d.p3d_shader_uniform_vec4(uniforms, RadiusWaveMaterial.WaveInfo, 0.5, 0.5, 0., 0., false);

        // let varying = BABYLON.Constants.VARYING_POSITION_WS | BABYLON.Constants.VARYING_NORMAL_WS | BABYLON.Constants.VARYING_UV | BABYLON.Constants.ARYING_COLOR;
        let binds_defines_base = BABYLON.Constants.BIND_EFFECT_VALUE | BABYLON.Constants.BIND_EFFECT_TEXTURE;

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec4", "v_color"],
            ["vec2", "v_uv"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });

        BABYLON.Engine.RegistShader(
            RadiusWaveMaterial.ShaderName,
            uniforms,
            `
            `,
            `
layout(location = 0) out vec4 gl_FragColor;

vec2 RadialBlur(vec2 uv, vec4 waveInfo, vec2 screenInfo) {
    vec2 center = uv * 2. - vec2(1.);
    center.y *= screenInfo.y / screenInfo.x;
    float len = abs(length(center));
    float temp = (waveInfo.y - waveInfo.x) / 2.0;
    float temp2 = min(temp, 0.1);
    float weight = waveInfo.w * smoothstep(waveInfo.x, waveInfo.x + temp2, len) * (1.0 - smoothstep(waveInfo.y - temp, waveInfo.y, len));
    // weight = weight < 0.0 ? 0.0 : weight;
    len = mix(mix(len - waveInfo.x, 0.0, step(waveInfo.y, len)), 0., step(len, waveInfo.x));
    // len = len < waveInfo.x ? 0.0 : (len > waveInfo.y ? 0.0 : len - waveInfo.x);
    len = len / (waveInfo.y - waveInfo.x) * waveInfo.z;
    float diff = weight * sin(len * 3.141592653589793);
    return uv + vec2(diff);
}
            `,
            `
gl_Position = vec4(A_POSITION.xy * 2.0, 0., 1.0);
v_uv    = A_UV;
v_color = A_COLOR4;
            `,
            `
vec4 baseColor          = v_color;
float alpha             = 1.0;
alpha                   *= v_color.a;

vec4 mainTextureColor   = mainTexture(RadialBlur(v_uv, matParam.waveInfo, matParam.screenInfo), vec2(0., 0.), matParam);
baseColor.rgb           *= mainTextureColor.rgb * mainStrength(matParam);
alpha                   *= mainTextureColor.a;

gl_FragColor = vec4(baseColor.rgb, alpha);     
            `,
            includes,
            ``,
            varyings,
            binds_defines_base
        );

    }
    constructor(name: string, scene: BABYLON.Scene, _pass: BABYLON.Constants.PassTag) {
        super(name, scene, RadiusWaveMaterial.ShaderName, _pass);
    }
}

export class RadiusWave extends PostprocessBase implements IPostProcess {
    public mat: RadiusWaveMaterial;
    constructor(pass_tag: BABYLON.Constants.PassTag = BABYLON.Constants.PASS_TAG_06, transparent: boolean = true) {
        RadiusWaveMaterial.regist();
        super('RadiusWave', pass_tag, transparent);
        let mat = new RadiusWaveMaterial(`RadiusWaveMaterial`, this.scene.impl, pass_tag);
        this.mat = mat;
        this.quad.setMaterial(this.mat);
    }
    apply(param: IRadiusWave, source: (BABYLON.Texture | BABYLON.CustomRenderTarget)[], target: ITempFrameBuffer | BABYLON.TAutoRenderTarget | null, _ctx?: ManagedWebGLRenderingContext) {
        this.mat.mainTexture = source[0];
        if (source[0] instanceof BABYLON.Texture) {
            p3d.p3d_material_uniform_vec2(BABYLON.Engine.commands(), this.mat.entity, ExtensionTools.getAtom(RadiusWaveMaterial.WaveInfo), 1.0, 1.0);
        } else {
            p3d.p3d_material_uniform_vec2(BABYLON.Engine.commands(), this.mat.entity, ExtensionTools.getAtom(RadiusWaveMaterial.WaveInfo), param.asCircle ? source[0].width : 1.0, param.asCircle ? source[0].height : 1.0);
        }
        p3d.p3d_material_uniform_vec4(BABYLON.Engine.commands(), this.mat.entity, ExtensionTools.getAtom(RadiusWaveMaterial.WaveInfo), Math.min(param.waveStartRadius, param.waveEndRadius), Math.max(param.waveStartRadius, param.waveEndRadius), param.waveCounter, param.waveWeight);

        super._apply(source, target, _ctx);
    }
    dispose() {
        this.mat.dispose();
        super.dispose();
    }
}