import { p3d } from "../render/types";
import { BABYLON } from "../render3d/babylon";
import { IPostProcess, PostprocessBase } from "./base";
import { KeyCopy, ICopy } from "./interface/copy";
import { ExtensionTools } from "../base";
import { ManagedWebGLRenderingContext } from "./webgl";
import { ITempFrameBuffer } from "../interface/base";
import { Engine, ITransition, UniformMainRenderTarget } from "../render/index";

export class TransitionMaterial extends BABYLON.EffectMaterial {
    public static ShaderName = "TransitionMaterial";
    public set mainTexture(val: BABYLON.Texture | BABYLON.CustomRenderTarget) {
        UniformMainRenderTarget.texture(this.entity, val);
    }
    public baseColor(r: number, g: number, b: number, a: number) {
        UniformMainRenderTarget.color(this.entity, r, g, b, a);
    }

    public static RECT = "uRECT";
    public tilloffset(uScale: number, vScale: number, uOffset: number, vOffset: number) {
        p3d.p3d_material_uniform_vec4(Engine.commands(), this.entity, ExtensionTools.getAtom(TransitionMaterial.RECT), uScale, vScale, uOffset, vOffset);
    }
    public static regist() {
        if (BABYLON.Engine.CheckShader(TransitionMaterial.ShaderName)) { return };
        let uniforms: p3d.MaterialUniformDefines = p3d.MaterialUniformDefines.create();
        let includes: p3d.NodematerialIncludes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacity);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformMainRenderTarget.BLOCK_NAME));
        p3d.p3d_shader_uniform_vec4(uniforms, TransitionMaterial.RECT, 1, 1, 0, 0, false);

        // let varying = BABYLON.Constants.VARYING_POSITION_WS | BABYLON.Constants.VARYING_NORMAL_WS | BABYLON.Constants.VARYING_UV | BABYLON.Constants.ARYING_COLOR;
        let binds_defines_base = BABYLON.Constants.BIND_EFFECT_VALUE | BABYLON.Constants.BIND_EFFECT_TEXTURE;

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec4", "v_color"],
            ["vec2", "v_uv"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });

        BABYLON.Engine.RegistShader(
            TransitionMaterial.ShaderName,
            uniforms,
            `
            `,
            `
            layout(location = 0) out vec4 gl_FragColor;
            `,
            `
            vec2 xy = A_POSITION.xy + vec2(0.5, 0.5);
            xy = xy * matParam.uRECT.xy + matParam.uRECT.zw;
            gl_Position = vec4((xy - vec2(0.5, 0.5)) * 2.0, 0., 1.0);

            v_uv    = A_UV;
            v_color = A_COLOR4;
            `,
            `
            vec4 baseColor = v_color;
            float alpha = opacity(matParam);
            alpha   *= v_color.a;
        
            vec4 mainTextureColor   = mainTexture(v_uv, vec2(0., 0.), matParam);
            baseColor.rgb           *= mainTextureColor.rgb * mainStrength(matParam);
            alpha                   *= mainTextureColor.a;
        
            if (cutoff(alpha, matParam)) {
                discard;
            }

            vec4 finalColor 		= vec4(
                mainColor(matParam) * baseColor.rgb,
                alpha
            );
        
            finalColor				= max(finalColor, 0.0);
        
            // Premulty
            finalColor.rgb          *= finalColor.a;
        
            gl_FragColor = finalColor;        
            `,
            includes,
            ``,
            varyings,
            binds_defines_base
        );
    }
    constructor(name: string, scene: BABYLON.Scene, _pass: BABYLON.Constants.PassTag) {
        super(name, scene, TransitionMaterial.ShaderName, _pass);
    }
}

export class Transition extends PostprocessBase implements IPostProcess {
    public mat: TransitionMaterial;
    constructor(pass_tag: BABYLON.Constants.PassTag = BABYLON.Constants.PASS_TAG_06, transparent: boolean = true, blendmode: number = 0) {
        TransitionMaterial.regist();
        super(KeyCopy, pass_tag, transparent);
        let mat = new TransitionMaterial(KeyCopy, this.scene.impl, pass_tag);
        this.mat = mat;
        this.quad.setMaterial(this.mat);
        if (transparent) {
            this.quad.applyAlphaMode(blendmode, pass_tag);
        }
    }
    apply(param: ITransition, source: (BABYLON.Texture | BABYLON.CustomRenderTarget)[], target: ITempFrameBuffer | BABYLON.TAutoRenderTarget | null, _ctx?: ManagedWebGLRenderingContext) {
        this.mat.mainTexture = source[0];
        this.mat.alpha = param.alpha;
        BABYLON.UniformMainTexture.level(this.mat.entity, param.alpha);
        // this.renderer.reset();
        super._apply(source, target, _ctx);
    }
    dispose() {
        this.mat.dispose();
        super.dispose();
    }
}


export class Transition extends PostProcess {
    public static shaderName = KeyTransition;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${Transition.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, Transition.getVertexString(), Transition.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static Transition = 'uTransition';

    public apply(param: ITransition, sources: RenderTexture[], target: RenderTarget, ctx: ManagedWebGLRenderingContext): ShaderCompiler {
        super.apply(param, sources, target, ctx);

        const gl = this.canvasOrContext.gl;
        let shader = Transition.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        shader.setUniformf(Transition.Transition, param.transition);

        let dataBuffer = this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey);
        PostProcess.bindData(this.canvasOrContext, dataBuffer, shader);
        PostProcess.drawData(this.canvasOrContext, dataBuffer);
        this.afterDraw();
    
        return shader;
    }

    protected static getFragmentString() {
        return `
        #define SHADER_NAME fragment:radial_wave
        precision highp float;

        varying vec2 sampleCenter;
        uniform sampler2D textureSampler;

        uniform float ${Transition.Transition};

        float easeOutBack(float x) {
            float c1 = 1.70158;
            float c3 = c1 + 1.;

            return 1. + c3 * pow(x - 1., 3.) + c1 * pow(x - 1., 2.);
        }
        vec4 Transition(sampler2D tex, vec2 uv) {

            vec4 acumulateColor = vec4(0.0);

            vec2 center = uv * 2.0 - 1.0; // [0, 1] -> [-1, 1]

            float len = abs(length(center));

            float t = ${Transition.Transition};
            float a = min(1., t * 2.);
            float b = min(1., max(0., (t * 2. - 1.)));
            b = easeOutBack(b);

            vec2 scl0 = (center / (1.0 + a*a*a * 10. * len) + 1.0 ) * 0.5;

            float c = len; // max(abs(center.x), abs(center.y));
            // c = (2. - c);
            c = pow(c - 1., 5.) + 1.;
            vec2 scl1 = mix(center, normalize(center) * c, 1.-b);
            scl1 = ( scl1 + 1. ) * 0.5;
            uv = mix(scl0, mix(scl1, uv, step(1., t)), step(0.5, t));
            return texture2D(tex, uv);
        }

        void main(void) {
            gl_FragColor = Transition(textureSampler, sampleCenter);
        }
        `;
    }
}