
import { ExtensionTools } from "../base";
import { BABYLON } from "../render3d/babylon";
import { IPostProcess, PostprocessBase } from "./base";
import { KeyVignette, IVignette } from "./interface/vignette";
import { p3d } from "../render/types";
import { ITempFrameBuffer } from "../interface/base";
import { ManagedWebGLRenderingContext } from "./webgl";

export class VignetteMaterial extends BABYLON.MainTextureMaterial {
    public static ShaderName = KeyVignette;
    public static Vignette = 'uVignette';
    public set mainTexture(val: BABYLON.Texture | BABYLON.CustomRenderTarget) {
        BABYLON.UniformMainTexture.texture(this.entity, val);
    }
    public baseColor(r: number, g: number, b: number, a: number) {
        BABYLON.UniformMainTexture.color(this.entity, r, g, b, a);
    }
    public static regist() {
        if (BABYLON.Engine.CheckShader(VignetteMaterial.ShaderName)) { return };
        let uniforms: p3d.MaterialUniformDefines = p3d.MaterialUniformDefines.create();
        let includes: p3d.NodematerialIncludes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacity);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);

        p3d.p3d_shader_uniform_vec4(uniforms, VignetteMaterial.Vignette, 0.5, 0.75, 1., 0., false);

        // let varying = BABYLON.Constants.VARYING_POSITION_WS | BABYLON.Constants.VARYING_NORMAL_WS | BABYLON.Constants.VARYING_UV | BABYLON.Constants.ARYING_COLOR;
        let binds_defines_base = BABYLON.Constants.BIND_EFFECT_VALUE | BABYLON.Constants.BIND_EFFECT_TEXTURE;

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec4", "v_color"],
            ["vec2", "v_uv"],
            ["vec2", "sampleCenter"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });

        BABYLON.Engine.RegistShader(
            VignetteMaterial.ShaderName,
            uniforms,
            `
            `,
            `
            layout(location = 0) out vec4 gl_FragColor;
            float Vignette(vec4 vignette, vec2 uv) {
                float dist = distance(uv, vec2(0.0, 0.0));
                return smoothstep(vignette.x, vignette.y, dist * vignette.z);
            }
            `,
            `
            gl_Position = vec4(A_POSITION.xy * 2.0, 0., 1.0);
            sampleCenter = A_POSITION.xy;

            v_uv    = A_UV;
            v_color = A_COLOR4;
            `,
            `
            vec4 baseColor = v_color;
            float alpha = opacity(matParam);
            alpha   *= v_color.a;
            
            float vignetteResult = Vignette(${VignetteMaterial.Vignette}, sampleCenter);
        
            vec4 mainTextureColor   = mainTexture(v_uv, vec2(0., 0.), matParam);
            baseColor.rgb           *= mainTextureColor.rgb * mainStrength(matParam);
            alpha                   *= mainTextureColor.a;

            if (cutoff(alpha, matParam)) {
                discard;
            }

            vec4 finalColor 		= vec4(
                mainColor() * vignetteResult + baseColor.rgb * (1.0 - vignetteResult),
                alpha
            );
        
            finalColor				= max(finalColor, 0.0);

            // Premulty
            finalColor.rgb          *= finalColor.a;
        
            gl_FragColor = finalColor;        
            `,
            includes,
            ``,
            varyings,
            binds_defines_base
        );

    }
    constructor(name: string, scene: BABYLON.Scene, _pass: BABYLON.Constants.PassTag) {
        super(name, scene, VignetteMaterial.ShaderName, _pass);
    }
}

export class Vignette extends PostprocessBase implements IPostProcess {
    public mat: VignetteMaterial;
    constructor(pass_tag: BABYLON.Constants.PassTag = BABYLON.Constants.PASS_TAG_06, transparent: boolean = true) {
        VignetteMaterial.regist();
        super(KeyVignette, pass_tag, transparent);
        let mat = new VignetteMaterial(KeyVignette, this.scene.impl, pass_tag);
        this.mat = mat;
        this.quad.setMaterial(this.mat);
    }
    apply(param: IVignette, source: (BABYLON.Texture | BABYLON.CustomRenderTarget)[], target: ITempFrameBuffer | BABYLON.TAutoRenderTarget | null, _ctx?: ManagedWebGLRenderingContext) {
        this.mat.mainTexture = source[0];
        this.mat.baseColor(param.vignetteColor[0], param.vignetteColor[1], param.vignetteColor[2], 1);
        p3d.p3d_material_uniform_vec4(BABYLON.Engine.commands(), this.mat.entity, ExtensionTools.getAtom(VignetteMaterial.Vignette), param.begin, param.end, param.scale, 1.);

        super._apply(source, target, _ctx);
    }
    dispose() {
        this.mat.dispose();

        super.dispose();
    }
}