import { BlendMode, Restorable } from "./base";

export class ManagedWebGLRenderingContext {
    public static readonly contentModes = ["webgl", "experimental-webgl", "webgl2", "webkit-3d", "moz-webgl"];

    private static pool: Map<HTMLCanvasElement, ManagedWebGLRenderingContext> = new Map();
    public static get(canvas: HTMLCanvasElement, contextConfig: any = { alpha: "true" }) {
        let result = ManagedWebGLRenderingContext.pool.get(canvas);
        if (!result) {
            result = new ManagedWebGLRenderingContext(canvas, contextConfig);
            ManagedWebGLRenderingContext.pool.set(canvas, result);
        }

        return result;
    }

    public canvas: HTMLCanvasElement;
    public gl: WebGLRenderingContext;
    private restorables = new Array<Restorable>();

    constructor(canvasOrContext: HTMLCanvasElement | WebGLRenderingContext, contextConfig: any = { alpha: "true" }) {
        // 为了兼容 Native端，Native端的canvasOrContext 不是HtmlCanvasElement
        // 注：微信小游戏平台，下面判断返回false：canvasOrContext["getContext"] instanceof Function
        this.gl = <any>{};
        this.canvas = <HTMLCanvasElement>this.gl.canvas;
    }

    addRestorable(restorable: Restorable) {
        this.restorables.push(restorable);
    }

    removeRestorable(restorable: Restorable) {
        let index = this.restorables.indexOf(restorable);
        if (index > -1) {
            this.restorables.splice(index, 1);
        }
    }

    alphaBlendState(alphaMode: number) {
        // const gl: WebGLRenderingContext = this.gl;

        // switch (alphaMode) {
        //     case (2): {
        //         gl.blendFuncSeparate(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA, gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);
        //         break;
        //     }
        //     case (7): {
        //         gl.blendFuncSeparate(gl.ONE, gl.ONE_MINUS_SRC_ALPHA, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
        //         break;
        //     }
        //     case (8): {
        //         gl.blendFuncSeparate(gl.ONE, gl.ONE_MINUS_SRC_ALPHA, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
        //         break;
        //     }
        //     default: {
        //         gl.blendFuncSeparate(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA, gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);
        //         break;
        //     }
        // }
    }
    public dispose() {
        ManagedWebGLRenderingContext.pool.delete(this.canvas);
    }
}

export class WebGLBlendModeConverter {
    static ZERO = 0;
    static ONE = 1;
    static SRC_COLOR = 0x0300;
    static ONE_MINUS_SRC_COLOR = 0x0301;
    static SRC_ALPHA = 0x0302;
    static ONE_MINUS_SRC_ALPHA = 0x0303;
    static DST_ALPHA = 0x0304;
    static ONE_MINUS_DST_ALPHA = 0x0305;
    static DST_COLOR = 0x0306

    static getDestGLBlendMode(blendMode: BlendMode) {
        switch (blendMode) {
            case BlendMode.Normal: return WebGLBlendModeConverter.ONE_MINUS_SRC_ALPHA;
            case BlendMode.Additive: return WebGLBlendModeConverter.ONE;
            case BlendMode.Multiply: return WebGLBlendModeConverter.ONE_MINUS_SRC_ALPHA;
            case BlendMode.Screen: return WebGLBlendModeConverter.ONE_MINUS_SRC_ALPHA;
            default: throw new Error("Unknown blend mode: " + blendMode);
        }
    }

    static getSourceGLBlendMode(blendMode: BlendMode, premultipliedAlpha: boolean = false) {
        switch (blendMode) {
            case BlendMode.Normal: return premultipliedAlpha ? WebGLBlendModeConverter.ONE : WebGLBlendModeConverter.SRC_ALPHA;
            case BlendMode.Additive: return premultipliedAlpha ? WebGLBlendModeConverter.ONE : WebGLBlendModeConverter.SRC_ALPHA;
            case BlendMode.Multiply: return WebGLBlendModeConverter.DST_COLOR;
            case BlendMode.Screen: return WebGLBlendModeConverter.ONE;
            default: throw new Error("Unknown blend mode: " + blendMode);
        }
    }
}