
import { Engine } from "../../engines/engine";
import { ITargetAnimation, ITargetAnimationProperty } from "../../../interface/index";
import { p3d } from "../../types";
import { ExtensionTools } from "../../../base";

export enum EEasingMode {
    None            = 0x00,
    BackIn          = 0x01,
    BackOut         = 0x02,
    BackInOut       = 0x03,
    CircleIn        = 0x04,
    CircleOut       = 0x05,
    CircleInOut     = 0x06,
    CubicIn         = 0x07,
    CubicOut        = 0x08,
    CubicInOut      = 0x09,
    SineIn          = 0x11,
    SineOut         = 0x12,
    SineInOut       = 0x13,
    QuadIn          = 0x14,
    QuadOut         = 0x15,
    QuadInOut       = 0x16,
    QuartIn         = 0x17,
    QuartOut        = 0x18,
    QuartInOut      = 0x19,
    QuintIn         = 0x21,
    QuintOut        = 0x22,
    QuintInOut      = 0x23,
    ExpoIn          = 0x24,
    ExpoOut         = 0x25,
    ExpoInOut       = 0x26,
    ElasticIn       = 0x27,
    ElasticOut      = 0x28,
    ElasticInOut    = 0x29,
    BounceIn        = 0x31,
    BounceOut       = 0x32,
    BounceInOut     = 0x33,
}

export class FrameCurve {
    public static readonly FramePerSecond = 120;
    private static AnimeCurveCounter: number = 0;
    public static AnimeCurveIDs: Map<string, number> = new Map();
    public static GetCurveID(name: string, force: boolean = false): number {
        let id = 0;
        if (force) {
            this.AnimeCurveCounter++;
            id = this.AnimeCurveCounter;
            FrameCurve.AnimeCurveIDs.set(name, id);
            return id;
        } else {
            id = this.AnimeCurveIDs.get(name);
            if (id == undefined) {
                this.AnimeCurveCounter++;
                id = this.AnimeCurveCounter;
                FrameCurve.AnimeCurveIDs.set(name, id);
            }
            return id;
        }
    }

    private static MAX_FRAME_CURVE_DATA: number = 1024 * 1024;
    public static Buffer: Float32Array;

    public static Init() {
        if (!FrameCurve.Buffer) {
            FrameCurve.Buffer = new Float32Array(this.MAX_FRAME_CURVE_DATA);
        }
    }

    public static check(id: number, property: p3d.EAnimePropertyID): boolean {
        return p3d.p3d_anime_curve_query(Engine.native(), Engine.param(), id, property);
    }
    public targetAnimation(group: number, target: number): ITargetAnimation {
        return <ITargetAnimationProperty>{ group, target, property: this._property, curve: this.id };
    }

    protected _frameIndexs: number[] = [];
    protected _frames: number[] = [];
    protected _mode: p3d.EAnimeCurve;
    protected _designFramePerSecond: number;
    protected _totalFrames: number;
    protected _easingmode: number;
    protected _from: number[];
    protected _scalar: number[];
    protected _x1: number;
    protected _y1: number;
    protected _x2: number;
    protected _y2: number;
    protected _property: p3d.EAnimePropertyID;
    public get property() {
        return this._property;
    }
    public get key() {
        return this._key;
    }
    private _id: number;
    public get id() {
        if (this._id == undefined) {
            this._id = p3d.p3d_animation_curve_id(ExtensionTools.getAtom(this._key));
        }
        return this._id;
    }
    public set_id_by_gltf(group_index: number, channel_index: number, gltf: p3d.GLTFRes) {
        this._id = p3d.p3d_animation_curve_id_bygltf(gltf, group_index, channel_index);
    }
    constructor(protected _key: string, designFramePerSecond: number) {
        this._designFramePerSecond = designFramePerSecond;
    }
    protected apply(buffer: Float32Array) : number {
        switch (this._mode) {
            case p3d.EAnimeCurve.FrameValues:
            case p3d.EAnimeCurve.FrameValuesStep: {
                if (this._frames.length + 1 < buffer.length) {
                    buffer[0] = this._designFramePerSecond;
                    let idx = 1;
                    // this._frameIndexs.forEach((v) => {
                    //     buffer[idx] = v;
                    //     idx += 1;
                    // });
                    this._frames.forEach((v) => {
                        buffer[idx] = v;
                        idx += 1;
                    });
                    return idx;
                }
                break;
            }
            case p3d.EAnimeCurve.EasingCurve: {
                if (this._from && this._scalar) {
                    buffer[0] = this._designFramePerSecond;
                    buffer[1] = this._totalFrames;
                    buffer[2] = this._easingmode || EEasingMode.None;
                    let idx = 3;
                    this._from.forEach((v) => {
                        buffer[idx] = v;
                        idx += 1;
                    });
                    this._scalar.forEach((v) => {
                        buffer[idx] = v;
                        idx += 1;
                    });

                    return idx;
                }
                break;
            }
            case p3d.EAnimeCurve.MinMaxCurve: {
                if (this._from && this._scalar && this._frames.length + this._from.length + this._scalar.length + 1 < buffer.length ) {
                    buffer[0] = this._designFramePerSecond;
                    let idx = 1;
                    this._from.forEach((v) => {
                        buffer[idx] = v;
                        idx += 1;
                    });
                    this._scalar.forEach((v) => {
                        buffer[idx] = v;
                        idx += 1;
                    });
                    this._frameIndexs.forEach((v) => {
                        buffer[idx] = v;
                        idx += 1;
                    });
                    this._frames.forEach((v) => {
                        buffer[idx] = v;
                        idx += 1;
                    });

                    return idx;
                }
                break;
            }
            case p3d.EAnimeCurve.CubicBezierCurve: {
                if (this._from && this._scalar) {
                    buffer[0] = this._designFramePerSecond;
                    buffer[1] = this._totalFrames;
                    let idx = 2;
                    this._from.forEach((v) => {
                        buffer[idx] = v;
                        idx += 1;
                    });
                    this._scalar.forEach((v) => {
                        buffer[idx] = v;
                        idx += 1;
                    });
                    buffer[idx] = this._x1; idx += 1;
                    buffer[idx] = this._y1; idx += 1;
                    buffer[idx] = this._x2; idx += 1;
                    buffer[idx] = this._y2; idx += 1;

                    return idx;
                }
                break;
            }
            case p3d.EAnimeCurve.GLTFCubicSpline: {
                if (this._frames.length + 1 < buffer.length) {
                    let idx = 0;
                    buffer[idx] = this._designFramePerSecond;
                    idx += 1;
                    this._frameIndexs.forEach((v) => {
                        buffer[idx] = v;
                        idx += 1;
                    });
                    this._frames.forEach((v) => {
                        buffer[idx] = v;
                        idx += 1;
                    });
                    return idx;
                }
                break;
            }
        }
        
        return 0;
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        return false;
    }
    public build() {
        let len = this.apply(FrameCurve.Buffer);
        if (len > 0) {
            p3d.p3d_anime_curve_create(Engine.native(), Engine.param(), this.id, this._property, FrameCurve.Buffer.slice(0, len), this._mode);
        }
    }
}

export class FrameCurveFrames extends FrameCurve {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.FrameValues;
    constructor(key: string, designFPS: number) {
        super(key, designFPS);
    }
}
export class FrameCurveFramesStep extends FrameCurve {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.FrameValuesStep;
    constructor(key: string, designFPS: number) {
        super(key, designFPS);
    }
}
export class FrameCurveFramesPIGLTFCubicSpline extends FrameCurve {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.GLTFCubicSpline;
    constructor(key: string, designFPS: number) {
        super(key, designFPS);
    }
}
export class FrameCurveFramesGLTFCubicSpline extends FrameCurve {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.GLTFCubicSpline;
    constructor(key: string, designFPS: number) {
        super(key, designFPS);
    }
}

export class FrameCurveV1Frames extends FrameCurveFrames {
    public addKey(frame: number, val: number) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, val);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 1;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let x = values[i * size + 0];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), x * times[i]);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV1FramesStep extends FrameCurveFramesStep {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.FrameValuesStep;
    public addKey(frame: number, val: number) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, val);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 1;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let x = values[i * size + 0];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), x * times[i]);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV1GLTFCubicSpline extends FrameCurveFramesGLTFCubicSpline {
    public addKey(frame: number, in_val: number, val: number, out_val: number) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, in_val, val, out_val);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 3 * 1;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let ix = values[i * size + 0];
                    let x = values[i * size + 1];
                    let ox = values[i * size + 2];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), ix, x, ox);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV1PIGLTFCubicSpline extends FrameCurveFramesPIGLTFCubicSpline {
    public addKey(frame: number, in_x: number, x: number) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, in_x, x, in_x);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 2 * 1;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let ix = values[i * size + 0];
                    let x = values[i * size + 1];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), ix, x);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV1Easing extends FrameCurve {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.EasingCurve;
    constructor(key: string, designFPS: number, totalSeconds: number, easing: EEasingMode, from: number, scalar: number) {
        super(key, designFPS);
        this._totalFrames = Math.round(designFPS * totalSeconds);
        this._easingmode = easing;
        this._from = [from];
        this._scalar = [scalar];
    }
}
export class FrameCurveV1Bezier extends FrameCurve {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.CubicBezierCurve;
    constructor(key: string, designFPS: number, totalSeconds: number, from: number, scalar: number, x1: number, y1: number, x2: number, y2: number) {
        super(key, designFPS);
        this._totalFrames = Math.round(designFPS * totalSeconds);
        this._from = [from];
        this._scalar = [scalar];
        this._x1 = x1;
        this._x2 = x2;
        this._y1 = y1;
        this._y2 = y2;
    }
}

export class FrameCurveV2Frames extends FrameCurveFrames {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.FrameValues;
    public addKey(frame: number, x: number, y: number) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, x, y);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 2;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let x = values[i * size + 0], y = values[i * size + 1];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), x, y);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV2FramesStep extends FrameCurveFramesStep {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.FrameValuesStep;
    public addKey(frame: number, x: number, y: number) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, x, y);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 2;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let x = values[i * size + 0], y = values[i * size + 1];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), x, y);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV2GLTFCubicSpline extends FrameCurveFramesGLTFCubicSpline {
    public addKey(frame: number, in_x: number, in_y: number, x: number, y: number, out_x: number, out_y: number) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, in_x, in_y, x, y, out_x, out_y);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 3 * 2;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let ix = values[i * size + 0], iy = values[i * size + 1];
                    let x = values[i * size + 2], y = values[i * size + 3];
                    let ox = values[i * size + 4], oy = values[i * size + 5];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), ix, iy, x, y, ox, oy);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV2PIGLTFCubicSpline extends FrameCurveFramesPIGLTFCubicSpline {
    public addKey(frame: number, in_x: number, in_y: number, x: number, y: number) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, in_x, in_y, x, y, in_x, in_y);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 2 * 2;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let ix = values[i * size + 0], iy = values[i * size + 1];
                    let x = values[i * size + 2], y = values[i * size + 3];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), ix, iy, x, y);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV2Easing extends FrameCurve {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.EasingCurve;
    constructor(key: string, designFPS: number, totalSeconds: number, easing: EEasingMode, from: [number, number], scalar: [number, number]) {
        super(key, designFPS);
        this._totalFrames = Math.round(designFPS * totalSeconds);
        this._easingmode = easing;
        this._from = from;
        this._scalar = scalar;
    }
}
export class FrameCurveV2Bezier extends FrameCurve {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.CubicBezierCurve;
    constructor(key: string, designFPS: number, totalSeconds: number, from: [number, number], scalar: [number, number], x1: number, y1: number, x2: number, y2: number) {
        super(key, designFPS);
        this._totalFrames = Math.round(designFPS * totalSeconds);
        this._from = from;
        this._scalar = scalar;
        this._x1 = x1;
        this._x2 = x2;
        this._y1 = y1;
        this._y2 = y2;
    }
}

export class FrameCurveV3Frames extends FrameCurveFrames {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.FrameValues;
    public addKey(frame: number, x: number, y: number, z: number) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, x, y, z);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 3;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let x = values[i * size + 0], y = values[i * size + 1], z = values[i * size + 2];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), x, y, z);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV3FramesStep extends FrameCurveFramesStep {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.FrameValuesStep;
    public addKey(frame: number, x: number, y: number, z: number) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, x, y, z);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 3;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let x = values[i * size + 0], y = values[i * size + 1], z = values[i * size + 2];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), x, y, z);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV3GLTFCubicSpline extends FrameCurveFramesGLTFCubicSpline {
    public addKey(frame: number, in_x: number, in_y: number, in_z: number, x: number, y: number, z: number, out_x: number, out_y: number, out_z: number) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, in_x, in_y, in_z, x, y, z, out_x, out_y, out_z);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 3 * 3;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let ix = values[i * size + 0], iy = values[i * size + 1], iz = values[i * size + 2];
                    let x = values[i * size + 3], y = values[i * size + 4], z = values[i * size + 5];
                    let ox = values[i * size + 6], oy = values[i * size + 7], oz = values[i * size + 8];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), ix, iy, iz, x, y, z, ox, oy, oz);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV3PIGLTFCubicSpline extends FrameCurveFramesPIGLTFCubicSpline {
    public addKey(frame: number, in_x: number, in_y: number, in_z: number, x: number, y: number, z: number) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, in_x, in_y, in_z, x, y, z, in_x, in_y, in_z);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 2 * 3;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let ix = values[i * size + 0], iy = values[i * size + 1], iz = values[i * size + 2];
                    let x = values[i * size + 3], y = values[i * size + 4], z = values[i * size + 5];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), ix, iy, iz, x, y, z);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV3Easing extends FrameCurve {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.EasingCurve;
    constructor(key: string, designFPS: number, totalSeconds: number, easing: EEasingMode, from: [number, number, number], scalar: [number, number, number]) {
        super(key, designFPS);
        this._totalFrames = Math.round(designFPS * totalSeconds);
        this._easingmode = easing;
        this._from = from;
        this._scalar = scalar;
    }
}
export class FrameCurveV3Bezier extends FrameCurve {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.CubicBezierCurve;
    constructor(key: string, designFPS: number, totalSeconds: number, from: [number, number, number], scalar: [number, number, number], x1: number, y1: number, x2: number, y2: number) {
        super(key, designFPS);
        this._totalFrames = Math.round(designFPS * totalSeconds);
        this._from = from;
        this._scalar = scalar;
        this._x1 = x1;
        this._x2 = x2;
        this._y1 = y1;
        this._y2 = y2;
    }
}

export class FrameCurveV4Frames extends FrameCurveFrames {
    public addKey(frame: number, x: number, y: number, z: number, w: number) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, x, y, z, w);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 4;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let x = values[i * size + 0], y = values[i * size + 1], z = values[i * size + 2], w = values[i * size + 3];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), x, y, z, w);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV4FramesStep extends FrameCurveFramesStep {
    public addKey(frame: number, x: number, y: number, z: number, w: number) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, x, y, z, w);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 4;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let x = values[i * size + 0], y = values[i * size + 1], z = values[i * size + 2], w = values[i * size + 3];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), x, y, z, w);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV4GLTFCubicSpline extends FrameCurveFramesGLTFCubicSpline {
    public addKey(frame: number, in_x: number, in_y: number, in_z: number, in_w: number, x: number, y: number, z: number, w: number, out_x: number, out_y: number, out_z: number, out_w: number, ) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, in_x, in_y, in_z, in_w, x, y, z, w, out_x, out_y, out_z, out_w);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 3 * 4;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let ix = values[i * size + 0], iy = values[i * size + 1], iz = values[i * size + 2], iw = values[i * size + 3];
                    let x = values[i * size + 4], y = values[i * size + 5], z = values[i * size + 6], w = values[i * size + 7];
                    let ox = values[i * size + 8], oy = values[i * size + 9], oz = values[i * size + 10], ow = values[i * size + 11];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), ix, iy, iz, iw, x, y, z, w, ox, oy, oz, ow);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV4PIGLTFCubicSpline extends FrameCurveFramesPIGLTFCubicSpline {
    public addKey(frame: number, in_x: number, in_y: number, in_z: number, in_w: number, x: number, y: number, z: number, w: number) {
        if (this._frameIndexs.includes(frame)) {
            return;
        }
        this._frameIndexs.push(frame);
        this._frames.push(frame, in_x, in_y, in_z, in_w, x, y, z, w, in_x, in_y, in_z, in_w);
    }
    public format(times: ArrayLike<number>, values: ArrayLike<number>): boolean {
        let frames = times.length;
        let size = 2 * 4;
        if (values.length >= frames * size) {
            if (!FrameCurve.check(this.id, this.property)) {
                for (let i = 0; i < frames; i++) {
                    let ix = values[i * size + 0], iy = values[i * size + 1], iz = values[i * size + 2], iw = values[i * size + 3];
                    let x = values[i * size + 4], y = values[i * size + 5], z = values[i * size + 6], w = values[i * size + 7];
                    this.addKey(Math.round(times[i] * this._designFramePerSecond), ix, iy, iz, iw, x, y, z, w);
                }
                this.build();
            }
            return true;
        } else {
            return false;
        }
    }
}
export class FrameCurveV4Easing extends FrameCurve {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.EasingCurve;
    constructor(key: string, designFPS: number, totalSeconds: number, easing: EEasingMode, from: [number, number, number, number], scalar: [number, number, number, number]) {
        super(key, designFPS);
        this._totalFrames = Math.round(designFPS * totalSeconds);
        this._easingmode = easing;
        this._from = from;
        this._scalar = scalar;
    }
}
export class FrameCurveV4Bezier extends FrameCurve {
    protected _mode: p3d.EAnimeCurve = p3d.EAnimeCurve.CubicBezierCurve;
    constructor(key: string, designFPS: number, totalSeconds: number, from: [number, number, number, number], scalar: [number, number, number, number], x1: number, y1: number, x2: number, y2: number) {
        super(key, designFPS);
        this._totalFrames = Math.round(designFPS * totalSeconds);
        this._from = from;
        this._scalar = scalar;
        this._x1 = x1;
        this._x2 = x2;
        this._y1 = y1;
        this._y2 = y2;
    }
}
