import { Skeleton } from './skeleton';
import { Engine } from '../engines/engine';

/**
 * 骨骼的 各动作矩阵缓存数据的描述信息
 */
export interface ISkeletonMatrixCacheInfo {
    [actionName: string]: ISkeletonMatrixDataInfo;
}
/**
 * 目标动作 矩阵缓存数据的描述信息
 */
export interface ISkeletonMatrixDataInfo {
    /**
     * 缓存数据的长度 - 1 - 1 个浮点数
     */
    dataLength: number;
    /**
     * 开始时间
     */
    startSecond: number;
    /**
     * 结束时间
     */
    endSecond: number;
    /**
     * 总帧数
     */
    totalFrameCount: number;
    /**
     * 目标数据起点 在数据流的偏移
     */
    dataStart: number;
    /**
     * 目标数据终点 在数据流的偏移
     */
    dataEnd: number;
    /**
     * 矩阵烘培精度 - 每秒采样烘培帧数
     */
    precision: number;
}

/**
 * 目标骨骼 矩阵烘培缓存表
 * * 提供指定 动作名，指定帧序号 缓存矩阵 的获取
 */
export class SkeletonMatrixCacheTable {
    /**
     * 各动作 的 每帧序号的 缓存矩阵 记录表
     * @hidden
     */
    private _matrixMap: Map<string, Map<number, Float32Array>> = new Map();
    private _cacheInfo: ISkeletonMatrixCacheInfo = <any>{};
    private _matrixData: Uint8Array | undefined;
    /**
     * 骨骼名称
     * @hidden
     */
    private _skeletonName: string;
    /**
     * 骨骼名称
     */
    public get skeletonName() {
        return this._skeletonName;
    }
    /**
     * sklai 文件路径
     * @hidden
     */
    private _ascriptionFile: string;
    /**
     * sklai 文件路径 - 每骨骼对应一个文件
     */
    public get ascriptionFile() {
        return this._ascriptionFile;
    }
    /**
     * 缓存信息在运行时 对应的 骨骼的 多个实例
     */
    private _refSkeletons: Skeleton[] = [];
    /**
     * Babylon 引擎
     */
    private engine: Engine;
    /**
     * 创建/初始化 目标骨骼的矩阵缓存表
     * @param ascriptionFile 缓存文件路径
     * @param targetSkeletonName 目标骨骼名称
     * @param engine 引擎实例
     * @param cacheInfo 各动作的 缓存数据的描述信息
     * @param data 缓存数据所在数据流
     */
    public constructor(ascriptionFile: string, targetSkeletonName: string, engine: Engine, cacheInfo?: ISkeletonMatrixCacheInfo, data?: Uint8Array) {
        this.engine = engine;
        // this.engine.skeletonCacheMatrixList.push(this);

        this._ascriptionFile = ascriptionFile;
        this._skeletonName = targetSkeletonName;
        if (cacheInfo && data) {
            for (const key in cacheInfo) {
                this.cache(key, cacheInfo[key], data);
            }
        }
    }
    /**
     * 缓存单个动作矩阵数据
     * @param actionName 动作名称
     * @param cacheDataInfo 缓存数据描述信息
     * @param data 缓存数据所在数据流
     */
    public cache(actionName: string, cacheDataInfo: ISkeletonMatrixDataInfo, data: Uint8Array) {
        this._cacheInfo[actionName] = cacheDataInfo;
        this._matrixData = data;
    }
    /**
     * 获取指定动作，指定帧序号 的矩阵数据
     * @param actionName 目标动作
     * @param frame 目标帧序号
     */
    public getCache(actionName: string, frame: number) {
        let cacheDataInfo = this._cacheInfo[actionName];
        if (this._matrixData && cacheDataInfo) {
            const frameSize = cacheDataInfo.dataLength / cacheDataInfo.totalFrameCount;
            let start = (cacheDataInfo.dataStart + frameSize * frame) * 4;
            let end = start + frameSize * 4;
            let floatArray: Float32Array;
            if (ArrayBuffer.isView(this._matrixData)) {
                floatArray = new Float32Array(this._matrixData.slice(start, end).buffer);
            } else {
                floatArray = new Float32Array((<ArrayBuffer>this._matrixData).slice(start, end));
            }
            return floatArray;
        }
        else {
            return;
        }
    }
    /**
     * 缓存表作用于骨骼
     * @param skeleton 目标骨骼
     */
    public useForSkeleton(skeleton: Skeleton) {
        if (this._refSkeletons.indexOf(skeleton) < 0) {
            this._refSkeletons.push(skeleton);
        }
        // if (skeleton.overrideMesh) {
        //     skeleton.overrideMesh.onDisposeObservable.addOnce(() => {
        //         this.unuseFromSkeleton(skeleton);
        //     });
        // }
    }
    /**
     * 目标骨骼销毁时，解除作用
     * @param skeleton 目标骨骼
     */
    public unuseFromSkeleton(skeleton: Skeleton) {
        const index = this._refSkeletons.indexOf(skeleton);
        if (index >= 0) {
            this._refSkeletons.splice(index, 1);
        }

        if (this._refSkeletons.length <= 0) {
            this.dispose();
        }
    }
    /**
     * 销毁
     */
    public dispose() {
        this._matrixMap.forEach((map) => {
            map.clear();
        });
        this._matrixMap.clear();
        this._matrixData = undefined;
        // const index = this.engine.skeletonCacheMatrixList.indexOf(this);
        // this.engine.skeletonCacheMatrixList.splice(index, 1);
    }
}