import { Entity, Engine } from "../engines/engine";
import { Scene } from "../scene";
import { Node } from "../node";
import { Matrix, Quaternion, Vector3 } from "../../maths/index";
import { Constants, Option, GlContext, ICamera, IEntity } from "../../interface/index";
import { IEntityD3, p3d } from "../types";
import { TransformNode } from "../meshes/transformNode";


export class Camera extends TransformNode implements ICamera, IEntity {
    /**
     * This is the default projection mode used by the cameras.
     * It helps recreating a feeling of perspective and better appreciate depth.
     * This is the best way to simulate real life cameras.
     */
    public static PERSPECTIVE_CAMERA = Constants.PERSPECTIVE_CAMERA;
    /**
     * This helps creating camera with an orthographic mode.
     * Orthographic is commonly used in engineering as a means to produce object specifications that communicate dimensions unambiguously, each line of 1 unit length (cm, meter..whatever) will appear to have the same length everywhere on the drawing. This allows the drafter to dimension only a subset of lines and let the reader know that other lines of that length on the drawing are also that length in reality. Every parallel line in the drawing is also parallel in the object.
     */
    public static ORTHOGRAPHIC_CAMERA: Constants.FREECAMERA_MODE  = Constants.ORTHOGRAPHIC_CAMERA;
    /**
     * This is the default FOV mode for perspective cameras.
     * This setting aligns the upper and lower bounds of the viewport to the upper and lower bounds of the camera frustum.
     */
    public static FOVMODE_VERTICAL_FIXED = Constants.FOVMODE_VERTICAL_FIXED;
    /**
     * This setting aligns the left and right bounds of the viewport to the left and right bounds of the camera frustum.
     */
    public static FOVMODE_HORIZONTAL_FIXED = Constants.FOVMODE_HORIZONTAL_FIXED;
    metadata: any = {};
    protected camera: Node;
    constructor(name: string, scene: Scene) {
        super(name, scene);
        // if (!customRender) {
        //     this.multiRenders();
        // }

        let entity = p3d.p3d_camera(Engine.native(), Engine.commands(), scene.entity);
        this.camera = new Node(name, scene, entity);
        this.camera.parent = this;
        
        this.active = true;
        this.minZ = 1;
        this.maxZ = 1001;
    }

    createRenderer(name: string, pass_tag: Constants.PassTag, transparent: boolean) : [number, string] {
        name = `V` + this.camera.uniqueId + name;
        let entity = p3d.p3d_create_render(Engine.native(), Engine.commands(), this.camera.entity, name, pass_tag, transparent);
        return [entity, name];
    }
    createPickRay(x: number, y: number, result: Float32Array): boolean {
        return p3d.p3d_create_pickingray(Engine.native(), Engine.param(), this.camera.entity, x, y, result);
    }

    dirtyLayerMask(): void {
        p3d.p3d_layermask(Engine.commands(), this.camera.entity, this._layerMask);
    }
    protected _layerMask: number = 0b11111111_11111111_11111111_11111111;
    public get layerMask() {
        return this._layerMask;
    }
    public set layerMask(val: number) {
        if (this._layerMask != val) {
            this._layerMask = val;
            // this._scene.cmdsLayerMask.set(this.camera.entity, this);
            p3d.p3d_layermask(Engine.commands(), this.camera.entity, this._layerMask);
        }
    }

    _onFrameUpdate(): void {
        if (this._isDisposed) {
            return;
        }
        super._onFrameUpdate();

        if (this._nearFarDirty) {
            p3d.p3d_camera_nearfar(Engine.commands(), this.camera.entity, Math.min(this._minZ, this._maxZ), Math.max(this._minZ, this._maxZ));
            this._nearFarDirty = false;
        }
    }

    protected _active: boolean = false;
    public get active() {
        return this._active;
    }
    public set active(val: boolean) {
        // if (val) {
        //     (<any>this._scene)._activeCamera = this;
        // }
        this._active = val;
        p3d.p3d_camera_active(Engine.commands(), this.camera.entity, val);
    }

    protected _needQueryViewMatrix: boolean = true;
    protected _viewMatrix: Matrix = Matrix.Identity();
    protected _projectMatrix: Matrix = Matrix.Identity();
    protected _transformMatrix: Matrix = Matrix.Identity();
    public getViewMatrix() {
        this._queryCameraMatrix();
        return this._viewMatrix;
    }

    public getProjectMatrix() {
        this._queryCameraMatrix();
        return this._projectMatrix;
    }

    public getTransformationMatrix() {
        this._queryCameraMatrix();
        return this._transformMatrix;
    }

    protected _queryCameraMatrix() {
        this._isSynchronized();
        if (this._needQueryViewMatrix) {
            if (p3d.p3d_query_view_matrix(Engine.native(), Engine.param(), this.camera.entity, Node.MatrixReceive)) {
                Matrix.FromArrayToRef(Node.MatrixReceive, 0, this._viewMatrix);
            }
            if (p3d.p3d_query_project_matrix(Engine.native(), Engine.param(), this.camera.entity, Node.MatrixReceive)) {
                Matrix.FromArrayToRef(Node.MatrixReceive, 0, this._projectMatrix);
            }
            if (p3d.p3d_query_viewproject_matrix(Engine.native(), Engine.param(), this.camera.entity, Node.MatrixReceive)) {
                Matrix.FromArrayToRef(Node.MatrixReceive, 0, this._transformMatrix);
            }
            this._needQueryViewMatrix = false;
        }
    }
    public forceInclude(mesh: IEntityD3, doadd: boolean = true) {
        p3d.p3d_viewer_force_include(Engine.commands(), this.camera.entity, mesh.entity, doadd);
    }

    public render(key?: string, ctx?: GlContext) {
        if (this.isDisposed()) {
            return;
        }

        ctx = this._scene.ctx;
    }

    public _renderX: number = 0;
    public _renderY: number = 0;
    public _renderWidth: number = 100;
    public _renderHeight: number = 100;
    protected _aspect: Option<number> = null;
    public set aspect(val: Option<number>) {
        if (this._aspect != val) {
            this._aspect = val;
            this._needQueryViewMatrix = true;
            p3d.p3d_camera_aspect(Engine.commands(), this.camera.entity, this._aspect);
        }
    }
    public get aspect(): Option<number> {
        return this._aspect;
    }

    protected _nearFarDirty: boolean = false;
    protected _minZ: number = 1;
    public set minZ(val: number) {
        val = Math.max(0.1, val);
        if (this._minZ != val) {
            this._minZ = val;
            this._nearFarDirty = true;
            this._needQueryViewMatrix = true;
            this._scene.cmdsForceDirty.set(this.camera.entity, this);
        }
    }
    
    protected _maxZ: number = 1001;
    public set maxZ(val: number) {
        val = Math.max(0.1, val);
        if (this._maxZ != val) {
            this._maxZ = val;
            this._nearFarDirty = true;
            this._needQueryViewMatrix = true;
            this._scene.cmdsForceDirty.set(this.camera.entity, this);
        }
    }

    dispose(): void {
        this.camera.dispose();
        super.dispose();
        this.camera = undefined;
    }

    /**
     * @deprecated
     * @param canvas 
     * @param flag 
     */
    attachControl(canvas: HTMLCanvasElement, flag: boolean = false) {
        //
    }
    public _isSynchronized(): boolean {
        let result = super._isSynchronized();
        if (result) {
            this._needQueryViewMatrix = true;
        }

        return result;
    }
}
