
import { DefaultKey, IAnimationAttach
    , ICameraState, IDisposable, IMaterialState2, IMeshState, IPerformance, IResourceState, ISceneState, ITextureLoaderState, ITransformState
} from "../../interface/index";
import { FrameCurve, Observable, Vector3 } from "../index";
import { EngineStore } from "./engineStore";
import { EntityID, EngineCapabilities, Nullable, IMemory } from "../entity";
import { IEntityD3, ISceneD3, IDeviceLimits, IDeviceFeatures, IDeviceDownLevelFlags, IAnimationEventDataEnd, IAnimationEventDataFrame, IAnimationEventDataLoop, IAnimationEventDataStart, IAnimationGroup3D, } from "../types";
import { ExtensionTools } from "../../base";
import { DomManagement } from "../misc/domManagement";
import { IActionSetScene3D, ICommandsExchangeD3, INativeEngine, IMaterialUniformDefines, INodematerialIncludes, IP3DShaderMeta, IP3DShaderVaryings, p3d } from "../../context";


export interface IMaterialRegistInfo {
    uniforms: IMaterialUniformDefines,
    vs_define_code: string, fs_define_code: string, vs_code: string, fs_code: string,
    includes: INodematerialIncludes,
    instance_code: string, varyings: IP3DShaderVaryings, binds_defines_base?: number
}

export class Entity implements IDisposable, IAnimationAttach {
    private static Temp: Float64Array = new Float64Array(1);
    public static ToString(entity: number) {
        Entity.Temp[0] = entity;
        return `${new Uint32Array(Entity.Temp.buffer)[0]}v${new Uint32Array(Entity.Temp.buffer)[1]}`;
    }
    public static Dispose(entity: number) {
        if (entity != 0) {
            p3d.p3d_dispose(Engine.commands(), entity);
        }
    }
    getEngine() {
        return Engine.getEngine();
    }

    public static DIRTY_MAT_FLOAT               = 0b0000_0000_0000_0000_0000_0000_0000_0001;
    public static DIRTY_MAT_VEC2                = 0b0000_0000_0000_0000_0000_0000_0000_0010;
    public static DIRTY_MAT_VEC4                = 0b0000_0000_0000_0000_0000_0000_0000_0100;
    public static DIRTY_MAT_UINT                = 0b0000_0000_0000_0000_0000_0000_0000_1000;
    public static DIRTY_MAT_TEXTURE             = 0b0000_0000_0000_0000_0000_0000_0001_0000;
    public static DIRTY_MAT_SHADER              = 0b0000_0000_0000_0000_0000_0000_0010_0000;

    public static DIRTY_NODE_PARENT             = 0b0000_0000_0000_0000_0000_0000_0000_0001;
    public static DIRTY_NODE_POSITION           = 0b0000_0000_0000_0000_0000_0000_0000_0010;
    public static DIRTY_NODE_EULER              = 0b0000_0000_0000_0000_0000_0000_0000_0100;
    public static DIRTY_NODE_SCALING            = 0b0000_0000_0000_0000_0000_0000_0000_1000;
    public static DIRTY_NODE_ROTATION           = 0b0000_0000_0000_0000_0000_0000_0001_0000;
    public static DIRTY_NODE_LAYERMASK          = 0b0000_0000_0000_0000_0000_0000_0010_0000;
    public static DIRTY_NODE_ENABLE             = 0b0000_0000_0000_0000_0000_0000_0100_0000;
    public static DIRTY_NODE_PICK               = 0b0000_0000_0000_0000_0000_0000_1000_0000;

    public static DIRTY_MESH_MATERIAL           = 0b0000_0000_0000_0000_0000_0000_0000_0001;
    public static DIRTY_MESH_RENDER_QUENE       = 0b0000_0000_0000_0000_0000_0000_0000_0010;
    public static DIRTY_MESH_RENDER_ALIGN       = 0b0000_0000_0000_0000_0000_0000_0000_0100;
    public static DIRTY_MESH_BLEND              = 0b0000_0000_0000_0000_0000_0000_0000_1000;
    public static DIRTY_MESH_CULLMODE           = 0b0000_0000_0000_0000_0000_0000_0001_0000;
    public static DIRTY_MESH_FRONTFACE          = 0b0000_0000_0000_0000_0000_0000_0010_0000;
    public static DIRTY_MESH_DEPTH_TEST         = 0b0000_0000_0000_0000_0000_0000_0100_0000;
    public static DIRTY_MESH_DEPTH_WRITE        = 0b0000_0000_0000_0000_0000_0000_1000_0000;

    public static DIRTY_CAMERA_SIZE             = 0b0000_0000_0000_0000_0000_0000_0000_0001;
    public static DIRTY_CAMERA_TARGET           = 0b0000_0000_0000_0000_0000_0000_0000_0010;
    public static DIRTY_CAMERA_FOV              = 0b0000_0000_0000_0000_0000_0000_0000_0100;

    protected _entity: EntityID;
    /**
     * get entity
     */
    get entity(): EntityID {
        return this._entity;
    }
    get entityString() {
        // return `D${p3d.get_entity_offset(this._entity)}v${p3d.get_entity_version(this._entity)}`;
        return `D` + this.uniqueId;
    }
    constructor(entity: EntityID) {
        this._entity = entity;
        // console.warn("Create " + this.uniqueID);
    }
    // onDisposed: Observable<number>;
    
    protected _animationGroups: Map<number, IAnimationGroup3D> = new Map();
    public get animationGroupMap() {
        return this._animationGroups;
    }

    public get uniqueId() {
        return Entity.ToString(this._entity);
    }

    public readonly onDisposeObservable: Observable<Entity> = new Observable();

    protected _disposed: boolean = false;
    dispose(disposeEntity: boolean = true): void {
        if (this._disposed) {
            return;
        }

        this.onDisposeObservable.notifyObservers(this);

        Entity.Temp[0] = this._entity;
        // console.warn("Dispose " + this.uniqueID);

        this._animationGroups.forEach((item) => {
            item.dispose();
        });
        this._animationGroups.clear();
        this._animationGroups = undefined;

        this._disposed = true;

        if (disposeEntity) {
            this._disposeCall();
        }

        this.onDisposeObservable.clear();
        // this._entity = 0;
    }
    protected _disposeCall() {
        Entity.Dispose(this._entity);
    }
    get _isDisposed() {
        return this._disposed;
    };
}

export class DefaultKeyRes implements IDisposable {
    private _dk: DefaultKey;
    /**
     * get entity
     */
    get dk(): DefaultKey {
        return this._dk;
    }
    constructor(dk: DefaultKey) {
        this._dk = dk;
    }

    protected _disposed: boolean = false;
    dispose(): void {
        if (this.isDisposed) {
            return;
        }

        this._disposed = true;

        if (this._dk != 0) {
            // p3d.p3d_dispose(Engine.native(), Engine.param(), this._entity);
        }

        this.onDisposeObservable.notifyObservers(this);
        this._dk = 0;
    }
    get isDisposed() {
        return this._disposed;
    }
    get _isDisposed() {
        return this._disposed;
    }
    public readonly onDisposeObservable: Observable<DefaultKeyRes> = new Observable();
}

const compressTextureFormatInfo = {
    astc: 'ktx',
    s3tc: 'ktx'
};


export class Engine {
    private static _errorFlag: boolean = true;
    private static _errorInfo: Uint32Array = new Uint32Array(4096);
    private static _stateInfo: Float32Array = new Float32Array(64);
    private static _memoryInfo: Float64Array = new Float64Array(256);
    private static _engine: Engine;
    public static getEngine(): Engine {
        return Engine._engine;
    }
    private static _DeviceInfo() {
        p3d.p3d_device_limis(Engine._engine._nativeengine, Engine._errorInfo);
        let i = 0;
        let data = Engine._errorInfo;
        let features: IDeviceFeatures = <any>{};
        let downlevels: IDeviceDownLevelFlags = <any>{};
        let limits: IDeviceLimits = <any>{};

                  features.ADDRESS_MODE_CLAMP_TO_ZERO = data[i] == 1;
        i += 1;   features.BGRA8UNORM_STORAGE = data[i] == 1;
        i += 1;   features.BUFFER_BINDING_ARRAY = data[i] == 1;
        i += 1;   features.CLEAR_TEXTURE = data[i] == 1;
        i += 1;   features.CONSERVATIVE_RASTERIZATION = data[i] == 1;
        i += 1;   features.DEPTH32FLOAT_STENCIL8 = data[i] == 1;
        i += 1;   features.DEPTH_CLIP_CONTROL = data[i] == 1;
        i += 1;   features.DUAL_SOURCE_BLENDING = data[i] == 1;
        i += 1;   features.FLOAT32_FILTERABLE = data[i] == 1;
        i += 1;   features.INDIRECT_FIRST_INSTANCE = data[i] == 1;
        i += 1;   features.MAPPABLE_PRIMARY_BUFFERS = data[i] == 1;
        i += 1;   features.MULTIVIEW = data[i] == 1;
        i += 1;   features.MULTI_DRAW_INDIRECT = data[i] == 1;
        i += 1;   features.MULTI_DRAW_INDIRECT_COUNT = data[i] == 1;
        i += 1;   features.PARTIALLY_BOUND_BINDING_ARRAY = data[i] == 1;
        i += 1;   features.PIPELINE_STATISTICS_QUERY = data[i] == 1;
        i += 1;   features.POLYGON_MODE_LINE = data[i] == 1;
        i += 1;   features.POLYGON_MODE_POINT = data[i] == 1;
        i += 1;   features.PUSH_CONSTANTS = data[i] == 1;
        i += 1;   features.RAY_QUERY = data[i] == 1;
        i += 1;   features.RAY_TRACING_ACCELERATION_STRUCTURE = data[i] == 1;
        i += 1;   features.RG11B10UFLOAT_RENDERABLE = data[i] == 1;
        i += 1;   features.SAMPLED_TEXTURE_AND_STORAGE_BUFFER_ARRAY_NON_UNIFORM_INDEXING = data[i] == 1;
        i += 1;   features.SHADER_EARLY_DEPTH_TEST = data[i] == 1;
        i += 1;   features.SHADER_F16 = data[i] == 1;
        i += 1;   features.SHADER_F64 = data[i] == 1;
        i += 1;   features.SHADER_I16 = data[i] == 1;
        i += 1;   features.SHADER_PRIMITIVE_INDEX = data[i] == 1;
        i += 1;   features.SHADER_UNUSED_VERTEX_OUTPUT = data[i] == 1;
        i += 1;   features.SPIRV_SHADER_PASSTHROUGH = data[i] == 1;
        i += 1;   features.STORAGE_RESOURCE_BINDING_ARRAY = data[i] == 1;
        i += 1;   features.TEXTURE_ADAPTER_SPECIFIC_FORMAT_FEATURES = data[i] == 1;
        i += 1;   features.TEXTURE_BINDING_ARRAY = data[i] == 1;
        i += 1;   features.TEXTURE_COMPRESSION_ASTC = data[i] == 1;
        i += 1;   features.TEXTURE_COMPRESSION_ASTC_HDR = data[i] == 1;
        i += 1;   features.TEXTURE_COMPRESSION_BC = data[i] == 1;
        i += 1;   features.TEXTURE_COMPRESSION_ETC2 = data[i] == 1;
        i += 1;   features.TEXTURE_FORMAT_16BIT_NORM = data[i] == 1;
        i += 1;   features.TEXTURE_FORMAT_NV12 = data[i] == 1;
        i += 1;   features.TIMESTAMP_QUERY = data[i] == 1;
        i += 1;   features.TIMESTAMP_QUERY_INSIDE_PASSES = data[i] == 1;
        i += 1;   features.UNIFORM_BUFFER_AND_STORAGE_TEXTURE_ARRAY_NON_UNIFORM_INDEXING = data[i] == 1;
        i += 1;   features.VERTEX_ATTRIBUTE_64BIT = data[i] == 1;
        i += 1;   features.VERTEX_WRITABLE_STORAGE = data[i] == 1;
        
        i += 1; downlevels.ANISOTROPIC_FILTERING = data[i] == 1;
        i += 1; downlevels.BASE_VERTEX = data[i] == 1;
        i += 1; downlevels.BUFFER_BINDINGS_NOT_16_BYTE_ALIGNED = data[i] == 1;
        i += 1; downlevels.COMPARISON_SAMPLERS = data[i] == 1;
        i += 1; downlevels.COMPUTE_SHADERS = data[i] == 1;
        i += 1; downlevels.CUBE_ARRAY_TEXTURES = data[i] == 1;
        i += 1; downlevels.DEPTH_BIAS_CLAMP = data[i] == 1;
        i += 1; downlevels.DEPTH_TEXTURE_AND_BUFFER_COPIES = data[i] == 1;
        i += 1; downlevels.FRAGMENT_STORAGE = data[i] == 1;
        i += 1; downlevels.FRAGMENT_WRITABLE_STORAGE = data[i] == 1;
        i += 1; downlevels.FULL_DRAW_INDEX_UINT32 = data[i] == 1;
        i += 1; downlevels.INDEPENDENT_BLEND = data[i] == 1;
        i += 1; downlevels.INDIRECT_EXECUTION = data[i] == 1;
        i += 1; downlevels.MULTISAMPLED_SHADING = data[i] == 1;
        i += 1; downlevels.NON_POWER_OF_TWO_MIPMAPPED_TEXTURES = data[i] == 1;
        i += 1; downlevels.READ_ONLY_DEPTH_STENCIL = data[i] == 1;
        i += 1; downlevels.SURFACE_VIEW_FORMATS = data[i] == 1;
        i += 1; downlevels.UNRESTRICTED_EXTERNAL_TEXTURE_COPIES = data[i] == 1;
        i += 1; downlevels.UNRESTRICTED_INDEX_BUFFER = data[i] == 1;
        i += 1; downlevels.VERTEX_STORAGE = data[i] == 1;
        i += 1; downlevels.VIEW_FORMATS = data[i] == 1;
        i += 1; downlevels.WEBGPU_TEXTURE_FORMAT_SUPPORT = data[i] == 1;
        
        i += 1; limits.max_bind_groups = data[i];
        i += 1; limits.max_bindings_per_bind_group = data[i];
        i += 1; limits.max_non_sampler_bindings = data[i];
        i += 1; limits.max_sampled_textures_per_shader_stage = data[i];
        i += 1; limits.max_samplers_per_shader_stage = data[i];
        i += 1; limits.max_storage_buffer_binding_size = data[i];
        i += 1; limits.max_texture_array_layers = data[i];
        i += 1; limits.max_texture_dimension_1d = data[i];
        i += 1; limits.max_texture_dimension_2d = data[i];
        i += 1; limits.max_texture_dimension_3d = data[i];
        i += 1; limits.max_uniform_buffer_binding_size = data[i];
        i += 1; limits.max_uniform_buffers_per_shader_stage = data[i];
        i += 1; limits.max_vertex_attributes = data[i];
        i += 1; limits.max_vertex_buffer_array_stride = data[i];
        i += 1; limits.max_vertex_buffers = data[i];
        i += 1; limits.min_storage_buffer_offset_alignment = data[i];
        i += 1; limits.min_uniform_buffer_offset_alignment = data[i];

        Engine._deviceDownLevels = downlevels;
        Engine._deviceFatures = features;
        Engine._deviceLimits = limits;

        let caps: EngineCapabilities = <any>{};
        caps.astc = features.TEXTURE_COMPRESSION_ASTC;
        caps.canUseTimestampForTimerQuery = features.TIMESTAMP_QUERY;
        caps.colorBufferFloat = features.RG11B10UFLOAT_RENDERABLE;
        caps.depthTextureExtension = features.DEPTH32FLOAT_STENCIL8;
        caps.fragmentDepthSupported = downlevels.DEPTH_TEXTURE_AND_BUFFER_COPIES;
        caps.highPrecisionShaderSupported = features.SHADER_F64;
        caps.instancedArrays = features.INDIRECT_FIRST_INSTANCE;
        // caps.maxAnisotropy = limits.;
        // caps.maxCombinedTexturesImageUnits = ;
        caps.maxCubemapTextureSize = limits.max_texture_dimension_3d;
        caps.maxFragmentUniformVectors = limits.max_bind_groups * limits.max_bindings_per_bind_group;
        // caps.maxMSAASamples
        caps.maxRenderTextureSize = limits.max_texture_dimension_2d;
        caps.maxSamples = limits.max_samplers_per_shader_stage;
        caps.maxTextureSize = limits.max_texture_dimension_2d;
        caps.maxTexturesImageUnits = limits.max_sampled_textures_per_shader_stage;
        // caps.maxVaryingVectors
        caps.maxVertexAttribs = limits.max_vertex_attributes;
        // caps.maxVertexTextureImageUnits
        caps.maxVertexUniformVectors = limits.max_non_sampler_bindings;
        
        caps.etc2 = features.TEXTURE_COMPRESSION_ETC2;
        caps.multiview = features.MULTIVIEW;
        caps.textureFloat = features.FLOAT32_FILTERABLE;
        caps.textureFloatLinearFiltering = features.FLOAT32_FILTERABLE;
        // caps.timerQuery = features.TIMESTAMP_QUERY;
        caps.uintIndices = downlevels.FULL_DRAW_INDEX_UINT32;

        Engine._caps = caps;
    
        return;
    }
    public static Initial(canvas: HTMLCanvasElement) {
        if (!Engine._engine) {
            Vector3.EngineStore = EngineStore;

            Engine._engine = new Engine(canvas);
            Engine._engine._nativeengine = window["engine"];

            Engine._DeviceInfo();

            // 压缩纹理适配
            const compressed_tex_format = ExtensionTools.QuerySupportCompressedTextureFormat();
            Engine._engine._textureFormatInUse = compressTextureFormatInfo[compressed_tex_format];
            // logger.warn(`TextureFormatInUse : ${compressed_tex_format}`);

            p3d.p3d_init_engine(Engine._engine._nativeengine);
            Engine._engine._params = p3d.ActionSetScene3D.create(Engine._engine._nativeengine);
            Engine._engine._commands = p3d.CommandsExchangeD3.create();
            // getGlobal().setPermanent(() => {
            //     Engine._engine?._renderLoop();
            // });
            ExtensionTools.setPermanentBefore(() => {
                let engine = Engine._engine;
                if (engine) {
                    let now = Date.now();
                    engine._deltaTime = now - engine._time;
                    engine._time = now;
                }
                Engine.beforeCall();
                engine._renderLoop();
            });

            FrameCurve.Init();

            (<any>window).engine3d = Engine._engine;
        }
    }
    public static GetRenderContent(): {
        contextWidth: number;
        contextHeight: number;
        renderX: number;
        renderY: number;
        renderWidth: number;
        renderHeight: number;
        layoutSize: any;
        styleTop: number;
        styleBottom: number;
        styleLeft: number;
        styleSize: {
            width: number,
            height: number,
        };
        styleRight: number;
    } {
        return undefined;
    }
    // private static cmdsTargetAnimation: [number, number, number, number, number][];
    // public static addCmdTargetAnimation(curve_key: number, property: number, id_scene: number, group: number, curve_target: number) {
    //     this.cmdsTargetAnimation.push([curve_key, property, id_scene, group, curve_target]);
    // }

    public static RecordMatrialsMap: Map<string, IP3DShaderMeta> = new Map();
    public static WaitMatrialsMap: Map<string, IMaterialRegistInfo > = new Map();
    public static CheckShader(shaderName: string) {
        return Engine.RecordMatrialsMap.get(shaderName) || Engine.WaitMatrialsMap.get(shaderName);
    }
    public static RegistShader(
        key: string, uniforms: IMaterialUniformDefines,
        vs_define_code: string, fs_define_code: string, vs_code: string, fs_code: string,
        includes: INodematerialIncludes,
        instance_code: string, varyings: IP3DShaderVaryings, binds_defines_base?: number
    ) {
        let params = { uniforms, vs_define_code, fs_define_code, vs_code, fs_code, includes, instance_code, varyings, binds_defines_base };
        Engine.WaitMatrialsMap.set(key, params );
    }
    private static _RegistShaders() {
        Engine.WaitMatrialsMap.forEach((item, key) => {
            let result = p3d.p3d_regist_material(
                Engine.native(), Engine.param(),
                key, item.uniforms, item.vs_define_code, item.fs_define_code, item.vs_code, item.fs_code, item.includes,
                item.instance_code, item.varyings, item.binds_defines_base
            );
            Engine.RecordMatrialsMap.set(key, result);
            
            item.varyings.free && item.varyings.free();
            item.uniforms.free && item.uniforms.free();
            item.includes.free && item.includes.free();
        });

        Engine.WaitMatrialsMap.clear();
    }

    private static _beforeCalls: Map<Function, Function> = new Map();
    public static registBeforeRun(call: Function) {
        Engine._beforeCalls.set(call, call);
    }
    public static unregistBeforeRun(call: Function) {
        Engine._beforeCalls.delete(call);
    }
    public static engineActive(flag: boolean) {
        p3d.p3d_engine_state(Engine.native(), Engine.param(), flag);
    }
    public static beforeCall() {
        Engine._RegistShaders();

        Engine._engine?._scenes.forEach((scene) => {
            scene.run();
        });
        Engine._beforeCalls.forEach((call) => {
            try {
                call();
            } catch (e) {
                console.error(e);
            }
        });

        Engine.disposeScene.forEach((scene) => {
            p3d.p3d_scene_dispose(Engine.commands(), scene);
        });
        Engine.disposeScene.clear();

        p3d.p3d_commands_exchange(Engine.native(), Engine.param(), Engine.commands());
    }
    public static disposeScene: Set<number> = new Set();
    private static _afterCalls: Map<Function, Function> = new Map();
    public static registAfterRun(call: Function) {
        Engine._afterCalls.set(call, call);
    }
    public static unregistAfterRun(call: Function) {
        Engine._afterCalls.delete(call);
    }
    public static afterCall() {
        Engine._engine.animationGroupEnvents();

        Engine._afterCalls.forEach((call) => {
            try {
                call();
            } catch (e) {
                console.error(e);
            }
        })
    }
    public static native() {
        return Engine._engine?._nativeengine;
    }
    public static commands(): ICommandsExchangeD3 {
        return Engine._engine?._commands;
    }
    public static param(): IActionSetScene3D {
        return Engine._engine?._params;
    }
    public static DebugState(flag: boolean) {
        p3d.p3d_engine_debug(Engine.native(), Engine.param(), flag);
    }
    public static ResourceState(state: IResourceState, performance: IPerformance) {
        p3d.p3d_query_resource_state(Engine.native(), Engine.param(), Engine.commands(), Engine._stateInfo);
        state._0_count_bindbuffer           = Engine._stateInfo[0];
        state._1_count_bindgroup            = Engine._stateInfo[1];
        state._2_count_gltf                 = Engine._stateInfo[2];
        state._3_count_imgtexture           = Engine._stateInfo[3];
        state._4_count_shader               = Engine._stateInfo[4];
        state._5_count_pipeline             = Engine._stateInfo[5];
        state._6_size_vertexbuffer          = Engine._stateInfo[6];
        state._7_count_vertexbuffer         = Engine._stateInfo[7];
        state._8_count_shadermeta           = Engine._stateInfo[8];
        state._9_mem_shadermeta             = Engine._stateInfo[9];
        state._10_mem_shader                = Engine._stateInfo[10];
        state._11_mem_bindbuffer            = Engine._stateInfo[11];
        state._12_mem_imgtexture            = Engine._stateInfo[12];
        
        state._14_count_passset0            = Engine._stateInfo[14];
        state._15_count_passset1            = Engine._stateInfo[15];
        state._16_count_passset2            = Engine._stateInfo[16];
        state._17_count_passbindgroups      = Engine._stateInfo[17];
        state._18_count_passshader          = Engine._stateInfo[18];
        state._19_count_passpipeline        = Engine._stateInfo[19];
        state._20_count_passdraw            = Engine._stateInfo[20];

        state._21_count_passmat             = Engine._stateInfo[21];
        state._22_count_passtexs            = Engine._stateInfo[22];
        state._23_count_vertex              = Engine._stateInfo[23];
        state._24_mem_world                 = Engine._stateInfo[24];
        state._25                 = Engine._stateInfo[25];
        state._26                 = Engine._stateInfo[26];
        state._27                 = Engine._stateInfo[27];
        state._28                 = Engine._stateInfo[28];

        p3d.p3d_query_performance_state(Engine.native(), Engine.param(), Engine._stateInfo);
        performance._0_animation            = Engine._stateInfo[0];
        performance._1_animationgroup       = Engine._stateInfo[1];
        performance._2_particlesystem       = Engine._stateInfo[2];
        performance._3_worldmatrix          = Engine._stateInfo[3];
        performance._4_culling              = Engine._stateInfo[4];
        performance._5_gltfanaly            = Engine._stateInfo[5];
        performance._6_drawobjs             = Engine._stateInfo[6];
        performance._7_uniformupdate        = Engine._stateInfo[7];
        performance._8_uniformbufferupdate  = Engine._stateInfo[8];
        performance._9_trailrenderer        = Engine._stateInfo[9];
    }
    public static Memory(state: IMemory) {
        let count = p3d.p3d_query_resource_memory(Engine.native(), Engine.param(), Engine._memoryInfo);
        state.itemcount = count;
        state.data = Engine._memoryInfo;
    }
    public static SceneState(scene: Entity, state: ISceneState) {
        p3d.p3d_query_scene_state(Engine.native(), Engine.param(), scene.entity, Engine._stateInfo);
        state._0_drawcalls                  = Engine._stateInfo[0];
        state._1_count_vertex               = Engine._stateInfo[1];
        state._2_count_particlesystem       = Engine._stateInfo[2];
        state._3_count_particles            = Engine._stateInfo[3];
        state._4_count_animationgroup       = Engine._stateInfo[4];
        state._5_count_trail                = Engine._stateInfo[5];
        state._6_count_trail_points         = Engine._stateInfo[6];
    }
    public static MaterialState(state: IMaterialState2) {
        p3d.p3d_material_state(Engine.native(), Engine.param(), Engine._stateInfo);
        state._0_count                      = Engine._stateInfo[0];
        state._1_ready                      = Engine._stateInfo[1];
        state._2_tex0                       = Engine._stateInfo[2];
        state._3_tex0_ready                 = Engine._stateInfo[3];
        state._4_tex1                       = Engine._stateInfo[4];
        state._5_tex1_ready                 = Engine._stateInfo[5];
        state._6_tex2                       = Engine._stateInfo[6];
        state._7_tex2_ready                 = Engine._stateInfo[7];
        state._8_tex3                       = Engine._stateInfo[8];
        state._9_tex3_ready                 = Engine._stateInfo[9];
        state._10_tex4                      = Engine._stateInfo[10];
        state._11_tex4_ready                = Engine._stateInfo[11];
        state._12_tex5                      = Engine._stateInfo[12];
        state._13_tex5_ready                = Engine._stateInfo[13];
        state._14_tex6                      = Engine._stateInfo[14];
        state._15_tex6_ready                = Engine._stateInfo[15];
    }
    public static MeshState(state: IMeshState, scene: IEntityD3) {
        p3d.p3d_mesh_state(Engine.native(), Engine.param(), scene?.entity, Engine._stateInfo);
        state._0_abstruct_mesh_count        = Engine._stateInfo[0];
        state._1_abstruct_mesh_count_enable = Engine._stateInfo[1];
        state._2_mesh_count                 = Engine._stateInfo[2];
        state._3_mesh_count_geometry_ready  = Engine._stateInfo[3];
        state._4_instance                   = Engine._stateInfo[4];
    }
    public static TransformState(state: ITransformState, scene: IEntityD3) {
        p3d.p3d_transform_state(Engine.native(), Engine.param(), scene?.entity, Engine._stateInfo);
        state._0_count                      = Engine._stateInfo[0];
        state._1_enable                     = Engine._stateInfo[1];
        state._2_global_eanble              = Engine._stateInfo[2];
        state._3_calc_local_matrix          = Engine._stateInfo[3];
        state._4_calc_world_matrix          = Engine._stateInfo[4];
        state._5_max_level                  = Engine._stateInfo[5];
    }
    public static CameraState(state: ICameraState, camera: IEntityD3) {
        p3d.p3d_camera_state(Engine.native(), Engine.param(), camera?.entity, Engine._stateInfo);
        state._0_includes                   = Engine._stateInfo[0];
        state._1_culling                    = Engine._stateInfo[1];
        state._2_culling_time               = Engine._stateInfo[2];
    }
    public static TextureLoaderState(state: ITextureLoaderState) {
        p3d.p3d_texture_loader_state(Engine.native(), Engine.param(), Engine._stateInfo);
        state._0_image_count                = Engine._stateInfo[0];
        state._1_image_fail                 = Engine._stateInfo[1];
        state._2_image_success              = Engine._stateInfo[2];
        state._3_image_waiting              = Engine._stateInfo[3];
        state._4_texview_count              = Engine._stateInfo[4];
        state._5_texview_fail               = Engine._stateInfo[5];
        state._6_texview_success            = Engine._stateInfo[6];
        state._7_texview_waiting            = Engine._stateInfo[7];
        state._8_imgtexture                 = Engine._stateInfo[8];
        state._9_imgtexture_mem             = Engine._stateInfo[9];
        state._10_imgtextureview            = Engine._stateInfo[10];
        state._11_imgtextureview_mem        = Engine._stateInfo[11];
    }
    public static ErrorState(func: (entity: number, error: number) => void, flag) {
        Engine._errorFlag = flag;
        let count = p3d.p3d_errors(Engine.native(), Engine.param(), Engine._errorInfo, Engine._errorFlag);
        for (let i = 0; i < count / 2; i++) {
            func(Engine._errorInfo[i * 2 + 0], Engine._errorInfo[i * 2 + 1]);
        }
    }
    /**
     * 纹理动态合并配置
     * @param format 纹理格式
     * @param maxlayer 纹理数组level
     * @param maxsize 纹理尺寸
     * @param maxcount 最大合并数目
     * @description 
     * * 目标 Texture 需要设置 cancombine 为 true,
     * * maxlayer 不能为 1, 不能被 6 整除
     */
    public static RegistTextureCombine(format: number, maxlayer: number, maxsize: number, maxcount: number) {
        p3d.p3d_texture_combine_param(Engine.native(), format, maxlayer, maxsize, maxcount);
    }
    public static queryChildren(id: number, info: Float64Array): number {
        return p3d.p3d_query_children(Engine.native(), Engine.param(), id, info);
    }
    public static queryMeshInfo(id: number, info: Uint32Array): boolean {
        return p3d.p3d_query_mesh_info(Engine.native(), Engine.param(), id, info);
    }
    public static queryMaterialTextureInfo(id: number, info: Uint32Array): boolean {
        return p3d.p3d_query_material_info(Engine.native(), Engine.param(), id, info);
    }
    public static _caps: EngineCapabilities;
    public static _deviceLimits: IDeviceLimits;
    public static _deviceFatures: IDeviceFeatures;
    public static _deviceDownLevels: IDeviceDownLevelFlags;
    public static CreateVertexBuffer(key: string, data: Float32Array) {
        p3d.p3d_create_vertex_buffer(Engine.native(), Engine.param(), key, data, data.length);
    }
    public static CreateIndiceBuffer(key: string, data: Uint16Array) {
        p3d.p3d_create_indices_buffer(Engine.native(), Engine.param(), key, data, data.length);
    }
    public static UpdateDataTexture(key: string, data: Uint8Array, xoffset: number, yoffset: number, width: number, height: number, aspect?: number) {
        p3d.p3d_update_data_texture(Engine.commands(), ExtensionTools.getAtom(key), data, xoffset, yoffset, width, height, aspect);
    }
    public static CreateDataTexture(key: string, width: number, height: number, format: number, aspect?: number) {
        p3d.p3d_create_data_texture(Engine.commands(), ExtensionTools.getAtom(key), width, height, format, aspect);
    }
    public static DeleteDataTexture(key: string) {
        p3d.p3d_remove_data_texture(Engine.commands(), ExtensionTools.getAtom(key));
    }

    /**
     * Ask the browser to promote the current element to pointerlock mode
     * @param element defines the DOM element to promote
     */
    static _RequestPointerlock(element: HTMLElement): void {
        element.requestPointerLock = element.requestPointerLock || (<any>element).msRequestPointerLock || (<any>element).mozRequestPointerLock || (<any>element).webkitRequestPointerLock;
        if (element.requestPointerLock) {
            element.requestPointerLock();
        }
    }
    
    public get _deviceLimits() { return Engine._deviceLimits };
    public get _deviceFatures() { return Engine._deviceFatures };
    public get _deviceDownLevels() { return Engine._deviceDownLevels };
    public get _caps() { return Engine._caps };

    private _nativeengine: INativeEngine;
    private _params: IActionSetScene3D;
    private _commands: ICommandsExchangeD3;
    private _scenes: Map<ISceneD3, ISceneD3> = new Map();
    private _time: number = Date.now();
    public get time() {
        return this._time;
    }
    // FPS
    private _fps = 60;
    private _deltaTime: number = 0;

    // Deterministic lockstepMaxSteps
    private _deterministicLockstep: boolean = false;
    private _lockstepMaxSteps: number = 0;
    private _timeStep: number = 1 / 60;

    _textureFormatInUse: string;

    /**
     * Observable event triggered each time the canvas loses focus
     */
    public onCanvasBlurObservable = new Observable<Engine>();

    /**
     * Observable event triggered each time the canvas gains focus
     */
    public onCanvasFocusObservable = new Observable<Engine>();
    /**
     * Observable event triggered each time the canvas receives pointerout event
     */
    public onCanvasPointerOutObservable = new Observable<PointerEvent>();

    /**
     * Turn this value on if you want to pause FPS computation when in background
     */
    public disablePerformanceMonitorInBackground = false;
    protected _windowIsBackground = false;
    private _pointerLockRequested: boolean;
    /**
     * Gets a boolean indicating if the engine is currently rendering in fullscreen mode
     */
    public isFullscreen = false;
    /**
     * Gets a boolean indicating if the pointer is currently locked
     */
    public isPointerLock = false;
    // Focus
    private _onFocus: () => void;
    private _onBlur: () => void;
    private _onCanvasPointerOut: (event: PointerEvent) => void;
    private _onCanvasBlur: () => void;
    private _onCanvasFocus: () => void;

    private _onFullscreenChange: () => void;
    private _onPointerLockChange: () => void;
    protected _renderingCanvas: Nullable<HTMLCanvasElement>;
    /**
     * Gets host window
     * @returns the host window object
     */
    public getHostWindow(): Nullable<Window> {
        if (!DomManagement.IsWindowObjectExist()) {
            return null;
        }

        if (this._renderingCanvas && this._renderingCanvas.ownerDocument && this._renderingCanvas.ownerDocument.defaultView) {
            return this._renderingCanvas.ownerDocument.defaultView;
        }

        return window;
    }

    constructor(canvas: HTMLCanvasElement) {
        
        this._onCanvasFocus = () => {
            this.onCanvasFocusObservable.notifyObservers(this);
        };

        this._onCanvasBlur = () => {
            this.onCanvasBlurObservable.notifyObservers(this);
        };

        canvas?.addEventListener("focus", this._onCanvasFocus);
        canvas?.addEventListener("blur", this._onCanvasBlur);
        
        this._onBlur = () => {
            if (this.disablePerformanceMonitorInBackground) {
                // this._performanceMonitor.disable();
            }
            this._windowIsBackground = true;
        };

        this._onFocus = () => {
            if (this.disablePerformanceMonitorInBackground) {
                // this._performanceMonitor.enable();
            }
            this._windowIsBackground = false;
        };

        this._onCanvasPointerOut = (ev) => {
            this.onCanvasPointerOutObservable.notifyObservers(ev);
        };

        canvas?.addEventListener("pointerout", this._onCanvasPointerOut);
        
        if (DomManagement.IsWindowObjectExist()) {
            let hostWindow = this.getHostWindow()!;
            hostWindow.addEventListener("blur", this._onBlur);
            hostWindow.addEventListener("focus", this._onFocus);

            let anyDoc = document as any;

            // Fullscreen
            this._onFullscreenChange = () => {

                if (anyDoc.fullscreen !== undefined) {
                    this.isFullscreen = anyDoc.fullscreen;
                } else if (anyDoc.mozFullScreen !== undefined) {
                    this.isFullscreen = anyDoc.mozFullScreen;
                } else if (anyDoc.webkitIsFullScreen !== undefined) {
                    this.isFullscreen = anyDoc.webkitIsFullScreen;
                } else if (anyDoc.msIsFullScreen !== undefined) {
                    this.isFullscreen = anyDoc.msIsFullScreen;
                }

                // Pointer lock
                if (this.isFullscreen && this._pointerLockRequested && canvas) {
                    Engine._RequestPointerlock(canvas);
                }
            };

            document?.addEventListener("fullscreenchange", this._onFullscreenChange, false);
            document?.addEventListener("mozfullscreenchange", this._onFullscreenChange, false);
            document?.addEventListener("webkitfullscreenchange", this._onFullscreenChange, false);
            document?.addEventListener("msfullscreenchange", this._onFullscreenChange, false);

            // Pointer lock
            this._onPointerLockChange = () => {
                this.isPointerLock = (anyDoc.mozPointerLockElement === canvas ||
                    anyDoc.webkitPointerLockElement === canvas ||
                    anyDoc.msPointerLockElement === canvas ||
                    anyDoc.pointerLockElement === canvas
                );
            };

            document?.addEventListener("pointerlockchange", this._onPointerLockChange, false);
            document?.addEventListener("mspointerlockchange", this._onPointerLockChange, false);
            document?.addEventListener("mozpointerlockchange", this._onPointerLockChange, false);
            document?.addEventListener("webkitpointerlockchange", this._onPointerLockChange, false);

            // // Create Audio Engine if needed.
            // if (!Engine.audioEngine && options.audioEngine && Engine.AudioEngineFactory) {
            //     Engine.audioEngine = Engine.AudioEngineFactory(this.getRenderingCanvas());
            // }
        }
    }

    public _renderLoop() {
        // let now = Date.now();
        // this._deltaTime = now - this._time;
        // this._time = now;

        // Engine.beforeCall();

        // this._scenes.forEach((scene) => {
        //     scene.render();
        // });

        Engine.afterCall();
    }
    /**
     * Gets the time spent between current and previous frame
     * @returns a number representing the delta time in ms
     */
    public getDeltaTime() {
        return this._deltaTime;
    }
    
    /**
     * Gets a boolean indicating that the engine is running in deterministic lock step mode
     * @see https://doc.babylonjs.com/babylon101/animations#deterministic-lockstep
     * @returns true if engine is in deterministic lock step mode
     */
    public isDeterministicLockStep = function () {
        return this._deterministicLockstep;
    };
    /**
     * Gets the max steps when engine is running in deterministic lock step
     * @see https://doc.babylonjs.com/babylon101/animations#deterministic-lockstep
     * @returns the max steps
     */
    public getLockstepMaxSteps = function () {
        return this._lockstepMaxSteps;
    };
    /**
     * Returns the time in ms between steps when using deterministic lock step.
     * @returns time step in (ms)
     */
    public getTimeStep = function () {
        return this._timeStep * 1000;
    };

    private animationEventCount = 512;
    private animationEventReceive: Float64Array = new Float64Array(this.animationEventCount * 4);
    public animationStartEvents: Map<number, IAnimationEventDataStart> = new Map();
    public animationEndEvents: Map<number, IAnimationEventDataEnd> = new Map();
    public animationLoopEvents: Map<number, IAnimationEventDataLoop> = new Map();
    public animationFrameEvents: Map<number, IAnimationEventDataFrame> = new Map();
    private animationGroupEnvents() {
        let eventCount = 1;
        while (eventCount > 0) {
            eventCount = p3d.p3d_query_anime_events(Engine.native(), Engine.param(), this.animationEventReceive, this.animationEventCount);
            for (let i = 0; i < eventCount; i++) {
                let idx = i * 4;
                let target = this.animationEventReceive[idx + 0];
                let group = this.animationEventReceive[idx + 1];
                let tag = this.animationEventReceive[idx + 2];
                let data = this.animationEventReceive[idx + 3];
    
                if (tag == 1) {
                    this.animationStartEvents.set(group, { group: target, target: target });
                } else if (tag  == 2) {
                    this.animationEndEvents.set(group, { group: target, target: target, });
                } else if (tag == 4) {
                    this.animationLoopEvents.set(group, { group: target, target: target, loop: data });
                } else if (tag == 8) {
                    this.animationFrameEvents.set(group, { group: target, target: target, data: data });
                }
            }
        }
    }

    /**
     * Indicates if the z range in NDC space is 0..1 (value: true) or -1..1 (value: false)
     */
    public readonly isNDCHalfZRange: boolean = false;
    
    /**
     * Gets or sets the epsilon value used by collision engine
     */
    public static CollisionsEpsilon = 0.001;

    _unpackFlipY() {
        
    }
}