import { Engine } from "../engines/engine";
import { Nullable, Observable, Observer } from "../../interface";
import { IScene } from "../../interface/scene";
import { KeyboardEventTypes, KeyboardInfo, PointerEventTypes, PointerInfo, PointerInfoPre } from "../events/index";
import { PickingInfo } from "../collisions/pickingInfo";
import { AbstractMesh } from "../meshes/abstractMesh";
import { Vector2 } from "../../maths/math.vector";
import { EventData, EventListener, KeyboardEventListener } from "../../interface/event";
import { ISceneD3 } from "../types";
import { DomManagement } from "../misc/domManagement";

/** @hidden */
class _ClickInfo {
    private _singleClick = false;
    private _doubleClick = false;
    private _hasSwiped = false;
    private _ignore = false;

    public get singleClick(): boolean {
        return this._singleClick;
    }
    public get doubleClick(): boolean {
        return this._doubleClick;
    }
    public get hasSwiped(): boolean {
        return this._hasSwiped;
    }
    public get ignore(): boolean {
        return this._ignore;
    }

    public set singleClick(b: boolean) {
        this._singleClick = b;
    }
    public set doubleClick(b: boolean) {
        this._doubleClick = b;
    }
    public set hasSwiped(b: boolean) {
        this._hasSwiped = b;
    }
    public set ignore(b: boolean) {
        this._ignore = b;
    }
}

export class InputManager {
    /** The distance in pixel that you have to move to prevent some events */
    public static DragMovementThreshold = 10; // in pixels
    /** Time in milliseconds to wait to raise long press events if button is still pressed */
    public static LongPressDelay = 500; // in milliseconds
    /** Time in milliseconds with two consecutive clicks will be considered as a double click */
    public static DoubleClickDelay = 300; // in milliseconds
    /** If you need to check double click without raising a single click at first click, enable this flag */
    public static ExclusiveDoubleClickMode = false;

    /** This is a defensive check to not allow control attachment prior to an already active one. If already attached, previous control is unattached before attaching the new one. */
    private _alreadyAttached = false;
    private _alreadyAttachedTo: Nullable<{
        addEventListener(type: string, listener: EventListener): void;
        removeEventListener(type: string, listener: EventListener): void;
    }>;
    private _alreadyAttachedToForKeyBoard: Nullable<{
        addEventListener(type: string, listener: KeyboardEventListener): void;
        removeEventListener(type: string, listener: KeyboardEventListener): void;
    }>;

    // Pointers
    private _wheelEventName = "";
    private _onPointerMove: EventListener;
    private _onPointerDown: EventListener;
    private _onPointerUp: EventListener;

    private _initClickEvent: (obs1: Observable<PointerInfoPre>, obs2: Observable<PointerInfo>, evt: PointerEvent, cb: (clickInfo: _ClickInfo, pickResult: Nullable<PickingInfo>) => void) => void;
    // private _initActionManager: (act: Nullable<AbstractActionManager>, clickInfo: _ClickInfo) => Nullable<AbstractActionManager>;
    private _delayedSimpleClick: (btn: number, clickInfo: _ClickInfo, cb: (clickInfo: _ClickInfo, pickResult: Nullable<PickingInfo>) => void) => void;
    private _delayedSimpleClickTimeout: number;
    private _previousDelayedSimpleClickTimeout: number;
    private _meshPickProceed = false;

    private _previousButtonPressed: number;
    private _currentPickResult: Nullable<PickingInfo> = null;
    private _previousPickResult: Nullable<PickingInfo> = null;
    private _totalPointersPressed = 0;
    private _doubleClickOccured = false;

    private _pointerOverMesh: Nullable<AbstractMesh>;

    private _pickedDownMesh: Nullable<AbstractMesh>;
    private _pickedUpMesh: Nullable<AbstractMesh>;

    private _pointerX: number = 0;
    private _pointerY: number = 0;
    private _unTranslatedPointerX: number;
    private _unTranslatedPointerY: number;
    private _translatedPointerX: number = 0;
    private _translatedPointerY: number = 0;
    private _startingPointerPosition = new Vector2(0, 0);
    private _previousStartingPointerPosition = new Vector2(0, 0);
    private _startingPointerTime = 0;
    private _previousStartingPointerTime = 0;
    private _pointerCaptures: { [pointerId: number]: boolean } = {};

    private _meshUnderPointerId: Nullable<AbstractMesh>[] = [];
    
    // Keyboard
    private _onKeyDown: (evt: KeyboardEvent) => void;
    private _onKeyUp: (evt: KeyboardEvent) => void;
    private _keyboardIsAttached = false;
    private _onCanvasFocusObserver: Nullable<Observer<Engine>>;
    private _onCanvasBlurObserver: Nullable<Observer<Engine>>;

    private _scene: ISceneD3;

    /**
     * Creates a new InputManager
     * @param scene defines the hosting scene
     */
    public constructor(scene: ISceneD3) {
        this._scene = scene;
    }

    private _updatePointerPosition(evt: EventData): void {

        this._pointerX = evt.x;
        this._pointerY = evt.y;

        this._unTranslatedPointerX = this._pointerX;
        this._unTranslatedPointerY = this._pointerY;
        
    }

    /**
     * Attach events to the canvas (To handle actionManagers triggers and raise onPointerMove, onPointerDown and onPointerUp
     * @param attachUp defines if you want to attach events to pointerup
     * @param attachDown defines if you want to attach events to pointerdown
     * @param attachMove defines if you want to attach events to pointermove
     * @param elementToAttachTo defines the target DOM element to attach to (will use the canvas by default)
     */
    public attachControl(
        attachUp = true, attachDown = true, attachMove = true,
        elementToAttachTo: Nullable<{
            addEventListener(type: string, listener: EventListener, options?: boolean | AddEventListenerOptions): void;
            removeEventListener(type: string, listener: EventListener, options?: boolean | AddEventListenerOptions): void;
        }> = null,
        elementAttachedToForKeyBoard: Nullable<{
            addEventListener(type: string, listener: KeyboardEventListener, options?: boolean | AddEventListenerOptions): void;
            removeEventListener(type: string, listener: KeyboardEventListener, options?: boolean | AddEventListenerOptions): void;
        }> = null
    ): void {

        if (this._alreadyAttached) {
            this.detachControl();
        }
        this._alreadyAttachedTo = elementToAttachTo;
        this._alreadyAttachedToForKeyBoard = elementAttachedToForKeyBoard;
        
        let scene = this._scene;
        let engine = Engine.getEngine();

        this._onKeyDown = (evt: KeyboardEvent) => {
            let type = KeyboardEventTypes.KEYDOWN;
            // if (scene.onPreKeyboardObservable.hasObservers()) {
            //     let pi = new KeyboardInfoPre(type, evt);
            //     scene.onPreKeyboardObservable.notifyObservers(pi, type);
            //     if (pi.skipOnPointerObservable) {
            //         return;
            //     }
            // }

            if (scene.onKeyboardObservable.hasObservers()) {
                let pi = new KeyboardInfo(type, evt);
                scene.onKeyboardObservable.notifyObservers(pi, type);
            }

            // if (scene.actionManager) {
            //     scene.actionManager.processTrigger(Constants.ACTION_OnKeyDownTrigger, ActionEvent.CreateNewFromScene(scene, evt));
            // }
        };

        this._onKeyUp = (evt: KeyboardEvent) => {
            let type = KeyboardEventTypes.KEYUP;
            // if (scene.onPreKeyboardObservable.hasObservers()) {
            //     let pi = new KeyboardInfoPre(type, evt);
            //     scene.onPreKeyboardObservable.notifyObservers(pi, type);
            //     if (pi.skipOnPointerObservable) {
            //         return;
            //     }
            // }

            if (scene.onKeyboardObservable.hasObservers()) {
                let pi = new KeyboardInfo(type, evt);
                scene.onKeyboardObservable.notifyObservers(pi, type);
            }

            // if (scene.actionManager) {
            //     scene.actionManager.processTrigger(Constants.ACTION_OnKeyUpTrigger, ActionEvent.CreateNewFromScene(scene, evt));
            // }
        };
        let attachedFunction = () => {
            if (!elementAttachedToForKeyBoard || this._keyboardIsAttached) {
                return;
            }
            elementAttachedToForKeyBoard.addEventListener("keydown", this._onKeyDown, false);
            elementAttachedToForKeyBoard.addEventListener("keyup", this._onKeyUp, false);
            this._keyboardIsAttached = true;
        };

        // Keyboard events
        this._onCanvasFocusObserver = engine.onCanvasFocusObservable.add(
            (() => {

                if (document.activeElement === elementAttachedToForKeyBoard) {
                    attachedFunction();
                }
                return attachedFunction;
            })()
        );

        this._onCanvasBlurObserver = engine.onCanvasBlurObservable.add(() => {
            if (!elementAttachedToForKeyBoard) {
                return;
            }
            elementAttachedToForKeyBoard.removeEventListener("keydown", this._onKeyDown);
            elementAttachedToForKeyBoard.removeEventListener("keyup", this._onKeyUp);
            this._keyboardIsAttached = false;
        });

        attachedFunction();

        this._onPointerMove = (evt: EventData) => {
            let scene = this._scene;
            // preserve compatibility with Safari when pointerId is not present
            if (evt.pointerId === undefined) {
                (evt as any).pointerId = 0;
            }
            // (<EventData>evt).stopPropagation = false;
            if (this._currentPickResult && (<EventData>evt).stopPropagation != true) {
                let ev = new PointerInfo(PointerEventTypes.POINTERMOVE, evt, this._currentPickResult);
                let parent = ev.pickInfo.pickedMesh;
                while (parent != undefined && (<EventData>ev.event).stopPropagation != true) {
                    parent.onPointerObservable.notifyObservers(ev);
                    parent = <any>parent.parent;
                }
                
                // scene.onPointerUp(evt, this._currentPickResult, PointerEventTypes.POINTERUP);
            }
        }
        
        this._onPointerDown = (evt: EventData) => {
            this._totalPointersPressed++;
            this._pickedDownMesh = null;
            this._meshPickProceed = false;

            let scene = this._scene;
            // preserve compatibility with Safari when pointerId is not present
            if (evt.pointerId === undefined) {
                (evt as any).pointerId = 0;
            }
            
            this._updatePointerPosition(evt);

            this._startingPointerPosition.x = this._pointerX;
            this._startingPointerPosition.y = this._pointerY;
            this._startingPointerTime = Date.now();

            this._pointerCaptures[evt.pointerId] = true;

            // (<EventData>evt).stopPropagation = false;
            if (scene.clientRect && (<EventData>evt).stopPropagation != true) {

                this._translatedPointerX = (this._pointerX - scene.clientRect.left) / scene.clientRect.width ;
                this._translatedPointerY = (this._pointerY - scene.clientRect.top ) / scene.clientRect.height;
                this._translatedPointerX = this._translatedPointerX * 2 - 1;
                this._translatedPointerY = (this._translatedPointerY * 2 - 1) * -1;
            
                var pickResult = <PickingInfo>scene.pickWithColliderInfo(this._translatedPointerX, this._translatedPointerY, undefined, true);
                this._currentPickResult = pickResult;

                if (this._currentPickResult) {
                    let ev = new PointerInfo(PointerEventTypes.POINTERDOWN, evt, this._currentPickResult);
                    let parent = ev.pickInfo.pickedMesh;
                    while (parent != undefined && (<EventData>ev.event).stopPropagation != true) {
                        parent.onPointerObservable.notifyObservers(ev);
                        parent = <any>parent.parent;
                    }
                }
                // scene.onPointerDown(evt, pickResult, PointerEventTypes.POINTERDOWN);

            }
        }

        this._onPointerUp = (evt: EventData) => {
            let scene = this._scene;
            if (this._totalPointersPressed === 0) {
                // We are attaching the pointer up to windows because of a bug in FF
                return; // So we need to test it the pointer down was pressed before.
            }

            this._totalPointersPressed--;
            this._pickedUpMesh = null;
            this._meshPickProceed = false;

            // preserve compatibility with Safari when pointerId is not present
            if (evt.pointerId === undefined) {
                (evt as any).pointerId = 0;
            }

            this._updatePointerPosition(evt);

            this._pointerCaptures[evt.pointerId] = false;

            // (<EventData>evt).stopPropagation = false;
            if (this._currentPickResult && (<EventData>evt).stopPropagation != true) {
                let ev = new PointerInfo(PointerEventTypes.POINTERUP, evt, this._currentPickResult);
                let parent = ev.pickInfo.pickedMesh;
                while (parent != undefined && ev.event.stopPropagation != true) {
                    parent.onPointerObservable?.notifyObservers(ev);
                    parent = <any>parent.parent;
                }
                
                // scene.onPointerUp(evt, this._currentPickResult, PointerEventTypes.POINTERUP);
            }

            this._currentPickResult = null;

            if (!this._pointerCaptures[evt.pointerId]) {
                return;
            }
        }
        
        var eventPrefix = "pointer";

        // // Check if pointer events are supported
        // if (DomManagement.IsWindowObjectExist() && !window.PointerEvent && DomManagement.IsNavigatorAvailable() && !navigator.pointerEnabled) {
        //     eventPrefix = "mouse";
        // }

        // // Special Fallback MacOS Safari...
        // if (
        //     // engine._badDesktopOS &&
        //     // !engine._badOS &&
        //     // And not ipad pros who claim to be macs...
        //     !(document && "ontouchend" in document)
        // ) {
        //     eventPrefix = "mouse";
        // }

        if (attachMove) {
            elementToAttachTo.addEventListener(eventPrefix + "move", <any>this._onPointerMove, false);

            // Wheel
            this._wheelEventName =
                "onwheel" in document.createElement("div")
                    ? "wheel" // Modern browsers support "wheel"
                    : (<any>document).onmousewheel !== undefined
                    ? "mousewheel" // Webkit and IE support at least "mousewheel"
                    : "DOMMouseScroll"; // let's assume that remaining browsers are older Firefox

            elementToAttachTo.addEventListener(this._wheelEventName, <any>this._onPointerMove, false);
        }

        if (attachDown) {
            elementToAttachTo.addEventListener(eventPrefix + "down", <any>this._onPointerDown, false);
        }

        if (attachUp) {
            // let hostWindow = this._scene.getEngine().getHostWindow();
            // if (hostWindow) {
                elementToAttachTo.addEventListener(eventPrefix + "up", <any>this._onPointerUp, false);
            // }
        }
        
        this._alreadyAttached = true;
    }

    public detachControl() {
        const eventPrefix = "pointer";
        let engine = Engine.getEngine();

        if (!this._alreadyAttachedTo) {
            return;
        }

        if (!this._alreadyAttached) {
            return;
        }

        // Pointer
        this._alreadyAttachedTo.removeEventListener(eventPrefix + "move", <any>this._onPointerMove);
        this._alreadyAttachedTo.removeEventListener(this._wheelEventName, <any>this._onPointerMove);
        this._alreadyAttachedTo.removeEventListener(eventPrefix + "down", <any>this._onPointerDown);
        window.removeEventListener(eventPrefix + "up", <any>this._onPointerUp);

        // Blur / Focus
        if (this._onCanvasBlurObserver) {
            engine.onCanvasBlurObservable.remove(this._onCanvasBlurObserver);
        }

        if (this._onCanvasFocusObserver) {
            engine.onCanvasFocusObservable.remove(this._onCanvasFocusObserver);
        }

        // Keyboard
        this._alreadyAttachedToForKeyBoard?.removeEventListener("keydown", this._onKeyDown);
        this._alreadyAttachedToForKeyBoard?.removeEventListener("keyup", this._onKeyUp);

        // // Cursor
        // if (!this._scene.doNotHandleCursors) {
        //     this._alreadyAttachedTo.style.cursor = this._scene.defaultCursor;
        // }

        this._alreadyAttached = false;
    }
}