import { Engine } from "../engines/engine";
import { Scene } from "../scene";
import { EffectMaterial } from "./material";
import { Texture } from "./Textures/texture";
import { UniformMainTexture, UniformMainTextureUVOffsetSpeed } from "./node_material/mainTextureBlock";
import { Color3 } from "../../maths/math.color";
import { VARYING_COLOR, VARYING_NORMAL_WS, VARYING_POSITION_WS, VARYING_UV, VARYING_UV2, VARYING_UV3, VARYING_UV4 } from "./node_material";
import { Vector4 } from "../../maths/math.vector";
import { Constants } from "../../interface/index";
import { p3d } from "../types";
import { ExtensionTools } from "../../base";

export class DistortionUVMaterial extends EffectMaterial {
    public static ShaderName = "DistortionUVMaterial";
    
    public static KEY_distortionInfo = "distortionInfo";
    public static KEY_control = "control";
    public static KEY_MaskTilloff = "uMaskTilloff";
    public static KEY_MaskTexture = "uMaskTex";
    public static KEY_MaskUVSpeed = "uMaskUVOS";

    constructor(name: string, scene: Scene, shaderName = DistortionUVMaterial.ShaderName, _pass: Constants.PassTag = Constants.PASS_TAG_06) {
        super(name, scene, shaderName, _pass);
    }

    private _distortionInfo: Vector4 = new Vector4(0, 0, 0, 0);
    private _control: Vector4 = new Vector4(0, 0, 0, 0);
    public _diffuseColor: Color3 = new Color3(1, 1, 1);

    private _diffuseColorDirty: boolean = true;
    public get emissiveColor() {
        return this._diffuseColor;
    }
    public set emissiveColor(value: Color3) {
        this._diffuseColor = value;
        this._diffuseColorDirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }


    private _infoDirty: boolean = true;
    public get distortionX() {
        return this._distortionInfo.x;
    }
    public set distortionX(value: number) {
        this._distortionInfo.x = value;
        this._infoDirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }
    public get distortionY() {
        return this._distortionInfo.y;
    }
    public set distortionY(value: number) {
        this._distortionInfo.y = value;
        this._infoDirty = true;
    }
    public get distortionSpeedX() {
        return this._distortionInfo.z;
    }
    public set distortionSpeedX(value: number) {
        this._distortionInfo.z = value;
        this._infoDirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }
    public get distortionSpeedY() {
        return this._distortionInfo.w;
    }
    public set distortionSpeedY(value: number) {
        this._distortionInfo.w = value;
        this._infoDirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }

    private _controlDirty: boolean = true;
    public get maskFlowMode() {
        return this._control.y;
    }
    public set maskFlowMode(value: number) {
        this._control.y = value;
        this._controlDirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }
    public get vertexColorFactor() {
        return this._control.z;
    }
    public set vertexColorFactor(value: number) {
        this._control.z = value;
        this._controlDirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }
    public get opacityFromMaskBlue() {
        return this._control.w > 0;
    }
    public set opacityFromMaskBlue(value: boolean) {
        this._control.w = value ? 1 : 0;
        this._controlDirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }

    private _maskTexture: Texture;
    private _maskTextureDirty: boolean = false;
    public set maskTexture(val: Texture) {
        if (this._maskTexture != val) {
            this._maskTexture = val;
            this._maskTextureDirty = true;
            if (val) val._material = this;
        }
    }
    public get maskTexture() {
        this._scene.cmdsForceDirty.set(this.entity, this);
        return this._maskTexture;
    }

    private _diffuseTexture: Texture;
    private _diffuseTextureDirty: boolean = false;
    public set diffuseTexture(val: Texture) {
        if (this._diffuseTexture != val) {
            this._diffuseTexture = val;
            this._diffuseTextureDirty = true;
            this._scene.cmdsForceDirty.set(this.entity, this);
            if (val) val._material = this;
        }
    }
    public get diffuseTexture() {
        this._scene.cmdsForceDirty.set(this.entity, this);
        return this._diffuseTexture;
    }

    _onFrameUpdate() {

        if (this._diffuseColorDirty || (this._diffuseTextureDirty || (this._diffuseTexture && this._diffuseTexture._levelDirty))) {
            UniformMainTexture.color(this.entity, this._diffuseColor.r, this._diffuseColor.g, this._diffuseColor.b, this._diffuseTexture?.level || 1.);
        }
        if (this._diffuseTexture) {
            if (this._diffuseTextureDirty || this._diffuseTexture._tilloffDirty) {
                UniformMainTexture.tilloff(this.entity, this._diffuseTexture.uScale, this._diffuseTexture.vScale, this._diffuseTexture.uOffset, this._diffuseTexture.vOffset);
            }
            if (this._diffuseTextureDirty || this._diffuseTexture._offsetSpeedDirty) {
                UniformMainTextureUVOffsetSpeed.value(this.entity, this._diffuseTexture.offsetSpeedU, this._diffuseTexture.offsetSpeedV);
            }
            if (this._diffuseTextureDirty) {
                UniformMainTexture.texture(this.entity, this._diffuseTexture);
            }
        }
        
        if (this._maskTexture) {
            if (this._maskTextureDirty || this._maskTexture._tilloffDirty) {
                p3d.p3d_material_uniform_vec4(
                    Engine.commands(), this.entity, ExtensionTools.getAtom(DistortionUVMaterial.KEY_MaskTilloff),
                    this._maskTexture.uScale, this._maskTexture.vScale, this._maskTexture.uOffset, this._maskTexture.vOffset
                );
            }
            if (this._maskTextureDirty || this._maskTexture._offsetSpeedDirty) {
                p3d.p3d_material_uniform_vec2(
                    Engine.commands(), this.entity, ExtensionTools.getAtom(DistortionUVMaterial.KEY_MaskUVSpeed),
                    this._maskTexture.offsetSpeedU, this._maskTexture.offsetSpeedV
                );
            }
            if (this._maskTextureDirty) {
                let val = this._maskTexture;
                // console.warn("Text Compressed: " + val.compressed);
                p3d.p3d_material_uniform_tex(
                    Engine.commands(), this.entity, ExtensionTools.getAtom(DistortionUVMaterial.KEY_MaskTexture),
                    ExtensionTools.getAtom(val.url), val.srgb, val.compressed, val.filter, val.address_mode_u, val.address_mode_v, val.address_mode_w,
                    val.mag_filter, val.min_filter, val.mipmap_filter, val.anisotropy_clamp, val.border_color, val.isfile, val.cancombine, val.compare
                );
            }
        }

        if (this._controlDirty) {
            p3d.p3d_material_uniform_vec4(
                Engine.commands(), this.entity, ExtensionTools.getAtom(DistortionUVMaterial.KEY_control),
                this._control.x, this._control.y, this._control.z, this._control.w
            );
        }

        if (this._infoDirty) {
            p3d.p3d_material_uniform_vec4(
                Engine.commands(), this.entity, ExtensionTools.getAtom(DistortionUVMaterial.KEY_distortionInfo),
                this._distortionInfo.x, this._distortionInfo.y, this._distortionInfo.z, this._distortionInfo.w
            );
        }

        this._maskTextureDirty = false;
        this._diffuseTextureDirty = false;
        this._diffuseColorDirty = false;
        this._controlDirty = false;
        this._infoDirty = false;
    }

    public static regist() {
        if (Engine.CheckShader(this.ShaderName)) {
            return;
        }
        let uniforms = p3d.MaterialUniformDefines.create();
        let includes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockFloat);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockFresnel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorSpace);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockViewDirection);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformMainTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTextureUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacity);

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec3", "v_pos"],
            ["vec3", "v_normal"],
            ["vec4", "v_color"],
            ["vec2", "v_uv"],
            ["vec2", "v_uv2"],
            ["vec2", "v_uv3"],
            ["vec2", "v_uv4"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });
        let varying = VARYING_POSITION_WS | VARYING_NORMAL_WS | VARYING_UV | VARYING_COLOR | VARYING_UV2 | VARYING_UV3 | VARYING_UV4;
        let binds_defines_base = Constants.BIND_MODEL | Constants.BIND_VIEWER | Constants.BIND_EFFECT_VALUE | Constants.BIND_EFFECT_TEXTURE;

        p3d.p3d_shader_uniform_vec4(uniforms, DistortionUVMaterial.KEY_distortionInfo, 0, 0, 0, 0, false);
        p3d.p3d_shader_uniform_vec4(uniforms, DistortionUVMaterial.KEY_control, 0, 0, 0, 0, false);
        p3d.p3d_shader_uniform_tex(uniforms, DistortionUVMaterial.KEY_MaskTexture, true, Constants.SHADER_STAGE_VERTEXFRAGMENT, Constants.DEFAULT_TEXTURE_BLACK, Constants.TextureViewDimensionD2);
        p3d.p3d_shader_uniform_vec4(uniforms, DistortionUVMaterial.KEY_MaskTilloff, 1, 1, 0, 0, false);
        p3d.p3d_shader_uniform_vec2(uniforms, DistortionUVMaterial.KEY_MaskUVSpeed, 0, 0, false);

        Engine.RegistShader(
            this.ShaderName,
            uniforms,
            `
            `,
            `
    layout(location = 0) out vec4 gl_FragColor;
            `,
            `
    mat4 finalWorld = PI_ObjectToWorld;
    vec4 position =  vec4(A_POSITION, 1.);
    vec4 worldPos =  finalWorld * position;
    gl_Position = PI_MATRIX_VP * worldPos;
    v_pos = worldPos.xyz;
    mat3 normalWorld = mat3(finalWorld);
    v_normal = vec3(finalWorld * vec4(A_NORMAL, 1.));
    v_color = A_COLOR4;
    float progress  = PI_Time.y; // ${DistortionUVMaterial.KEY_control}.x;
    float maskValueFlow = matParam.${DistortionUVMaterial.KEY_control}.y;
    float maskUVFlow = 1.0 - maskValueFlow;
    vec2 diffValue = vec2(progress, progress) * matParam.${DistortionUVMaterial.KEY_distortionInfo}.zw;
    v_uv = A_UV;
    v_uv2 = A_UV;
    v_uv2 = v_uv2 + vec2(maskUVFlow, maskUVFlow) * diffValue;
    v_uv2 = v_uv2 * matParam.${DistortionUVMaterial.KEY_MaskTilloff}.xy + matParam.${DistortionUVMaterial.KEY_MaskTilloff}.zw;
    v_uv3 = vec2(maskValueFlow, maskValueFlow) * diffValue;
    v_uv4 = A_UV;
            `,
            `
    float alpha         = opacity(matParam);
    alpha               *= v_color.a;
    vec4 baseColor      = mix(vec4(1.0), v_color, matParam.${DistortionUVMaterial.KEY_control}.z);
    baseColor.a         *= alpha;
    vec4 textureMask    = Get${DistortionUVMaterial.KEY_MaskTexture}(v_uv2, vec2(0., 0.), vec4(1., 1., 0., 0.));
    vec2 uvDiff         = textureMask.rg * 2. - vec2(1., 1.) + v_uv3;
    vec4 textureColor   = mainTexture(v_uv, applyUVOffsetSpeed(matParam.uMainUVOS) + uvDiff * ${DistortionUVMaterial.KEY_distortionInfo}.xy, matParam);
    baseColor.rgb       *= textureColor.rgb * mainStrength(matParam) * mainColor(matParam);
    baseColor.a         *= textureColor.a;
    if (matParam.${DistortionUVMaterial.KEY_control}.w > 0.0) {
        baseColor *= Get${DistortionUVMaterial.KEY_MaskTexture}(v_uv4, vec2(0., 0.), vec4(1., 1., 0., 0.)).b;
    }
    if (cutoff(baseColor.a, matParam)) {
        discard;
    }
    gl_FragColor = baseColor;    
            `,
            includes,
            ``,
            varyings,
            binds_defines_base
        );
    }
}