import { Engine } from "../engines/engine";
import { Scene } from "../scene";
import { UniformMainTexture } from "./node_material/mainTextureBlock";
import { UniformEmissiveTexture } from "./node_material/emissiveBlock";
import { VARYING_COLOR, VARYING_NORMAL_WS, VARYING_POSITION_WS, VARYING_UV, VARYING_UV2, VARYING_UV3 } from "./node_material";
import { UniformOpacityTexture } from "./node_material/opacityBlock";
import { MainOpacityMaterial } from "./main_opacity";
import { FresnelParameters } from "./fresnel";
import { UniformEmissiveFresnel } from "./node_material/fresnelBlock";
import { Constants } from "../../interface/index";
import { p3d } from "../types";
import { ExtensionTools } from "../../base";

export class MainOpacityEmissiveFresnelMaterial extends MainOpacityMaterial {
    public static ShaderName = "MainOpacityEmissiveFresnelMaterial"
    constructor(name: string, scene: Scene, shaderName = MainOpacityEmissiveFresnelMaterial.ShaderName, _pass: Constants.PassTag = Constants.PASS_TAG_06) {
        super(name, scene, shaderName, _pass);
    }

    private _emissiveFresnelParameters: FresnelParameters = new FresnelParameters();
    public set emissiveFresnelParameters(val: FresnelParameters) {
        this._emissiveFresnelParameters = val;
        (<any>val)._dirty = true;
        this._scene.cmdsForceDirty.set(this.entity, this);
    }
    public get emissiveFresnelParameters() {
        this._scene.cmdsForceDirty.set(this.entity, this);
        return this._emissiveFresnelParameters;
    }

    _onFrameUpdate() {
        super._onFrameUpdate();

        if (this._emissiveFresnelParameters.dirty) {
            this._emissiveFresnelParameters._update();

            let leftColor = (<any>this._emissiveFresnelParameters)._leftColor;
            let rightColor = (<any>this._emissiveFresnelParameters)._rightColor;

            UniformEmissiveFresnel.left(this.entity, leftColor.r, leftColor.g, leftColor.b);
            UniformEmissiveFresnel.right(this.entity, rightColor.r, rightColor.g, rightColor.b);
            UniformEmissiveFresnel.biasAndPower(this.entity, this._emissiveFresnelParameters.bias, this._emissiveFresnelParameters.power);
        }
    }

    public static regist() {
        if (Engine.CheckShader(this.ShaderName)) {
            return;
        }
        let uniforms = p3d.MaterialUniformDefines.create();
        let includes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockFloat);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockFresnel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorSpace);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockViewDirection);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformMainTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTextureUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacity);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformOpacityTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacityTextureUVOffsetSpeed);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformEmissiveTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveTextureUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockEmissiveFresnel);

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec3", "v_pos"],
            ["vec3", "v_normal"],
            ["vec4", "v_color"],
            ["vec2", "v_uv"],
            ["vec2", "v_uv2"],
            ["vec2", "v_uv3"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });
        let varying = VARYING_POSITION_WS | VARYING_NORMAL_WS | VARYING_UV | VARYING_COLOR | VARYING_UV2 | VARYING_UV3;
        let binds_defines_base = Constants.BIND_MODEL | Constants.BIND_VIEWER | Constants.BIND_EFFECT_VALUE | Constants.BIND_EFFECT_TEXTURE;

        Engine.RegistShader(
            this.ShaderName,
            uniforms,
            `
            `,
            `
            layout(location = 0) out vec4 gl_FragColor;
            `,
            `
            mat4 finalWorld = PI_ObjectToWorld;
        
            vec4 position =  vec4(A_POSITION, 1.);
            vec4 worldPos =  finalWorld * position;
            // vec4 worldPos =  position;
        
            gl_Position = PI_MATRIX_VP * worldPos;

            // gl_Position = position;
        
            v_pos = worldPos.xyz;
        
            mat3 normalWorld = mat3(finalWorld);
            v_normal = normalWorld * A_NORMAL;
            
            v_uv    = A_UV;
            v_uv2   = A_UV;
            v_uv3   = A_UV;
            v_color = A_COLOR4;
            `,
            `
            vec4 baseColor = v_color;
            float alpha = opacity(matParam);
            alpha   *= v_color.a;
        
            float glossiness 			= 0.;
            vec3 specularColor 		    = vec3(0., 0., 0.);
            vec3 diffuseColor 		    = vec3(1., 1., 1.);
            vec3 baseAmbientColor		= vec3(1., 1., 1.);
            vec4 refractionColor		= vec4(0., 0., 0., 1.);
            vec4 reflectionColor 		= vec4(0., 0., 0., 1.);
            
            vec3 normalW 				= normalize(v_normal);
            vec3 viewDirectionW		    = WorldSpaceViewDir(v_pos);
            float NdotVunclamped 		= dot(normalW, viewDirectionW);
            float NdotV					= absEps(NdotVunclamped);
            vec3 V 					    = viewDirectionW;
            vec3 N 					    = normalW;
            float absNdV                = abs(NdotVunclamped);
        
            // ----------------------------------------------------------
            vec3 diffuseBase 					= vec3(0., 0., 0.);
            vec3 specularBase 					= vec3(0., 0., 0.);
        
            vec4 mainTextureColor   = mainTexture(v_uv, applyUVOffsetSpeed(matParam.uMainUVOS), matParam);
            baseColor.rgb           *= mainTextureColor.rgb * mainStrength(matParam) * mainColor(matParam);
            alpha                   *= mainTextureColor.a;
        
            // diffuseColor            = mainColor(matParam);
        
            vec4 opacityData        = opacityTexture(v_uv2, applyUVOffsetSpeed(matParam.uOpacityUVOS), matParam);
            alpha                   *= opacityChannel(opacityData, matParam);
        
            if (cutoff(alpha, matParam)) {
                discard;
            }
        
            vec3 emissiveColor      = emissiveColor(matParam);
            emissiveColor           *= emissiveTexture(v_uv3, applyUVOffsetSpeed(matParam.uEmissiveUVOS), matParam.).rgb * emissiveStrength(matParam);
            emissiveColor           = emissiveFresnel(absNdV, emissiveColor, matParam);
        
            vec3 finalSpecular 		= specularBase * specularColor;
            vec3 finalDiffuse       = (diffuseBase * diffuseColor + emissiveColor) * baseColor.rgb;
            
            vec4 finalColor 		= vec4(
                finalDiffuse * baseAmbientColor
                + 
                finalSpecular
                +
                refractionColor.rgb
                + 
                reflectionColor.rgb
                ,
                alpha
            );
        
            finalColor				= max(finalColor, 0.0);
        
            // Premulty
            // finalColor.rgb          *= finalColor.a;
        
            gl_FragColor = finalColor;        
            `,
            includes,
            ``,
            varyings,
            binds_defines_base
        );
    }
}