import { Scene } from "../scene";
import { Texture } from "./Textures/texture";
import { Material } from "./material";
import { UniformMainTexture } from "./node_material/mainTextureBlock";
import { VARYING_COLOR, VARYING_NORMAL_WS, VARYING_POSITION_WS, VARYING_UV, VARYING_UV2, VARYING_UV3 } from "./node_material";
import { Engine } from "../engines/engine";
import { Constants } from "../../interface/index";
import { CustomRenderTarget } from "../../render/renderer";
import { p3d } from "../types";
import { ExtensionTools } from "../../base";

// export class MainTextureMaterial extends Material {
//     constructor(name: string, scene: Scene, pass: number = 0) {
//         super(name, scene, "UnlitShader", pass);
//     }
//     public set mainTexture(val: Texture) {
//         MainTextureBlock.texture(this.entity, val);
//     }
//     public baseColor(r: number, g: number, b: number, a: number) {
//         MainTextureBlock.color(this.entity, r, g, b, a);
//     }
// }

export class MainTextureMaterial extends Material {
    public static ShaderName = "MainTexture";
    public set mainTexture(val: Texture | CustomRenderTarget) {
        UniformMainTexture.texture(this.entity, val);
    }
    public baseColor(r: number, g: number, b: number, a: number) {
        UniformMainTexture.color(this.entity, r, g, b, a);
    }
    public static regist() {
        if (Engine.CheckShader(MainTextureMaterial.ShaderName)) { return };
        let uniforms = p3d.MaterialUniformDefines.create();
        let includes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockColorGray);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockTextureChannel);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockUVOffsetSpeed);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockCutoff);
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockOpacity);
        // p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTexture);
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformMainTexture.BLOCK_NAME));
        p3d.p3d_material_include(includes, p3d.BuiltinNodeMaterialBlock.BlockMainTextureUVOffsetSpeed);

        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec3", "v_pos"],
            ["vec3", "v_normal"],
            ["vec4", "v_color"],
            ["vec2", "v_uv"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });
        let varying = VARYING_POSITION_WS | VARYING_NORMAL_WS | VARYING_UV | VARYING_COLOR;
        let binds_defines_base = Constants.BIND_MODEL | Constants.BIND_VIEWER | Constants.BIND_EFFECT_VALUE | Constants.BIND_EFFECT_TEXTURE;

        Engine.RegistShader(
            MainTextureMaterial.ShaderName,
            uniforms,
            `
            `,
            `
            layout(location = 0) out vec4 gl_FragColor;
            `,
            `
            mat4 finalWorld = PI_ObjectToWorld;
        
            vec4 position =  vec4(A_POSITION, 1.);
            vec4 worldPos =  finalWorld * position;
            // vec4 worldPos =  position;
        
            gl_Position = PI_MATRIX_VP * worldPos;

            // gl_Position = position;
        
            v_pos = worldPos.xyz;
        
            mat3 normalWorld = mat3(finalWorld);
            v_normal = A_NORMAL;
            
            v_uv    = A_UV;
            v_color = A_COLOR4;
            `,
            `
            vec4 baseColor = v_color;
            float alpha = opacity();
            alpha   *= v_color.a;
        
            float glossiness 			= 0.;
            vec3 specularColor 		    = vec3(0., 0., 0.);
            vec3 diffuseColor 		    = vec3(0., 0., 0.);
            vec3 baseAmbientColor		= vec3(1., 1., 1.);
            vec4 refractionColor		= vec4(0., 0., 0., 1.);
            vec4 reflectionColor 		= vec4(0., 0., 0., 1.);
        
            // ----------------------------------------------------------
            vec3 diffuseBase 					= vec3(0., 0., 0.);
            vec3 specularBase 					= vec3(0., 0., 0.);
        
            vec4 mainTextureColor   = mainTexture(v_uv, applyUVOffsetSpeed(matParam.uMainUVOS), matParam);
            baseColor.rgb           *= mainTextureColor.rgb * mainStrength(matParam);
            alpha                   *= mainTextureColor.a;
        
            if (cutoff(alpha, matParam)) {
                discard;
            }
        
            vec3 emissiveColor      = mainColor(matParam);
        
            vec3 finalSpecular 		= specularBase * specularColor;
            vec3 finalDiffuse       = (diffuseBase * diffuseColor + emissiveColor) * baseColor.rgb;
            
            vec4 finalColor 		= vec4(
                finalDiffuse * baseAmbientColor
                + 
                finalSpecular
                +
                refractionColor.rgb
                + 
                reflectionColor.rgb
                ,
                alpha
            );
        
            finalColor				= max(finalColor, 0.0);
        
            // Premulty
            finalColor.rgb          *= finalColor.a;
        
            gl_FragColor = finalColor;        
            `,
            includes,
            ``,
            varyings,
            binds_defines_base
        );
    }
    constructor(name: string, scene: Scene, shadername: string = MainTextureMaterial.ShaderName, _pass: Constants.PassTag = Constants.PASS_TAG_06) {
        super(name, scene, _pass);
        Material.modify(this, shadername);
    }
}