import { Entity, Engine } from "../engines/engine";
import { Scene } from "../scene";
import { UniformOpacity } from "./node_material/opacityBlock";
import { UniformCutoff } from "./node_material/cutoffBlock";
import { p3d, IMaterialD3 } from "../types";
import { ExtensionTools } from "../../base";
import { Constants } from "../../interface/index";
import { UniformMainTexture } from "./node_material/mainTextureBlock";

export class Material extends Entity implements IMaterialD3 {
    public static readonly ALPHA_DISABLE = 0;
    public static readonly ALPHA_ADD = 1;
    public static readonly ALPHA_COMBINE = 2;
    public static readonly ALPHA_SUBTRACT = 3;
    public static readonly ALPHA_MULTIPLY = 4;
    public static readonly ALPHA_MAXIMIZED = 5;
    public static readonly ALPHA_ONEONE = 6;
    public static readonly ALPHA_PREMULTIPLIED = 7;
    public static readonly ALPHA_PREMULTIPLIED_PORTERDUFF = 8;
    public static readonly ALPHA_INTERPOLATE = 9;
    public static readonly ALPHA_SCREENMODE = 10;
    public static readonly PI_ALPHA_PREMULTIPLIED = 11;
    public static readonly ALPHA_ONEONE_ONEONE = 11;
    public static readonly ALPHA_ALPHATOCOLOR = 12;
    public static readonly ALPHA_REVERSEONEMINUS = 13;
    public static readonly ALPHA_SRC_DSTONEMINUSSRCALPHA = 14;
    public static readonly ALPHA_ONEONE_ONEZERO = 15;
    public static readonly ALPHA_EXCLUSION = 16;
    public static readonly PI_ALPHA_SRCCOLOR_ONE = 20;
    public static readonly PI_ALPHA_ONE_ONEMINUSSRCALPHA = 21;
    public static readonly PI_ALPHA_SRCCOLOR_ONEMINUSSRCALPHA = 22;
    public static readonly PI_ALPHA_SRCCOLOR_ONEMINUSSRCCOLOR = 23;

    public static readonly OPAQUE	= 0;
    public static readonly ALPHATEST = 1;	
    public static readonly ALPHABLEND = 2;
    public static readonly ALPHATESTANDBLEND	= 3;

    public static readonly MATERIAL_OPAQUE	= 0;
    public static readonly MATERIAL_ALPHATEST = 1;	
    public static readonly MATERIAL_ALPHABLEND = 2;
    public static readonly MATERIAL_ALPHATESTANDBLEND	= 3;

    /**
     * Returns the triangle fill mode
     */
    public static readonly TriangleFillMode = 0;
    /**
     * Returns the wireframe mode
     */
    public static readonly WireFrameFillMode = 1;
    /**
     * Returns the point fill mode
     */
    public static readonly PointFillMode = 2;
    /**
     * Returns the point list draw mode
     */
    public static readonly PointListDrawMode = 3;
    /**
     * Returns the line list draw mode
     */
    public static readonly LineListDrawMode = 4;
    /**
     * Returns the line loop draw mode
     */
    public static readonly LineLoopDrawMode = 5;
    /**
     * Returns the line strip draw mode
     */
    public static readonly LineStripDrawMode = 6;

    /**
     * Returns the triangle strip draw mode
     */
    public static readonly TriangleStripDrawMode = 7;
    /**
     * Returns the triangle fan draw mode
     */
    public static readonly TriangleFanDrawMode = 8;

    /**
     * Stores the clock-wise side orientation
     */
    public static readonly ClockWiseSideOrientation = 0;
    /**
     * Stores the counter clock-wise side orientation
     */
    public static readonly CounterClockWiseSideOrientation = 1;

    public static KEY_VERTEX_ABOUT = "uVertexApplyMode";
    public static MODE_VERTEX_COLOR = 1;
    public static MODE_VERTEX_ALPHA = 2;
    public static ApplyVertexAbout(uniforms) {
        p3d.p3d_shader_uniform_uint(uniforms, Material.KEY_VERTEX_ABOUT, 0, false);
    }

    metadata: any = {};
    protected _scene: Scene;
    protected _name: string;
    protected _shader: string;
    protected _pass: Constants.PassTag;
    public get pass() {
        return this._pass;
    }
    protected _meshes: Set<number> = new Set();
    public get meshes() {
        return this._meshes.keys();
    }
    constructor(name: string, scene: Scene, _pass: Constants.PassTag) {
        // let atom = ExtensionTools.getAtom(name);
        let entity = p3d.p3d_entity(Engine.native());
        super(entity);

        this._pass = _pass;
        this._name = name;
        this._scene = scene;
        this._scene.materials.set(this.entity, this);
    }
    public get shader() {
        return this._shader;
    }
    public static modify(mat: Material, shader: string) {
        if (mat._isDisposed) {
            return;
        }

        // if (mat._shader != shader || mat._pass != pass) {
            // if (window.mmm == undefined) {
            //     window.mmm = 0;
            // }
            // window.mmm += 1;
            mat._shader = shader;
            let key = ExtensionTools.getAtom(shader);
            p3d.p3d_material_shader(Engine.commands(), mat.entity, key);
            // console.error("Material: ", mat.entity);
            // key.free();
        // }
    }
    public dispose(forceDisposeEffect?: boolean, forceDisposeTextures?: boolean): void {
        this._scene.materials.delete(this.entity);
        super.dispose();
    }

    public mainColor(r: number, g: number, b: number) {
    }

    protected _alpha: number = 1;
    public set alpha(val: number) {
        if (this._alpha != val) {
            this._alpha = val;
            p3d.p3d_material_uniform_float(Engine.commands(), this.entity, ExtensionTools.getAtom(UniformOpacity.KEY_ALPHA), this._alpha);
        }
    }
    public get alpha() {
        return this._alpha;
    }

    protected _alphaCutoff: number = 0;
    public set alphaCutOff(val: number) {
        if (this._alphaCutoff != val) {
            this._alphaCutoff = val;
            p3d.p3d_material_uniform_float(Engine.commands(), this.entity, ExtensionTools.getAtom(UniformCutoff.KEY_VALUE), this._alphaCutoff);
        }
    }
    public get alphaCutOff() {
        return this._alphaCutoff;
    }

    protected _useVertexMode: number = 0;
    protected _useVertexAlpha: boolean;
    set useVertexAlpha(val: boolean) {
        this._useVertexAlpha = val;
        let old = (this._useVertexMode & Material.MODE_VERTEX_ALPHA) == Material.MODE_VERTEX_ALPHA;
        if (val != old) {
            if (old) {
                this._useVertexMode = this._useVertexMode - Material.MODE_VERTEX_ALPHA;
            } else {
                this._useVertexMode = this._useVertexMode + Material.MODE_VERTEX_ALPHA;
            }
            p3d.p3d_material_uniform_uint(Engine.commands(), this.entity, ExtensionTools.getAtom(Material.KEY_VERTEX_ABOUT), this._useVertexMode);
        }
    }
    get useVertexAlpha() {
        return this._useVertexAlpha;
    }

    set useVertexColor(val: boolean) {
        let old = (this._useVertexMode & Material.MODE_VERTEX_COLOR) == Material.MODE_VERTEX_COLOR;
        if (val != old) {
            if (old) {
                this._useVertexMode = this._useVertexMode - Material.MODE_VERTEX_COLOR;
            } else {
                this._useVertexMode = this._useVertexMode + Material.MODE_VERTEX_COLOR;
            }
            p3d.p3d_material_uniform_uint(Engine.commands(), this.entity, ExtensionTools.getAtom(Material.KEY_VERTEX_ABOUT), this._useVertexMode);
        }
    }

    protected _resultAlphaPremultiply: boolean = false;
    get resultAlphaPremultiply() {
        return this._resultAlphaPremultiply;
    }
    set resultAlphaPremultiply(val: boolean) {
        this._resultAlphaPremultiply = val;
    }

    forceDirty() {
        this._scene?.cmdsForceDirty.set(this._entity, this);
    }

    _onFrameUpdate() {

    }

    // /**
    //  * @deprecated
    //  */
    // transparencyMode: number;
    // /**
    //  * @deprecated
    //  */
    // alphaMode: number;
    // /**
    //  * @deprecated
    //  */
    // disableDepthWrite: boolean;
}

export class EffectMaterial extends Material {
    constructor(name: string, scene: Scene, shaderName: string, _pass: Constants.PassTag) {
        super(name, scene, _pass);
        Material.modify(this, shaderName);
    }
}

export class DefaultMaterial extends Material {
    constructor(name: string, scene: Scene, _pass: Constants.PassTag = Constants.PASS_TAG_03) {
        // debugger;
        super(name, scene, _pass);
        Material.modify(this, "Default");
    }
}

export class ShaderToy {
    public static regist(shaderName: string, fsCode: string, textures: string[]) {
    
        if (Engine.CheckShader(shaderName)) { return };
        let uniforms = p3d.MaterialUniformDefines.create();
        let includes = p3d.NodematerialIncludes.create();
        p3d.p3d_material_include_by_name(includes, ExtensionTools.getAtom(UniformMainTexture.BLOCK_NAME));
        
        textures.forEach((texName) => {
            p3d.p3d_shader_uniform_tex(uniforms, texName, true, Constants.SHADER_STAGE_VERTEXFRAGMENT, Constants.DEFAULT_TEXTURE_BLACK, Constants.TextureViewDimensionD2);
        });
    
        let varyings = p3d.P3DShaderVaryings.create();
        [
            ["vec2", "v_uv"],
        ].forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });
        let binds_defines_base = Constants.BIND_MODEL | Constants.BIND_VIEWER | Constants.BIND_EFFECT_VALUE;
        if (textures.length > 0) {
            binds_defines_base |= Constants.BIND_EFFECT_TEXTURE;
        }
    
        Engine.RegistShader(
            shaderName,
            uniforms,
            `
            `,
            `
            layout(location = 0) out vec4 gl_FragColor;
            `,
            `
            mat4 finalWorld = PI_ObjectToWorld;
        
            vec4 position =  vec4(A_POSITION, 1.);
            vec4 worldPos =  finalWorld * position;
            gl_Position = PI_MATRIX_VP * worldPos;

            v_uv    = A_UV;
            `,
            fsCode,
            includes,
            ``,
            varyings,
            binds_defines_base
        );
    }
}
