import { Material } from "./material";
import { Engine } from "../engines/engine";
import { Scene } from "../scene";
import { Constants } from "../../interface/index";
import { p3d } from "../types";
import { ExtensionTools } from "../../base";

export type TVaryingFormat = "vec4" | "vec3" | "vec2" | "float" | "uint" | "int";

export class ShaderMaterial extends Material {
    public static Counter: number = 0;
    
    public uniforms = p3d.MaterialUniformDefines.create();
    public includes = p3d.NodematerialIncludes.create();
    public varyings: [string, string][] = [];

    constructor(name: string, scene: Scene, _pass: Constants.PassTag) {
        super(name, scene, _pass);
    }
    public addVarying(key: string, format: TVaryingFormat) {
        this.varyings.push([format, key]);
    }

    private reset() {
        if (this.uniforms) {
            this.uniforms.free && this.uniforms.free();
            this.uniforms = undefined;
        }

        this.varyings.length = 0;
        this.uniforms = p3d.MaterialUniformDefines.create();
        p3d.p3d_material_includes_reset(this.includes);
    }
    public setTexture(key: string) {
        p3d.p3d_shader_uniform_tex(this.uniforms, key, true, Constants.SHADER_STAGE_FRAGMENT, Constants.DEFAULT_TEXTURE_WHITE, Constants.TextureViewDimensionD2);
    }
    public setFloat(key: string, value: number, caninstance: boolean = false) {
        p3d.p3d_shader_uniform_float(this.uniforms, key, value, caninstance);
    }
    public setFloatx2(key: string, v1: number, v2: number, caninstance: boolean = false) {
        p3d.p3d_shader_uniform_vec2(this.uniforms, key, v1, v2, caninstance);
    }
    public setFloatx4(key: string, v1: number, v2: number, v3: number, v4: number, caninstance: boolean = false) {
        p3d.p3d_shader_uniform_vec4(this.uniforms, key, v1, v2, v3, v4, caninstance);
    }
    public modify(
        fs_defined_code: string,
        fs_main_code: string,
    ) {
        let shader = `Shader${ShaderMaterial.Counter++}`;
        if (this._isDisposed) {
            return;
        }

        let varyings = p3d.P3DShaderVaryings.create();
        this.varyings.forEach((v) => {
            p3d.p3d_varying(varyings, ExtensionTools.getAtom(v[1]), ExtensionTools.getAtom(v[0]));
        });
        let binds_defines_base = Constants.BIND_MODEL | Constants.BIND_VIEWER | Constants.BIND_EFFECT_VALUE | Constants.BIND_EFFECT_TEXTURE;

        Engine.RegistShader(
            shader,
            this.uniforms,
            `
            `,
            `
            ${fs_defined_code}
            layout(location = 0) out vec4 gl_FragColor;
            `,
            `
            mat4 finalWorld = PI_ObjectToWorld;
        
            vec4 position =  vec4(A_POSITION, 1.);
            vec4 worldPos =  finalWorld * position;
        
            gl_Position = vec4(position.x * 2., position.y * 2., 0.1, 1.);
            // gl_Position.z = gl_Position.z * 0.5 + 0.5;
        
            v_pos = gl_Position.xyz;
        
            mat3 normalWorld = mat3(finalWorld);
            v_normal = A_NORMAL;
            
            v_uv = A_UV;
            v_color = A_COLOR4;
            `,
            fs_main_code,
            this.includes,
            ``,
            varyings,
            binds_defines_base
        );
        this.uniforms = undefined;
        this.includes = undefined;

        Material.modify(this, shader);

        this.reset();
    }
    public dispose(): void {
        super.dispose();
        this.uniforms?.free && this.uniforms.free();
        this.uniforms = undefined;
        this.includes?.free && this.includes.free();
        this.includes = undefined;
    }
}
