
import { Entity, Engine } from "../engines/engine";
import { Node, Transform } from "../node";
import { Material } from "../materials/material";
import { Quaternion, Vector3 } from "../../maths/index";
import { _MeshCollisionData } from "../collisions/meshCollisionData";
import { Constants, IAbstractMesh, ICollisionResponse, EInstanceKind, Nullable, IBoundingBox } from "../../interface/index";
import { BoundingInfo } from "../culling/boundinginfo";
import { p3d } from "../types";
import { Scene } from "../scene";

export class AbstractMesh extends Transform implements ICollisionResponse, IAbstractMesh {
    public static readonly CULLING_STRATEGY_ALWAYS = 0;
    public static readonly CULLING_STRATEGY_OPTIMIC = 1;
    public static readonly CULLING_STRATEGY_STANDARD = 2;

    constructor(name: string, scene: Scene, entity: number) {
        super(name, scene, entity);
        scene.meshes.set(this.entity, this);
        this.alwaysSelectAsActiveMesh = true;
    }

    dirtyLayerMask(): void {
        p3d.p3d_layermask(Engine.commands(), this.entity, this._layerMask);
    }
    protected _layerMask: number = 0b11111111_11111111_11111111_11111111;
    public get layerMask() {
        return this._layerMask;
    }
    public set layerMask(val: number) {
        if (this._layerMask != val) {
            this._layerMask = val;
            // this._scene.cmdsLayerMask.set(this._entity, this);
            p3d.p3d_layermask(Engine.commands(), this._entity, this._layerMask);
        }
    }

    protected _renderingGroupId: number = 0;
    public get renderingGroupId() {
        return this._renderingGroupId;
    }
    public set renderingGroupId(val: number) {
        if (this._renderingGroupId != val) {
            this._renderingGroupId = Math.floor(val);
            this._scene.cmdsMeshRenderQuene.set(this._entity, this);
        }
    }

    protected _alphaIndex: number = 0;
    public get alphaIndex() {
        return this._alphaIndex;
    }
    public set alphaIndex(val: number) {
        if (this._alphaIndex != val) {
            this._alphaIndex = Math.floor(val);
            this._scene.cmdsMeshRenderQuene.set(this._entity, this);
        }
    }

    
    protected _boundingInfo: IBoundingBox;
    get boundingBox() {
        return this._boundingInfo;
    }

    get material(): Material {
        return undefined;
    }
    protected _isVisible: boolean;
    get isVisible() {
        return this._isVisible;
    }
    set isVisible(val: boolean) {
        this._isVisible = val;
    }

    metadata: any = {};
    //
    protected _scalingMode: Constants.ScalingMode;
    public get scalingMode() {
        return this._scalingMode;
    }
    public set scalingMode(val: Constants.ScalingMode) {
        this._scalingMode = val;
        p3d.p3d_abstruct_mesh_scaling_mode(Engine.commands(), this.entity, val);
    }

    public set alwaysSelectAsActiveMesh(val: boolean) {
        if (val) {
            p3d.p3d_mesh_bounding_cullingmode(Engine.commands(), this.entity, AbstractMesh.CULLING_STRATEGY_ALWAYS);
        } else {
            p3d.p3d_mesh_bounding_cullingmode(Engine.commands(), this.entity, this._cullingStrategy);
        }
    }

    protected _cullingStrategy: Constants.CULLING_STRATEGY = AbstractMesh.CULLING_STRATEGY_OPTIMIC;
    public set cullingStrategy(val: Constants.CULLING_STRATEGY) {
        if (this.cullingStrategy != val) {
            this._cullingStrategy = val;
            p3d.p3d_mesh_bounding_cullingmode(Engine.commands(), this.entity, this._cullingStrategy);
        }
    }

    // protected _velocity: [number,];
    // public get renderAlignment() {
    //     return this._renderalignment;
    // }
    // public set renderAlignment(val: p3d.ERenderAlignment) {
    //     this._renderalignment = val;
    //     p3d.p3d_mesh_render_alignment(Engine.native(), Engine.param(), this.entity, val);
    // }

    // Collisions
    protected _meshCollisionData = new _MeshCollisionData();
    /**
     * Gets or sets a collision mask used to mask collisions (default is -1).
     * A collision between A and B will happen if A.collisionGroup & b.collisionMask !== 0
     */
    public get collisionMask(): number {
        return this._meshCollisionData._collisionMask;
    }

    public set collisionMask(mask: number) {
        this._meshCollisionData._collisionMask = !isNaN(mask) ? mask : -1;
    }

    /**
     * Gets or sets a collision response flag (default is true).
     * when collisionResponse is false, events are still triggered but colliding entity has no response
     * This helps creating trigger volume when user wants collision feedback events but not position/velocity
     * to respond to the collision.
     */
    public get collisionResponse(): boolean {
        return this._meshCollisionData._collisionResponse;
    }

    public set collisionResponse(response: boolean) {
        this._meshCollisionData._collisionResponse = response;
    }
    /**
     * Gets or sets the current collision group mask (-1 by default).
     * A collision between A and B will happen if A.collisionGroup & b.collisionMask !== 0
     */
    public get collisionGroup(): number {
        return this._meshCollisionData._collisionGroup;
    }

    public set collisionGroup(mask: number) {
        this._meshCollisionData._collisionGroup = !isNaN(mask) ? mask : -1;
    }
    
    /**
     * Gets or sets current surrounding meshes (null by default).
     *
     * By default collision detection is tested against every mesh in the scene.
     * It is possible to set surroundingMeshes to a defined list of meshes and then only these specified
     * meshes will be tested for the collision.
     *
     * Note: if set to an empty array no collision will happen when this mesh is moved.
     */
    public get surroundingMeshes() {
        return this._meshCollisionData._surroundingMeshes;
    }

    public set surroundingMeshes(meshes: Nullable<ICollisionResponse[]>) {
        this._meshCollisionData._surroundingMeshes = meshes;
    }

    public dispose(disposeEntity: boolean = true): void {
        this._scene?.meshes.delete(this.entity);
        super.dispose(disposeEntity);    
    }
}

export type InstancedDesc = EInstanceKind[];

// export class InstancedBuffer implements IInstancedBuffer {
    // protected _r: number                = 1.;
    // protected _g: number                = 1.;
    // protected _b: number                = 1.;
    // protected _alpha: number            = 1.0;

    // protected _uScale: number           = 1.0;
    // protected _vScale: number           = 1.0;
    // protected _uOffset: number          = 0.0;
    // protected _vOffset: number          = 0.0;

    // protected _insV4Ax: number           = 1.0;
    // protected _insV4Ay: number           = 1.0;
    // protected _insV4Az: number           = 1.0;
    // protected _insV4Aw: number           = 1.0;

    // protected _insV4Bx: number           = 1.0;
    // protected _insV4By: number           = 1.0;
    // protected _insV4Bz: number           = 1.0;
    // protected _insV4Bw: number           = 1.0;

    // protected _insV4Cx: number           = 1.0;
    // protected _insV4Cy: number           = 1.0;
    // protected _insV4Cz: number           = 1.0;
    // protected _insV4Cw: number           = 1.0;

    // protected _insV4Dx: number           = 1.0;
    // protected _insV4Dy: number           = 1.0;
    // protected _insV4Dz: number           = 1.0;
    // protected _insV4Dw: number           = 1.0;

    // // protected _worldMatrix: Matrix      = Matrix.Identity();
    // public rgb(r: number, g: number, b: number) {
    //     this._r = r;
    //     this._g = g;
    //     this._b = b;
    //     this._instance.scene?.cmdsInsRGB.set(this.entity, this);
    // }
    // public set color(val: Color4) {
    //     this._r = val.r;
    //     this._g = val.g;
    //     this._b = val.b;
    //     this._alpha = val.a;
    //     this._instance.scene?.cmdsInsRGB.set(this.entity, this);
    //     this._instance.scene?.cmdsInsAlpha.set(this.entity, this);
    // }
    // public set alpha(val: number) {
    //     this._alpha = val;
    //     this._instance.scene?.cmdsInsAlpha.set(this.entity, this);
    // }
    // public tilloff(uScale: number, vScale: number, uOffset: number, vOffset: number) {
    //     this._uScale = uScale;
    //     this._vScale = vScale;
    //     this._uOffset = uOffset;
    //     this._vOffset = vOffset;
    //     this._instance.scene?.cmdsInsTilloff.set(this.entity, this);
    // }
    // public insVec4A(x: number, y: number, z: number, w: number) {
    //     this._insV4Ax = x;
    //     this._insV4Ay = y;
    //     this._insV4Az = z;
    //     this._insV4Aw = w;
    // }
    // public insVec4B(x: number, y: number, z: number, w: number) {
    //     this._insV4Bx = x;
    //     this._insV4By = y;
    //     this._insV4Bz = z;
    //     this._insV4Bw = w;
    // }
    // public insVec4C(x: number, y: number, z: number, w: number) {
    //     this._insV4Bx = x;
    //     this._insV4By = y;
    //     this._insV4Bz = z;
    //     this._insV4Bw = w;
    // }
    // public insVec4D(x: number, y: number, z: number, w: number) {
    //     this._insV4Bx = x;
    //     this._insV4By = y;
    //     this._insV4Bz = z;
    //     this._insV4Bw = w;
    // }
    // constructor(protected _instance: Node) {

    // }
    // get uniqueID() {
    //     return this._instance.uniqueID;
    // }
    // onDisposeObserver: Observable<number> = new Observable();
    // dispose(): void {
    //     this.onDisposeObserver.notifyObservers(this.entity);
    //     this._instance = undefined;
    // }
    
    // get entity() {
    //     return this._instance.entity;
    // }
    // dirtyInsColor(data: Float64Array, offset: number): boolean {
    //     if (this._instance.isDisposed()) {
    //         return false;
    //     } else {
    //         data[offset + 0] = this._instance.entity;
    //         data[offset + 1] = this._r;
    //         data[offset + 2] = this._g;
    //         data[offset + 3] = this._b;
    //         return true;
    //     }
    // }
    // dirtyInsAlpha(data: Float64Array, offset: number): boolean {
    //     if (this._instance.isDisposed()) {
    //         return false;
    //     } else {
    //         data[offset + 0] = this._instance.entity;
    //         data[offset + 1] = this._alpha;
    //         return true;
    //     }
    // }
    // dirtyInsTilloff(data: Float64Array, offset: number): boolean {
    //     if (this._instance.isDisposed()) {
    //         return false;
    //     } else {
    //         data[offset + 0] = this._instance.entity;
    //         data[offset + 1] = this._uScale;
    //         data[offset + 2] = this._vScale;
    //         data[offset + 3] = this._uOffset;
    //         data[offset + 4] = this._vOffset;
    //         return true;
    //     }
    // }
    // dirtyInsVec4A(data: Float64Array, offset: number): boolean {
    //     throw new Error("Method not implemented.");
    // }
    // dirtyInsVec4B(data: Float64Array, offset: number): boolean {
    //     throw new Error("Method not implemented.");
    // }
    // dirtyInsVec4C(data: Float64Array, offset: number): boolean {
    //     throw new Error("Method not implemented.");
    // }
    // dirtyInsVec4D(data: Float64Array, offset: number): boolean {
    //     throw new Error("Method not implemented.");
    // }
// }