import { Node } from "../node";
import { Material } from "../materials/material";
import { Scene } from "../scene";
import { Engine, Entity } from "../engines/engine";
import { AbstractMesh } from "./abstractMesh";
import { InstancedMesh } from "./instancedMesh";
import { Constants, IInstancedMesh, IInstanceSource, InstanceAttribute } from "../../interface/index";
import { BoundingInfo } from "../culling/boundinginfo";
import { EntityID, IBlend } from "../entity";
import { IMeshD3, p3d } from "../types";
import { ExtensionTools } from "../../base";
import { VertexData } from "./vertexData";
import { IGeometryMeta } from "../../context";

export class Mesh extends AbstractMesh implements IInstanceSource, IMeshD3 {

    /**
     * Mesh side orientation : usually the external or front surface
     */
    public static readonly FRONTSIDE = VertexData.FRONTSIDE;

    /**
     * Mesh side orientation : usually the internal or back surface
     */
    public static readonly BACKSIDE = VertexData.BACKSIDE;
    /**
     * Mesh side orientation : both internal and external or front and back surfaces
     */
    public static readonly DOUBLESIDE = VertexData.DOUBLESIDE;
    /**
     * Mesh side orientation : by default, `FRONTSIDE`
     */
    public static readonly DEFAULTSIDE = VertexData.DEFAULTSIDE;

    /**
     * Mesh cap setting : no cap
     */
    public static readonly NO_CAP = 0;
    /**
     * Mesh cap setting : one cap at the beginning of the mesh
     */
    public static readonly CAP_START = 1;
    /**
     * Mesh cap setting : one cap at the end of the mesh
     */
    public static readonly CAP_END = 2;
    /**
     * Mesh cap setting : two caps, one at the beginning  and one at the end of the mesh
     */
    public static readonly CAP_ALL = 3;

    public readonly _instancestates: Map<string, InstanceAttribute> = new Map();
    public _cacheInstances: number[] = [];
    /**
     * 
     * @param name 
     * @param scene 
     * @param instancestate 
     * @param instances_use_common_buffer 
     * @param entity 外部可能拓展 Mesh 的实现预留 Entity 的外部创建, 通常情况不做设置，内部自动创建
     */
    constructor(name: string, scene: Scene, public readonly instancestates: InstanceAttribute[] = [], public readonly instances_use_common_buffer: boolean = true, instancematrix: boolean = false, entity?: number) {
        let instances = p3d.VInstanceAttributes.create(instancematrix);
        instancestates.forEach((item) => {
            p3d.p3d_instance_attribute(instances, ExtensionTools.getAtom(item.key), ExtensionTools.getAtom(item.code), ExtensionTools.getAtom(item.foruniformKey || ""), item.vtype);
        });

        if (entity == undefined) {
            entity = p3d.p3d_mesh(Engine.native(), Engine.commands(), scene.entity, instances, !instances_use_common_buffer);
        }
        instances.free && instances.free();

        super(name, scene, entity);
        instancestates.forEach((item) => {
            this._instancestates.set(item.key, item);
        });
        // TODO
        this.subMeshes = [this];
        this.frontFace = Constants.CCW;
    }

    _dirtyMesh: number = 0;
    applyBlend() {
        p3d.p3d_mesh_blend( Engine.commands(), this.entity, this.blend_enable, this.src_color, this.dst_color, this.src_alpha, this.dst_alpha, this.opt_color, this.opt_alpha, Constants.PASS_TAG_01 );
        p3d.p3d_mesh_blend( Engine.commands(), this.entity, this.blend_enable, this.src_color, this.dst_color, this.src_alpha, this.dst_alpha, this.opt_color, this.opt_alpha, Constants.PASS_TAG_02 );
        p3d.p3d_mesh_blend( Engine.commands(), this.entity, this.blend_enable, this.src_color, this.dst_color, this.src_alpha, this.dst_alpha, this.opt_color, this.opt_alpha, Constants.PASS_TAG_03 );
        p3d.p3d_mesh_blend( Engine.commands(), this.entity, this.blend_enable, this.src_color, this.dst_color, this.src_alpha, this.dst_alpha, this.opt_color, this.opt_alpha, Constants.PASS_TAG_04 );
        p3d.p3d_mesh_blend( Engine.commands(), this.entity, this.blend_enable, this.src_color, this.dst_color, this.src_alpha, this.dst_alpha, this.opt_color, this.opt_alpha, Constants.PASS_TAG_05 );
        p3d.p3d_mesh_blend( Engine.commands(), this.entity, this.blend_enable, this.src_color, this.dst_color, this.src_alpha, this.dst_alpha, this.opt_color, this.opt_alpha, Constants.PASS_TAG_06 );
        p3d.p3d_mesh_blend( Engine.commands(), this.entity, this.blend_enable, this.src_color, this.dst_color, this.src_alpha, this.dst_alpha, this.opt_color, this.opt_alpha, Constants.PASS_TAG_07 );
        p3d.p3d_mesh_blend( Engine.commands(), this.entity, this.blend_enable, this.src_color, this.dst_color, this.src_alpha, this.dst_alpha, this.opt_color, this.opt_alpha, Constants.PASS_TAG_08 );
    }

    applyAlphaMode(mode: number, pass: Constants.PassTag) {
        let result: IBlend = <any>{};
        ExtensionTools.alphaMode(mode, result);
        p3d.p3d_mesh_blend( Engine.commands(), this.entity, result.blend_enable, result.src_color, result.dst_color, result.src_alpha, result.dst_alpha, result.opt_color, result.opt_alpha, pass );
    }

    forceDirty() {
        this._scene?.cmdsForceDirty.set(this._entity, this);
    }

    private __geometryID: number;
    public get geometry() {
        return this.__geometryID;
    }

    set boundingBox(val: BoundingInfo) {
        this._boundingInfo = val;
    }

    /**
     * @hidden
     */
    public _instances: Map<EntityID, InstancedMesh> = new Map();
    get instances() {
        return this._instances;
    }
    createInstance(name: string): InstancedMesh {
        if (this._isDisposed) {
            return;
        }

        let instance = new InstancedMesh(name, this);

        return instance;
    }

    /**
     * @deprecated
     * @param key 
     * @param strip 
     */
    registerInstancedBuffer(key: string, strip: number) {

    }

    protected _morphTargetsInflueneWeights: [number, number, number, number] = [0, 0, 0, 0];
    public morphTargetInfluenceWeight(idx: number, weight: number) {
        this._morphTargetsInflueneWeights[idx] = weight;
    }
    protected _morphTargetsInfluence: number = 0;
    public set morphTargetsInfluence(val: number) {
        this._morphTargetsInfluence = val;
        p3d.p3d_mesh_morphtargetinfluence(Engine.commands(), this._entity, this._morphTargetsInflueneWeights[0] * this._morphTargetsInfluence, this._morphTargetsInflueneWeights[1] * this._morphTargetsInfluence, this._morphTargetsInflueneWeights[2] * this._morphTargetsInfluence, this._morphTargetsInflueneWeights[3] * this._morphTargetsInfluence);
    }
    public get morphTargetsInfluence(): number {
        return this._morphTargetsInfluence;
    }

    private _materials: Map<Constants.PassTag, Material> = new Map();

    protected _material: Material;
    public get material() {
        return this._material;
    }
    public set material(mat: Material) {
        (<any>mat)._meshes.add(this.entity);
        this.setMaterial(mat);
    }
    public setMaterial(materail: Material) {
        let oldmat = this._materials.get(materail.pass);
        if (oldmat != materail) {
            (<any>oldmat)?._meshes.delete(this.entity);
            this._material = materail;
            this._materials.set(materail.pass, materail);
            p3d.p3d_material_apply(Engine.commands(), materail.entity, this.entity, materail.pass);
        }
        // TODO
    }

    private _alpha: number = 1;

    public set visibility(val: number) {

    }
    public get visibility() {
        return this._alpha;
    }

    private _renderalignment: Constants.RenderAlignment;
    public get renderAlignment() {
        return this._renderalignment;
    }
    public set renderAlignment(val: Constants.RenderAlignment) {
        if (this._renderalignment != val) {
            this._renderalignment = val;
            this._scene.cmdsMeshRenderAlign.set(this._entity, this);
        }
    }

    private _receiveShadows: boolean;
    public get receiveShadows() {
        return this._receiveShadows;
    }
    public set receiveShadows(val: boolean) {
        this._receiveShadows = val;
        console.warn("TODO");
    }

    private _depthTestMode: Constants.CompareFunction = Constants.LEQUAL;
    public set depthTest(val: Constants.CompareFunction) {
        if (this._depthTestMode !== val) {
            this._depthTestMode = val;
            this._scene.cmdsMeshDepthTest.set(this._entity, this);
        }
    }
    public get depthTest() {
        return this._depthTestMode;
    }

    private _depthWrite: boolean = true;
    public set depthWrite(val: boolean) {
        this._depthWrite = val;
        this._scene.cmdsMeshDepthWrite.set(this._entity, this);
    }
    public get depthWrite() {
        return this._depthWrite;
    }
    

    private _blendMode: number = Material.OPAQUE;
    public set transparencyMode(val: number) {
        this._blendMode = val;
    }

    protected src_color: Constants.BlendFactor = Constants.ONE;
    protected dst_color: Constants.BlendFactor = Constants.ONE_MINUS_SRC_ALPHA;
    protected src_alpha: Constants.BlendFactor = Constants.ONE;
    protected dst_alpha: Constants.BlendFactor = Constants.ONE_MINUS_SRC_ALPHA;
    protected opt_color: Constants.BlendOperation = Constants.FUNC_ADD;
    protected opt_alpha: Constants.BlendOperation = Constants.FUNC_ADD;
    protected blend_enable: boolean = false;

    private _alphaMode: number = 0;
    public set alphaMode(val: number) {
        this._alphaMode = val;
        ExtensionTools.alphaMode(this._alphaMode, <any>this);

        this._scene.cmdsMeshBlend.set(this._entity, this);
    }
    
    private _cullMode: Constants.CullMode = Constants.BACK;
    public set cullMode(val: Constants.CullMode) {
        if (this._cullMode != val) {
            this._cullMode = val;
            this._scene.cmdsMeshCullMode.set(this._entity, this);
        }
    }
    public get cullMode() {
        return this._cullMode;
    }

    private _stencilTest: boolean = true;
    public set stencilTest(val: boolean) {
        this._stencilTest = val;
        // this._scene.cmdsMeshStencilTest
    }
    public get stencilTest() {
        return this._stencilTest;
    }

    private _frontFace: Constants.FrontFace = Constants.CW;
    public set frontFace(val: Constants.FrontFace) {
        if (this._frontFace != val) {
            this._frontFace = val;
            this._scene.cmdsMeshFrontFace.set(this._entity, this);
        }
    }
    public get frontFace() {
        return this._frontFace;
    }

    private _topolygon: Constants.TopolygonMode;
    private _topolygonDirty: boolean = false;
    public set topolygon(val: Constants.TopolygonMode) {
        this._topolygon = val;
        this._topolygonDirty = true;
        this._scene.cmdsForceDirty.set(this._entity, this);
    }
    public get topolygon() {
        return this._topolygon;
    }
    
    private _polygon: Constants.PolygonMode;
    private _polygonDirty: boolean = false;
    public set polygon(val: Constants.PolygonMode) {
        this._polygon = val;
        this._polygonDirty = true;
        this._scene.cmdsForceDirty.set(this._entity, this);
    }
    public get polygon() {
        return this._polygon;
    }

    /**
     * @deprecated 
     */
    public readonly subMeshes: Mesh[];

    private _doNotSyncBoundingInfo: boolean = true;
    public set doNotSyncBoundingInfo(val: boolean) {
        this._doNotSyncBoundingInfo = val;
    }
    public get doNotSyncBoundingInfo() {
        return this._doNotSyncBoundingInfo;
    }

    private _indexRangeDirty: boolean = false;
    private _indexStart: number = null;
    private _indexCount: number = null;
    public set indexStart(val: number | undefined) {
        this._indexStart = val;
        this._indexRangeDirty = true;
        this._scene.cmdsForceDirty.set(this._entity, this);
    }
    public get indexStart() {
        return this._indexStart;
    }
    public set indexCount(val: number | undefined) {
        this._indexCount = val;
        this._indexRangeDirty = true;
        this._scene.cmdsForceDirty.set(this._entity, this);
    }
    public get indexCount() {
        return this._indexCount;
    }

    private _vertexRangeDirty: boolean = false;
    private _vertexStart: number = null;
    private _vertexCount: number = null;
    public set vertexStart(val: number | undefined) {
        this._vertexStart = val;
        this._vertexRangeDirty = true;
        this._scene.cmdsForceDirty.set(this._entity, this);
    }
    public get vertexStart() {
        return this._vertexStart;
    }
    public set vertexCount(val: number | undefined) {
        this._vertexCount = val;
        this._vertexRangeDirty = true;
        this._scene.cmdsForceDirty.set(this._entity, this);
    }
    public get vertexCount() {
        return this._vertexCount;
    }

    _onFrameUpdate(): void {
        if (this._disposed) {
            return;
        }

        super._onFrameUpdate();

        if (this._indexRangeDirty) {
            p3d.p3d_mesh_indexrange(Engine.commands(), this.entity, this._indexStart, this._indexStart + this._indexCount);
            this._indexRangeDirty = false;
        }
        if (this._vertexRangeDirty) {
            p3d.p3d_mesh_vertexrange(Engine.commands(), this.entity, this._vertexStart, this._vertexCount);
            this._vertexRangeDirty = false;
        }
        if (this._topolygonDirty) {
            p3d.p3d_mesh_topology(Engine.commands(), this.entity, this._topolygon, Constants.PASS_TAG_01);
            p3d.p3d_mesh_topology(Engine.commands(), this.entity, this._topolygon, Constants.PASS_TAG_02);
            p3d.p3d_mesh_topology(Engine.commands(), this.entity, this._topolygon, Constants.PASS_TAG_03);
            p3d.p3d_mesh_topology(Engine.commands(), this.entity, this._topolygon, Constants.PASS_TAG_04);
            p3d.p3d_mesh_topology(Engine.commands(), this.entity, this._topolygon, Constants.PASS_TAG_05);
            p3d.p3d_mesh_topology(Engine.commands(), this.entity, this._topolygon, Constants.PASS_TAG_06);
            p3d.p3d_mesh_topology(Engine.commands(), this.entity, this._topolygon, Constants.PASS_TAG_07);
            p3d.p3d_mesh_topology(Engine.commands(), this.entity, this._topolygon, Constants.PASS_TAG_08);
            this._topolygonDirty = false;
        }
        if (this._polygonDirty) {
            p3d.p3d_mesh_polygon_mode(Engine.commands(), this.entity, this._polygon, Constants.PASS_TAG_01);
            p3d.p3d_mesh_polygon_mode(Engine.commands(), this.entity, this._polygon, Constants.PASS_TAG_02);
            p3d.p3d_mesh_polygon_mode(Engine.commands(), this.entity, this._polygon, Constants.PASS_TAG_03);
            p3d.p3d_mesh_polygon_mode(Engine.commands(), this.entity, this._polygon, Constants.PASS_TAG_04);
            p3d.p3d_mesh_polygon_mode(Engine.commands(), this.entity, this._polygon, Constants.PASS_TAG_05);
            p3d.p3d_mesh_polygon_mode(Engine.commands(), this.entity, this._polygon, Constants.PASS_TAG_06);
            p3d.p3d_mesh_polygon_mode(Engine.commands(), this.entity, this._polygon, Constants.PASS_TAG_07);
            p3d.p3d_mesh_polygon_mode(Engine.commands(), this.entity, this._polygon, Constants.PASS_TAG_08);
            this._polygonDirty = false;
        }
    }
    
    checkCollisions: boolean;
    _checkCollision(data: any) {
        
    }

    dispose(doNotRecurse?: boolean, disposeMaterialAndTextures?: boolean): void {
        if (this._disposed) {
            return;
        }
        this.instances.forEach((ins) => {
            (<InstancedMesh>ins).dispose();
        });
        this.instances.clear();
        this._cacheInstances.forEach(function(entity) {
            Entity.Dispose(entity);
        });
        this._cacheInstances.length = 0;
        super.dispose();

        this._materials.forEach((v) => {
            (<any>v)?._meshes.delete(this.entity);
            v.dispose();
        });
    }
}

export class BoxBuilder {
    public static CreateBox(name: string, param: { size: number }, scene: Scene, instancestate: InstanceAttribute[] = [], instancematrix: boolean = false, instances_use_common_buffer: boolean = true) {
        let mesh = new Mesh(name, scene, instancestate, instances_use_common_buffer, instancematrix);
        let meta = p3d.GeometryMeta.create_box();
        p3d.p3d_mesh_geometry(Engine.native(), Engine.commands(), mesh.entity, meta);
        meta.free && meta.free();

        return mesh;
    }
    public static meta(): IGeometryMeta {
        return p3d.GeometryMeta.create_box();
    }
}

export class QuadBuilder {
    public static CreateQuad(name: string, param: { size: number }, scene: Scene, instancestate: InstanceAttribute[] = [], instancematrix: boolean = false, instances_use_common_buffer: boolean = true) {
        let mesh = new Mesh(name, scene, instancestate, instances_use_common_buffer, instancematrix);
        mesh.cullMode = 0;
        mesh.frontFace = Constants.CCW;
        let meta = p3d.GeometryMeta.create_quad();
        p3d.p3d_mesh_geometry(Engine.native(), Engine.commands(), mesh.entity, meta);
        meta.free && meta.free();

        return mesh;
    }
    public static meta(): IGeometryMeta {
        return p3d.GeometryMeta.create_quad();
    }
}

export class PlaneBuilder extends QuadBuilder {
    public static CreatePlane(name: string, opt: { size?: number, width?: number, height?: number }, scene: Scene, instancestate: InstanceAttribute[] = [], instancematrix: boolean = false, instances_use_common_buffer: boolean = true) {
        let mesh = new Mesh(name, scene, instancestate, instances_use_common_buffer, instancematrix);
        mesh.cullMode = 0;
        mesh.frontFace = Constants.CCW;
        let meta = p3d.GeometryMeta.create_quad();
        p3d.p3d_mesh_geometry(Engine.native(), Engine.commands(), mesh.entity, meta);
        meta.free && meta.free();

        return mesh;
    }
}

Node.prototype.getChildMeshes = function(this: Node) {
    let result = [];

    (<any>this)._children.forEach((item) => {
        if (item instanceof AbstractMesh) {
            result.push(item);
        }
    });

    return result;
}