import { Entity, Engine } from "./engines/engine";
import { IEntity, INode, IPickInfo, ITransformNode } from "../interface/index";
import { Scene } from "./scene";
import { MathDataPool, Matrix, Quaternion, Space, Vector3 } from "../maths/index";
import { IFrameUpdate, INodeColliderD3, INodeD3, ITreeNode, p3d } from "./types";
import { EntityID, Observable } from "./entity";
import { PointerInfo } from "./events/pointerEvents";

export class Node extends Entity implements IFrameUpdate, INode, IEntity, ITreeNode, INodeColliderD3 {
    public static Info: Float64Array = new Float64Array(4096);
    protected static MatrixReceive: Float32Array = new Float32Array(16);

    public static SetPosition(node: Node, x: number, y: number, z: number) {
        if (node._isDisposed) {
            return;
        }
        // if (node._position.x == x && node._position.y == y && node._position.z == z) return;
        node._position.copyFromFloats(x, y, z);
        node._position._isDirty = false;
        node._worldMatrixDirty = true;
        p3d.p3d_local_position(Engine.commands(), node._entity, x, y, z);
    }

    public readonly _scene: Scene;
    protected _name: string;
    protected _id: any;
    protected _parentNode: Node;
    public rayID: any;
    constructor(name: string, scene: Scene, entity: number) {
        if (scene._isDisposed) {
            debugger;
        }
        super(entity);
        this._name = name;
        this._id = name;
        this._scene = scene;
        // TODO
        this._scene.transforms.set(this.entity, this);
        this._scene._rootNodeMap.set(this.entity, this);
        this.onPointerObservable = new Observable<PointerInfo>();
    }

    getChildren(predicate?: (node: ITreeNode) => boolean, directDescendantsOnly?: boolean): ITreeNode[] {
        if (this._isDisposed) {
            return [];
        } else {
            let result = [];
            this._children.forEach((key) => {
                // let node = this._scene.transforms.get(key.entity);
                // if (node) {
                //     result.push(node);
                // }
                result.push(key);
            });
            // let count = Engine.queryChildren(this._entity, Node.Info);
            // for (let i = 0; i < count; i++) {
            //     let key = Node.Info[i];
            //     let node = this._scene.transforms.get(key);
            //     if (node) {
            //         result.push(node);
            //     }
            // }
            return result;
        }
    }
    dirtyPosition(): void {
        p3d.p3d_local_position(Engine.commands(), this.entity, this._position.x, this._position.y, this._position.z);
    }
    dirtyPick(): void {
        throw new Error("Method not implemented.");
    }
    // dirtyParent(): void {
    //     p3d.p3d_transform_node_parent(Engine.commands(), this.entity, this._parent ? this._parent.entity : this._scene.entity);
    // }
    dirtyEnable(): void {
        p3d.p3d_node_enable(Engine.commands(), this.entity, this._isEnabled);
    }

    public get id() {
        return this._id;
    }
    public set id(val: any) {
        this._id = val;
    }
    public get name() {
        return this._name;
    }
    public set name(val: string) {
        this._name = val;
    }
    
    protected _children: Map<EntityID, Node> = new Map();
    public get parent() {
        return this._parentNode;
    }
    public set parent(val: INode) {
        if (this._disposed) {
            return;
        }

        if (val?.isDisposed()) {
            return;
        }

        if (this._parentNode) {
            (<Node>this._parentNode)?._children.delete(this.entity);
        }

        if (this._parentNode !== val) {
            this._parentNode = <Node>val;
            // console.warn(`Parent ${val?.uniqueID} Child ${this.uniqueID}`);
            (<Node>this._parentNode)?._children.set(this.entity, this);
            // this._scene.cmdsParent.set(this._entity, this);
            if (this._parentNode) {
                p3d.p3d_transform_node_parent(Engine.commands(), this.entity, this._parentNode.entity);
            } else if (this._scene._isDisposed == false) {
                p3d.p3d_transform_node_parent(Engine.commands(), this.entity, this._scene.entity);
            } else {
                debugger;
            }
            this._worldMatrixDirty = true;
        }

        if (this._parentNode) {
            this._scene._rootNodeMap.delete(this._entity);
        } else {
            this._scene._rootNodeMap.set(this._entity, this);
        }
        
        // Enabled state
        this._syncParentEnabledState();
    }
    public setParent(val: INodeD3) {
        this.parent = val;
    }

    protected __collider: {
        minimum: Vector3;
        maximum: Vector3;
        sphereTreshold: number;
    };
    getCollider(): {
        minimum: Vector3;
        maximum: Vector3;
        sphereTreshold: number;
    } {
        return this.__collider;
    }
    setCollider(val: {
        minimum: Vector3;
        maximum: Vector3;
        sphereTreshold: number;
    }) {
        this.__collider = val;
        this._scene.cmdsCollider.set(this._entity, this);
    }

    protected _position: Vector3 = MathDataPool.Vector3(0, 0, 0);
    public set position(val: Vector3) {
        this.setPosition(val.x, val.y, val.z);
    }
    public get position() {
        this._scene.cmdsForceDirty.set(this.entity, this);
        return this._position;
    }
    public setPosition(x: number, y: number, z: number) {
        Node.SetPosition(this, x, y, z);
    }

    public getScene() {
        return this._scene;
    }

    public getChildMeshes(): Node[] {
        return [];
    }

    public getChildTransformNodes(): ITransformNode[] {
        let result = [];
    
        (<any>this)._children.forEach((item) => {
            result.push(item);
        });
    
        return result;
    }

    dispose(disposeEntity: boolean = true): void {
        if (this._disposed) {
            return;
        }

        this._children.forEach((v) => {
            v.dispose();
        });

        this._scene._rootNodeMap.delete(this._entity);
        this._scene.transforms.delete(this.entity);

        this.onPointerObservable.clear();
        super.dispose(disposeEntity);

        MathDataPool.Recycle(this._globalPosition   );
        MathDataPool.Recycle(this._worldMatrix      );
        MathDataPool.Recycle(this.__localMatrix     );

        this._globalPosition   = undefined;
        this._worldMatrix      = undefined;
        this.__localMatrix     = undefined;

        this._name = undefined;
        // this._scene = undefined;
        this._parentNode = undefined;
    }

    isDisposed(): boolean {
        return this._isDisposed;
    }

    protected __worldMatrixDirty: boolean = true;
    protected set _worldMatrixDirty(val: boolean) {
        this.__worldMatrixDirty = val;
        if (val) {
            this.__childUpdateId();
        }
    }
    _onFrameUpdate(): void {
        if (this._isDisposed) {
            return;
        }

        if (this._position._isDirty) {
            this._position._isDirty = false;
            this._worldMatrixDirty = true;
            p3d.p3d_local_position(Engine.commands(), this._entity, this._position._x, this._position._y, this._position._z);
        }
    }

    private _globalPosition: Vector3 = MathDataPool.Vector3(0, 0, -1);
    public get globalPosition() {
        this._computeWorldMatrix();
        return this._globalPosition.clone();
    }
    public getAbsolutePosition() {
        return this.globalPosition;
    }

    private _computeWorldMatrix() {
        this.isSynchronized();
        if (this.__worldMatrixDirty) {
            if (p3d.p3d_query_world_matrix(Engine.native(), Engine.param(), this.entity, Node.MatrixReceive)) {
                Matrix.FromArrayToRef(Node.MatrixReceive, 0, this._worldMatrix);
            }
            if (p3d.p3d_query_local_matrix(Engine.native(), Engine.param(), this.entity, Node.MatrixReceive)) {
                Matrix.FromArrayToRef(Node.MatrixReceive, 0, this.__localMatrix);
            }
            this._worldMatrix.decompose(undefined, undefined, this._globalPosition);
            this._worldMatrixDirty = false;
        }
    }

    private _isEnabled: boolean = true;
    setEnabled(val: boolean) {
        this._isEnabled = val;
        p3d.p3d_node_enable(Engine.commands(), this.entity, this._isEnabled);
        this._syncParentEnabledState();
    }
    isEnabled(checkAncestors: boolean = true) {
        if (checkAncestors === false) {
            return this._isEnabled;
        }

        if (!this._isEnabled) {
            return false;
        }

        return this._isParentEnabled;

        if (this._isEnabled == false) {
            return false;
        }
        if (p3d.p3d_query_transform_state(Engine.native(), Engine.param(), this.entity, Node.MatrixReceive)) {
            return Node.MatrixReceive[1] > 0;
        }
        return false;
    }
    get enable() {
        return this._isEnabled;
    }

    private _worldMatrix: Matrix = MathDataPool.Matrix();
    public computeWorldMatrix() {
        this._computeWorldMatrix();
        return this._worldMatrix.clone();
    }

    private __localMatrix: Matrix = MathDataPool.Matrix();
    public get _localMatrix() {
        this._computeWorldMatrix();
        return this.__localMatrix.clone();
    }

    /// Pick
    private _isPickable: boolean = false;
    /**
     * Gets or sets a boolean indicating if the mesh can be picked (by scene.pick for instance or through actions). Default is false
     */
    public set isPickable(val: boolean) {
        this._isPickable = val;
    }
    public get isPickable() {
        return this._isPickable;
    }

    private _pickInfo: IPickInfo;
    public setPickInfo(val: IPickInfo) {
        this._pickInfo = val;
    }

    /**
     * For Ts Animation
     */
    animations: any[] = [];
    
    /** @hidden */
    public _getDescendants(results: INode[], directDescendantsOnly: boolean = false, predicate?: (node: INode) => boolean): void {
        if (!this._children) {
            return;
        }

        this._children.forEach((item) => {
            if (!predicate || predicate(item)) {
                results.push(item);
            }

            if (!directDescendantsOnly) {
                item._getDescendants(results, false, predicate);
            }
        })
    }

    /**
     * Will return all nodes that have this node as ascendant
     * @param directDescendantsOnly defines if true only direct descendants of 'this' will be considered, if false direct and also indirect (children of children, an so on in a recursive manner) descendants of 'this' will be considered
     * @param predicate defines an optional predicate that will be called on every evaluated child, the predicate must return true for a given child to be part of the result, otherwise it will be ignored
     * @return all children nodes of all types
     */
    public getDescendants(directDescendantsOnly?: boolean, predicate?: (node: Node) => boolean): Node[] {
        var results = new Array<Node>();

        this._getDescendants(results, directDescendantsOnly, predicate);

        return results;
    }

    private _isParentEnabled = true;
    /**
     * @hidden 
     * 递归遍历子节点同步其 isEnabled
     */
    protected _syncParentEnabledState() {
        this._isParentEnabled = this._parentNode ? this._parentNode.isEnabled() : true;

        if (this._children) {
            this._children.forEach((c) => {
                c._syncParentEnabledState(); // Force children to update accordingly
            });
        }
    }

    private _parentUpdateId = -1;
    /** @hidden */
    public _childUpdateId = -1;
    /** @hidden */
    public _cache: any = {};

    /** @hidden */
    public _markSyncedWithParent() {
        if (this._parentNode) {
            this._parentUpdateId = this._parentNode._childUpdateId;
        }
    }

    /** @hidden */
    public isSynchronizedWithParent(): boolean {
        if (!this._parentNode) {
            return true;
        }

        if (this._parentUpdateId !== this._parentNode._childUpdateId) {
            return false;
        }

        return this._parentNode.isSynchronized();
    }

    // override it in derived class if you add new variables to the cache
    /** @hidden */
    public _isSynchronized(): boolean {
        return true;
    }

    /** @hidden */
    public isSynchronized(): boolean {
        if (this._cache.parent != this._parentNode) {
            this._cache.parent = this._parentNode;
            return false;
        }

        if (this._parentNode && !this.isSynchronizedWithParent()) {
            return false;
        }

        let result = this._isSynchronized();
        if (result) {
            this.__worldMatrixDirty = result;
        }
        return result;
    }

    protected __childUpdateId() {
        this._childUpdateId ++;
    }
    
    /**
     * Observable event triggered each time an input event is received from the rendering canvas
     */
    public onPointerObservable: Observable<PointerInfo>;
}

export class IsometryNode extends Node {
    public static TmpQuaternionA = new Quaternion();
    public static TmpQuaternionB = new Quaternion();
    public static TmpVector3A = new Vector3();
    public static TmpVector3B = new Vector3();
    public static TmpVector3C = new Vector3();
    public static SetRotation(node: IsometryNode, x: number, y: number, z: number) {
        if (node._isDisposed) {
            return;
        }
        // if (node._rotation.x == x && node._rotation.y == y && node._rotation.z == z) return;
        node._rotation.copyFromFloats(x, y, z);
        node._rotation._isDirty = false;
        node._worldMatrixDirty = true;
        Quaternion.FromEulerVectorToRef(node._rotation, node._rotationQuaternion);
        node._rotationQuaternion._isDirty = false;
        p3d.p3d_local_quaternion(Engine.commands(), node.entity, node._rotationQuaternion.x, node._rotationQuaternion.y, node._rotationQuaternion.z, node._rotationQuaternion._w);
    }
    public static SetRotationQuaternion(node: IsometryNode, i: number, j: number, k: number, w: number) {
        if (node._isDisposed) {
            return;
        }
        node._rotationQuaternion.copyFromFloats(i, j, k, w);
        node._rotationQuaternion.toEulerAnglesToRef(node._rotation);
        node._rotationQuaternion._isDirty = false;
        node._rotation._isDirty = false;
        node._worldMatrixDirty = true;
        p3d.p3d_local_quaternion(Engine.commands(), node.entity, node._rotationQuaternion.x, node._rotationQuaternion.y, node._rotationQuaternion.z, node._rotationQuaternion._w);
    }
    protected _rotation: Vector3 = MathDataPool.Vector3(0, 0, 0);
    public set rotation(val: Vector3) {
        this.setRotation(val.x, val.y, val.z);
    }
    public get rotation() {
        this._scene.cmdsForceDirty.set(this.entity, this);
        return this._rotation;
    }
    public setRotation(x: number, y: number, z: number) {
        IsometryNode.SetRotation(this, x, y, z);
        return this;
    }
    protected _rotationQuaternion: Quaternion = MathDataPool.Quaternion(0, 0, 0, 1);
    public set rotationQuaternion(val: Quaternion) {
        this.setRotationQuaternion(val.x, val.y, val.z, val.w);        
    }
    public get rotationQuaternion() {
        this._scene.cmdsForceDirty.set(this.entity, this);
        return this._rotationQuaternion;
        // p3d.p3d_local_euler(Engine.native(), Engine.param(), this.entity, this._r_x, this._r_y, this._r_z);
    }
    public setRotationQuaternion(i: number, j: number, k:  number, w: number) {
        IsometryNode.SetRotationQuaternion(this, i, j, k, w);
        return this;
    }
    _onFrameUpdate(): void {
        if (this._isDisposed) {
            return;
        }
        super._onFrameUpdate();
        if (this._rotation._isDirty) {
            this.setRotation(this._rotation.x, this._rotation.y, this._rotation.z);
        }
        if (this._rotationQuaternion._isDirty) {
            this.setRotationQuaternion(this._rotationQuaternion.x, this._rotationQuaternion.y, this._rotationQuaternion.z, this._rotationQuaternion.w);
        }
    }
    public rotate(axis: Vector3, amount: number, space: number = Space.LOCAL) {
        if (this._isDisposed) {
            return;
        }
        axis.normalize();
        if (!space || space === Space.LOCAL) {
            Quaternion.RotationAxisToRef(axis, amount, Transform.TmpQuaternionA);
            this._rotationQuaternion.multiplyToRef(Transform.TmpQuaternionA, Transform.TmpQuaternionB);
            this.setRotationQuaternion(Transform.TmpQuaternionB.x, Transform.TmpQuaternionB.y, Transform.TmpQuaternionB.z, Transform.TmpQuaternionB.w);
        }
        else {
            console.error("Not Support Rotate In WorldSpace")
            // if (this.parent) {
            //     var invertParentWorldMatrix = TmpVectors.Matrix[0];
            //     (<TransformNode>this.parent).getWorldMatrix().invertToRef(invertParentWorldMatrix);
            //     axis = Vector3.TransformNormal(axis, invertParentWorldMatrix);
            // }
            // rotationQuaternion = Quaternion.RotationAxisToRef(axis, amount, TransformNode._rotationAxisCache);
            // rotationQuaternion.multiplyToRef(this.rotationQuaternion, this.rotationQuaternion);
        }
        return this;
    }

    /**
     * Sets this transform node rotation to the given local axis.
     * @param localAxis the axis in local space
     * @param yawCor optional yaw (y-axis) correction in radians
     * @param pitchCor optional pitch (x-axis) correction in radians
     * @param rollCor optional roll (z-axis) correction in radians
     * @returns this TransformNode
     */
    public setDirection(localAxis: Vector3, yawCor: number = 0, pitchCor: number = 0, rollCor: number = 0): IsometryNode {
        var yaw = -Math.atan2(localAxis.z, localAxis.x) + Math.PI / 2;
        var len = Math.sqrt(localAxis.x * localAxis.x + localAxis.z * localAxis.z);
        var pitch = -Math.atan2(localAxis.y, len);
        if (this.rotationQuaternion) {
            Quaternion.RotationYawPitchRollToRef(yaw + yawCor, pitch + pitchCor, rollCor, Transform.TmpQuaternionA);
            this.setRotationQuaternion(Transform.TmpQuaternionA.x, Transform.TmpQuaternionA.y, Transform.TmpQuaternionA.z, Transform.TmpQuaternionA.w);
        }
        else {
            this.setRotation(pitch + pitchCor, yaw + yawCor, rollCor);
        }
        return this;
    }

    lookAt(target: Vector3) {

    }
    dispose(disposeEntity?: boolean): void {
        super.dispose(disposeEntity);
        MathDataPool.Recycle(this._rotation);
        this._rotation = undefined;
        MathDataPool.Recycle(this._rotationQuaternion);
        this._rotationQuaternion = undefined;
    }
}

export class Transform extends IsometryNode {
    public static SetScaling(node: Transform, x: number, y: number, z: number) {
        if (node._isDisposed) {
            return;
        }
        // if (node._scaling.x == x && node._scaling.y == y && node._scaling.z == z) return;
        node._scaling.copyFromFloats(x, y, z);
        node._scaling._isDirty = false;
        node._worldMatrixDirty = true;
        p3d.p3d_local_scaling(Engine.commands(), node._entity, x, y, z);
    }

    protected _scaling: Vector3 = MathDataPool.Vector3(1, 1, 1);
    public set scaling(val: Vector3) {
        Transform.SetScaling(this, val.x, val.y, val.z);
    }
    public get scaling() {
        this._scene.cmdsForceDirty.set(this.entity, this);
        return this._scaling;
    }
    public setScaling(x: number, y: number, z: number) {
        Transform.SetScaling(this, x, y, z);
        return this;
    }
    _onFrameUpdate(): void {
        if (this._isDisposed) {
            return;
        }
        super._onFrameUpdate();

        if (this._scaling._isDirty) {
            let x = this._scaling.x;
            let y = this._scaling.y;
            let z = this._scaling.z;
            if (Math.abs(x) < this._scene.miniScalingScalar) {
                x = this._scene.miniScalingScalar * (x < 0 ? -1 : 1);
            }
            if (Math.abs(y) < this._scene.miniScalingScalar) {
                y = this._scene.miniScalingScalar * (y < 0 ? -1 : 1);
            }
            if (Math.abs(z) < this._scene.miniScalingScalar) {
                z = this._scene.miniScalingScalar * (z < 0 ? -1 : 1);
            }
            this.setScaling(x, y, z);
        }
    }
    dispose(disposeEntity?: boolean): void {
        super.dispose(disposeEntity);
        MathDataPool.Recycle(this._scaling);
        this._scaling = undefined;
    }
}
