
import { ICamera, ICustomRenderTarget, TAutoRenderTarget, Option, Constants, ITempFrameBuffer, GlContext, IRenderer, PIWebGLFramebuffer } from "../interface/index";
import { Scene } from "./scene";
import { Camera } from "./cameras/camera";
import { Engine, Entity } from "./engines/engine";
import { Light } from "./lights";
import { ShadowGenerator } from "./shadowgenerator/index";
import { IRenderPipelineD3, IRendererD3, p3d } from "./types";

export class Renderer extends Entity implements IRendererD3 {
    public readonly pass: Constants.PassTag;
    constructor(camera: Camera, key: string, pass_tag: Constants.PassTag, transparent: boolean) {
        let [entity, name] = camera.createRenderer(key, pass_tag, transparent);
        super(entity);
        this.pass = pass_tag;
        this.name = name;
    }
    name?: string;
    _onFrameUpdate(): void {
        throw new Error("Method not implemented.");
    }
    protected _autoClearColor: boolean;
    autoClearColor(flag: boolean): void {
        this._autoClearColor = flag;
        p3d.p3d_render_auto_clear_color(Engine.commands(), this.entity, flag);
    }
    autoClearDepth(flag: boolean): void {
        p3d.p3d_render_auto_clear_depth(Engine.commands(), this.entity, flag);
    }
    autoClearStencil(flag: boolean): void {
        p3d.p3d_render_auto_clear_stencil(Engine.commands(), this.entity, flag);
    }
    clearColor(r: number, g: number, b: number, a: number): void {
        p3d.p3d_render_clear_color(Engine.commands(), this.entity, r, g, b, a * 255);
    }
    clearDepth(depth: number): void {
        p3d.p3d_render_clear_depth(Engine.commands(), this.entity, depth);
    }
    clearStencil(value: number): void {
        p3d.p3d_render_clear_stencil(Engine.commands(), this.entity, value);
    }
    setRenderTarget(target: ITempFrameBuffer | TAutoRenderTarget): void {
        if ((<TAutoRenderTarget>target)?.autoTarget) {
            p3d.p3d_render_target_auto(Engine.commands(), this.entity, (<TAutoRenderTarget>target).width, (<TAutoRenderTarget>target).height, (<TAutoRenderTarget>target).colorFormat, (<TAutoRenderTarget>target).depthStencilFormat);
        } else {
            p3d.p3d_render_target_key(Engine.commands(), this.entity, (<PIWebGLFramebuffer>(<ICustomRenderTarget>target)?.frameBuffer)?.key);
        }
    }
    viewport(x: number, y: number, width: number, height: number) {
        p3d.p3d_render_viewport(Engine.commands(), this.entity, x, y, width, height);
    }
    dispose(): void {
        // debugger;
        super.dispose();
    }
    private _next: Renderer;
    next(item: Renderer) {
        if (this._next != item) {
            if (this._next != undefined) {
                console.warn(`Dis ${this.name} XX ${this._next?.name}`);
                p3d.p3d_render_graphic(Engine.commands(), this.entity, this._next.entity, true);
            }
            this._next = item;
            if (this._next != undefined) {
                p3d.p3d_render_graphic(Engine.commands(), this.entity, this._next.entity, false);
            }
            // console.warn("Next: " + this._next?.name);
            console.warn(`${this.name} >> ${this._next?.name}`);
        }
    }
    private _pre: Renderer;
    pre(item: Renderer) {
        if (this._pre != item) {
            if (this._pre != undefined) {
                console.warn(`Dis ${this._pre?.name} XX ${this.name}`);
                p3d.p3d_render_graphic(Engine.commands(), this._pre.entity, this.entity, true);
            }
            this._pre = item;
            if (this._pre != undefined) {
                p3d.p3d_render_graphic(Engine.commands(), this._pre.entity, this.entity, false);
            }
            console.warn(`${this._pre?.name} >> ${this.name}`);
        }
    }
    reset() {
        if (this._next != undefined) {
            console.warn(`Dis ${this.name} XX ${this._next?.name}`);
            p3d.p3d_render_graphic(Engine.commands(), this.entity, this._next.entity, true);
        }
        if (this._pre != undefined) {
            console.warn(`Dis ${this._pre?.name} XX ${this.name}`);
            p3d.p3d_render_graphic(Engine.commands(), this._pre.entity, this.entity, true);
        }
        this._next = undefined;
        this._pre = undefined;
    }
}

export class CustomRenderTarget implements ITempFrameBuffer {
    private _key: number;
    public get key() {
        return this._key;
    }
    constructor(
        public readonly width: number,
        public readonly height: number,
        public readonly colorFormat: number,
        public readonly depthStencilFormat: number,
        public readonly filter: number,
        public readonly address: number,
        public readonly anisotropy_clamp: number,
        public readonly screen: boolean = false,
    ) {
        if (screen == false) {
            this._key = p3d.p3d_create_render_target(Engine.native(), Engine.param(), colorFormat, depthStencilFormat, width, height, filter, address, anisotropy_clamp);
        }
        this.rect = [0, 0, width, height];
        this._framebuffer = { key: this._key };
    }
    rect: [number, number, number, number];
    protected _framebuffer: PIWebGLFramebuffer;
    get frameBuffer(): PIWebGLFramebuffer {
        return this._framebuffer;
    }
    ctx?: GlContext;
    public dispose() {
        p3d.p3d_dispose_render_target(Engine.native(), Engine.param(), this._key);
    }
}

export class ForwardRenderPipeline implements IRenderPipelineD3 {
    public readonly predepthRenderer: Option<Renderer>;
    public readonly opaqueRenderer: Option<Renderer>;
    public readonly opaqueTextureRenderer: Option<Renderer>;
    public readonly skyRenderer: Option<Renderer>;
    public readonly transparentRenderer: Option<Renderer>;
    public readonly shadowGenerator: Option<ShadowGenerator>;

    public readonly PASS_SHADOW = Constants.PASS_TAG_01;
    public readonly PASS_PREDEPTH = Constants.PASS_TAG_02;
    public readonly PASS_OPAQUE = Constants.PASS_TAG_03;
    public readonly PASS_SKY = Constants.PASS_TAG_04;
    public readonly PASS_OPAQUE_TEXTURE = Constants.PASS_TAG_05;
    public readonly PASS_TRANSPARENT = Constants.PASS_TAG_06;
    private _renders: IRendererD3[] = [];
    constructor(
        public readonly scene: Scene,
        public readonly maincamera: Camera,
        public readonly shadowLight: Light,
        public readonly predepthEanble = true,
        public readonly opaqueEanble = true,
        public readonly opaqueTextureEanble = true,
        public readonly skyEanble = true,
        public readonly transparentEanble = true,
    ) {
        if (this.shadowLight) {
            this.shadowGenerator = new ShadowGenerator(scene, shadowLight, this.PASS_SHADOW);
            this._renders.push(this.shadowGenerator);
        }
        if (this.predepthEanble) {
            this.predepthRenderer = new Renderer(maincamera, "PreDepth", this.PASS_PREDEPTH, false);
            this._renders.push(this.predepthRenderer);
        }
        if (this.opaqueEanble) {
            this.opaqueRenderer = new Renderer(maincamera, "Opaque", this.PASS_OPAQUE, false);
            this._renders.push(this.opaqueRenderer);
        }
        if (this.skyEanble) {
            this.skyRenderer = new Renderer(maincamera, "Sky", this.PASS_SKY, false);
            this._renders.push(this.skyRenderer);
        }
        if (this.opaqueTextureEanble) {
            this.opaqueTextureRenderer = new Renderer(maincamera, "OpaqueTexture", this.PASS_OPAQUE_TEXTURE, false);
            this._renders.push(this.opaqueTextureRenderer);
        }
        if (this.transparentEanble) {
            this.transparentRenderer = new Renderer(maincamera, "Transparent", this.PASS_TRANSPARENT, true);
            this._renders.push(this.transparentRenderer);
        }
        
        let len = this._renders.length;
        for (let i = 1; i < len; i++) {
            let before: Renderer = <any>this._renders[i - 1];
            let after: Renderer = <any>this._renders[i];
            p3d.p3d_render_graphic(Engine.commands(), before.entity, after.entity, false);
        }
    }
    get first() {
        return this._renders[0];
    }
    get final(): Renderer {
        return <Renderer>this._renders[this._renders.length - 1];
    }
    public insert_between(renderer: number, before: number, after: number) {
        p3d.p3d_render_graphic(Engine.commands(), before, after, true);
        p3d.p3d_render_graphic(Engine.commands(), before, renderer, false);
        p3d.p3d_render_graphic(Engine.commands(), renderer, after, false);
    }
    dispose() {
        this._renders.forEach((item) => {
            item.dispose();
        });
        this._renders.length = 0;
    }
    public get renderers() {
        return this._renders;
    }
    public queryPass(queue: number) {
        if (queue < 2400) {
            return this.PASS_OPAQUE;
        } else if (queue < 3000) {
            return this.PASS_SKY;
        } else {
            return this.PASS_TRANSPARENT;
        }
    }
}