import { Entity, Engine } from "./engines/engine";
import { Constants, GlContext, IShadowGenerator, ITargetAnimation, Nullable, Observable } from "../interface/index";
import { Plane } from "../maths/math.plane";
import { Camera } from "./cameras/camera";
import { IClipPlanesHolder } from "./misc/interfaces/iClipPlanesHolder";
import { ICollisionCoordinator } from "./collisions/collisionCoordinator";
import { PickingInfo } from "./collisions/pickingInfo";
import { Color3 } from "../maths/math.color";
import { IAnimationGroup3D, IFrameUpdate, IInstancedMeshD3, ILayerMaskD3, IMaterialD3, IMeshD3, IMeshRenderQueueD3, INodeColliderD3, INodeD3
    , IParticleSystemD3, IPickingInfo, IRenderPipelineD3, ISceneD3, ISprite3D, p3d
} from "./types";
import { EntityID, Option } from "./entity";
import { ExtensionTools } from "../base";
import { Matrix, Vector3 } from "../maths/math.vector";
import { Ray } from "./culling/ray";
import { InputManager } from "./inputs/inputManager";
import { PointerEventTypes } from "./events/pointerEvents";
import { EventData } from "../interface/event";
import { KeyboardInfo } from "./events/keyboardEvents";


/** Interface defining initialization parameters for Scene class */
export interface SceneOptions {
    /**
     * Defines that scene should keep up-to-date a map of geometry to enable fast look-up by uniqueId
     * It will improve performance when the number of geometries becomes important.
     */
    useGeometryUniqueIdsMap?: boolean;

    /**
     * Defines that each material of the scene should keep up-to-date a map of referencing meshes for fast diposing
     * It will improve performance when the number of mesh becomes important, but might consume a bit more memory
     */
    useMaterialMeshMap?: boolean;

    /**
     * Defines that each mesh of the scene should keep up-to-date a map of referencing cloned meshes for fast diposing
     * It will improve performance when the number of mesh becomes important, but might consume a bit more memory
     */
    useClonedMeshMap?: boolean;

    /** Defines if the creation of the scene should impact the engine (Eg. UtilityLayer's scene) */
    virtual?: boolean;
}

export class Scene extends Entity implements IClipPlanesHolder, ISceneD3 {
    public static _PickResultCache: Float64Array = new Float64Array(16);
    
    /** The fog is deactivated */
    public static readonly FOGMODE_NONE: number = 0;
    /** The fog density is following an exponential function */
    public static readonly FOGMODE_EXP: number = 1;
    /** The fog density is following an exponential function faster than FOGMODE_EXP */
    public static readonly FOGMODE_EXP2: number = 2;
    /** The fog density is following a linear function. */
    public static readonly FOGMODE_LINEAR: number = 3;
    public static readonly FOGMODE_HEIGHT_BASE: number = 4;
    public static readonly FOGMODE_ALTITUDE_BASE: number = 4;

    public static readonly BOUNDING_CULLING_MODE_NONE = 0;
    public static readonly BOUNDING_CULLING_MODE_QUAD_TREE = 1;
    public static readonly BOUNDING_CULLING_MODE_QCT_TREE = 2;
    
    public static readonly BOUNDING_COLLIDER_MODE_NONE = 0;
    public static readonly BOUNDING_COLLIDER_MODE_QUAD_TREE = 1;
    public static readonly BOUNDING_COLLIDER_MODE_QCT_TREE = 2;
    
    public static readonly BOUNDING_CULLING_PRAMA: Int32Array = new Int32Array(9);

    // PI_END
    /**
     * Gets or sets the minimum deltatime when deterministic lock step is enabled
     * @see https://doc.babylonjs.com/babylon101/animations#deterministic-lockstep
     */
    public static MinDeltaTime = 1.0;
    /**
     * Gets or sets the maximum deltatime when deterministic lock step is enabled
     * @see https://doc.babylonjs.com/babylon101/animations#deterministic-lockstep
     */
    public static MaxDeltaTime = 1000.0;

    public static MAX_UINT16_PARAM_RECORD       = 1 * 1024;
    public static MAX_TRANSFORM_PARAM_RECORD    = 1 * 1024;
    public static TRANSFORM_PARAM_v1_SIZE       = 2;
    public static TRANSFORM_PARAM_v2_SIZE       = 3;
    public static TRANSFORM_PARAM_v3_SIZE       = 4;
    public static TRANSFORM_PARAM_v4_SIZE       = 5;
    public static MAX_TRANSFORM_PARAM_RECORD_V1 = Math.floor(Scene.MAX_TRANSFORM_PARAM_RECORD / Scene.TRANSFORM_PARAM_v1_SIZE);
    public static MAX_TRANSFORM_PARAM_RECORD_V2 = Math.floor(Scene.MAX_TRANSFORM_PARAM_RECORD / Scene.TRANSFORM_PARAM_v2_SIZE);
    public static MAX_TRANSFORM_PARAM_RECORD_V3 = Math.floor(Scene.MAX_TRANSFORM_PARAM_RECORD / Scene.TRANSFORM_PARAM_v3_SIZE);
    public static MAX_TRANSFORM_PARAM_RECORD_V4 = Math.floor(Scene.MAX_TRANSFORM_PARAM_RECORD / Scene.TRANSFORM_PARAM_v4_SIZE);
    public static TRANSFORM_PARAM_BUFFER        = new Float64Array(Scene.MAX_TRANSFORM_PARAM_RECORD);
    public static UINT16_PARAM_BUFFER           = new Uint16Array(Scene.MAX_UINT16_PARAM_RECORD);
    public static SPRIT_FRAME_PARAM_RECORD      = 16;
    public static SPRIT_FRAME_PARAM_BUFFER      = new Uint16Array(Scene.SPRIT_FRAME_PARAM_RECORD);
    
    /**
     * For Ts Animation
     */
    _activeAnimatables: any[] = [];

    /**
     * Factory used to create the a collision coordinator.
     * @returns The collision coordinator
     */
    public static CollisionCoordinatorFactory(): ICollisionCoordinator {
        throw new Error("DefaultCollisionCoordinator");
    }

    metadata: any = {};
    clipPlane: Plane;
    clipPlane2: Plane;
    clipPlane3: Plane;
    clipPlane4: Plane;
    clipPlane5: Plane;
    clipPlane6: Plane;
    private _name: string;
    public ctx: GlContext;
    materials: Map<EntityID, IMaterialD3> = new Map();
    textures: Map<IFrameUpdate, IFrameUpdate> = new Map();
    meshes: Map<EntityID, IFrameUpdate> = new Map();
    transforms: Map<EntityID, INodeD3> = new Map();
    _rootNodeMap: Map<EntityID, INodeD3> = new Map();
    cacheTransformNodes: number[] = [];

    animationGroupsRS: Map<EntityID, IAnimationGroup3D> = new Map();
    useRightHandedSystem: boolean;

    defaultMaterial: IFrameUpdate;

    animationFPS: number = 60;

    cmdsForceDirty: Map<EntityID, IFrameUpdate>                 = new Map();

    cmdsLayerMask: Map<EntityID, ILayerMaskD3>                  = new Map();
    cmdsCollider: Map<EntityID, INodeColliderD3>                = new Map();

    cmdsMeshMaterial: Map<EntityID, IMeshD3>                    = new Map();
    cmdsMeshRenderQuene: Map<EntityID, IMeshRenderQueueD3>      = new Map();
    cmdsMeshRenderAlign: Map<EntityID, IMeshD3>                 = new Map();
    cmdsMeshBlend: Map<EntityID, IMeshD3>                       = new Map();
    cmdsMeshCullMode: Map<EntityID, IMeshD3>                    = new Map();
    cmdsMeshFrontFace: Map<EntityID, IMeshD3>                   = new Map();
    cmdsMeshDepthTest: Map<EntityID, IMeshD3>                   = new Map();
    cmdsMeshDepthWrite: Map<EntityID, IMeshD3>                  = new Map();
    cmdsMeshStencilTest: Map<EntityID, IMeshD3>                 = new Map();
    cmdsMeshInstanceAttr: Map<EntityID, IInstancedMeshD3>       = new Map();

    // cmdsInsRGB: Map<EntityID, IDirtyInstenceColor>          = new Map();
    // cmdsInsAlpha: Map<EntityID, IDirtyInstenceAlpha>        = new Map();
    // cmdsInsTilloff: Map<EntityID, IDirtyInstenceTilloff>    = new Map();
    
    cmdsSprite: Map<EntityID, ISprite3D>                        = new Map();
    cmdsParticlePlay: Map<EntityID, IParticleSystemD3>   = new Map();
    cmdsParticleSpeed: Map<EntityID, IParticleSystemD3>  = new Map();
    
    cmdsShadowBase: Map<EntityID, IShadowGenerator>         = new Map();
    cmdsShadowFrustum: Map<EntityID, IShadowGenerator>      = new Map();

    cmdsTargetAnimation: ITargetAnimation[]  = [];
    cmdsAnimationAttach: [number, number][]  = [];

    onBeforeStepObservable: Observable<Scene> = new Observable();
    onAfterStepObservable: Observable<Scene> = new Observable();
    onAfterAnimationsObservable: Observable<Scene> = new Observable();
    onBeforeRenderObservable: Observable<Scene> = new Observable();

    public miniScalingScalar = 0.0001;

    // auto
    private _engine: Engine;

    /** @hidden */
    public _active: boolean  = true;

    private _ambientColor: Color3;
    
    public get ambientColor() : Color3 {
        return this._ambientColor;
    }

    public set ambientColor(v : Color3) {
        this._ambientColor = v;
        p3d.p3d_scene_ambientcolor(Engine.commands(),  this.entity, this._ambientColor.r, this._ambientColor.g, this._ambientColor.b);
    }

    public set ambientIntensity(v : number) {
        p3d.p3d_scene_ambientintensity(Engine.commands(),  this.entity, v);
    }

    private _fogColor: Color3;
    public set fogColor(v : Color3) {
        this._fogColor = v;
        p3d.p3d_scene_fogcolor(Engine.commands(),  this.entity, this._fogColor.r, this._fogColor.g, this._fogColor.b);
    }
    public get fogColor() {
        return this._fogColor;
    }

    private _fogDirty: boolean = false;
    private _fogMode: number = 0;
    public set fogMode(v : number) {
        this._fogMode = v;
        this._fogDirty = true;
    }
    public get fogMode() {
        return this._fogMode;
    }

    private _fogEnabled: boolean;
    public set fogEnabled(v : boolean) {
        this._fogEnabled = v;
        this._fogDirty = true;
    }
    public get fogEnabled() {
        return this._fogEnabled;
    }

    private _fogDensity: number = 0.1;
    public set fogDensity(v : number) {
        this._fogDensity = v;
        this._fogDirty = true;
    }
    public get fogDensity() {
        return this._fogDensity;
    }
    
    private _fogStart: number = 1000;
    public set fogStart(v : number) {
        this._fogStart = v;
        this._fogDirty = true;
    }
    public get fogStart() {
        return this._fogStart;
    }
    
    private _fogEnd: number = 0;
    public set fogEnd(v : number) {
        this._fogEnd = v;
        this._fogDirty = true;
    }
    public get fogEnd() {
        return this._fogEnd;
    }

    private _fogDensityFalloff: number = 0;
    public set fogDensityFalloff(v : number) {
        this._fogDensityFalloff = v;
        this._fogDirty = true;
    }
    public get fogDensityFallof() {
        return this._fogDensityFalloff;
    }

    private _fogAltitudeBase: number = 0;
    public set fogAltitudeBase(v : number) {
        this._fogAltitudeBase = v;
        this._fogDirty = true;
    }
    public get fogAltitudeBase() {
        return this._fogAltitudeBase;
    }

    set shadowsEnabled(val: boolean) {
        // todo
    }
    set customLODSelector(val: Function) {
        // todo
    }

    constructor(name: string, boundingcullingmode: number = Scene.BOUNDING_CULLING_MODE_NONE, collidermode: number = Scene.BOUNDING_CULLING_MODE_NONE) {
        // let shadowcfg = p3d.PassCfg.create(p3d.RenderFormat.Rgba16Float, p3d.DepthStencilFormat.Depth32, false);
        // let opaque = p3d.PassCfg.create(p3d.RenderFormat.Rgba8Unorm, p3d.DepthStencilFormat.Depth32, false);
        // let transparent = p3d.PassCfg.create(p3d.RenderFormat.Rgba8Unorm, p3d.DepthStencilFormat.Depth32, true);
        let entity = p3d.p3d_scene(
            Engine.native(),
            Engine.commands(), 
            boundingcullingmode,
            collidermode,
            Scene.BOUNDING_CULLING_PRAMA
        );
        super(entity);
        this._name = name;
        this._engine = Engine.getEngine();
        (<any>Engine.getEngine())._scenes.set(this, this);

        // Engine.registBeforeRun(this._checkMaterials);
        // Engine.registBeforeRun(this._checkTextures);
        // Engine.registBeforeRun(this._checkTransforms);

        Engine.registAfterRun(this._checkAnimationGroups);
    }
    clientRect: { left: number; top: number; width: number; height: number; } = { left: 0, top: 0, width: 100, height: 100 };
    onPointerMove: (evt: EventData, pickInfo: IPickingInfo, type: PointerEventTypes) => void;
    onPointerDown: (evt: EventData, pickInfo: IPickingInfo, type: PointerEventTypes) => void;
    onPointerUp: (evt: EventData, pickInfo: Option<IPickingInfo>, type: PointerEventTypes) => void;
    onPointerPick: (evt: EventData, pickInfo: IPickingInfo) => void;

    public getEngine() {
        return Engine.getEngine();
    }

    _onFrameUpdate(): void {
        this.onBeforeRenderObservable.notifyObservers(this);
        if (this._fogDirty) {
            if (this._fogEnabled && this._fogMode != Scene.FOGMODE_NONE) {
                if (this._fogMode == Scene.FOGMODE_EXP) {
                    p3d.p3d_scene_fogparam(Engine.commands(),  this.entity, Scene.FOGMODE_EXP, this._fogDensity, 0., 0.);
                } else if (this._fogMode == Scene.FOGMODE_EXP2) {
                    p3d.p3d_scene_fogparam(Engine.commands(),  this.entity, Scene.FOGMODE_EXP2, this._fogDensity, 0., 0.);
                } else if (this._fogMode == Scene.FOGMODE_LINEAR) {
                    p3d.p3d_scene_fogparam(Engine.commands(),  this.entity, Scene.FOGMODE_LINEAR, this._fogStart, this._fogEnd, 0.);
                } else if (this._fogMode == Scene.FOGMODE_HEIGHT_BASE) {
                    p3d.p3d_scene_fogparam(Engine.commands(),  this.entity, Scene.FOGMODE_HEIGHT_BASE, this._fogAltitudeBase, this._fogDensity, this._fogDensityFalloff);
                }
            } else {
                p3d.p3d_scene_fogparam(Engine.commands(),  this.entity, Scene.FOGMODE_NONE, 0., 0., 0.);
            }
            this._fogDirty = false;
        }
    }

    private _activeCamera: Camera;
    public get activeCamera() {
        return this._activeCamera;
    }
    public set activeCamera(val: Camera) {
        this._activeCamera = val;
    }

    private _animationTimeScale: number = 1.;
    public set animationTimeScale(val: number) {
        this._animationTimeScale = val;
    }
    public get animationTimeScale() {
        return this._animationTimeScale;
    }

    private _deltaTime: number = 0;
    /**
     * @returns number of ms
     */
    public get deltaTime() {
        return this._deltaTime;
    }
    public useConstantAnimationDeltaTime: boolean = false;
    public physicsEnabled: boolean = false;
    // Deterministic lockstep
    private _timeAccumulator = 0;
    private _currentStepId = 0;
    private _currentInternalStep = 0;
    private _animationRatio: number = 0;

    private _animationTimeLast: number = 0;
    private __animationTime: number = 0;
    /**
     * @returns number of ms
     */
    public get _animationTime() {
        return this.__animationTime;
    }

    private _animationsEnabled: boolean = true;
    public set animationsEnabled(val: boolean) {
        this._animationsEnabled = val;
        p3d.p3d_scene_animation_enable(Engine.commands(),  this.entity, this._animationsEnabled);
    }
    public get animationsEnabled() {
        return this._animationsEnabled;
    }

    private anime() {
        if (this._engine.isDeterministicLockStep()) {
            var deltaTime = Math.max(Scene.MinDeltaTime, Math.min(this._engine.getDeltaTime(), Scene.MaxDeltaTime)) + this._timeAccumulator;
            var defaultFrameTime = this._engine.getTimeStep();
            var defaultFPS = (1000.0 / defaultFrameTime) / 1000.0;
            var stepsTaken = 0;
            var maxSubSteps = this._engine.getLockstepMaxSteps();
            var internalSteps = Math.floor(deltaTime / defaultFrameTime);
            internalSteps = Math.min(internalSteps, maxSubSteps);
            while (deltaTime > 0 && stepsTaken < internalSteps) {
                this.onBeforeStepObservable.notifyObservers(this);
                // Animations
                this._animationRatio = defaultFrameTime * defaultFPS;
                this.__animate();
                this.onAfterAnimationsObservable.notifyObservers(this);
                // Physics
                if (this.physicsEnabled) {
                    this._advancePhysicsEngineStep(defaultFrameTime);
                }
                this.onAfterStepObservable.notifyObservers(this);
                this._currentStepId++;
                stepsTaken++;
                deltaTime -= defaultFrameTime;
            }
            this._timeAccumulator = deltaTime < 0 ? 0 : deltaTime;
        }
        else {
            // Animations
            var deltaTime = this.useConstantAnimationDeltaTime ? 16 : Math.max(Scene.MinDeltaTime, Math.min(this._engine.getDeltaTime(), Scene.MaxDeltaTime));
            this._animationRatio = deltaTime * (60.0 / 1000.0);
            this.__animate();
            this.onAfterAnimationsObservable.notifyObservers(this);
            // Physics
            if (this.physicsEnabled) {
                this._advancePhysicsEngineStep(deltaTime);
            }
        }
    }

    private __animate() {
        this._deltaTime = this.useConstantAnimationDeltaTime ? 16.0 : this._engine.getDeltaTime() * this.animationTimeScale;
        if (this.animationsEnabled == false) {
            this._deltaTime = 0;
        }
        this.__animationTime += this._deltaTime;
        p3d.p3d_scene_time(Engine.commands(),  this.entity, this.__animationTime);
        this._animate();
    }

    /**
     * For Ts Animation
     */
    public _animate() {

    }
    getNodeByID(id: number) {
        return this.transforms.get(id);
    }
    public render() {
        if (this._active) {
            if (!this.isDisposed && this._activeCamera) {
                this._activeCamera.render(null, this.ctx);
                // if (this.renderPipeline && this.renderPipeline.renderers) {
                //     this.ctx?.canvas?.setBrush(this.renderPipeline.renderers[this.renderPipeline.renderers.length - 1].entity);
                // }
            }
        }
    }

    public run() {
        if (!this.isDisposed) {
            this._onFrameUpdate();
            this.cmdsForceDirty.forEach((node) => {
                node._onFrameUpdate();
            });
            this.cmdsForceDirty.clear();

            this.textures.forEach((item) => {
                item._onFrameUpdate();
            });
            this.textures.clear();

            this.anime();

            this._checkNodes();
            this._checkMesh();
            this._checkCamera();
            this._checkMaterials();
            this._checkShadow();
            this._checkAniamtion();
        }
        if (this._active) {
            
        }
    }

    private _checkMaterials = () => {
    }
    private _checkNodes = () => {
        let index = 0;
        
        index = 0;
        this.cmdsLayerMask.forEach(function(item) {
            p3d.p3d_layermask(Engine.commands(), item.entity, item.layerMask);
        });
        this.cmdsLayerMask.clear();
        
        index = 0;
        this.cmdsCollider.forEach((item) => {
            let isPickable = item.isPickable;
            let val = item.getCollider();
            if (val) {
                p3d.p3d_collider(Engine.commands(), this.entity, val.minimum.x, val.minimum.y, val.minimum.z, val.maximum.x, val.maximum.y, val.maximum.z, val.sphereTreshold);
            } else {

            }
        });
        this.cmdsCollider.clear();
    }
    private _checkMesh = () => {
        let index = 0;
        this.cmdsParticlePlay.forEach(function(item) {
            if (item.isPlaying()) {
                p3d.p3d_particle_system_start(Engine.commands(),  item.entity);
            } else {
                p3d.p3d_particle_system_stop(Engine.commands(),  item.entity);
            }
        });
        this.cmdsParticlePlay.clear();
        
        this.cmdsParticleSpeed.forEach(function(item) {
            p3d.p3d_particle_system_timescale(Engine.commands(),  item.entity,item.speedRatio);
        });
        this.cmdsParticleSpeed.clear();

        this.cmdsMeshBlend.forEach(function(item) {
            item.applyBlend();
        });
        this.cmdsMeshBlend.clear();
        
        this.cmdsMeshCullMode.forEach(function(item) {
            p3d.p3d_mesh_cull_mode(Engine.commands(), item.entity, item.cullMode, Constants.PASS_TAG_01);
            p3d.p3d_mesh_cull_mode(Engine.commands(), item.entity, item.cullMode, Constants.PASS_TAG_02);
            p3d.p3d_mesh_cull_mode(Engine.commands(), item.entity, item.cullMode, Constants.PASS_TAG_03);
            p3d.p3d_mesh_cull_mode(Engine.commands(), item.entity, item.cullMode, Constants.PASS_TAG_04);
            p3d.p3d_mesh_cull_mode(Engine.commands(), item.entity, item.cullMode, Constants.PASS_TAG_05);
            p3d.p3d_mesh_cull_mode(Engine.commands(), item.entity, item.cullMode, Constants.PASS_TAG_06);
            p3d.p3d_mesh_cull_mode(Engine.commands(), item.entity, item.cullMode, Constants.PASS_TAG_07);
            p3d.p3d_mesh_cull_mode(Engine.commands(), item.entity, item.cullMode, Constants.PASS_TAG_08);
        });
        this.cmdsMeshCullMode.clear();

        this.cmdsMeshDepthTest.forEach(function(item) {
            p3d.p3d_mesh_depth_compare(Engine.commands(), item.entity, item.depthTest, Constants.PASS_TAG_01);
            p3d.p3d_mesh_depth_compare(Engine.commands(), item.entity, item.depthTest, Constants.PASS_TAG_02);
            p3d.p3d_mesh_depth_compare(Engine.commands(), item.entity, item.depthTest, Constants.PASS_TAG_03);
            p3d.p3d_mesh_depth_compare(Engine.commands(), item.entity, item.depthTest, Constants.PASS_TAG_04);
            p3d.p3d_mesh_depth_compare(Engine.commands(), item.entity, item.depthTest, Constants.PASS_TAG_05);
            p3d.p3d_mesh_depth_compare(Engine.commands(), item.entity, item.depthTest, Constants.PASS_TAG_06);
            p3d.p3d_mesh_depth_compare(Engine.commands(), item.entity, item.depthTest, Constants.PASS_TAG_07);
            p3d.p3d_mesh_depth_compare(Engine.commands(), item.entity, item.depthTest, Constants.PASS_TAG_08);
        });
        this.cmdsMeshDepthTest.clear();
        
        this.cmdsMeshDepthWrite.forEach(function(item) {
            p3d.p3d_mesh_depth_write(Engine.commands(), item.entity, item.depthWrite, Constants.PASS_TAG_01);
            p3d.p3d_mesh_depth_write(Engine.commands(), item.entity, item.depthWrite, Constants.PASS_TAG_02);
            p3d.p3d_mesh_depth_write(Engine.commands(), item.entity, item.depthWrite, Constants.PASS_TAG_03);
            p3d.p3d_mesh_depth_write(Engine.commands(), item.entity, item.depthWrite, Constants.PASS_TAG_04);
            p3d.p3d_mesh_depth_write(Engine.commands(), item.entity, item.depthWrite, Constants.PASS_TAG_05);
            p3d.p3d_mesh_depth_write(Engine.commands(), item.entity, item.depthWrite, Constants.PASS_TAG_06);
            p3d.p3d_mesh_depth_write(Engine.commands(), item.entity, item.depthWrite, Constants.PASS_TAG_07);
            p3d.p3d_mesh_depth_write(Engine.commands(), item.entity, item.depthWrite, Constants.PASS_TAG_08);
        });
        this.cmdsMeshDepthWrite.clear();
        
        this.cmdsMeshStencilTest.forEach(function(item) {
            //
        });
        this.cmdsMeshStencilTest.clear();
    
        this.cmdsMeshFrontFace.forEach(function(item) {
            p3d.p3d_mesh_frontface(Engine.commands(), item.entity, item.frontFace, Constants.PASS_TAG_01);
            p3d.p3d_mesh_frontface(Engine.commands(), item.entity, item.frontFace, Constants.PASS_TAG_02);
            p3d.p3d_mesh_frontface(Engine.commands(), item.entity, item.frontFace, Constants.PASS_TAG_03);
            p3d.p3d_mesh_frontface(Engine.commands(), item.entity, item.frontFace, Constants.PASS_TAG_04);
            p3d.p3d_mesh_frontface(Engine.commands(), item.entity, item.frontFace, Constants.PASS_TAG_05);
            p3d.p3d_mesh_frontface(Engine.commands(), item.entity, item.frontFace, Constants.PASS_TAG_06);
            p3d.p3d_mesh_frontface(Engine.commands(), item.entity, item.frontFace, Constants.PASS_TAG_07);
            p3d.p3d_mesh_frontface(Engine.commands(), item.entity, item.frontFace, Constants.PASS_TAG_08);
        });
        this.cmdsMeshFrontFace.clear();
        
        this.cmdsMeshRenderAlign.forEach(function(item) {
            p3d.p3d_mesh_render_alignment(Engine.commands(), item.entity, item.renderAlignment);
        });
        this.cmdsMeshRenderAlign.clear();
        
        this.cmdsMeshMaterial.forEach((item) => {
            // item.dirtyMeshMaterial();
            // item._dirtyMesh ^= Entity.DIRTY_MESH_MATERIAL;
        });
        this.cmdsMeshMaterial.clear();

        index = 0;
        this.cmdsMeshRenderQuene.forEach(function(item) {
            // item.dirtyMeshRenderQuene();
            // item._dirtyMesh ^= Entity.DIRTY_MESH_RENDER_QUENE;
            Scene.TRANSFORM_PARAM_BUFFER[index * Scene.TRANSFORM_PARAM_v2_SIZE + 0] = item.entity;
            Scene.TRANSFORM_PARAM_BUFFER[index * Scene.TRANSFORM_PARAM_v2_SIZE + 1] = item.renderingGroupId;
            Scene.TRANSFORM_PARAM_BUFFER[index * Scene.TRANSFORM_PARAM_v2_SIZE + 2] = item.alphaIndex;
            index ++;
            if (index == Scene.MAX_TRANSFORM_PARAM_RECORD_V2) {
                p3d.p3d_mesh_render_queue_arr(Engine.commands(),  Scene.TRANSFORM_PARAM_BUFFER, index * Scene.TRANSFORM_PARAM_v2_SIZE);
                index = 0;
            }
        });
        index > 0 && p3d.p3d_mesh_render_queue_arr(Engine.commands(),  Scene.TRANSFORM_PARAM_BUFFER, index * Scene.TRANSFORM_PARAM_v2_SIZE);
        this.cmdsMeshRenderQuene.clear();

        this.cmdsMeshInstanceAttr.forEach(function(item) {
            item._applyInstanceAttribute();
        });
        this.cmdsMeshInstanceAttr.clear();

        this.cmdsSprite.forEach(function(item) {
            item._applySpriteData(Scene.SPRIT_FRAME_PARAM_BUFFER, 0);
        });
        this.cmdsSprite.clear();
    }
    private _checkCamera = () => {

    }
    private _checkShadow = () => {
        this.cmdsShadowBase.forEach((item, id) => {
            p3d.p3d_shadow_base_param(Engine.commands(),  id, item.bias, item.normalBias, item.depthScale);
        });
        this.cmdsShadowBase.clear();

        this.cmdsShadowFrustum.forEach((item, id) => {
            p3d.p3d_shadow_frustum(Engine.commands(),  id, item.frustumSize, item.minZ, item.maxZ);
        });
        this.cmdsShadowFrustum.clear();
    }
    private _checkAniamtion = () => {
        this.cmdsTargetAnimation.forEach(item => {
            // if ((<ITargetAnimationProperty>item).target) {
            //     p3d.p3d_property_target_animation(Engine.native(), Engine.commands(),  item.curve, (<ITargetAnimationProperty>item).property, 0, item.group, (<ITargetAnimationProperty>item).target);
            // } else if ((<ITargetAnimationUniform>item).uniformProperty) {
            //     p3d.p3d_uniform_target_animation(
            //         Engine.commands(), (<ITargetAnimationUniform>item).uniformTarget, item.group,
            //         getAtom((<ITargetAnimationUniform>item).uniformProperty), item.curve
            //     );
            // } else if ((<ITargetAnimationAttribute>item).attributeProperty) {
            //     p3d.p3d_uniform_target_animation(
            //         Engine.commands(), (<ITargetAnimationAttribute>item).attributeTarget, item.group,
            //         getAtom((<ITargetAnimationAttribute>item).attributeProperty), item.curve
            //     );
            // }
        });
        this.cmdsTargetAnimation.length = 0;
    }
    private _checkAnimationGroups = () => {
        let engine = Engine.getEngine();
        engine.animationStartEvents.forEach((v, k) => {
            let group = this.animationGroupsRS.get(k);
            if (group) {
                group._onStart(v);
                engine.animationStartEvents.delete(k);
            }
        });
        engine.animationLoopEvents.forEach((v, k) => {
            let group = this.animationGroupsRS.get(k);
            if (group) {
                group._onLoop(v);
                engine.animationLoopEvents.delete(k);
            }
        });
        engine.animationFrameEvents.forEach((v, k) => {
            let group = this.animationGroupsRS.get(k);
            if (group) {
                group._onFrame(v);
                engine.animationFrameEvents.delete(k);
            }
        });
        engine.animationEndEvents.forEach((v, k) => {
            let group = this.animationGroupsRS.get(k);
            if (group) {
                group._onEnd(v);
                engine.animationEndEvents.delete(k);
            }
        });
    }

    get isDisposed() {
        return this._disposed;
    }

    dispose(): void {
        
        (<any>Engine.getEngine())._scenes.delete(this);

        // super.dispose();
        if (this.isDisposed) {
            return;
        }

        this.onDisposeObservable.notifyObservers(this);

        this.cacheTransformNodes.forEach(function(item) {
            p3d.p3d_dispose(Engine.commands(), item);
        });
        this.cacheTransformNodes.length = 0;

        this.transforms.forEach((item) => {
            item.dispose();
        });

        this.materials.forEach((mat) => {
            mat.dispose();
        });
        this.materials = undefined;

        this.textures.forEach((tex) => {
            // tex.dispose();
        });
        this.textures = undefined;
        
        // Engine.unregistBeforeRun(this._checkMaterials);
        // Engine.unregistBeforeRun(this._checkTextures);
        // Engine.unregistBeforeRun(this._checkTransforms);

        this.onBeforeRenderObservable.clear();
        this.onDisposeObservable.clear();
        this.onKeyboardObservable.clear();
        this.onBeforeStepObservable.clear();
        this.onAfterStepObservable.clear();

        this._disposed = true;

        // console.error("Not Impl Scene Dispose !!!");
        Engine.disposeScene.add(this.entity);
        // p3d.p3d_scene_dispose(Engine.commands(),  this.entity);
        (<any>this)._entity = 0;
    }

    public getAnimationRatio() {
        return this._animationRatio;
    }

    /** @hidden */
    public _advancePhysicsEngineStep = function (step) {
        // Do nothing. Code will be replaced if physics engine component is referenced
    };

    public set environmentBRDFTexture(url: string) {
        p3d.p3d_scene_brdf_texture(Engine.commands(),  this.entity, ExtensionTools.getAtom(url), false);
    }

    public set environmentTexture(url: string) {
        p3d.p3d_scene_env_texture(Engine.commands(),  this.entity, ExtensionTools.getAtom(url), false);
    }

    public renderPipeline: IRenderPipelineD3;

    public queryPassTag = (renderQueue: number): Constants.PassTag => {
        if (this.renderPipeline) {
            return this.renderPipeline.queryPass(renderQueue);
        } else {
            if (renderQueue < 2400) {
                return Constants.PASS_TAG_03;
            } else {
                return Constants.PASS_TAG_06;
            }
        }
    }

    boundingBoxDebug(display: boolean, pass: number): void {
        if (this._disposed) { return; }
        p3d.p3d_scene_boundingbox(Engine.commands(), this._entity, display, pass);
    }

    displayScreenRect: { x: number, y: number, w: number, h: number, width: number, height: number } = { x: 0, y: 0, w: 100, h: 100, width: 100, height: 100  }
    /**
     * 
     * @param x 
     * @param y 
     * @param world 未使用 - x,y 为屏幕 0-1 坐标系坐标
     * @param camera 
     * @param cameraViewSpace 
     * @returns 
     */
    createPickingRay(x: number, y: number, world: Matrix, camera: Nullable<Camera>, cameraViewSpace = false): Ray {
        if (!camera) {
            if (!this.activeCamera) {
                return undefined;
            }
    
            camera = this.activeCamera;
        }

        let info = Engine.GetRenderContent();
        x -= info.styleLeft;
        y -= info.styleTop;
        x /= (info.styleSize.width - info.styleRight - info.styleLeft);
        y /= (info.styleSize.height - info.styleBottom - info.styleTop);
        
        if (camera.createPickRay(x, y, Ray._Cache) == true) {
            let origin      = new Vector3(Ray._Cache[0], Ray._Cache[1], Ray._Cache[2]);
            let direction   = new Vector3(Ray._Cache[3], Ray._Cache[4], Ray._Cache[5]);
            return new Ray(origin, direction);
        }
    
        return undefined;
    }

    public pickWithColliderInfo(x: number, y: number, camera?: Camera) {
        let now = Date.now();
        // Dummy info if picking as not been imported
        const pi = new PickingInfo();

        if (!camera) {
            if (!this.activeCamera) {
                return undefined;
            }
    
            camera = this.activeCamera;
        }

        pi.hit = false;
        if (p3d.p3d_scene_pick(Engine.native(), Engine.param(), this.entity, (<any>camera).camera.entity, x, y, false, Scene._PickResultCache)) {
            pi.hit = true;
            let entity      = Scene._PickResultCache[ 0];
            let bybounding  = Scene._PickResultCache[ 1] > 0;
            let minx        = Scene._PickResultCache[ 2];
            let miny        = Scene._PickResultCache[ 3];
            let minz        = Scene._PickResultCache[ 4];
            let maxx        = Scene._PickResultCache[ 5];
            let maxy        = Scene._PickResultCache[ 6];
            let maxz        = Scene._PickResultCache[ 7];
            let detail      = Scene._PickResultCache[ 8] > 0;
            let detailx     = Scene._PickResultCache[ 9];
            let detaily     = Scene._PickResultCache[10];
            let detailz     = Scene._PickResultCache[11];
            pi.pickedMesh   = this.transforms.get(entity);
        }
        pi._pickingUnavailable = true;

        console.warn("Pick: " + (Date.now() - now));
        return pi;
    }

    /** @hidden */
    public _inputManager = new InputManager(this);

    // //#region 
    // /**
    //  * @deprecated Camera.shadowsEnabled
    //  */
    // shadowsEnabled: boolean;
    // /**
    //  * @deprecated Camera.XXX
    //  */
    // autoClear: boolean;
    // /**
    //  * @deprecated Camera.XXX
    //  */
    // autoClearDepthAndStencil: boolean;

    /**
     * @deprecated 
     */
    preventDefaultOnPointerDown: boolean;
    /**
     * @deprecated 
     */
    preventDefaultOnPointerUp: boolean;
    /**
     * @deprecated 
     */
    blockMaterialDirtyMechanism: boolean;
    /**
     * @deprecated 
     */
    onBeforeDrawPhaseObservable: Observable<void>;
    
    /**
     * Observable event triggered each time an keyboard event is received from the hosting window
     */
    onKeyboardObservable = new Observable<KeyboardInfo>();
    //#endregion
}
