
/**
 *
 */
import * as BABYLON from './context';
import { Option, Nullable } from '../interface/index';
import { AnimConstants, IAnimationScene } from '../animation_ts/interface';
import { Animatable, Animation, AnimationGroup, RuntimeAnimation } from '../animation_ts/index';
import { Matrix, Quaternion, TmpVectors, Vector3 } from './context';
import { SmartArrayNoDuplicate } from '../render/misc/smartArray';

import { Engine } from './engine';
import { ISpriteAtlas, ISpriteManager, ISpriteScene } from '../interface/sprite';
import { SpriteManager } from '../sprite/sprite_manager';
import { ExtensionTools, ResTab } from '../interface/index';
import { GlobalAtlasManager } from '../sprite/sprite_atlas';

/**
 * 封装项目层使用的 场景命中信息
 * * 底层处理过命中检测 - 只通过包围盒检测
 */
// tslint:disable-next-line:interface-name
export interface IRayCastInfo extends BABYLON.PickingInfo {
    hit: boolean;
    pickedPoint: Option<BABYLON.Vector3>;
    /**
     * 命中的目标mesh
     */
    target: BABYLON.INodeD3;
    rayID: number;
}

// 重载BABYLON.Scene， 释放资源
export class Scene extends BABYLON.Scene implements ISpriteScene, IAnimationScene, BABYLON.ITreeContext, BABYLON.ITreeNode {
    public static IDCounter: number = 0;
    public resTab: ResTab;
    public piID: number;

    onAfterRenderObservable: BABYLON.Observable<Scene> = new BABYLON.Observable();

    attachControl(down: boolean, up: boolean, move: boolean) {

    }
    /**
     * @deprecated
     */
    pi_autoApplyLayerMask: boolean;
    //保证循环粒子中一定有粒子存活
    /**
     * @deprecated
     */
    loopParticleMustHasAlive: boolean;
    /**
     * @deprecated
     */
    skeletonMeshHasRotate: boolean;

    updateTransformMatrix(force: boolean) {

    }
    getTransformMatrix(): BABYLON.Matrix {
        return this.activeCamera.getTransformationMatrix();
    }


    constructor(engine: BABYLON.Engine, opt?: BABYLON.SceneOptions) {
        super(`${Scene.IDCounter++}`);
        this.resTab = ExtensionTools.CreateResTab();
        this.piID = Scene.IDCounter++;
        this.onBeforeRenderObservable.add(() => {
            this._updateSprites()
        });
        ExtensionTools.setPermanent(this.cleanSpiteManagerNoSprite);
    }
    getChildren(predicate?: (node: BABYLON.ITreeNode) => boolean, directDescendantsOnly?: boolean): BABYLON.ITreeNode[] {
        if (this._isDisposed) {
            return [];
        } else {
            return this.rootNodes;
        }
    }
    get name() {
        return this.uniqueId;
    }
    get rootNodes(): BABYLON.ITreeNode[] {
        let result = [];
        for (let node of this._rootNodeMap.values()) {
            result.push(node)
        }
        return result;
    }

    //#region Sprite
    public spriteUpdateTime = 0;
    private _spriteManagers: Map<string, ISpriteManager> = new Map();
    registSpritAtlas(cfg: ISpriteAtlas, maxSprite?: number, material?: BABYLON.Material, pass?: BABYLON.Constants.PassTag): void {
        let atlasName = cfg.image;
        if (this._spriteManagers.get(atlasName) == undefined) {
            let atlas = GlobalAtlasManager.getAtlas(cfg);
            let item = new SpriteManager(this, atlas, maxSprite, material, pass);
            this._spriteManagers.set(atlasName, item);
        }
    }
    getSpritManager(atlas: string): ISpriteManager {
        atlas = GlobalAtlasManager.queryRealAtlasName(atlas);
        let result = this._spriteManagers.get(atlas);
        if (result == undefined) {
            let cfg = GlobalAtlasManager.getAtlas(atlas);
            if (cfg) {
                this.registSpritAtlas(cfg.atlas);
            }
            result = this._spriteManagers.get(atlas);
        }
        return result;
    }
    delSpritManager(managr: ISpriteManager): void {
        this._spriteManagers.delete(managr.atlas.image);
    }
    /**
     * 清除没有Sprite的SpriteManager - 会清除内部 Mesh,Texture
     */
    public cleanSpiteManagerNoSpriteMS: number = 1000;
    private _lastCleanSpiteManagerNoSpriteMS: number = Date.now();
    cleanSpiteManagerNoSprite = () => {
        if (this.getEngine().time - this._lastCleanSpiteManagerNoSpriteMS > this.cleanSpiteManagerNoSpriteMS) {
            this._spriteManagers.forEach((item) => {
                if ((<SpriteManager>item).spriteCount() == 0) {
                    item.dispose();
                }
            });
            this._lastCleanSpiteManagerNoSpriteMS = this.getEngine().time;
        }
    }
    private _updateSprites = () => {
        let time = (window.performance || Date).now();
        this._spriteManagers.forEach((item) => {
            (<SpriteManager>item)._onFrameUpdate();
        });
        this.spriteUpdateTime = (window.performance || Date).now() - time;
    }
    //#endregion
    _activeRequests: any[] = [];
    importedMeshesFiles: any[] = [];
    geometriesByUniqueId: { [index: string]: string } = {}

    /**
     * 3D 场景点击命中结果
     * @return
     * * .pickedMesh.link: 命中的mesh
     */
    public pick3D(x: number, y: number, camera?: BABYLON.Camera) {
        const pickRes: BABYLON.PickingInfo = this.pickWithColliderInfo(
            x, 
            y,
            camera
        );
        (<IRayCastInfo>pickRes).target = <any>pickRes.pickedMesh;
        (<IRayCastInfo>pickRes).rayID  = (<IRayCastInfo>pickRes).target ? (<any>(<IRayCastInfo>pickRes).target)?.rayID : null;

        return (<IRayCastInfo>pickRes);
    }
    dispose(): void {
        this.resTab.clear();
        ExtensionTools.clearPermanent(this.cleanSpiteManagerNoSprite);
        super.dispose();
    }

    //#region Animation TS
    /** @hidden */
    public _registeredForLateAnimationBindings = new SmartArrayNoDuplicate<any>(256);
    public animationGroupsTS: AnimationGroup[] = [];
    addAnimationGroup(group: AnimationGroup) {
        this.animationGroupsTS.push(group);
    }
    _animate(): void {
        if (!this.animationsEnabled) {
            return;
        }
    
        // // Getting time
        // // var now = PrecisionDate.Now;
        // if (!this._animationTimeLast) {
        //     if (this._pendingData.length > 0) {
        //         return;
        //     }
        //     this._animationTimeLast = now;
        // }
    
        // this.deltaTime = this.useConstantAnimationDeltaTime ? 16.0 : (now - this._animationTimeLast) * this.animationTimeScale;
        // this._animationTimeLast = now;
        // this._animationTime += this.deltaTime;
    
        const animatables = this._activeAnimatables;
        if (animatables.length === 0) {
            return;
        }
        const animationTime = this._animationTime;
    
        for (let index = 0; index < animatables.length; index++) {
            let animatable = animatables[index];
    
            if (!animatable._animate(animationTime) && animatable.disposeOnEnd) {
                index--; // Array was updated
            }
        }

        // Late animation bindings
        this._processLateAnimationBindings();
    }
    beginWeightedAnimation(target: any, from: number, to: number, weight = 1.0, loop?: boolean, speedRatio: number = 1.0,
        onAnimationEnd?: () => void, animatable?: Animatable, targetMask?: (target: any) => boolean, onAnimationLoop?: () => void, isAdditive = false): Animatable {
    
        let returnedAnimatable = this.beginAnimation(target, from, to, loop, speedRatio, onAnimationEnd, animatable, false, targetMask, onAnimationLoop, isAdditive);
        returnedAnimatable.weight = weight;
    
        return returnedAnimatable;
    }
    beginAnimation(target: any, from: number, to: number, loop?: boolean, speedRatio: number = 1.0,
        onAnimationEnd?: () => void, animatable?: Animatable, stopCurrent = true,
        targetMask?: (target: any) => boolean, onAnimationLoop?: () => void, isAdditive = false): Animatable {
    
        if (from > to && speedRatio > 0) {
            speedRatio *= -1;
        }
    
        if (stopCurrent) {
            this.stopAnimation(target, undefined, targetMask);
        }
    
        if (!animatable) {
            animatable = new Animatable(this, target, from, to, loop, speedRatio, onAnimationEnd, undefined, onAnimationLoop, isAdditive);
        }
    
        const shouldRunTargetAnimations = targetMask ? targetMask(target) : true;
        // Local animations
        if (target.animations && shouldRunTargetAnimations) {
            animatable.appendAnimations(target, target.animations);
        }
    
        // Children animations
        if (target.getAnimatables) {
            var animatables = target.getAnimatables();
            for (var index = 0; index < animatables.length; index++) {
                this.beginAnimation(animatables[index], from, to, loop, speedRatio, onAnimationEnd, animatable, stopCurrent, targetMask, onAnimationLoop);
            }
        }
    
        animatable.reset();
    
        return animatable;
    }
    beginHierarchyAnimation(target: any, directDescendantsOnly: boolean, from: number, to: number, loop?: boolean, speedRatio: number = 1.0,
        onAnimationEnd?: () => void, animatable?: Animatable, stopCurrent = true,
        targetMask?: (target: any) => boolean, onAnimationLoop?: () => void, isAdditive = false): Animatable[] {
    
        let children = target.getDescendants(directDescendantsOnly);
    
        let result = [];
        result.push(this.beginAnimation(target, from, to, loop, speedRatio, onAnimationEnd, animatable, stopCurrent, targetMask, undefined, isAdditive));
        for (var child of children) {
            result.push(this.beginAnimation(child, from, to, loop, speedRatio, onAnimationEnd, animatable, stopCurrent, targetMask, undefined, isAdditive));
        }
    
        return result;
    }
    beginDirectAnimation(target: any, animations: Animation[], from: number, to: number, loop?: boolean, speedRatio?: number, onAnimationEnd?: () => void, onAnimationLoop?: () => void, isAdditive = false): Animatable {
        if (speedRatio === undefined) {
            speedRatio = 1.0;
        }
    
        if (from > to && speedRatio > 0) {
            speedRatio *= -1;
        }
    
        var animatable = new Animatable(this, target, from, to, loop, speedRatio, onAnimationEnd, animations, onAnimationLoop, isAdditive);
    
        return animatable;
    }
    beginDirectHierarchyAnimation(target: BABYLON.Node, directDescendantsOnly: boolean, animations: Animation[], from: number, to: number, loop?: boolean, speedRatio?: number, onAnimationEnd?: () => void, onAnimationLoop?: () => void, isAdditive = false): Animatable[] {
        let children = target.getDescendants(directDescendantsOnly);
    
        let result = [];
        result.push(this.beginDirectAnimation(target, animations, from, to, loop, speedRatio, onAnimationEnd, onAnimationLoop, isAdditive));
        for (var child of children) {
            result.push(this.beginDirectAnimation(child, animations, from, to, loop, speedRatio, onAnimationEnd, onAnimationLoop, isAdditive));
        }
    
        return result;
    }
    getAnimatableByTarget(target: any): Nullable<Animatable> {
        for (var index = 0; index < this._activeAnimatables.length; index++) {
            if (this._activeAnimatables[index].target === target) {
                return this._activeAnimatables[index];
            }
        }
    
        return null;
    }
    getAllAnimatablesByTarget(target: any): Array<Animatable> {
        let result = [];
        for (var index = 0; index < this._activeAnimatables.length; index++) {
            if (this._activeAnimatables[index].target === target) {
                result.push(this._activeAnimatables[index]);
            }
        }
    
        return result;
    }
    stopAnimation(target: any, animationName?: string, targetMask?: (target: any) => boolean): void {
        var animatables = this.getAllAnimatablesByTarget(target);
    
        for (var animatable of animatables) {
            animatable.stop(animationName, targetMask);
        }
    }
    stopAllAnimations(): void {
        if (this._activeAnimatables) {
            for (let i = 0; i < this._activeAnimatables.length; i++) {
                this._activeAnimatables[i].stop();
            }
            this._activeAnimatables = [];
        }
    
        for (var group of this.animationGroupsTS) {
            group.stop();
        }
    }
    _registerTargetForLateAnimationBinding(runtimeAnimation: RuntimeAnimation, originalValue: any): void {
        let target = runtimeAnimation.target;
        this._registeredForLateAnimationBindings.pushNoDuplicate(target);
    
        if (!target._lateAnimationHolders) {
            target._lateAnimationHolders = {};
        }
    
        if (!target._lateAnimationHolders[runtimeAnimation.targetPath]) {
            target._lateAnimationHolders[runtimeAnimation.targetPath] = {
                totalWeight: 0,
                totalAdditiveWeight: 0,
                animations: [],
                additiveAnimations: [],
                originalValue: originalValue
            };
        }
    
        if (runtimeAnimation.isAdditive) {
            target._lateAnimationHolders[runtimeAnimation.targetPath].additiveAnimations.push(runtimeAnimation);
            target._lateAnimationHolders[runtimeAnimation.targetPath].totalAdditiveWeight += runtimeAnimation.weight;
        } else {
            target._lateAnimationHolders[runtimeAnimation.targetPath].animations.push(runtimeAnimation);
            target._lateAnimationHolders[runtimeAnimation.targetPath].totalWeight += runtimeAnimation.weight;
        }
    }
    _processLateAnimationBindingsForMatrices(holder: {
        totalWeight: number,
        totalAdditiveWeight: number,
        animations: RuntimeAnimation[],
        additiveAnimations: RuntimeAnimation[],
        originalValue: Matrix
    }): any {
        if (holder.totalWeight === 0 && holder.totalAdditiveWeight === 0) {
            return holder.originalValue;
        }
    
        let normalizer = 1.0;
        let finalPosition = TmpVectors.Vector3[0];
        let finalScaling = TmpVectors.Vector3[1];
        let finalQuaternion = TmpVectors.Quaternion[0];
        let startIndex = 0;
        let originalAnimation = holder.animations[0];
        let originalValue = holder.originalValue;
    
        var scale = 1;
        let skipOverride = false;
        if (holder.totalWeight < 1.0) {
            // We need to mix the original value in
            scale = 1.0 - holder.totalWeight;
            originalValue.decompose(finalScaling, finalQuaternion, finalPosition);
        } else {
            startIndex = 1;
            // We need to normalize the weights
            normalizer = holder.totalWeight;
            scale = originalAnimation.weight / normalizer;
            if (scale == 1) {
                if (holder.totalAdditiveWeight) {
                    skipOverride = true;
                } else {
                    return originalAnimation.currentValue;
                }
            }
    
            originalAnimation.currentValue.decompose(finalScaling, finalQuaternion, finalPosition);
        }
    
        // Add up the override animations
        if (!skipOverride) {
            finalScaling.scaleInPlace(scale);
            finalPosition.scaleInPlace(scale);
            finalQuaternion.scaleInPlace(scale);
    
            for (var animIndex = startIndex; animIndex < holder.animations.length; animIndex++) {
                var runtimeAnimation = holder.animations[animIndex];
                if (runtimeAnimation.weight === 0) {
                    continue;
                }
    
                var scale = runtimeAnimation.weight / normalizer;
                let currentPosition = TmpVectors.Vector3[2];
                let currentScaling = TmpVectors.Vector3[3];
                let currentQuaternion = TmpVectors.Quaternion[1];
    
                runtimeAnimation.currentValue.decompose(currentScaling, currentQuaternion, currentPosition);
                currentScaling.scaleAndAddToRef(scale, finalScaling);
                currentQuaternion.scaleAndAddToRef(scale, finalQuaternion);
                currentPosition.scaleAndAddToRef(scale, finalPosition);
            }
        }
    
        // Add up the additive animations
        for (let animIndex = 0; animIndex < holder.additiveAnimations.length; animIndex++) {
            var runtimeAnimation = holder.additiveAnimations[animIndex];
            if (runtimeAnimation.weight === 0) {
                continue;
            }
    
            let currentPosition = TmpVectors.Vector3[2];
            let currentScaling = TmpVectors.Vector3[3];
            let currentQuaternion = TmpVectors.Quaternion[1];
    
            runtimeAnimation.currentValue.decompose(currentScaling, currentQuaternion, currentPosition);
            currentScaling.multiplyToRef(finalScaling, currentScaling);
            Vector3.LerpToRef(finalScaling, currentScaling, runtimeAnimation.weight, finalScaling);
            finalQuaternion.multiplyToRef(currentQuaternion, currentQuaternion);
            Quaternion.SlerpToRef(finalQuaternion, currentQuaternion, runtimeAnimation.weight, finalQuaternion);
            currentPosition.scaleAndAddToRef(runtimeAnimation.weight, finalPosition);
    
        }
    
        let workValue = originalAnimation ? originalAnimation._animationState.workValue : TmpVectors.Matrix[0].clone();
        Matrix.ComposeToRef(finalScaling, finalQuaternion, finalPosition, workValue);
        return workValue;
    }
    _processLateAnimationBindingsForQuaternions(holder: {
        totalWeight: number,
        totalAdditiveWeight: number,
        animations: RuntimeAnimation[],
        additiveAnimations: RuntimeAnimation[],
        originalValue: Quaternion
    }, refQuaternion: Quaternion): Quaternion {
        if (holder.totalWeight === 0 && holder.totalAdditiveWeight === 0) {
            return refQuaternion;
        }
    
        let originalAnimation = holder.animations[0];
        let originalValue = holder.originalValue;
        let cumulativeQuaternion = refQuaternion;
    
        if (holder.totalWeight === 0 && holder.totalAdditiveWeight > 0) {
            cumulativeQuaternion.copyFrom(originalValue);
        } else if (holder.animations.length === 1) {
            Quaternion.SlerpToRef(originalValue, originalAnimation.currentValue, Math.min(1.0, holder.totalWeight), cumulativeQuaternion);
    
            if (holder.totalAdditiveWeight === 0) {
                return cumulativeQuaternion;
            }
        } else if (holder.animations.length > 1) {
            // Add up the override animations
            let normalizer = 1.0;
            let quaternions: Array<Quaternion>;
            let weights: Array<number>;
    
            if (holder.totalWeight < 1.0) {
                let scale = 1.0 - holder.totalWeight;
    
                quaternions = [];
                weights = [];
    
                quaternions.push(originalValue);
                weights.push(scale);
            } else {
                if (holder.animations.length === 2) { // Slerp as soon as we can
                    Quaternion.SlerpToRef(holder.animations[0].currentValue, holder.animations[1].currentValue, holder.animations[1].weight / holder.totalWeight, refQuaternion);
    
                    if (holder.totalAdditiveWeight === 0) {
                        return refQuaternion;
                    }
                }
    
                quaternions = [];
                weights = [];
                normalizer = holder.totalWeight;
            }
    
            for (var animIndex = 0; animIndex < holder.animations.length; animIndex++) {
                let runtimeAnimation = holder.animations[animIndex];
                quaternions.push(runtimeAnimation.currentValue);
                weights.push(runtimeAnimation.weight / normalizer);
            }
    
            // https://gamedev.stackexchange.com/questions/62354/method-for-interpolation-between-3-quaternions
    
            let cumulativeAmount = 0;
            for (var index = 0; index < quaternions.length;) {
                if (!index) {
                    Quaternion.SlerpToRef(quaternions[index], quaternions[index + 1], weights[index + 1] / (weights[index] + weights[index + 1]), refQuaternion);
                    cumulativeQuaternion = refQuaternion;
                    cumulativeAmount = weights[index] + weights[index + 1];
                    index += 2;
                    continue;
                }
                cumulativeAmount += weights[index];
                Quaternion.SlerpToRef(cumulativeQuaternion, quaternions[index], weights[index] / cumulativeAmount, cumulativeQuaternion);
                index++;
            }
        }
    
        // Add up the additive animations
        for (let animIndex = 0; animIndex < holder.additiveAnimations.length; animIndex++) {
            let runtimeAnimation = holder.additiveAnimations[animIndex];
            if (runtimeAnimation.weight === 0) {
                continue;
            }
    
            cumulativeQuaternion.multiplyToRef(runtimeAnimation.currentValue, TmpVectors.Quaternion[0]);
            Quaternion.SlerpToRef(cumulativeQuaternion, TmpVectors.Quaternion[0], runtimeAnimation.weight, cumulativeQuaternion);
        }
    
        return cumulativeQuaternion!;
    }
    _processLateAnimationBindings(): void {
        if (!this._registeredForLateAnimationBindings.length) {
            return;
        }
        for (var index = 0; index < this._registeredForLateAnimationBindings.length; index++) {
            var target = this._registeredForLateAnimationBindings.data[index];
    
            for (var path in target._lateAnimationHolders) {
                var holder = target._lateAnimationHolders[path];
                let originalAnimation: RuntimeAnimation = holder.animations[0];
                let originalValue = holder.originalValue;
    
                let matrixDecomposeMode = AnimConstants.AllowMatrixDecomposeForInterpolation && originalValue.m; // ie. data is matrix
    
                let finalValue: any = target[path];
                if (matrixDecomposeMode) {
                    finalValue = this._processLateAnimationBindingsForMatrices(holder);
                } else {
                    let quaternionMode = originalValue.w !== undefined;
                    if (quaternionMode) {
                        finalValue = this._processLateAnimationBindingsForQuaternions(holder, finalValue || Quaternion.Identity());
                    } else {
    
                        let startIndex = 0;
                        let normalizer = 1.0;
    
                        if (holder.totalWeight < 1.0) {
                            // We need to mix the original value in
                            if (originalAnimation && originalValue.scale) {
                                finalValue = originalValue.scale(1.0 - holder.totalWeight);
                            } else if (originalAnimation) {
                                finalValue = originalValue * (1.0 - holder.totalWeight);
                            } else if (originalValue.clone) {
                                finalValue = originalValue.clone();
                            } else {
                                finalValue = originalValue;
                            }
                        } else if (originalAnimation) {
                            // We need to normalize the weights
                            normalizer = holder.totalWeight;
                            let scale = originalAnimation.weight / normalizer;
                            if (scale !== 1) {
                                if (originalAnimation.currentValue.scale) {
                                    finalValue = originalAnimation.currentValue.scale(scale);
                                } else {
                                    finalValue = originalAnimation.currentValue * scale;
                                }
                            } else {
                                finalValue = originalAnimation.currentValue;
                            }
    
                            startIndex = 1;
                        }
    
                        // Add up the override animations
                        for (var animIndex = startIndex; animIndex < holder.animations.length; animIndex++) {
                            var runtimeAnimation = holder.animations[animIndex];
                            var scale = runtimeAnimation.weight / normalizer;
    
                            if (!scale) {
                                continue;
                            } else if (runtimeAnimation.currentValue.scaleAndAddToRef) {
                                runtimeAnimation.currentValue.scaleAndAddToRef(scale, finalValue);
                            } else {
                                finalValue += runtimeAnimation.currentValue * scale;
                            }
                        }
    
                        // Add up the additive animations
                        for (let animIndex = 0; animIndex < holder.additiveAnimations.length; animIndex++) {
                            var runtimeAnimation = holder.additiveAnimations[animIndex];
                            var scale: number = runtimeAnimation.weight;
    
                            if (!scale) {
                                continue;
                            } else if (runtimeAnimation.currentValue.scaleAndAddToRef) {
                                runtimeAnimation.currentValue.scaleAndAddToRef(scale, finalValue);
                            } else {
                                finalValue += runtimeAnimation.currentValue * scale;
                            }
                        }
                    }
                }
                target[path] = finalValue;
            }
    
            target._lateAnimationHolders = {};
        }
        this._registeredForLateAnimationBindings.reset();
    }
    //#endregion
}