import { BABYLON } from "../render3d/babylon";
import { RibbonBuilder } from "./ribbon";
import { IPath, IRibbonRenderer } from "./interface";
import { CustomGeometry, CustomScene } from "../index";

export class RibbonRender implements IRibbonRenderer {
    /**
     * 最大点数
     */
    private readonly maxCount: number;
    /**
     * 两点之间最小有效距离
     */
    private readonly minimumDistance = 0.02;
    /**
     * 水波扩散宽度
     */
    private size = 0.35;
    /**
     * 脏更新标志
     */
    private dirty = false;
    /**
     * 水波流动速度
     */
    private readonly textureUOffSpeed = 0.10;
    private readonly textureVOffSpeed = 0.10;

    private _image: string;
    public get image() {
        return this._image;
    }
    public set image(path: string) {
        if (this._image != path) {
            this._image = path;
            if (this.material.diffuseTexture) {
                this.material.diffuseTexture.updateURL(this._image);
            } else {
                let texture = new BABYLON.Texture(null, this.scene.impl, true, false);
                texture.updateURL(this._image);
                this.material.diffuseTexture = texture;
            }
        }
    }

    public get alphaIndex() {
        return this.lineSource.alphaIndex;
    }
    public set alphaIndex(val: number) {
        this.lineSource.alphaIndex = val;
    }

    constructor(private readonly scene: CustomScene, maxCount = 128) {
        this.maxCount = maxCount;

        this.createLineSourceMesh();
        this.scene.registerBeforeRenderCall(this.render);
    }

    /**
     * 添加一个点
     */
    public addPoint(newPoint: BABYLON.Vector3) {
        this.basePath.push(newPoint);
        this.dirty = true;
    }

    public resetPoint(newPoint: BABYLON.Vector3[]) {
        this.basePath = newPoint;
        this.dirty = true;
    }

    public updateScale(uScale: number, vScale: number) {
        let texture = <BABYLON.Texture>(<BABYLON.StandardMaterial>this.lineSource?.material)?.diffuseTexture;
        if (texture) {
            texture.uScale = uScale;
            texture.vScale = vScale;
        }
    }

    //#region 调试数据
    private isDebug: boolean = false;
    private debugBoxSource: BABYLON.Mesh;
    private readonly debugBoxs: BABYLON.InstancedMesh[] = [];
    private createDebugBox() {
        let debugBoxMaterial = new BABYLON.DefaultMaterial("debugMaterial", this.scene.impl, this.scene.renderpipeline.PASS_OPAQUE);
        this.debugBoxSource = BABYLON.BoxBuilder.CreateBox("Box", this.scene.impl);
        this.debugBoxSource.setMaterial(debugBoxMaterial);
        let ins = this.debugBoxSource.createInstance(`${-1}`);
        ins.parent = this.debugBoxSource;
        this.isDebug = true;
    }
    //#endregion

    /**
     * 外部逻辑添加的点
     */
    private basePath: BABYLON.Vector3[] = [];
    /**
     * 构建出来的水波点
     */
    private readonly resultPath: IPath = { points: [] };
    private lineSource: BABYLON.Mesh;
    private geometry: CustomGeometry;
    private material: BABYLON.StandardEffectMaterial;
    private createLineSourceMesh() {
        let material = new BABYLON.StandardEffectMaterial("meshSourceMaterial", this.scene.impl);
        // material.backFaceCulling = false;
        material.emissiveColor.copyFromFloats(1, 1, 1);
        this.lineSource = new BABYLON.Mesh("meshSource", this.scene.impl);
        this.lineSource.setMaterial(material);
        this.geometry = new CustomGeometry(
            this.lineSource, BABYLON.VertexBuffer.UniqueBufferID(),
            [BABYLON.Constants.VERTEX_ATTR_POSITION, BABYLON.Constants.VERTEX_ATTR_COLOR4, BABYLON.Constants.VERTEX_ATTR_UV],
            this.maxCount * 2, this.maxCount * 2, true
        );
        this.lineSource.alwaysSelectAsActiveMesh = true;
    }

    private destroyLineSourceMesh() {
        this.resultPath.points = [];
        if (this.lineSource) {
            this.lineSource.dispose();
            this.lineSource = null;
        }
    }

    public init() {
        // this.createDebugBox();
        // this.createQuadSourceMesh();
    }

    public destroy() {
        this.destroyLineSourceMesh();
        this.scene.unregisterBeforeRenderCall(this.render);
    }

    /**
     * 渲染
     */
    private render = () => {
        let texture = <BABYLON.Texture>(<BABYLON.StandardMaterial>this.lineSource?.material)?.diffuseTexture;
        if (texture) {
            texture.uOffset = Date.now() % (1000 / this.textureUOffSpeed) / (1000 / this.textureUOffSpeed);
            texture.vOffset = Date.now() % (1000 / this.textureVOffSpeed) / (1000 / this.textureVOffSpeed);
        }

        if (this.dirty && this.basePath.length >= 2) {
            RibbonBuilder.FormatPathForceNormal(this.basePath, this.size, this.resultPath, false, this.minimumDistance, new BABYLON.Vector3(0, 0, 1));
            this.dirty = false;
            this.updateLines();
            this.updateDebug();
        }
    }

    /**
     * 更新线模式
     */
    private updateLines() {
        let indices = this.geometry.indices();
        let [positions, strip, posOffset] = this.geometry.vertices(BABYLON.Constants.VERTEX_ATTR_POSITION);
        let [colors, _, colorOffset] = this.geometry.vertices(BABYLON.Constants.VERTEX_ATTR_COLOR4);
        let [uvs, __, uvOffset] = this.geometry.vertices(BABYLON.Constants.VERTEX_ATTR_UV);
    
        let indicesCount = RibbonBuilder.Build(this.resultPath, 0, positions, colors, uvs, indices, strip, posOffset, colorOffset, uvOffset);
        this.geometry.update(indicesCount);
    }

    /**
     * 更新调试数据
     */
    private updateDebug() {
        // if (!this.isDebug) return;
        // let count = this.resultPath.points.length;
        // this.debugBoxs.forEach((v) => {
        //     v.dispose();
        // });
        // for (let i = 0; i < count; i++) {
        //     for (let j = 0; j < 3; j++) {
        //         let ins = this.debugBoxSource.createInstance(`${i}`);
        //         ins.position.copyFromFloats(this.positions[(i * 3 + j) * 3 + 0], this.positions[(i * 3 + j) * 3 + 1], this.positions[(i * 3 + j) * 3 + 2]);
        //         this.debugBoxs.push(ins);
        //         ins.parent = this.debugBoxSource;
        //     }
        // }
    }
}