import { TextureTillOffData, TextureTillOffDataScene } from "../material/standardmaterial";
import { IMeshMaterialProperty, IMeshRenderQueue, ITransformNode, ISprite, ISpriteManager, ITexturePackFrame, ITexturePackAtlas, ISpriteAtlasRects, ISpriteAtlasCells, ExtensionTools } from "../interface/index";
import { BABYLON } from "../render3d/babylon";
import { Scene } from "../render3d/scene";
import { SpriteManager } from "./sprite_manager";
import { PiCollider } from "../scene_pick";
import { ISprite3D, p3d } from "../render/types";
import { MathDataPool } from "../maths/index";
import { GlobalAtlasManager } from "./sprite_atlas";
import { IAtom } from "../context";

class Record<K, V> {
    private _dirty: boolean = false;
    private map: Map<K, V> = new Map();
    set(k: K, v: V) {
        this._dirty = true;
        this.map.set(k, v);
    }
    get(k: K) {
        return this.map.get(k);
    }
    delete(k: K) {
        this._dirty = this.map.delete(k);
    }
    iter() {
        return this.map.entries();
    }
    undirty() {
        this._dirty = false;
    }
    get dirty() {
        return this._dirty;
    }
}

export class Sprite extends BABYLON.InstancedMesh implements IMeshRenderQueue, ITransformNode, IMeshMaterialProperty, ISprite, ISprite3D {
    public static KeyAttrColor = "InsColor";
    public static KeyAttrUV = "InsTilloff";
    
    public static AtomKeyAttrUV: IAtom;
    public static RecordFrameIdx: Map<string, number> = new Map();

    private static Counter: number = Number.MIN_SAFE_INTEGER;
    private static Vx: BABYLON.Vector3 = new BABYLON.Vector3(1, 0, 0);
    private static Vy: BABYLON.Vector3 = new BABYLON.Vector3(0, 1, 0);
    public static Tmp: Uint16Array = new Uint16Array(14);
    protected _color: BABYLON.Color3 = MathDataPool.Color3(1, 1, 1);
    protected _applyColor() {
        let _r = Math.min(255, Math.max(0, this._color.r * 255));
        let _g = Math.min(255, Math.max(0, this._color.g * 255));
        let _b = Math.min(255, Math.max(0, this._color.b * 255));
        let _a = Math.min(255, Math.max(0, this._alpha * 255));
        this.instanceAttribute(Sprite.KeyAttrColor, _r, _g, _b, _a);
    }
    get color() {
        return this._color;
    }
    set color(val: BABYLON.Color3) {
        if (this._isDisposed) {
            // debugger;
            return;
        }
        if (this._color.r == val.r && this._color.g == val.g && this._color.b == val.b) return;
        this._color.copyFromFloats(val.r, val.g, val.b);
        this._applyColor();
        this.mgr?.dirty();
    }
    setColor(r: number, g: number, b: number) {
        if (this._isDisposed) return;
        if (this._color.r == r && this._color.g == g && this._color.b == b) return;
        this._color.copyFromFloats(r, g, b);
        this._applyColor();
        this.mgr?.dirty();
        return this;
    }
    protected _alpha: number = 1;
    get alpha() {
        return this._alpha;
    }
    set alpha(val: number) {
        if (this._isDisposed) return;
        if (this._alpha == val) return;
        this._alpha = val;
        this._applyColor();
        this.mgr?.dirty();
    }
    duration: number = 1.0;

    protected mgr: ISpriteManager;
    sortParam() {
        // this.computeWorldMatrix();
        return this.position.y;
    }

    setEnabled(val: boolean): void {
        super.setEnabled(val);
        this.mgr?.dirty();
    }
    isDisposed(): boolean {
        return this._isDisposed;
    }
    dispose(_doNotRecurse?: boolean, _disposeMaterialAndTextures?: boolean): void {
        if (this.mgr) {
            this.mgr.delSprite(this);
            this.mgr = undefined;
            super.dispose();
            MathDataPool.Recycle(this._color);
            this._color = undefined;
            this.tilloff?.dispose();
            this.tilloff = undefined;
        }
    }
    vertexPosition(data: Float32Array, offset: number, strip: number = 3, localoff: number = 0): number {
        throw new Error("Not Implements.");
    }

    vertexColor(data: Float32Array, offset: number, strip: number = 4, localoff: number = 0): number {
        throw new Error("Not Implements.");
    }

    vertexUV(data: Float32Array, offset: number, strip: number = 2, localoff: number = 0): number {
        throw new Error("Not Implements.");
    }

    protected tilloff: TextureTillOffData;
    public get uScale() {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        return this.tilloff.uScale;
    }
    public set uScale(val: number) {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        this.tilloff.uScale = val;
        this.tillOffset(this.tilloff.uScale, this.tilloff.vScale, this.tilloff.uOffset, this.tilloff.vOffset);
        this.mgr?.dirty();
    }
    public get vScale() {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        return this.tilloff.vScale;
    }
    public set vScale(val: number) {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        this.tilloff.vScale = val;
        this.tillOffset(this.tilloff.uScale, this.tilloff.vScale, this.tilloff.uOffset, this.tilloff.vOffset);
        this.mgr?.dirty();
    }
    public get uOffset() {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        return this.tilloff.uOffset;
    }
    public set uOffset(val: number) {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        this.tilloff.uOffset = val;
        this.tillOffset(this.tilloff.uScale, this.tilloff.vScale, this.tilloff.uOffset, this.tilloff.vOffset);
        this.mgr?.dirty();
    }
    public get vOffset() {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        return this.tilloff.vOffset;
    }
    public set vOffset(val: number) {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        this.tilloff.vOffset = val;
        this.tillOffset(this.tilloff.uScale, this.tilloff.vScale, this.tilloff.uOffset, this.tilloff.vOffset);
        this.mgr?.dirty();
    }
    public tillOffset(uScale: number, vScale: number, uOffset: number, vOffset: number) {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        this.tilloff.uScale = uScale;
        this.tilloff.vScale = vScale;
        this.tilloff.uOffset = uOffset;
        this.tilloff.vOffset = vOffset;

        this._scene.cmdsSprite.set(this._entity, this);
    }

    public readonly spriteId: number;
    protected _spriteFrame: string | number;
    constructor(protected atlas: string, _spriteFrame: string | number, scene: Scene, renderingGroupId: number = 0, alphaIndex: number = 999999) {
        let spriteId: number = Sprite.Counter++;
        let mgr = scene.getSpritManager(atlas);
        super(`${spriteId}`, (<SpriteManager>mgr).mesh);

        this.mgr = mgr;
        this.spriteId = spriteId;
        this.renderingGroupId = renderingGroupId;
        this.alphaIndex = alphaIndex;
        this.mgr?.addSprite(this);
        this._texturePackFrame = GlobalAtlasManager.getTexturePackFrame(this._spriteFrame, this.atlas);
        let renderMode = (<any>mgr)._atlasFrames.renderMode;
        if (renderMode == 1 || renderMode == 2 || renderMode == 3) {
            this._color.copyFromFloats(0, 0, 0);
        }
        this._applyColor();
        this.spriteFrame = _spriteFrame;
        this.alwaysSelectAsActiveMesh = true;
    }
    _onFrameUpdate(): boolean {
        super._onFrameUpdate();
        return this._animIsLoop !== undefined;
    }
    protected _texturePackFrame: ITexturePackFrame;
    get spriteFrame() {
        if (this._animIsLoop !== undefined) {
            if ((<ITexturePackAtlas>this.mgr?.atlas).meta) {
                let animations = (<ITexturePackAtlas>this.mgr?.atlas).animations;
                if (animations) {
                    if (this._animeName === undefined) {
                        for (let animeName in animations) {
                            if (animations[animeName].indexOf(<string>this._spriteFrame) >= 0) {
                                this._animeName = animeName;
                                break;
                            }
                        }
                    }
                    if (this._animeName !== undefined) {
                        this._animedTime += this._scene.deltaTime;
                        let frames = animations[this._animeName];
                        let runnedFrame = this._animedTime / (1000 * this.duration) * frames.length;
                        if (runnedFrame >= frames.length && this._animIsLoop === false) {
                            let frame = frames.length - 1;
                            this.spriteFrame = frames[frame];
                            this._animIsLoop = undefined;
                        } else {
                            let frame = Math.floor(runnedFrame) % frames.length;
                            this.spriteFrame = frames[frame];
                        }
                    }
                }
            } else if ((<ISpriteAtlasRects>this.mgr?.atlas).rects) {
                // this._animedTime += this._scene.deltaTime;
            } else {
                let atlas = <ISpriteAtlasCells>this.mgr?.atlas;
                let frames = atlas.hCount * atlas.vCount;
                this._animedTime += this._scene.deltaTime;
                let runnedFrame = this._animedTime / (1000 * this.duration) * frames;
                if (runnedFrame >= frames && this._animIsLoop === false) {
                    let frame = frames - 1;
                    this.spriteFrame = frame;
                    this._animIsLoop = undefined;
                } else {
                    let frame = Math.floor(runnedFrame) % frames;
                    this.spriteFrame = frame;
                }
            }
        }
        return this._spriteFrame;
    }
    set spriteFrame(val: string | number) {
        if (this.mgr == undefined) {
            debugger;
            return;
        }
        if (this._spriteFrame == val) return;
        this._spriteFrame = val;
        this._texturePackFrame = GlobalAtlasManager.getTexturePackFrame(this._spriteFrame, this.atlas);
        // let frame = this.mgr?.getFrame(this);

        if (!this._texturePackFrame) { return; }

        this._scene.cmdsSprite.set(this._entity, this);
        this.mgr.dirty();
    }
    _applySpriteData(data: Uint16Array, offset: number): number {
        if (Sprite.AtomKeyAttrUV == undefined) {
            Sprite.AtomKeyAttrUV = ExtensionTools.getAtom("InsTilloff");
        }
        let result = this._isDisposed == false && this.mgr != undefined;
        if (result) {
            if (this.tilloff) {
                let sw = 8192;
                let sh = 8192;
                let w = Math.floor(sw * this.tilloff.uScale);
                let h = Math.floor(sh * this.tilloff.vScale);
                let x = Math.floor(sw * this.tilloff.uOffset);
                let y = Math.floor(sh * this.tilloff.vOffset);
                data[offset +  0]  = 0;
                data[offset +  1]  = 0;
                data[offset +  2]  = w;
                data[offset +  3]  = h;
                data[offset +  4]  = 0;
                data[offset +  5]  = 0;
                data[offset +  6]  = w;
                data[offset +  7]  = h;
                data[offset +  8]  = x;
                data[offset +  9]  = y;
                data[offset + 10] = w;
                data[offset + 11] = h;
                data[offset + 12] = sw;
                data[offset + 13] = sh;
                p3d.p3d_sprite_frame_data(BABYLON.Engine.commands(), this.entity, Sprite.AtomKeyAttrUV, data);
                return this.entity;
            } else if (this._texturePackFrame) {
                let key = `${this.atlas}#${this._spriteFrame}`;
                let idx = Sprite.RecordFrameIdx.get(key);
                if (idx === undefined) {
                    if ((<any>this._texturePackFrame).toU16) {
                        (<any>this._texturePackFrame).toU16(data, offset);
                    } else {
                        data[offset +  0]  = this._texturePackFrame.rotated ? 1 : 0;
                        data[offset +  1]  = this._texturePackFrame.trimmed ? 1 : 0;
                        data[offset +  2]  = this._texturePackFrame.sourceSize.w;
                        data[offset +  3]  = this._texturePackFrame.sourceSize.h;
                        data[offset +  4]  = this._texturePackFrame.spriteSourceSize.x;
                        data[offset +  5]  = this._texturePackFrame.spriteSourceSize.y;
                        data[offset +  6]  = this._texturePackFrame.spriteSourceSize.w;
                        data[offset +  7]  = this._texturePackFrame.spriteSourceSize.h;
                        data[offset +  8]  = this._texturePackFrame.frame.x;
                        data[offset +  9]  = this._texturePackFrame.frame.y;
                        data[offset + 10]  = this._texturePackFrame.frame.w;
                        data[offset + 11]  = this._texturePackFrame.frame.h;
                    }
                    // Sprite.Tmp[0] = 0;
                    data[offset + 12] = (<ITexturePackAtlas>this.mgr.atlas).meta.size.w;
                    data[offset + 13] = (<ITexturePackAtlas>this.mgr.atlas).meta.size.h;
                    idx = p3d.p3d_record_sprite_frame_data(BABYLON.Engine.commands(), data);
                    Sprite.RecordFrameIdx.set(key, idx);
                }
                
                p3d.p3d_sprite_frame(BABYLON.Engine.commands(), this.entity, Sprite.AtomKeyAttrUV, idx);
                return this.entity;
            } else {
                return undefined;
            }
        } else {
            return undefined;
        }
    }

    private _animedTime: number = 0;
    private _animeName: string | number;
    private _animIsLoop: boolean;
    setAnim(isLoop: boolean) {
        this._animedTime = 0;
        if ((<ITexturePackAtlas>this.mgr?.atlas).meta) {
            if ((<ITexturePackAtlas>this.mgr?.atlas).animations) {
                this._animIsLoop = isLoop;
            }
        } else {
            this._animIsLoop = isLoop;
        }
        return this;
    }

    getCollider() {
        if (this.__collider == undefined) {
            this.setCollider(new PiCollider(new BABYLON.Vector3(-0.5, -0.5, -0.01), new BABYLON.Vector3(0.5, 0.5, 0.01)));
        }
        return this.__collider;
    }
}
