import { ISprite, ISpriteAtlas, ISpriteAtlasCells, ISpriteAtlasRects, ISpriteManager, ITexturePackAtlas, ITexturePackFrame } from "../interface/sprite";
import { CustomGeometry } from "../customgeometry/custom_geometry";
import { BABYLON } from "../render3d/babylon";
import { Constants } from "../interface/constants";
import { Scene } from "../render3d/scene";
import { Observable } from "../interface/observable";
import { Atlas, GlobalAtlasManager } from "./sprite_atlas";
import { ForwardRenderPipeline } from "../render";
import { TextureHSVMaterial } from "../material/hsv";
import { TextureHueSaturateMaterial } from "../material/hue_saturate";
import { Sprite } from "./sprite";
import { EAtlasRenderMode } from "../interface/texturepacker";
import { OverlayColorMaterial } from "../material/overlay_color";

export class SpriteManager implements ISpriteManager {
    public static MAX_SPRITE = 1000;
    get material() {
        return this._material;
    }
    _scene: Scene;
    _isDisposed: boolean = false;
    setEnabled(val: boolean): void {
        // throw new Error("Method not implemented.");
    }
    isDisposed(): boolean {
        return this._isDisposed;
    }
    dispose(): void {
        if (this._isDisposed) return;

        this._scene.delSpritManager(this);

        this._sprites.forEach((item) => {
            item.dispose(false, true);
        });

        this._material.dispose(false, true);
        this._mesh?.dispose(false, true);
        this._meshes.forEach((item) => {
            item.geo.dispose();
            item.mesh.dispose(false, true);
        });
    }

    protected _geometry: CustomGeometry;
    protected _material: BABYLON.Material;
    protected _sprites: Map<number, ISprite> = new Map();
    protected _meshes: Map<string, { mesh: BABYLON.Mesh, geo: CustomGeometry }> = new Map();
    protected _atlasFrames: Atlas;
    protected _dirty: boolean = false;
    public readonly atlasImage: string;
    public readonly atlas: ISpriteAtlas;
    protected alphaMode: number;
    protected _mesh: BABYLON.Mesh;
    public get mesh() {
        return this._mesh;
    }
    public static MaterialFactory: Map<EAtlasRenderMode, (scene: Scene, atlasImage: string, noMipmap: boolean, samplerMode: number, alphaMode: number, pass?: Constants.PassTag) => BABYLON.Material> = new Map();
    public static CreateMaterial(scene: Scene, atlasImage: string, noMipmap: boolean, samplerMode: number, alphaMode: number, pass?: Constants.PassTag, renderMode: EAtlasRenderMode = 0) {
        SpriteManager.Initial();
        let func = SpriteManager.MaterialFactory.get(renderMode) || SpriteManager.MaterialFactory.get(0);
        return func(scene, atlasImage, noMipmap, samplerMode, alphaMode, pass);
    }
    private static HasInitialed: boolean = false;
    public static Initial() {
        if (SpriteManager.HasInitialed) return;
        SpriteManager.HasInitialed = true;
        SpriteManager.MaterialFactory.set(1, function(scene: Scene, atlasImage: string, noMipmap: boolean, samplerMode: number, alphaMode: number, pass?: Constants.PassTag) {
            let _material = new TextureHSVMaterial(`${atlasImage}`, scene, undefined, pass);
            let texture = GlobalAtlasManager.getTexture(atlasImage, scene);
            if (texture == undefined) {
                texture = new BABYLON.Texture(null, scene, noMipmap, false, samplerMode || BABYLON.Texture.LINEAR_LINEAR);
                texture.updateURL(atlasImage);
            }
            texture.hasAlpha = true;
            texture.wrapU = BABYLON.Constants.CLAMP_ADDRESSMODE;
            texture.wrapV = BABYLON.Constants.CLAMP_ADDRESSMODE;
            _material.mainTexture = texture;
            // _material.emissiveColor = new BABYLON.Color3(1, 1, 1);
            // _material.disableDepthWrite = true;
            // _material.backFaceCulling = false;
            // _material.alphaMode = alphaMode;
            // _material.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
            return _material;
        });
        SpriteManager.MaterialFactory.set(2, function(scene: Scene, atlasImage: string, noMipmap: boolean, samplerMode: number, alphaMode: number, pass?: Constants.PassTag) {
            let _material = new TextureHueSaturateMaterial(`${atlasImage}`, scene, undefined, pass);
            let texture = GlobalAtlasManager.getTexture(atlasImage, scene);
            if (texture == undefined) {
                texture = new BABYLON.Texture(null, scene, noMipmap, false, samplerMode || BABYLON.Texture.LINEAR_LINEAR);
                texture.updateURL(atlasImage);
            }
            texture.hasAlpha = true;
            texture.wrapU = BABYLON.Constants.CLAMP_ADDRESSMODE;
            texture.wrapV = BABYLON.Constants.CLAMP_ADDRESSMODE;
            _material.diffuseTexture = texture;
            _material.emissiveColor = new BABYLON.Color3(1, 1, 1);
            // _material.disableDepthWrite = true;
            // _material.backFaceCulling = false;
            // _material.alphaMode = alphaMode;
            // _material.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
            return _material;
        });
        SpriteManager.MaterialFactory.set(3, function(scene: Scene, atlasImage: string, noMipmap: boolean, samplerMode: number, alphaMode: number, pass?: Constants.PassTag) {
            let _material = new OverlayColorMaterial(`${atlasImage}`, scene, undefined, pass);
            let texture = GlobalAtlasManager.getTexture(atlasImage, scene);
            if (texture == undefined) {
                texture = new BABYLON.Texture(null, scene, noMipmap, false, samplerMode || BABYLON.Texture.LINEAR_LINEAR);
                texture.updateURL(atlasImage);
            }
            texture.hasAlpha = true;
            texture.wrapU = BABYLON.Constants.CLAMP_ADDRESSMODE;
            texture.wrapV = BABYLON.Constants.CLAMP_ADDRESSMODE;
            _material.mainTexture = texture;
            _material.mainColor(1, 1, 1);
            // _material.disableDepthWrite = true;
            // _material.backFaceCulling = false;
            // _material.alphaMode = alphaMode;
            // _material.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
            return _material;
        });
        SpriteManager.MaterialFactory.set(0, function(scene: Scene, atlasImage: string, noMipmap: boolean, samplerMode: number, alphaMode: number, pass?: Constants.PassTag) {
            let _material = new BABYLON.MainOpacityMaterial(`${atlasImage}`, scene, undefined, pass);
            let texture = GlobalAtlasManager.getTexture(atlasImage, scene);
            if (texture == undefined) {
                texture = new BABYLON.Texture(null, scene, noMipmap, false, samplerMode || BABYLON.Texture.LINEAR_LINEAR);
                texture.updateURL(atlasImage);
            }
            texture.hasAlpha = true;
            texture.wrapU = BABYLON.Constants.CLAMP_ADDRESSMODE;
            texture.wrapV = BABYLON.Constants.CLAMP_ADDRESSMODE;
            _material.diffuseTexture = texture;
            // _material.disableLighting = true;
            // _material.emissiveColor = new BABYLON.Color3(1, 1, 1);
            // _material.disableDepthWrite = true;
            // _material.backFaceCulling = false;
            // _material.alphaMode = alphaMode;
            // _material.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
            // _material.useAlphaFromDiffuseTexture = true;
            // _material.cutoffEdgeAlphaDiff = 0.0;
            _material.alphaCutOff = 0.0;
            return _material;
        });
    }
    constructor(scene: Scene, atlas: Atlas, public readonly maxSprite: number = SpriteManager.MAX_SPRITE, material?: BABYLON.Material, private pass?: Constants.PassTag) {
        this._scene = scene;
        this._atlasFrames = atlas;
        this.atlas = atlas.atlas;
        this.atlasImage = this._atlasFrames.atlasImage;
        this.alphaMode = atlas.alphaMode;
        if (material) {
            this._material = <BABYLON.Material>material;
        } else {
            this._material = SpriteManager.CreateMaterial(this._scene, atlas.atlasImage, atlas.noMipmap, atlas.samplerMode, atlas.alphaMode, pass, atlas.renderMode);
        }
        this._mesh = BABYLON.QuadBuilder.CreateQuad(
            this.atlasImage, { size: 1 }, scene,
            [
                {
                    key: Sprite.KeyAttrColor, code: "A_COLOR4 *= InsColor;\n", vtype: Constants.INSTANCE_ATTR_UNORM8X4,
                },
                {
                    key: Sprite.KeyAttrUV, code: "A_UV = A_UV.xy * InsTilloff.xy + InsTilloff.zw;\n", vtype: Constants.INSTANCE_ATTR_UNORM16X4,
                }
            ],
            true, true
        );
        this._mesh.setMaterial(this._material);
        this._mesh.alphaMode = atlas.alphaMode;
        this._mesh.depthTest = BABYLON.Constants.ALWAYS;
        this._mesh.depthWrite = false;
        // this._mesh.alwaysSelectAsActiveMesh = true;
    }
    extendAtlasFrame(key: string | number, rect: [number, number, number, number]): void {
        this._atlasFrames.extendAtlasFrame(key, rect);
    }
    getFrame(item: ISprite): [number, number, number, number] {
        return undefined;
    }
    getTexturePackFrame(item: ISprite): ITexturePackFrame {
        return undefined;
    }
    addSprite(item: ISprite): boolean {
        if (this._sprites.size < this.maxSprite) {
            this._sprites.set(item.spriteId, item);
            this._dirty = true;
            return true;
        } else {
            return false;
        }
    }
    delSprite(item: ISprite): void {
        this._sprites.delete(item.spriteId);
        this._dirty = true;
    }
    dirty(): void {
        this._dirty = true;
    }
    spriteCount() {
        return this._sprites.size;
    }
    onDisposeObservable: Observable<number>;
    name?: string;
    _onFrameUpdate(): void {
    }
    /**
     * 默认param 为 sprite 的全局坐标的 y 值
     * @param list 
     */
    sort(list: { param: any; }[]): void {
        list.sort((a, b) => b.param - a.param);
    }
}
