import { SceneStruct } from "../scene_struct";
import { ModelObj } from "../model_obj";
import { IModelOpt, IPrefabCreation, ITexturePackAtlas, Constants } from "../interface/index";
import { BABYLON } from "../render3d/babylon";
import { GlobalAtlasManager } from "./sprite_atlas";
import { CustomScene, ModelPrefab, PrefabModelObjPool } from "../index";
import { Sprite } from "./sprite";

export interface ISpriteModelCfg extends IPrefabCreation {
    spriteFrame: string,
    spriteAtlas: string,
    distributions?: [[number, number, number], [number, number, number], [number, number, number]][];
    alphaMode: Constants.ALPHA_MODE,
    samplingMode: number,
}

export class SpriteModel extends ModelObj {
    public static registPrefabCfg(prefabPool: PrefabModelObjPool, modelName: string, spriteFrame: string, spriteAtlas: string, cacheCount: number = 4, alphaMode: Constants.ALPHA_MODE = 7, samplingMode: number = BABYLON.Texture.LINEAR_LINEAR, distributions?: [number, number, number][]) {
        prefabPool.registPrefabCfg(modelName, { path: "null", fileName: "null", single: false, cacheCount, spriteFrame, spriteAtlas, distributions, alphaMode, samplingMode });
        PrefabModelObjPool.registPrefabFactory(modelName, ModelPrefab.Creation);
        PrefabModelObjPool.registPrefabResFactory(modelName, SpriteModel.Creation);
    }

    public static Creation(meshName: string, scene: CustomScene, opt?: ISpriteModelCfg) {
        return new SpriteModel(meshName, scene, opt);
    }

    protected atlas: ITexturePackAtlas;
    constructor(meshName: string, scene: CustomScene, opt: ISpriteModelCfg) {
        let fileName = opt.fileName;
        opt.fileName = undefined;
        super(meshName, scene, opt);
        opt.fileName = fileName;
        let atlasName = opt.spriteAtlas || ((opt.path || "") + fileName);
        let pureName = atlasName.replace(".png", "").replace(".atlas", "");

        this._loadPromise = new Promise((resolve, reject) => {

            let atlas = GlobalAtlasManager.getAtlas(`${pureName}.png`);
            if (atlas) {
                this.atlas = <ITexturePackAtlas>atlas.atlas;
                // if (this.atlas.animations) {
                    let texture = new BABYLON.Texture(null, scene.impl, true, false, opt.samplingMode);
                    texture.wrapU = BABYLON.Constants.CLAMP_ADDRESSMODE;
                    texture.wrapV = BABYLON.Constants.CLAMP_ADDRESSMODE;
                    texture.updateURL(atlas.atlasImage);
                    // console.warn(`Load ${atlas.atlasImage}`);
                    // texture.onLoadObservable.addOnce(() => {
                        if (this.isDisposed == false) {
                            this.loadedSprite();
                            // console.warn(`Load Success ${atlas.atlasImage}`);
                            if (this.insertedCall) {
                                this.insertedCall(this);
                            }
                        }
                        resolve(null);
                        texture.dispose();
                    // });
                // } else {
                //     reject(`SpriteModel Atlas Not Found Animation! ${pureName}`);
                // }
            } else {
                GlobalAtlasManager.load(`${pureName}.atlas`).then(() => {
                    let atlas = GlobalAtlasManager.getAtlas(`${pureName}.png`);
                    if (atlas) {
                        this.atlas = <ITexturePackAtlas>atlas.atlas;
                        // if (this.atlas.animations) {
                            let texture = new BABYLON.Texture(null, scene.impl, true, false, opt.samplingMode);
                            texture.updateURL(atlas.atlasImage);
                            texture.wrapU = BABYLON.Constants.CLAMP_ADDRESSMODE;
                            texture.wrapV = BABYLON.Constants.CLAMP_ADDRESSMODE;
                            // console.warn(`Load ${atlas.atlasImage}`);
                            // texture.onLoadObservable.addOnce(() => {
                                if (this.isDisposed == false) {
                                    this.loadedSprite();
                                    // console.warn(`Load Success ${atlas.atlasImage}`);
                                    if (this.insertedCall) {
                                        this.insertedCall(this);
                                    }
                                }
                                resolve(null);
                                texture.dispose();
                            // });
                        // } else {
                        //     reject(`SpriteModel Atlas Not Found Animation! ${pureName}`);
                        // }
                    } else {
                        reject(`SpriteModel Atlas Not Found! ${pureName}`);
                    }
                }).catch((err) => {
                    reject(`SpriteModel Atlas Not Found! ${pureName}, Err: ${err}`);
                })
            }
        });
    }
    protected _spriteFrame: number = 0;
    protected get spriteFrame() {
        return this._spriteFrame;
    }
    protected set spriteFrame(val: number) {
        this._spriteFrame = val;
        if (this.animation && this.atlas.animations) {
            // let sprtie = <Sprite>this.rootImpl;
            let animations = this.atlas.animations[this.animation.name];
            if (animations) {
                this.distributions.forEach(sprtie => {
                    let idx = Math.min(Math.floor(val), animations.length - 1);
                    sprtie.spriteFrame = animations[idx];
                });
            }
        }
    }
    protected distributions: Sprite[] = [];
    protected loadedSprite() {
        let animationGroups = [];
        let frame: string;
        if (this.atlas.animations) {
            for (let animeName in this.atlas.animations) {
                let animations = this.atlas.animations[animeName];
                let animeGroup = new BABYLON.AnimationGroup(animeName, this.scene.impl);
                let animation = new BABYLON.Animation(animeName, "spriteFrame", 1, BABYLON.Animation.ANIMATIONTYPE_FLOAT);
                animation.setKeys([
                    {
                        frame: 0,
                        value: 0,
                    },
                    {
                        frame: animations.length / 30,
                        value: animations.length - 0.01
                    }
                ]);
                animeGroup.addTargetedAnimation(animation, this);
                animationGroups.push(animeGroup);
                if (frame == undefined) {
                    frame = animations[0];
                }
            }
        }
        let option = <ISpriteModelCfg>this.option;
        let rootImpl: BABYLON.Transform;
        if (option.distributions) {
            rootImpl = new BABYLON.TransformNode(this.atlas.image, this.scene.impl);
            rootImpl.setEnabled(false);

            option.distributions.forEach((param) => {
                let position = param[0];
                let scaling = param[1];
                let rotation = param[2];

                let item = new Sprite(this.atlas.image, frame, this.scene.impl);
                position && item.position.copyFromFloats(...position);
                scaling && item.scaling.copyFromFloats(...scaling);
                rotation && item.rotation.copyFromFloats(...rotation);
                item.parent = rootImpl;
                this.distributions.push(item);
            });

            this.loaded(rootImpl, [rootImpl], [], [], [], animationGroups, []);
        } else {
            rootImpl = new Sprite(this.atlas.image, frame, this.scene.impl);
            this.distributions.push(<Sprite>rootImpl);
            rootImpl.setEnabled(false);
            this.loaded(rootImpl, [rootImpl], [], [], [], animationGroups, []);
        }
    }
    protected _changeAlphaIndex() {
        if (this.isLoaded && this._meshes && this._alphaIndex != undefined) {
            // (<Sprite>this.rootImpl).alphaIndex = this._alphaIndex;
            this.distributions.forEach(v => {
                v.alphaIndex = this._alphaIndex;
            });
        }
    }
    protected _changeAlpha() {
        super._changeAlpha();
        if (!this.isLoaded || this.isDisposed) { return }
        if (this._alpha !== undefined) {
            // (<Sprite>this.rootImpl).alpha = this.alpha;
            this.distributions.forEach(v => {
                v.alpha = this._alpha;
            });
        }
    }
    protected _changeColor(): void {
        super._changeColor();
        if (!this.isLoaded || this.isDisposed) { return }
        if (this._color) {
            // (<Sprite>this.rootImpl).color = new BABYLON.Color3(...this.color);
            this.distributions.forEach(v => {
                v.color = new BABYLON.Color3(...this._color);
            });
        }
    }
    public loadAsync(opt: IModelOpt): Promise<ModelObj> {
        throw new Error("Not implement!");
    }
    protected _changeAnim(): void {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this.animOpt !== undefined) {
            if (this.animationMap.has(this.animOpt.animName)) {
                super._changeAnim();
            } else {
                this.distributions.forEach(sprtie => {
                    sprtie.spriteFrame = this.animOpt.animName;
                });
            }
        } else {
            if (this._animation) {
                this._stopAnimation(this._animation);
                this._animation = undefined;
            }
        }
    }
    protected _getColliderNode() {
        return this.distributions[0];
    }
}