import { BABYLON } from "../render3d/babylon";
import { SceneStruct } from "../scene_struct";
import { ITileAssetExtend, TileMapAsset } from "./assets";
import { MainOpacityMaterial, BABYLONAction, MATERIAL_ALPHABLEND, CustomGeometry, Constants } from "../index";
import { Tile } from "./tile";
import { TileObject } from "./tile_object";

export class TileObjectMesh {
    public mesh: BABYLON.Mesh;
    protected objects: Map<any, TileObject> = new Map();
    public dirty: boolean = true;
    public dirtyPosition: boolean = true;
    public dirtyColor: boolean = true;
    public dirtyUVs: boolean = true;
    protected customGeometry: CustomGeometry;
    public meshInitialOk: boolean = false;
    constructor(
        public scene: SceneStruct,
        public readonly layer: number,
        public readonly assetSource: string,
        public readonly maxCount = 2000,
    ) {
        this.mesh = new BABYLON.Mesh(`${this.assetSource}#${this.layer}`, this.scene.impl);
        this.customGeometry = new CustomGeometry(this.mesh, `${assetSource}#${layer}`, [Constants.VERTEX_ATTR_POSITION, Constants.VERTEX_ATTR_COLOR4, Constants.VERTEX_ATTR_UV], this.maxCount * Tile.POINT_COUNT, this.maxCount * Tile.TRIANGLE_COUNT);
    }
    get(id: any): TileObject | undefined {
        return this.objects.get(id);
    }
    add(value: TileObject) {
        this.dirty = true;
        return this.objects.set(value.id, value);
    }
    remove(id: any): TileObject | undefined {
        let old = this.objects.get(id);
        if (old) {
            this.objects.delete(id);
            this.dirty = true;
            return old;
        } else {
            return undefined;
        }
    }
    dispose() {
        this.objects.forEach((obj) => {
            obj.dispose();
        });
        this.mesh && this.mesh.dispose();
        this.mesh = undefined;
        this.objects = undefined;
    }
    protected _mesh(
        count: number,
        assetMap: TileMapAsset,
        materials: MainOpacityMaterial,
        parent: BABYLON.TransformNode,
    ) {

        if (this.meshInitialOk == false) {
            BABYLONAction.useMaterial(this.mesh, materials, MATERIAL_ALPHABLEND, Constants.ALPHA_COMBINE);
            this.mesh.alphaIndex = this.layer;
            this.mesh.alwaysSelectAsActiveMesh = true;
            this.mesh.doNotSyncBoundingInfo = true;
            BABYLONAction.alphaMode(this.mesh, 2);
            BABYLONAction.depthTest(this.mesh, false);
            BABYLONAction.depthWrite(this.mesh, false);
            // this.mesh.cullMode = BABYLON.CullMode.Off;
            // this.mesh.frontFace = BABYLON.FrontFace.Ccw;
            // this.mesh.hasVertexAlpha = true;

            this.mesh.parent = parent;

            this.meshInitialOk = true;
        }
        // if (count > this.mObjCount) {
        //     if (this.mesh) {
        //         this.mesh.dispose();
        //     }

        //     this.mObjCount = Math.ceil(count / 100) * 100;

        //     this.mPositions = new Float32Array(this.mObjCount * Tile.POINT_COUNT * Tile.POINT_POSITION_COUNT);
        //     this.mUVs = new Float32Array(this.mObjCount * Tile.POINT_COUNT * Tile.POINT_UV_COUNT);
        //     this.mIndices = new Uint16Array(this.mObjCount * Tile.TRIANGLE_COUNT * Tile.TRIANGLE_INDICES_COUNT);
        //     this.mColors = new Float32Array(this.mObjCount * Tile.POINT_COUNT * Tile.POINT_COLOR_COUNT);
        //     let len = this.mObjCount * Tile.POINT_COUNT * Tile.POINT_COLOR_COUNT;
        //     for (let i = 0; i < len; i++) {
        //         this.mColors[i] = 1.;
        //     }

        //     this.mesh = new BABYLON.Mesh(`${this.assetSource}#${this.layer}`, this.scene.impl);
        //     BABYLONAction.useMaterial(this.mesh, materials, MATERIAL_ALPHABLEND, ALPHA_COMBINE);
        //     this.mesh.alphaIndex = this.layer;
        //     this.mesh.alwaysSelectAsActiveMesh = true;
        //     this.mesh.doNotSyncBoundingInfo = true;
        //     this.mesh.hasVertexAlpha = true;

        //     this.mesh.parent = parent;

        //     return true;
        // }
        // return false;
    }
    protected _build(asset: ITileAssetExtend, events?: [number, number, number, number][]) {
        let index = 0;
        this.objects.forEach((obj) => {
            if (index >= this.maxCount) return;

            TileObject.positions(index, obj, ...this.customGeometry.vertices(Constants.VERTEX_ATTR_POSITION), events);

            {
                let [data, strip, offset] = this.customGeometry.vertices(Constants.VERTEX_ATTR_UV);
                obj._uvs(index, data, asset.columns, obj.state.cellID, asset.imageUSize, asset.imageVSize, strip, offset);
            }

            Tile.indices(index, this.customGeometry.indices());

            {
                let [data, strip, offset] = this.customGeometry.vertices(Constants.VERTEX_ATTR_COLOR4);
                Tile.colors(index, data, obj.r, obj.g, obj.b, obj.a, strip, offset);
            }


            index += 1;
        });
    }
    protected _buildPosition(
        events?: [number, number, number, number][]
    ) {
        let index = 0;
        this.objects.forEach((obj) => {
            if (index >= this.maxCount) return;
            TileObject.positions(index, obj, ...this.customGeometry.vertices(Constants.VERTEX_ATTR_POSITION), events);
            index += 1;
        });

        // this.mesh.updateVerticesData(BABYLON.VertexBuffer.PositionKind, this.mPositions);
        // this.mesh.setEnabled(true);
    }
    protected _buildUV(asset: ITileAssetExtend) {
        let index = 0;
        this.objects.forEach((obj) => {
            if (index >= this.maxCount) return;
            {
                let [data, strip, offset] = this.customGeometry.vertices(Constants.VERTEX_ATTR_UV);
                obj._uvs(index, data, asset.columns, obj.state.cellID, asset.imageUSize, asset.imageVSize, strip, offset);
            }

            index += 1;
        });

        // this.mesh.updateVerticesData(BABYLON.VertexBuffer.UVKind, this.mUVs);
        // this.mesh.setEnabled(true);
    }
    protected _buildColor() {
        let index = 0;
        this.objects.forEach((obj) => {
            if (index >= this.maxCount) return;
            {
                let [data, strip, offset] = this.customGeometry.vertices(Constants.VERTEX_ATTR_COLOR4);
                Tile.colors(index, data, obj.r, obj.g, obj.b, obj.a, strip, offset);
            }
            index += 1;
        });

        // this.mesh.updateVerticesData(BABYLON.VertexBuffer.ColorKind, this.mColors);
        // this.mesh.setEnabled(true);
    }
    protected _buildPositionAndUV(asset: ITileAssetExtend, events?: [number, number, number, number][]) {
        let index = 0;
        this.objects.forEach((obj) => {
            if (index >= this.maxCount) return;
            TileObject.positions(index, obj, ...this.customGeometry.vertices(Constants.VERTEX_ATTR_POSITION), events);

            {
                let [data, strip, offset] = this.customGeometry.vertices(Constants.VERTEX_ATTR_UV);
                obj._uvs(index, data, asset.columns, obj.state.cellID, asset.imageUSize, asset.imageVSize, strip, offset);
            }
            index += 1;
        });

        // this.mesh.updateVerticesData(BABYLON.VertexBuffer.PositionKind, this.mPositions);
        // this.mesh.updateVerticesData(BABYLON.VertexBuffer.UVKind, this.mUVs);
        // this.mesh.setEnabled(true);
    }
    build(
        tileMapAsset: TileMapAsset,
        materials: MainOpacityMaterial,
        parent: BABYLON.TransformNode,
        events?: [number, number, number, number][]
    ) {
        let count = this.objects.size;
        let enable = count > 0;
        let asset = tileMapAsset.assets.get(this.assetSource);

        this._mesh(count, tileMapAsset, materials, parent);

        if (events?.length > 0) {
            this.dirtyPosition = true;
        }

        if (this.dirty && enable) {
            this._build(asset, events);
        } else if (enable && this.dirtyPosition && this.dirtyUVs) {
            this._buildPositionAndUV(asset, events);
        } else if (enable && this.dirtyPosition) {
            this._buildPosition(events);
        } else if (enable && this.dirtyUVs) {
            this._buildUV(asset);
        }
        if (enable && this.dirtyColor) {
            this._buildColor();
        }
        this.customGeometry.update(count * Tile.TRIANGLE_COUNT * Tile.TRIANGLE_INDICES_COUNT);
        if (this.mesh) {
            this.mesh.setEnabled(enable);
        }
        // this.mesh.setEnabled(false);

        this.dirty = false;
        this.dirtyPosition = false;
        this.dirtyUVs = false;
        this.dirtyColor = false;
    }
}
